c--------------------------------------------------------------------
c eqStep_RNSF.f
c
c Driver for the equilibrium calculation. Allocates space, 
c initializes parameters (arrays) for NIPS, calls NIPS and 
c converts output from computational variables into a vector
c of mole-numbers.
c--------------------------------------------------------------------
c     
c INPUTS:            
c
c epsConc                       the smallest mole-number permitted
c piMulti                       number of multi-phases
c piPure                        number of pure phases
c nRQ                           number of equilibrium reactions (=< nR)
c nPhase(piMulti+piPure)        number of species in each phase
c phaseIdentity(nSpecies)       phase-index for each species
c reactionType(nR)              Reaction type indicator,
c                               (0 = equilibrium, 1 = kinetic)
c
c nipsIparm(*)
c nipsRparm(*)
c nipsCparm(*)
c formulaMatrix(nComp,nSpecies)
c V(nSpecies,nR)
c mu(nSpecies)
c compCharge(nComp)
c 
c
c INPUT/OUTPUT:
c
c nvec(nSpecies)                On call: GUESS for the mole-numbers 
c                               On exit: New EQUILIBRATED mole-numbers 
c
c OUTPUT:
c
c ierror                        Signal problem in nips (.ne.0)
c iterations                    Number of iterations taken in NIPS
c
c ADDITIONAL INFORMATION:
c
c The compilation of this rountine depends on the choice of version
c as specified in the Makefile (UNSF, SF or RNSF).
c     
c-----------------------------------------------------------------------

c SLB 14/10/97 receive initial guess 

      subroutine eqStep(nvec,
     1                  InitGuess,
     &     epsConc,piMulti,piPure,nRQ,nProducts,
     $     nPhase,nS, nC,phaseIdentity,reactionType,nipsIparm,
     $     formulaMatrix,V,mu,compCharge,nipsRparm,
     $     rwk,dimrwk,iwk,dimiwk,idebug,debugunit,
     $     iterations,ierror,nipsCparm)
       
      IMPLICIT NONE
 
      integer nS, nC, nProducts 
      integer nSpecies,nComp,nR
      common /chemSizes/nSpecies,nComp,nR

      integer piMulti,piPure,nRQ,nPhase(piMulti+piPure),
     $     phaseIdentity(nS),reactionType(nProducts),
     $     nipsIparm(*),dimiwk,dimrwk,iwk(dimiwk),
     $     ierror,locH,locWater,maxIter,nViol,idealFlag,scaleFlag,
     $     testSolFlag,interpFlag,hessFlag,iptr,iptr_NEXT,iwkSize,
     $     rptr,rptr_NEXT,rptr_sol,rwkSize,iterations,
     $     idebug,debugunit,i,j
      integer ind,nMinQ
      logical mineral

      real*8 nvec(nS),formulaMatrix(nC,nS),
     $     V(nS,nProducts),mu(nS),compCharge(nC),nipsRparm(*),
     $     rwk(dimrwk),epsConc,relTol,alpha,tauMin,rho,one,zero,bigExp

      parameter(one = 1.0, zero = 0.0, bigExp = -100.0)
      
      character*15 nipsCparm(*)

c     COMMON AREAS for NIPS package

      integer n,me,mi,neq,niparm,nrparm
      common /nips_sizes/n,me,mi,neq,niparm,nrparm

      real*8 gasconst,abstemp,simtemp,pressure
      logical ideal
      common /thermoparam/ gasconst,abstemp,pressure,ideal

c   SLB 14/10/97 declare initial guess
      real*8 InitGuess(nS)
      
      IF ( idebug .ge. 1 ) THEN
        write (debugunit,*) ' Here is InitGuess inside eqStep:'
       write (debugunit,'(1pe10.3)') (InitGuess(j),j=1,nSpecies)
      ENDIF
      nSpecies = nS
      nComp = nC
      nR = nProducts

      gasconst = 1.0
      abstemp = 1.0


c-----------------------------------------------------------------------
c     PACK PARAMETER VECTORS FOR NIPS/NLEQ
c
c     INTEGER VECTOR: iparam(niparam)
c     REAL*8  VECTOR: rparam(nrparam)


c-----------------------------------------------------------------------
c     /INTEGER/
c-----------------------------------------------------------------------
c
c     Find out how many equilibrium minerals there are
c
c     write(*,*) 'Packing mineral data'
      nMinQ = 0
      do i = 1,nR
         mineral = nPhase(phaseIdentity(nComp+i)).eq.1
         if (mineral.and.(reactionType(i).eq.0)) nMinQ = nMinQ + 1
      enddo

c     write(*,*) 'Done with mineral data',nMinQ
c
c     Pack a smaller problem corresponding to 
c     EQUILIBRIUM PRODUCTS only!
c

c      write(*,*)'nComp:', nC,' nComp:',nComp, nProducts, nRQ
c     write(*,*) (iwk(i),i=1,4)
      iwk(1) = nC
      iwk(2) = nC + nRQ
      iwk(3) = nMinQ
c     write(*,*) (iwk(i),i=1,4)

      iptr = 3

c     /Mineral list - indicates whether a given EQUILIBRIUM 
c     product-species is a mineral /

c     write(*,*) 'Packing phase data'
      do i = 1, nR
         if (reactionType(i).eq.0) then
            iptr = iptr + 1
            mineral = nPhase(phaseIdentity(nComp+i)).eq.1
            if (mineral) then
               iwk(iptr) = 1
            else
               iwk(iptr) = 0
            endif
         endif
      enddo

c     /Mineral table - index into EQUILIBRIUM product-species 
c     list of the jth EQUILIBRIUM mineral/

      ind = 0
      do i = 1, nR
         if (reactionType(i).eq.0) then
            ind = ind + 1
            mineral = nPhase(phaseIdentity(nComp+i)).eq.1
            if (mineral) then
               iptr = iptr + 1
               iwk(iptr) = ind
            endif
         endif
      enddo

      niparm = iptr
      iptr_NEXT = iptr + 1
      iwkSize = dimiwk - iptr_NEXT + 1

      if (iwkSize.le.0) goto 998

c-----------------------------------------------------------------------
c     /REAL*8/
c-----------------------------------------------------------------------    

      rptr = 1
      rwk(rptr) = epsConc

c
c     For non-linear equation algorithms, pack in 
c     the following data:
c
c     REDUCED nComp*(nComp+nRQ) formulaMatrix corresponding to equilibrium
c      system only.
c     Total moles NOT in the form of kinetic product species
c     REDUCED vector of chemical reference potentials mu0(nComp+nRQ)
c

c     /A (REDUCED)/
c     write(*,*) 'packing A matrix'
      do i = 1, nComp
         do j = 1, nComp
            rptr = rptr + 1
            rwk(rptr) = formulaMatrix(j,i)
         enddo
      enddo

      do i = 1, nR
         if (reactionType(i).eq.0) then
            do j = 1, nComp
               rptr = rptr + 1
               rwk(rptr) = formulaMatrix(j,nComp+i)
            enddo
         endif
      enddo

c     /Totals/

      do i = 1,nComp
         rwk(rptr+i) = nvec(i)
      enddo

c     write(*,*) 'Fixing totals'
      do i = 1,nR
         if (reactionType(i).eq.0) then
            do j = 1, nComp
               rwk(rptr+j) = rwk(rptr+j) + 
     $              formulaMatrix(j,nComp+i)*nvec(nComp+i)
            enddo
         endif
      enddo
      
      rptr = rptr + nComp


c     /mu (reduced)/

c     write(*,*) 'Packing Mu'
      do i = 1, nComp
         rptr = rptr + 1
         rwk(rptr) = mu(i)
      enddo

      do i = 1, nR
         if (reactionType(i).eq.0) then
            rptr = rptr + 1
            rwk(rptr) = mu(nComp+i)
         endif
      enddo
c     write(*,*) 'Done packing RNSF data'

c-----------------------------------------------------------------------
c     DETERMINE NIPS/NLEQ SIZES 
c
c     n      =  Number of primary unknowns
c     me     =  Number of equality constraints
c     mi     =  Number of inequality constraints
c     neq    =  Total number of unknowns (size of linear system)
c     niparm =  Number of integer parameters for NIPS
c     nrparm =  Number of real*8 parameters for NIPS
c-----------------------------------------------------------------------

      n =  nComp + nMinQ
      me = 0
      mi = nMinQ

      neq = n + me + mi

      nrparm     = rptr
      rptr_sol   = rptr + 1
      rptr_NEXT  = rptr_sol + neq
      rwkSize    = dimrwk - rptr_NEXT + 1
      if (rwkSize.le.0) goto 999

c      
c-----------------------------------------------------------------------
c     UNPACK SOME PARAMETERS USED EXCLUSIVELY BY NIPS
c-----------------------------------------------------------------------
c
c     /INTEGER/
      
      locH        = nipsIparm(1)
      locWater    = nipsIparm(2)
      maxIter     = nipsIparm(3)
      nViol       = nipsIparm(4)
      idealFlag   = nipsIparm(5)
      scaleFlag   = nipsIparm(6)
      testSolFlag = nipsIparm(7)
      interpFlag  = nipsIparm(8)
      hessFlag    = nipsIparm(9)

      if (idealFlag.eq.1) then
         ideal = .true.
      else
         ideal = .false.
      endif


c     /REAL*8/

      simtemp   = nipsRparm(1)
      pressure  = nipsRparm(2)
      relTol    = nipsRparm(3)
      alpha     = nipsRparm(4)
      tauMin    = nipsRparm(5)
      rho       = nipsRparm(6)


c-----------------------------------------------------------------------
c     WRITE VERSION INFORMATION ON DEBUG UNIT
c-----------------------------------------------------------------------

      if (idebug.gt.0) then
         write(debugunit,1100)
      endif         

c-----------------------------------------------------------------------
c     PREPARE INITIAL GUESS FOR PRIMARY UNKNOWNS
c-----------------------------------------------------------------------

c
c     THIS GUESS HAS TWO PARTS:
c

c     - COMPONENT-CONCENTRATIONS: NATURAL LOGARITHMS

      do i = 1,nComp
c SLB  14/10/97 use conc at previous time step as initial guess
c         if (nvec(i).le.zero) then
         if (InitGuess(i).le.zero) then
            IF ( idebug .ge. 1 ) THEN
              write (debugunit,*) ' InitGuess(',i,') = ',
     &                               InitGuess(i)
              write (debugunit,*) ' Replacing it with ln(x) = ',bigExp
            ENDIF
            rwk(rptr_sol + i - 1) = bigExp
         else
c SLB  14/10/97 use conc at previous time step as initial guess
c           rwk(rptr_sol + i - 1) = dlog(nvec(i))
           rwk(rptr_sol + i - 1) = dlog(InitGuess(i))
         endif
      enddo

c     - EQUILIBRIUM MINERALS: CONCENTRATIONS

      ind = rptr_sol + nComp - 1
      do i = 1,nR
         if ((reactionType(i).eq.0).and.
     $        (nPhase(phaseIdentity(nComp+i)).eq.1)) then
            ind = ind + 1
c SLB  14/10/97 use conc at previous time step as initial guess
c            rwk(ind) = dmax1(nvec(nComp+i), 1.1*epsConc)
            rwk(ind) = dmax1(InitGuess(nComp+i), 1.1*epsConc)
         endif
      enddo

c-----------------------------------------------------------------------
c     CALL NLEQ PACKAGE
c-----------------------------------------------------------------------
c     write(*,*) 'Calling nleq..'
      call nleq(rwk(rptr_sol),iwk(1),rwk(1),iwk(iptr_NEXT),iwkSize,
     $     rwk(rptr_NEXT),rwkSize,relTol,alpha,tauMin,rho,
     $     maxIter,nViol,scaleFlag,interpFlag,idebug,debugunit,
     $     iterations,ierror)

      if (ierror.ne.0 .or. iterations.eq.0) RETURN

c-----------------------------------------------------------------------
c     POST-PROCESSING OF SOLUTION BEFORE RETURN
c-----------------------------------------------------------------------
c     Convert from computational variables to molenumbers
c
c     write(*,*) 'Calling compToMoles'
      call compToMoles(nvec,rwk(rptr_sol),nSpecies,nComp,nR,nMinQ,
     $     formulaMatrix,mu,nPhase,phaseIdentity,reactionType,
     $     rwk(rptr_NEXT),iwk(iptr_NEXT))

      RETURN

c-----------------------------------------------------------------------
c     ERROR TRAP (WORKSPACE)
c-----------------------------------------------------------------------

 998  write(*,*)'Out of integer workspace in eqStep'
      stop
 999  write(*,*)'Out of real*8 workspace in eqStep'
      stop

 1000 format(/,'EQUILIBRIUM SUBSTEP, VERSION UNSF',/)
 1100 format(/,'EQUILIBRIUM SUBSTEP, VERSION RNSF',/)
 1200 format(/,'EQUILIBRIUM SUBSTEP, VERSION SF',/)

      end


