c--------------------------------------------------------------------
c eqStep_SF.f
c
c Driver for the equilibrium calculation. Allocates space, 
c initializes parameters (arrays) for NIPS, calls NIPS and 
c converts output from computational variables into a vector
c of mole-numbers.
c--------------------------------------------------------------------
c     
c INPUTS:            
c
c epsConc                       the smallest mole-number permitted
c piMulti                       number of multi-phases
c piPure                        number of pure phases
c nRQ                           number of equilibrium reactions (=< nR)
c nPhase(piMulti+piPure)        number of species in each phase
c phaseIdentity(nSpecies)       phase-index for each species
c reactionType(nR)              Reaction type indicator,
c                               (0 = equilibrium, 1 = kinetic)
c
c nipsIparm(*)
c nipsRparm(*)
c nipsCparm(*)
c formulaMatrix(nComp,nSpecies)
c V(nSpecies,nR)
c mu(nSpecies)
c compCharge(nComp)
c 
c
c INPUT/OUTPUT:
c
c nvec(nSpecies)                On call: GUESS for the mole-numbers 
c                               On exit: New EQUILIBRATED mole-numbers 
c
c OUTPUT:
c
c ierror                        Signal problem in nips (.ne.0)
c iterations                    Number of iterations taken in NIPS
c
c ADDITIONAL INFORMATION:
c
c The compilation of this rountine depends on the choice of version
c as specified in the Makefile (UNSF, SF or RNSF).
c     
c-----------------------------------------------------------------------

c SLB 14/10/97 receive initial guess 

      subroutine eqStep(nvec,
     1                  InitGuess,
     &     epsConc,piMulti,piPure,nRQ,nProducts,
     $     nPhase,nS, nC,phaseIdentity,reactionType,nipsIparm,
     $     formulaMatrix,V,mu,compCharge,nipsRparm,
     $     rwk,dimrwk,iwk,dimiwk,idebug,debugunit,
     $     iterations,ierror,nipsCparm)
       
      IMPLICIT NONE
 
      integer nS, nC, nProducts 
      integer nSpecies,nComp,nR
      common /chemSizes/nSpecies,nComp,nR

      integer piMulti,piPure,nRQ,nPhase(piMulti+piPure),
     $     phaseIdentity(nS),reactionType(nProducts),
     $     nipsIparm(*),dimiwk,dimrwk,iwk(dimiwk),
     $     ierror,locH,locWater,maxIter,nViol,idealFlag,scaleFlag,
     $     testSolFlag,interpFlag,hessFlag,iptr,iptr_NEXT,iwkSize,
     $     rptr,rptr_NEXT,rptr_sol,rwkSize,iterations,
     $     idebug,debugunit,i,j
      integer rptr_VQ,rptr_n0

      real*8 nvec(nS),formulaMatrix(nC,nS),
     $     V(nS,nProducts),mu(nS),compCharge(nC),nipsRparm(*),
     $     rwk(dimrwk),epsConc,relTol,alpha,tauMin,rho,one,zero,bigExp

      parameter(one = 1.0, zero = 0.0, bigExp = -100.0)
      
      character*15 nipsCparm(*)

c     COMMON AREAS for NIPS package

      integer n,me,mi,neq,niparm,nrparm
      common /nips_sizes/n,me,mi,neq,niparm,nrparm

      real*8 gasconst,abstemp,simtemp,pressure
      logical ideal
      common /thermoparam/ gasconst,abstemp,pressure,ideal

c   SLB 14/10/97 declare initial guess
      real*8 InitGuess(nS)
      
      IF ( idebug .ge. 1 ) THEN
        write (debugunit,*) ' Here is InitGuess inside eqStep:'
       write (debugunit,'(1pe10.3)') (InitGuess(j),j=1,nSpecies)
      ENDIF
      nSpecies = nS
      nComp = nC
      nR = nProducts

      gasconst = 1.0
      abstemp = 1.0


c-----------------------------------------------------------------------
c     PACK PARAMETER VECTORS FOR NIPS/NLEQ
c
c     INTEGER VECTOR: iparam(niparam)
c     REAL*8  VECTOR: rparam(nrparam)


c-----------------------------------------------------------------------
c     /INTEGER/
c-----------------------------------------------------------------------

      iwk(1) = nRQ
      iwk(2) = nSpecies
      iwk(3) = nComp
      iwk(4) = piMulti
      iwk(5) = piPure
      iptr = 5

c     /nPhase/

      do i = 1, piMulti+piPure
         iptr = iptr + 1
         iwk(iptr) = nPhase(i)
      enddo

c     /phaseIdentity/

      do i = 1, nSpecies
         iptr = iptr + 1
         iwk(iptr) = phaseIdentity(i)
      enddo

c     /reactionType/

      do i = 1,nR
         iptr = iptr + 1
         iwk(iptr) = reactionType(i)
      enddo

      niparm = iptr
      iptr_NEXT = iptr + 1
      iwkSize = dimiwk - iptr_NEXT + 1

      if (iwkSize.le.0) goto 998

c-----------------------------------------------------------------------
c     /REAL*8/
c-----------------------------------------------------------------------    

      rptr = 1
      rwk(rptr) = epsConc

c
c     For the stoichiometric algorithm, pack in
c     the following data:
c
c     Stoichiometric matrix VQ for equilibrium sub-space
c     Chemical reference potentials mu0
c     Reference mole-vector n0 (satisfying A*n0 = b)
c     For safety, ensure that component & equilibrium products
c     in n0 are slightly larger than epsConc (for feasibility).
c
c
c     Note: These pointers are used by xsiToMoles.
c     
      rptr_VQ = 2
      rptr_n0 = rptr_VQ + nSpecies*nRQ + nSpecies

c     /V/

      do i = 1, nR
         if (reactionType(i).eq.0) then
            do j = 1, nSpecies
               rptr = rptr + 1
               rwk(rptr) = V(j,i)
            enddo
         endif
      enddo

c     /mu/

      do i = 1, nSpecies
         rptr = rptr + 1
         rwk(rptr) = mu(i)
      enddo

c     /n0/

      do i = 1,nComp
         rptr = rptr + 1 
         rwk(rptr) = dmax1(nvec(i), 1.01*epsConc)
      enddo

      do i = 1,nR
         rptr = rptr + 1 
         if (reactionType(i).eq.0) then
            rwk(rptr) = dmax1(nvec(nComp+i), 1.01*epsConc)
         else
            rwk(rptr) = nvec(nComp+i)
         endif
      enddo

c-----------------------------------------------------------------------
c     DETERMINE NIPS/NLEQ SIZES 
c
c     n      =  Number of primary unknowns
c     me     =  Number of equality constraints
c     mi     =  Number of inequality constraints
c     neq    =  Total number of unknowns (size of linear system)
c     niparm =  Number of integer parameters for NIPS
c     nrparm =  Number of real*8 parameters for NIPS
c-----------------------------------------------------------------------

      n =  nRQ
      me = 0
      mi = nRQ + nComp

      neq = n + me + mi

      nrparm     = rptr
      rptr_sol   = rptr + 1
      rptr_NEXT  = rptr_sol + neq
      rwkSize    = dimrwk - rptr_NEXT + 1
      if (rwkSize.le.0) goto 999

c      
c-----------------------------------------------------------------------
c     UNPACK SOME PARAMETERS USED EXCLUSIVELY BY NIPS
c-----------------------------------------------------------------------
c
c     /INTEGER/
      
      locH        = nipsIparm(1)
      locWater    = nipsIparm(2)
      maxIter     = nipsIparm(3)
      nViol       = nipsIparm(4)
      idealFlag   = nipsIparm(5)
      scaleFlag   = nipsIparm(6)
      testSolFlag = nipsIparm(7)
      interpFlag  = nipsIparm(8)
      hessFlag    = nipsIparm(9)

      if (idealFlag.eq.1) then
         ideal = .true.
      else
         ideal = .false.
      endif


c     /REAL*8/

      simtemp   = nipsRparm(1)
      pressure  = nipsRparm(2)
      relTol    = nipsRparm(3)
      alpha     = nipsRparm(4)
      tauMin    = nipsRparm(5)
      rho       = nipsRparm(6)


c-----------------------------------------------------------------------
c     WRITE VERSION INFORMATION ON DEBUG UNIT
c-----------------------------------------------------------------------

      if (idebug.gt.0) then
         write(debugunit,1200)
      endif         

c-----------------------------------------------------------------------
c     PREPARE INITIAL GUESS FOR PRIMARY UNKNOWNS
c-----------------------------------------------------------------------

c     THIS GUESS IS A VECTOR OF EXTENTS OF REACTION (ZERO)
      do i = 1,n
         rwk(rptr_sol + i - 1) = zero
      enddo

c-----------------------------------------------------------------------
c     CALL NIPS PACKAGE TO SOLVE MINIMIZATION PROBLEM
c-----------------------------------------------------------------------
c     write(*,*) 'Calling nips..'
      call nips(rwk(rptr_sol),iwk(1),rwk(1),iwk(iptr_NEXT),iwkSize,
     $     rwk(rptr_NEXT),rwkSize,relTol,alpha,tauMin,rho,
     $     maxIter,nViol,scaleFlag,testSolFlag,interpFlag,hessFlag,
     $     idebug,debugunit,iterations,ierror)

      if (ierror.ne.0 .or. iterations.eq.0) RETURN

c-----------------------------------------------------------------------
c     POST-PROCESSING OF SOLUTION BEFORE RETURN
c-----------------------------------------------------------------------

c     Use computed extents-of reactions to find corresponding 
c     mole-numbers

      call xsiToMoles(nvec,rwk(rptr_n0),rwk(rptr_sol),
     $     rwk(rptr_VQ),nSpecies,nRQ)

      RETURN

c-----------------------------------------------------------------------
c     ERROR TRAP (WORKSPACE)
c-----------------------------------------------------------------------

 998  write(*,*)'Out of integer workspace in eqStep'
      stop
 999  write(*,*)'Out of real*8 workspace in eqStep'
      stop

 1000 format(/,'EQUILIBRIUM SUBSTEP, VERSION UNSF',/)
 1100 format(/,'EQUILIBRIUM SUBSTEP, VERSION RNSF',/)
 1200 format(/,'EQUILIBRIUM SUBSTEP, VERSION SF',/)

      end


