c--------------------------------------------------------------------
c lnsrch.f
c     
c Backtracking linesearch algorithm, designed for the NLEQ arising
c  in the interior-point method.
c--------------------------------------------------------------------
c     
c INPUT:
c
c  dirder           Directional derivative in the Newton-direction
c  alpha            parameter for alpha-condition
c  interpolate      parameter controlling the type of line-search
c  oneTrip          if .TRUE.  no reduction of steplength occurs
c                   of .FALSE. a regular linesearch is performed
c
c
c INPUT/OUTPUT:
c
c x,z,g             On call: Current iterates
c                   On return: New iterates from line-search 
c
c residual          On call:    Damped newton-step
c                   On return : Residuals at the new point
c
c cost              On call: current value of meritfunction
c                   On return : new value of meritfunction
c
c OUTPUT:
c 
c retcode           Error (1) or succesful completion (0)
c
c ALGORITHM:
c     
c Back-tracking along Newton (descent) direction.
c For details, see Dennis & Schnabel's "Numerical Methods for
c Unconstrained Optimization and Nonlinear Equations", pp. 325-326.
c 
c Note: If flag interpolate = .true. a quadratic/cubic interpolation
c  method is used (pp. 325-326 above). Otherwise, a simple reduction
c  scheme of the form lambda := lambda*rho is used with some rho,
c  satisfying 0 < rho < 1.
c--------------------------------------------------------------------


      subroutine lnsrch(x,z,mu,residual,g,xt,zt,pnewton,
     $     iparam,rparam,rwk,dimrwk,iwk,dimiwk,cost,
     $     dirder,alpha,interpolate,lambda,oneTrip,retcode)


      IMPLICIT NONE
      INTRINSIC sqrt

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam

      integer retcode,maxstep,iter,i,iparam(niparam),
     $     dimiwk,dimrwk,iwk(dimiwk)


      real*8 x(n),xt(n),z(*),zt(*),g(*),mu,rparam(nrparam),
     $     rwk(dimrwk),residual(neq),pnewton(neq),alpha,dcompi,
     $     cost,costplus,lambda,dirder,term1,term2,term3,a,b,discr,
     $     lambdatmp,pcostplus,plambda,upperb,lowerb,rho,dnrm2

      logical interpolate,oneTrip,okcostplus

      parameter(
     $     maxstep = 100,
     $     upperb = 0.5d0,
     $     lowerb = 0.1d0,
     $     rho = 0.5d0)

      iter       = 1
      retcode    = 1
      lambda     = 1.0d0

c------------------------------------------------------------------------
c     TOP OF LINE-SEARCH LOOP
c------------------------------------------------------------------------

      call dcopy(neq,residual,1,pnewton,1)     

c------------------------------------------------------------------------
c     COMPUTE NEW CANDIDATE POINT
c------------------------------------------------------------------------

 100  do i = 1,neq
         dcompi = lambda*pnewton(i)
         if (i.le.n) then
            xt(i) = x(i) + dcompi
         elseif (i.gt.n .and. i.le. n+mi) then
            zt(i-n-me) = z(i-n-me) + dcompi
         endif
      enddo

c------------------------------------------------------------------------
c     EVALUATE RESIDUALS AT THE NEW CANDIDATE POINT
c     AND COMPUTE THE VALUE OF THE MERIT FUNCTION
c------------------------------------------------------------------------

      
      call constrEval(xt,g,iparam,rparam,rwk(1),dimrwk,iwk(1),dimiwk)
      call resEval(residual,xt,g,zt,mu,iparam,rparam,
     $     rwk(1),dimrwk,iwk(1),dimiwk)

      costplus   = dnrm2(neq,residual,1)**2

c     THIS REPRESENTS AN UPPER LIMIT FOR HOW POOR A STEP WE 
c     WILL EVER ACCEPT. IT SERVES AS A SAFEGUARD. THE PARAMETER 
c     SELECTION IS AD-HOC.

      okcostplus = (costplus.le.1.0d3*cost)

c------------------------------------------------------------------------
c      DETERMINE IF ALPHA-CONDITION IS SATISFIED FOR THIS LAMBDA 
c------------------------------------------------------------------------

      if (costplus .le. (cost + dirder*alpha*lambda)) then

c------------------------------------------------------------------------     
c     SOLUTION FOUND; ALGORITHM TERMINATES WITH OK FLAG
c------------------------------------------------------------------------  
 
         call dcopy(n,xt,1,x,1)     
         call dcopy(mi,zt,1,z,1)     
         cost = costplus
         retcode = 0
         RETURN

c------------------------------------------------------------------------          
c     RETURN WITH ERROR (NONSATISFACTION OF A-G COND) IN TWO CASES:
c
c     (1) WHEN ONLY ONE TRIP WAS PERMITTED (AND WE DON'T INCREASE TOO MUCH),
c             OR
c     (2) WHEN THE MAX NUMBER OF ITERATIONS IS EXCEEDED
c------------------------------------------------------------------------     

      elseif ((oneTrip.and.okcostplus).or.(iter.gt.maxstep)) then

         call dcopy(n,xt,1,x,1)     
         call dcopy(mi,zt,1,z,1)     
         cost = costplus
         retcode = 1
         RETURN

      else
c     
c       Find new lambda through back-tracking 
c     

         if (interpolate) then

c------------------------------------------------------------------------
c     USE QUADRATIC/CUBIC INTERPOLATION  
c------------------------------------------------------------------------

            if (iter .eq. 1) then
c     
c     First backtrack: use quadratic fit
c     
               lambdatmp = -dirder/(2.0*(costplus-cost-dirder))

            else
c     
c     All other backtracks: use qubic fit
c     
               term1 = costplus - cost - dirder*lambda
               term2 = pcostplus - cost - dirder*plambda
               term3 = 1.0d0/(lambda-plambda)
               a = term3*(term1/lambda**2 - term2/plambda**2)
               b = term3*(term2*lambda/plambda**2 - 
     $              term1*plambda/lambda**2)
               discr = b**2 - 3.0d0*a*dirder

               if (discr .gt. b**2) then
c     
c     Only one positive critical point
c     
                  lambdatmp = (-b + sign(1.0d0,a)*
     $                 sqrt(discr))/(3.0d0*a)
               else
c     
c     Two positive critical points, first is minimum
c     
                  lambdatmp = (-b - sign(1.0d0,a)*
     $                 sqrt(discr))/(3.0d0*a)
               endif
            endif
c     
c     Make sure new lambda satisfies constraints 
c     
            if (lambdatmp .gt. upperb*lambda) then
               lambdatmp = lambda*upperb
            elseif (lambdatmp .lt. lowerb*lambda) then
               lambdatmp = lambda*lowerb
            endif
            plambda = lambda
            lambda = lambdatmp
            pcostplus = costplus
            
            
         else

c------------------------------------------------------------------------
c     USE SIMPLE REDUCTION OF STEPLENGTH
c------------------------------------------------------------------------

            lambda = rho * lambda


         endif

c     Increase iteration counter and proceed with next iteration

         iter = iter + 1
         goto 100

      endif

      return
      end





