c--------------------------------------------------------------------
c nleqDriver.f
c
c Non-linear equations solver modified to include inequality constraint
c --------------------------------------------------------------------
c
c INPUTS/OUTPUTS:
c
c x(n)                On call:  Initial guess for primary variables 
c z(mi)                         No guess for multipliers
c
c                     On return: The solution.
c     
c OUTPUTS:
c
c returnflag          Indicates status upon completion
c                     0 - Normal completion
c                     1 - Exceeded maximum number of iterations
c                     2 - Exceeded maximum number of backtracks
c                     3 - Ill-conditioning of linear system 
c                         (not implemented)
c                     4 - Infeasible initial point
c
c BRIEF VARIABLE DEFINITIONS:
c
c x(n)                primary variables
c z(mi)               lagrange multipliers corresponding to the 
c                     inequality constraints
c
c matrix(neq,neq)     Holds the generally non-sparse matrix resulting when
c                     the (perturbed) Newton method is applied to the
c                     non-linear system of equations + complementarity
c
c residual(neq)       Holds the right-hand side that results in the
c                     perturbed Newton method.
c
c
c EXTERNAL ROUTINES:
c
c >> User-supplied
c
c residual             
c jacobian
c
c >> BLAS
c
c dnrm2                compute the 2-norm of a vector
c dcopy                copy vectors
c dgemv                matrix/vector multiplication + update
c
c --------------------------------------------------------------------

      subroutine nleqDriver(x,z,g,matrix,residual,tol,alpha,tauMin,rho,
     $     maxiter,nViol,interpolate,scaling,iparam,rparam,iwk,dimiwk,
     $     rwk,dimrwk,verbosity,outUnit,iter,returnflag)

      IMPLICIT NONE
      INTRINSIC sqrt

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam
      
      logical interpolate,scaling,lnsrchError,feasibleIC,
     $     oneTrip,converged,constraints

      integer i,dimiwk,dimrwk,iwk_SIZE,rwk_SIZE,ip_NEXT,rp_gstep,rp_xt,
     $     rp_zt,rp_step,rp_NEXT,iwk(dimiwk),iparam(niparam),verbosity,
     $     outUnit,returnflag,lflag,maxiter,nViol,violCtr,iter,job

      real*8 x(n),z(*),g(*),matrix(neq,neq),residual(neq),
     $     dnrm2,ddot,errorL2,rcond,cost,zero,one,tol,alpha,dirder,
     $     cc,rho,mu0Hi,mu0Lo,z0Hi,z0Lo,sum,mu,tauMin,
     $     rparam(nrparam),lambda_lnsrch,lambda_dmp,rwk(dimrwk)

      parameter (zero = 0.0d0, one = 1.0d0,
     $     z0Hi = 1.0d-1, z0Lo = 1.0d-3, 
     $     mu0Hi = 1.0d-3, mu0Lo = 1.0d-6)

c --------------------------------------------------------------------
c     INITIALIZE STORAGE FOR THE DRIVER 
c --------------------------------------------------------------------

c      /REAL WORKSPACE/

      rp_gstep  = 1
      rp_xt     = rp_gstep + max(mi,1)
      rp_zt     = rp_xt + n
      rp_step   = rp_zt + max(mi,1)
      rp_NEXT   = rp_step + 4*neq
      rwk_SIZE  = dimrwk - rp_NEXT + 1

      if (rwk_SIZE.le.0) then
         write(*,*)'OUT OF REAL MEMORY IN NLEQ DRIVER'
         stop
      endif

c     /INTEGER WORKSPACE/

      ip_NEXT = 1
      iwk_SIZE = dimiwk - ip_NEXT + 1

      if (iwk_SIZE.le.0) then
         write(*,*)'OUT OF INTEGER MEMORY IN NLEQ DRIVER'
         stop
      endif

c --------------------------------------------------------------------
c     INITIALIZE
c --------------------------------------------------------------------

      violCtr = 0
      lnsrchError = .false.
      iter = 0
      constraints = (mi.ne.0)
      
      if (constraints) then

         job = 0

c     Evaluate constraints
         
         call constrEval(x,g,iparam,rparam,rwk(rp_NEXT),
     $        rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)
         
c     Check the feasibility of the initial point
         
         feasibleIC = .true.
         do i = 1,mi
            feasibleIC = (g(i).gt.zero.and.feasibleIC)
         enddo

         if (.not.feasibleIC) then
            returnflag = 4
            RETURN
         endif

c     Select a guess for the multipliers and initial perturbation,
c     using complementarity condition. This is heuristic. 

         do i = 1,mi
            z(i) = dmin1(z0Hi, dmax1(z0Lo, 1.0/g(i)))
         enddo

         mu  = ddot(mi, g(1), 1, z(1), 1)
         mu  = dmin1(mu0Hi, dmax1(mu0Lo, mu))

      else

         mu  = zero
         job = 1

      endif

c     Evaluate residuals at (x0,z0)

      call resEval(residual,x,g,z,mu,iparam,rparam,rwk(rp_NEXT),
     $     rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)


c ---------------------------------------------------------------- 
c     CHECK ERROR AT INITIAL POINT
c     RETURN QUICKLY IF POSSIBLE
c ---------------------------------------------------------------- 

      call stopTestEval(residual,x,z,g,iparam,rparam,rwk(rp_NEXT),
     $     rwk_SIZE,iwk(ip_NEXT),iwk_SIZE,1,tol,constraints,
     $     converged,errorL2)

      if (converged) then
         if (verbosity.gt.0) write(outUnit,975)
         returnFlag = 0
         iter = 0
         RETURN
      endif

      if (verbosity.gt.0) write(outUnit,972) errorL2

      if (verbosity.eq.1) then
         write(outUnit,900)
      elseif (verbosity.ge.2) then
         write(outUnit,950)
      endif


c ----------------------------------------------------------------      
c     TOP OF THE MAIN LOOP
c ---------------------------------------------------------------       

 100  if (.not.(converged).and.(iter.le.maxiter).and. 
     $     .not.(lnsrchError)) then

c ---------------------------------------------------------------       
c  CHECK THE CENTRALITY OF THE CURRENT POINT
c
c   cc = || G(x)Ze - mu*e ||_2
c
c  IF current mu satisfies cc < mu 
c  THEN mu should be DECREASED by a factor rho in (0,1),
c  OTHERWISE mu remains UNCHANGED for next iteration.
c  IF a new mu is selected, THEN new residuals must be 
c  computed.
c ---------------------------------------------------------------       

         if (constraints) then
            sum = zero
            do i = 1,mi
               sum = sum + (g(i)*z(i) - mu)**2
            enddo
            cc = sqrt(sum)
            if (cc.lt.mu)  then
               mu = mu * rho
               call resEval(residual,x,g,z,mu,iparam,rparam,
     $              rwk(rp_NEXT),rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)
            endif
         endif

c --------------------------------------------------------------- 
c     EVALUATE JACOBIAN
c ---------------------------------------------------------------       

         call jacEval(matrix,x,z,g,iparam,rparam,
     $        rwk(rp_NEXT),rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)

c ---------------------------------------------------------------  
c     EVALUATE COST-FUNCTION AT CURRENT POINT 
c ---------------------------------------------------------------       

         cost = dnrm2(neq,residual,1)**2

c ---------------------------------------------------------------       
c     SOLVE THE LINEAR SYSTEM 
c     RESULT IS TRANSFERED TO RESIDUAL, MATRIX IS DESTROYED
c     NOTE: job = 0 non-symmetric solve, job = 1 symmetric solve
c ---------------------------------------------------------------       

         call linearSolver(matrix,residual,rcond,job,verbosity.ge.2,
     $        iwk(ip_NEXT),iwk_SIZE,rwk(rp_NEXT),rwk_SIZE,outUnit)

c ---------------------------------------------------------------       
c     PERFORM DAMPING OF STEP TO ENSURE FEASIBILITY.
c     THE STEP, CONTAINED IN 'residual' IS SCALED BY THE FACTOR
c     'lambda_dmp'.
c ---------------------------------------------------------------       
         
         if (constraints) then
            call damping(residual,x,z,g,rwk(rp_gstep),iparam,
     $           rparam,tauMin,lambda_dmp)
         else
            lambda_dmp = one
         endif

c     Compute the initial rate of decrese of the merit function.
c     The computed rate ACCOUNTS FOR THE DAMPING  

         dirder = -2.0d0*cost*lambda_dmp

c ---------------------------------------------------------------       
c     PERFORM LINESEARCH
c     Note that lflag = 0 signals normal completion
c               lflag = 1 signals a violation 
c
c     The user-specified parameter nViol specifies how
c     many consequtive violations will be accepted.
c
c     NOTE: This step will update the unknowns 'x' ,'z' and 'g'.
c           In addition, the vector 'residual' will be 
c           replaced by the value of the residuals at the 
c           new point (x,z).
c ---------------------------------------------------------------       

         if (violCtr.lt.nViol) then
            oneTrip = .true.
         else
            oneTrip = .false.
         endif

         call lnsrch(x,z,mu,residual,g,rwk(rp_xt),rwk(rp_zt),
     $        rwk(rp_step),iparam,rparam,rwk(rp_NEXT),
     $        rwk_SIZE,iwk(ip_NEXT),iwk_SIZE,cost,dirder,alpha,
     $        interpolate,lambda_lnsrch,oneTrip,lflag)
         
         if (lflag.eq.1) then
            violCtr = violCtr + 1 
         else
            violCtr = 0
         endif

         if (violCtr.gt.nViol) lnsrchError = .true.

c ---------------------------------------------------------------       
c     CHECK STOPPING CRITERIA
c ---------------------------------------------------------------       

         call stopTestEval(residual,x,z,g,iparam,rparam,
     $        rwk(rp_NEXT),rwk_SIZE,iwk(ip_NEXT),iwk_SIZE,0,
     $        tol,constraints,converged,errorL2)

c ---------------------------------------------------------------       
c     GENERATE OUTPUT FOR THIS ITERATION
c ---------------------------------------------------------------       

         iter = iter + 1

         if (verbosity.eq.1) then
            write(outUnit,1000) iter,cost,lambda_lnsrch
         elseif (verbosity.ge.2) then
            write(outUnit,1050) iter,cost,errorL2,
     $           lambda_lnsrch,lambda_dmp,mu,rcond
         endif


c ----------------------------------------------------------------      
c     END OF LOOP
c ----------------------------------------------------------------  

         goto 100
      endif


c ----------------------------------------------------------------      
c     DETERMINE RETURN STATUS
c ----------------------------------------------------------------           

      if (iter .gt. maxiter) then 
         returnflag = 1
      elseif (lnsrchError) then
         returnflag = 2
      else
         returnflag = 0
      endif

      
c ---------------------------------------------------------------- 
c     FORMAT STATEMENTS 
c     

 900  format(/,
     $     t10, ' -----------------------',/,
     $     t10, '| ITERATION INFORMATION |',/,
     $     t10, ' -----------------------'//,
     $     'Iteration',tr10,'MERITFCN',tr10,'LAMBDA')

 950  format(/,
     $     t15, ' ------------------------------',/,
     $     t15, '| FULL CONVERGENCE INFORMATION |',/,
     $     t15, ' ------------------------------'//,
     $     'Iteration',tr5,'MERITFCN',tr10,'L2-ERROR',
     $     tr8, 'LINESEARCH',tr6,'FEASIBILITY',tr9,
     $     'MU',tr13,'CONDITION')

 972  format(/,'Initial L2-error is',1x,e10.4)
 975  format(/,'This is the equilibrium step, VERSION RNSF', /,
     $     'The initial point is satisfactory')

 1000 format(t1,i3,tr13,e10.4,tr7,e10.4)
 1050 format(t1,i3,tr11,e10.4,5(tr7,e10.4))
 2000 format(/,a,tr1,i3)
      
      return
      end	





