c--------------------------------------------------------------------
c chempot.f
c
c Compute the chemical potential of a given multiple phase species.
c --------------------------------------------------------------------
c     
c INPUTS:
c     
c mu0(nSpecies)            Standard chemical potentials
c nVec(n)                  Composition vector
c nTot(pi)                 Vector of total moles 
c indx                     Index of species whose chemical potential 
c                          will be determined
c phaseIdentity(nSpecies)
c nPhase(pi)
c nSpecies                 Total number of species
c pi                       Number of phases
c
c OUTPUTS:
c     
c chempot                  Chemical potential of species nVec(indx)
c                          (scaled by R*T)
c --------------------------------------------------------------------
c   
c ALGORITHM:
c
c If ideal = .true.: 
c  
c The standard expression for chemical potential is used, i.e.
c     
c  chempot = mu0_i + RT*ln(x_i)
c
c If ideal = .false. we use
c
c chempot = mu0_i + RT*ln(gamma(x)*x_i)
c     
c where gamma_i is the ACTIVITY COEFFICIENT,
c which may be a function of the composition of the entire phase
c in which x_i participates.
c     
c     
c NOTE1: 
c Non-ideal case is currently NOT IMPLEMENTED!
c For reasons of efficiency, we don't test this.
c
c NOTE2:
c The pure phases have a COMPOSITION INDEPENDENT 
c chemical potential.  For those species,
c
c chempot = mu0_i
c
c NOTE3:
c We actually compute a scaled form of the potential, namely
c
c mu^hat = mu/RT
c
c This will scale the total energy content in the system (an extensive
c quantity) by the factor R*T, a transformation which leaves the solution
c unchanged. However, we save a few flops and get more 
c reasonable numbers.
c --------------------------------------------------------------------

      real*8 function chempot(mu0,nVec,nTot,indx,
     $     phaseIdentity,nPhase,nSpecies,pi)
      
      IMPLICIT NONE
      
      integer indx,pi,nSpecies,phaseIdentity(nSpecies),nPhase(pi),phase
      real*8  nVec(nSpecies),mu0(nSpecies),nTot(pi),molefrac

      real*8 gasconst,abstemp,pressure
      logical ideal
      common /thermoparam/ gasconst,abstemp,pressure,ideal

      phase = phaseIdentity(indx)

      if (nPhase(phase).gt.1) then

c--------------------------------------------------------------------
c     THIS INDEX CORRESPONDS TO A PARTICIPANT 
c     IN A MULTI-SPECIES PHASE
c--------------------------------------------------------------------
         
            molefrac = nVec(indx)/nTot(phase)
            chempot = mu0(indx) + dlog(molefrac)

      else

c--------------------------------------------------------------------
c     THIS INDEX CORRESPONDS TO A MINERAL
c--------------------------------------------------------------------

         chempot = mu0(indx)

      endif

      return
      end	


































































