c--------------------------------------------------------------------
c fdhleval.f
c
c Finite-difference evaluation of the hessian of the lagrangian
c--------------------------------------------------------------------     
c Note : The hessian of the lagrangian has the following form:
c     
c     
c hessl = hess(f) - sum_{i=1}^{me)[y(i)*hess(h_i)] - 
c sum_{i=1}^{mi)[z(i)*hess(g_i)] 
c     
c where {hess(f)}_ij : = d^2(f)/dx(i)dx(j) etc.
c     
c--------------------------------------------------------------------
c INPUTS:
c     
c x(n)
c y(me) 
c z(mi)
c n
c me
c mi
c    
c OUTPUTS:
c    
c hess(n,n)        The approximate hessian resulting 
c                  from a finite-difference
c                  approximation (symmetrized).
c     
c EXTERNAL ROUTINES:
c     
c gradeval         Evaluate gradients of f,h,g
c     
c     
c ALGORITHM:
c     
c The approximate Hessian A of LAGRANGIAN at the point x 
c is computed from
c     
c A(:,j) = [grad(x + h_j * e_j) - grad(x)]/h_j
c     
c A : = (A+A')/2
c    
c with grad() = gradl()
c     
c In the expressions above, e_j is a unit vector in the jth coordinate 
c direction, and the steplength h_j is determined by
c     
c h_j : = sqrt(m_eps) * x(j).
c     
c     
c (See "Numerical Methods for Unconstrained Optimization and 
c  Nonlinear Equations", Dennis&Schnable, Prentice-Hall,
c  pages 95-105  for details regarding this procedure 
c  and recommendations for the selection of the 
c  perturbation size.)
c     
c --------------------------------------------------------------------

      subroutine fdhleval(hess,x,y,z,gradf_plus,gradh_plus,gradg_plus,
     $     grad,iparam,rparam,rwk,dimrwk,iwk,dimiwk)

      IMPLICIT NONE
      INTRINSIC sqrt

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam
      
      integer i,j,l,iparam(niparam),dimrwk,dimiwk,iwk(dimiwk)
      real*8 hess(n,n),x(n),gradf_plus(n),grad(n),
     $     rparam(nrparam),rwk(dimrwk),
     $     eta,dsqeta,stepsize_j,tempxj,half,gradplus_i
      real*8 y(*),z(mi),gradh_plus(n,*),gradg_plus(n,mi)

      parameter(eta = 1.0d-16,half = 0.5d0)

      dsqeta = sqrt(eta)

      do j = 1,n

c--------------------------------------------------------------------     
c     PERTURB IN THE jth COORDINATE DIRECTION
c--------------------------------------------------------------------     

         if (dabs(x(j)).gt.eta) then
            stepsize_j = dsqeta*dabs(x(j))
         else
            stepsize_j = dmax1(0.01*dabs(x(j)),1.0d-30)
         endif

         tempxj = x(j)
         x(j) = x(j) +  stepsize_j

c--------------------------------------------------------------------     
c     CALCULATE THE NEW GRADIENTS 
c--------------------------------------------------------------------  
   
         call gradeval(gradf_plus,gradh_plus,gradg_plus,
     $        x,iparam,rparam,rwk,dimrwk,iwk,dimiwk)

c--------------------------------------------------------------------     
c     COMPUTE THE jth COLUMN OF HESSIAN
c--------------------------------------------------------------------
     
         do i = 1,n
c--------------------------------------------------------------------     
c     ASSEMBLE THE ith COMPONENT OF GRADIENT OF LAGRANGIAN AT NEW POINT
c--------------------------------------------------------------------     
            gradplus_i = gradf_plus(i)

            do l=1,me	
               gradplus_i =  gradplus_i - y(l)*gradh_plus(i,l)
            enddo
            do l=1,mi
               gradplus_i =  gradplus_i - z(l)*gradg_plus(i,l)
            enddo

c--------------------------------------------------------------------     
c     1st ORDER FINITE-DIFFERENCE APPROXIMATION 
c--------------------------------------------------------------------   
  
            hess(i,j) = (gradplus_i - grad(i))/stepsize_j

         enddo
c--------------------------------------------------------------------     
c      RE-SET x(j)
c--------------------------------------------------------------------     
         x(j) = tempxj

      enddo
      
c--------------------------------------------------------------------     
c     SYMMETRIZE HESSIAN APPROXIMATION
c--------------------------------------------------------------------     

c     FORM AVERAGES OF THE STRICT UPPER/LOWER TRIANGULAR PARTS

      do i = 1,n-1
         do j = i+1,n
            hess(i,j) = half*( hess(i,j) + hess(j,i) )
         enddo
      enddo

c     FILL IN

      do i = 2,n
         do j = 1,i-1 
            hess(i,j) = hess(j,i)
         enddo
      enddo


      return
      end	
