c--------------------------------------------------------------------
c nipsDriver.f
c
c Constrained optimization using the Interior-point Method.
c--------------------------------------------------------------------
c     
c This routines primary function is to return a solution to the 
c constrained optimization problem:
c
c          min f(x)
c          subject to h(x)   = 0
c                     g(x) > = 0.
c
c
c      Here, f : R^N -> R
c            h : R^N -> R^me
c            g : R^N -> R^mi
c
c     where n is the dimension of the problem, and me and mi
c     are the number of equality and inequality constraints,
c     respectively. The code attempts to find a KKT point for 
c     the above system.
c
c     For details on the Interior Point Method, see the Rice University 
c     Technical report TR92-40 (El-Bakry, Tapia, Zhang, Tsuchiya)
c     
c     Work on the code began August 1995
c
c     Primary authors:
c       Fredrik Saaf <fredrik@rice.edu>
c       Amr El-Bakry <elbakry@rice.edu>
c
c
c
c NOTE: This implementation tailored for SIMPLE BOUND TYPE INEQUALITIES.
c
c --------------------------------------------------------------------
c INPUTS: 
c
c x(n),y(me),z(mi)      Initial guess (for primal variables only)
c hanalytic             logical flag signaling how the Hessian of the
c                       Lagrangian is computed 
c                       .true. - analytic, must be supplied by user
c                       .false - finite-difference approximation
c
c INPUTS/OUTPUT:
c x(n),y(me),z(mi)      Solution 
c     
c OUTPUTS:
c returnflag          Indicates status upon completion
c                     0 - Normal completion
c                     1 - Exceeded maximum number of iterations
c                     2 - Exceeded maximum number of backtracks
c                     3 - Ill-conditioning of linear system 
c                         (not implemented)
c                     4 - Infeasible initial point
c
c BRIEF VARIABLE DEFINITIONS:
c
c x(n)                primary variables
c y(me)               lagrange multipliers corresponding to the 
c                     equality constraints
c z(mi)               lagrange multipliers corresponding to the 
c                     inequality constraints
c
c f                   objective function value
c h(me)               equality constraint vector
c g(mi)               inequality constraint vector
c
c gradf(n)            gradient of f 
c gradh(n,me)         gradient of equality constraint functions        
c gradg(n,mi)         gradient of inequality constraint functions        
c gradl(n)            gradient of the Lagrangian
c
c (Note: l(x,y,z) = f(x) - y'h(x) - z'g(x) => 
c        gradl = gradf - gradh*y - gradg*z )
c
c hessl(n,n)          Hessian of the Lagrangian
c
c (Note: hessl = hessf - hessh*y - hessg*z where 
c        {hessf}_(i,j) = d^2 f/dx_i dx_j etc.
c       The hessians can either be supplied analytically or be 
c       approximated, based on finite-differences)
c       
c matrix(neq,neq)     Holds the generally non-sparse matrix resulting when
c                     the (perturbed) Newton method is applied to the
c                     non-linear system of KKT conditions.
c
c residual(neq)        Holds the right-hand side that results in the
c                     perturbed Newton method.
c
c
c
c EXTERNAL ROUTINES:
c
c >> User-supplied
c
c func                 Evaluate the functions f(x),g(x) and h(x)
c grad                 Evaluate the gradients of f(x),g(x) and h(x)
c
c >> BLAS
c
c dnrm2                compute the 2-norm of a vector
c dcopy                copy vectors
c dgemv                matrix/vector multiplication + update
c
c --------------------------------------------------------------------

      subroutine nipsDriver(x,y,z,h,g,gradf,gradh,gradg,gradl,
     $     hessl,matrix,residual,tol,alpha,tauMin,rho,maxiter,
     $     nViol,hanalytic,interpolate,scaling,secSuffTest,iparam,
     $     rparam,iwk,dimiwk,rwk,dimrwk,verbosity,outUnit,
     $     iter,returnflag)

      IMPLICIT NONE
      INTRINSIC sqrt

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam
      
      logical hanalytic,interpolate,scaling,secSuffTest,
     $     lnsrchError,feasibleIC,oneTrip,converged

      integer i,dimiwk,dimrwk,iwk_SIZE,rwk_SIZE,ip_NEXT,rp_NEXT,
     $     iwk(dimiwk),iparam(niparam),verbosity,outUnit,returnflag,
     $     lflag,maxiter,nViol,violCtr,iter,rp_xt,rp_yt,rp_zt,
     $     rp_gf,rp_gh,rp_gg,rp_gstep

      real*8 x(n),y(*),z(mi),f,h(*),g(mi),gradf(n),gradh(n,*),
     $     gradg(n,mi),gradl(n),hessl(n,n),matrix(neq,neq),dnrm2,ddot,
     $     errorL2,l2norm,rcond,residual(neq),lambda_lnsrch,cost,zero,
     $     tol,alpha,dirder,cc,rho,sum,mu,tauMin,lambda_dmp,
     $     lambda,rwk(dimrwk),rparam(nrparam),z0Hi,z0Lo,mu0Hi,mu0Lo 

      parameter (zero = 0.0d0, 
     $     z0Hi  = 1.0d0, z0Lo = 1.0d-3,
     $     mu0Hi = 1.0d-3, mu0Lo = 1.0d-6)


c --------------------------------------------------------------------
c     INITIALIZE STORAGE FOR THE DRIVER 
c --------------------------------------------------------------------

c      /REAL WORKSPACE/

      rp_xt    = 1
      rp_yt    = rp_xt + n
      rp_zt    = rp_yt + max(me,1)
      rp_gf    = rp_zt + mi
      rp_gh    = rp_gf + n	
      rp_gg    = rp_gh + n*max(1,me)
      rp_gstep = rp_gg + n*mi
      rp_NEXT  = rp_gstep + mi

      rwk_SIZE = dimrwk - rp_NEXT + 1

      if (rwk_SIZE.le.0) then
         write(*,*)'OUT OF REAL MEMORY IN NIPS DRIVER'
         stop
      endif
c
c     /INTEGER WORKSPACE/
c
      ip_NEXT = 1
      iwk_SIZE = dimiwk - ip_NEXT + 1

      if (iwk_SIZE.le.0) then
         write(*,*)'OUT OF INTEGER MEMORY IN NIPS DRIVER'
         stop
      endif

c --------------------------------------------------------------------
c     INITIALIZE
c --------------------------------------------------------------------

c     EVALUATE ALL FUNCTIONS AT (x0)

      call funceval(f,h,g,x,iparam,rparam,rwk(rp_NEXT),
     $     rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)     

      call gradeval(gradf,gradh,gradg,x,iparam,rparam,
     $     rwk(rp_NEXT),rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)

c     CHECK THE FEASIBILITY OF THE INITIAL POINT

      feasibleIC = .true.
      do i = 1,mi
         feasibleIC = (g(i).gt.zero.and.feasibleIC)
      enddo

      if (.not.feasibleIC) then
         returnflag = 4
         RETURN
      endif

c     SELECT A GUESS FOR MULTIPLIERS (y) if applicable and (z)
c     and an INITIAL PERTURBATION (mu)

      if (me.gt.0) then
         do i = 1,me
            y(i) = 1.0d0
         enddo
      endif

c     Try to use complementarity as a guide for selecting z

      do i = 1,mi
         z(i) = dmin1(z0Hi, dmax1(z0Lo, 1.0/g(i)))
      enddo

      mu  = ddot(mi,g(1),1,z(1),1)
      mu  = dmin1(mu0Hi,dmax1(mu0Lo,mu))
      
c
c     EVALUATE GRADL AT (x0,y0,z0)
c
      call dcopy(n,gradf,1,gradl,1)
      if (me.gt.0) call dgemv('N',n,me,-1.0d0,gradh,
     $     n,y,1,1.0d0,gradl,1)
      call dgemv('N',n,mi,-1.0d0,gradg,n,z,1,1.0d0,gradl,1)

c
c     EVALUATE HESSIAN AT (x0,y0,z0) 
c
      if (hanalytic) then
         call hleval(hessl,x,y,z,iparam,rparam,rwk(rp_NEXT),
     $        rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)
      else
         call fdhleval(hessl,x,y,z,rwk(rp_gf),rwk(rp_gh),rwk(rp_gg),
     $        gradl,iparam,rparam,rwk(rp_NEXT),rwk_SIZE,
     $        iwk(ip_NEXT),iwk_SIZE)
      endif

c ---------------------------------------------------------------- 
c     CHECK ERROR AT INITIAL POINT
c     RETURN QUICKLY IF POSSIBLE
c ---------------------------------------------------------------- 

      call stopTestEval(f,h,g,gradf,gradh,gradg,gradl,x,y,z,iparam,
     $     rparam,rwk(rp_NEXT),rwk_SIZE,iwk(ip_NEXT),iwk_SIZE,1,
     $     converged,tol,errorL2)

      if (verbosity.gt.0) write(outUnit,972) errorL2

      if (converged) then
         if (verbosity.gt.0) write(outUnit,975)
         returnFlag = 0
         iter = 0
         RETURN
      endif

c     SET-UP

      violCtr = 0
      lnsrchError = .false.
      iter = 0

c     PRINT HEADER 

      if (verbosity.eq.1) then
         write(outUnit,900)
      elseif (verbosity.ge.2) then
         write(outUnit,950)
      endif

c ----------------------------------------------------------------      
c     TOP OF THE MAIN LOOP
c ---------------------------------------------------------------       

 100  if ((.not.converged).and.(iter.le.maxiter).and.
     $     (.not.lnsrchError)) then


c ---------------------------------------------------------------       
c  Define CENTRALITY CONDITION 
c  (here, w.r.t notion of "central path"),
c
c   cc = || G(x)Ze - mu*e ||_2 + || h(x) ||_2
c
c  IF current mu satisfies cc < mu 
c  THEN mu should be DECREASED by a factor rho in (0,1),
c  OTHERWISE mu remains UNCHANGED for next iteration.
c
c  This strategy is analogous to the notion of solving the 
c  logarithmic barrier formulation for a given mu up to the
c  tolerance specified by mu itself. When we've reached a point
c  within a neighborhood of mu of the central path, mu is reduced,
c  and we begin iterating on a problem closer to the one we
c  ultimately wish to solve.
c ---------------------------------------------------------------       

         sum = zero
         do i = 1,mi
            sum = sum + (g(i)*z(i) - mu)**2
         enddo
         
         if (me.gt.0) then
            cc = sqrt(sum) + dnrm2(me,h,1)
         else
            cc = sqrt(sum)
         endif

         if (cc.lt.mu)  mu = mu * rho

c ---------------------------------------------------------------  
c     EVALUATE COST-FUNCTION AT CURRENT POINT 
c ---------------------------------------------------------------       

         cost = l2norm(gradl,h,g,z,mu)**2

c --------------------------------------------------------------- 
c     SET UP MATRIX AND RIGHT-HAND SIDE FOR THE
c     PERTURBED NEWTON SYSTEM (MU.NEQ.0) 
c ---------------------------------------------------------------       

         call setupmat(matrix,hessl,gradh,gradg,g,z)
         call setuprhs(residual,h,g,gradl,z,mu)

c     PERFORM DIAGONAL SCALING OF DEPENDENT VARIABLES 
     
         if (scaling) call yscale(matrix,residual,hessl,h,g,z)

c ---------------------------------------------------------------       
c     SOLVE THE LINEAR SYSTEM 
c     RESULT IS TRANSFERED TO RESIDUAL, MATRIX IS DESTROYED
c     NOTE: job = 0 non-symmetric solve, job = 1 symmetric solve
c ---------------------------------------------------------------       

         call linearSolver(matrix,residual,rcond,0,verbosity.ge.2,
     $        iwk(ip_NEXT),iwk_SIZE,rwk(rp_NEXT),rwk_SIZE,outUnit)

c ---------------------------------------------------------------       
c     PERFORM DAMPING OF STEP TO ENSURE FEASIBILITY.
c     THE STEP, CONTAINED IN 'residual' IS SCALED BY THE FACTOR
c     'lambda_dmp'.
c ---------------------------------------------------------------       

         call damping(residual,x,z,g,rwk(rp_gstep),iparam,rparam,
     $        tauMin,lambda_dmp)

c     
c     Compute the initial rate of decrese of the merit function.
c     The computed rate accounts for the damping.  

         lambda = lambda_dmp
         dirder = -2.0d0*cost*lambda

c ---------------------------------------------------------------       
c     PERFORM LINESEARCH
c     Note that lflag = 0 signals normal completion
c               lflag = 1 signals a violation 
c
c     The user-specified parameter nViol specifies how
c     many consequtive violations will be accepted.
c
c
c -->   NOTE: This step will update 'x' ,'y', 'z' 
c                               and 'f', 'h', 'g'
c                               and  'gradf', 'gradh', 'gradg'
c                               and  'gradl'
c
c ---------------------------------------------------------------       

         if (violCtr.lt.nViol) then
            oneTrip = .true.
         else
            oneTrip = .false.
         endif
         
         call lnsrch(x,y,z,mu,residual,f,h,g,gradf,gradh,gradg,gradl,
     $        rwk(rp_xt),rwk(rp_yt),rwk(rp_zt),iparam,rparam,
     $        rwk(rp_NEXT),rwk_SIZE,iwk(ip_NEXT),iwk_SIZE,cost,
     $        dirder,alpha,interpolate,lambda_lnsrch,oneTrip,lflag)

         if (lflag.eq.1) then
            violCtr = violCtr + 1 
         else
            violCtr = 0
         endif

         if (violCtr.gt.nViol) lnsrchError = .true.

c --------------------------------------------------------------- 
c     UPDATE HESSIAN OF THE LAGRANGIAN BASED ON NEW ITERATE
c ---------------------------------------------------------------       

         if (hanalytic) then
            call hleval(hessl,x,y,z,iparam,rparam,rwk(rp_NEXT),
     $           rwk_SIZE,iwk(ip_NEXT),iwk_SIZE)
         else
            call fdhleval(hessl,x,y,z,rwk(rp_gf),rwk(rp_gh),rwk(rp_gg),
     $           gradl,iparam,rparam,rwk(rp_NEXT),rwk_SIZE,
     $           iwk(ip_NEXT),iwk_SIZE)
         endif

c ---------------------------------------------------------------       
c     CHECK STOPPING CRITERIA
c ---------------------------------------------------------------       

         call stopTestEval(f,h,g,gradf,gradh,gradg,gradl,x,y,z,
     $        iparam,rparam,rwk(rp_NEXT),rwk_SIZE,
     $        iwk(ip_NEXT),iwk_SIZE,0,converged,tol,errorL2)

c ---------------------------------------------------------------       
c     GENERATE OUTPUT FOR THIS ITERATION
c ---------------------------------------------------------------       

         iter = iter + 1

         if (verbosity.eq.1) then
            write(outUnit,1000) iter,f,cost,lambda_lnsrch
         elseif (verbosity.ge.2) then
            write(outUnit,1050) iter,f,cost,errorL2,
     $           lambda_lnsrch,lambda_dmp,mu,rcond
         endif

c ----------------------------------------------------------------      
c     END OF LOOP
c ----------------------------------------------------------------  

         goto 100
      endif


c ----------------------------------------------------------------      
c     DETERMINE RETURN STATUS
c ----------------------------------------------------------------           

      if (iter .gt. maxiter) then 
         returnflag = 1
      elseif (lnsrchError) then
         returnflag = 2
      else
         returnflag = 0
      endif


c ------------------------------------------------------------------     
c     FORMAT STATEMENTS
c     

 900  format(/,
     $     t10, '-------------------------',/,
     $     t10, '| ITERATION INFORMATION |',/,
     $     t10, '-------------------------'//,
     $     'Iteration',tr10,'f(x)',tr10,'COSTFCN',tr10,'LAMBDA')

 950  format(/,
     $     t15, '------------------------------',/,
     $     t15, '|FULL CONVERGENCE INFORMATION |',/,
     $     t15, '------------------------------'//,
     $     'Iteration',tr5,'f(x)',tr11,'COSTFCN',tr10,'L2-ERROR',
     $     tr8, 'LINESEARCH',tr6,'FEASIBILITY',tr9,
     $     'MU',tr13,'CONDITION')

 972  format(/,'Initial L2-error is',1x,e10.4)
 975  format(/,'The initial point is satisfactory: RETURN')

 1000 format(t1,i3,tr13,e10.4,2(tr7,e10.4))
 1050 format(t1,i3,tr8,e10.4,tr7,e10.4,5(tr7,e10.4))
      
      return
      end	





