c ------------------------------------------------------------------------
c secsuff.f
c
c Compute the eigenvalues of the projected Hessian to examine the  
c second-order necessary conditions at a given point.
c ------------------------------------------------------------------------
c     
c AUTHOR: Fredrik Saaf
c     
c VARIABLE DESCRIPTION:
c     
c INPUT:
c x(n), y(me), z(mi)
c rwk,nrwk 	     Real work array and its size
c iwk,niwk 	     Integer work array and its size
c iparam
c rparam
c
c
c OUTPUT:
c
c bindlist(mi)       List of approximately binding (1) or
c                     non-binding (0) inequalities.
c 
c ncActive           Number of (approximately) active constraints at x
c                    (me + number of active inequality constraints)
c
c ncRank             Number of linearly independent constraint-gradients
c                    (out of the 'nActive' possible ones)
c
c eigval(n)          Eigenvalues of the hessian of f projected onto 
c                    constraint manifold
c -------------------------------------------------------------------------
c
      subroutine secsuff(x,y,z,bindlist,ncActive,ncRank,eigval,
     $     iparam,rparam,rwk,dimrwk,iwk,dimiwk)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam
      
      integer i,iparam(niparam),dimiwk,iwk(dimiwk),dimrwk,
     $     rp_h,rp_g,rp_gradf,rp_gradh,rp_gradg,
     $     rp_hess,rp_xhess,rp_a,rp_vt,rp_s,rp_NEXT,rwk_SIZE,
     $     ncActive,ncRank,info,nbind,bindlist(mi)

      real*8 f,x(n),y(*),z(*),eigval(n),rparam(nrparam),rwk(dimrwk),gtol

      parameter (gtol = 1.0d-8)

c -------------------------------------------------------------------------
c     ALLOCATE REAL STORAGE
c -------------------------------------------------------------------------

      rp_h     =  1
      rp_g     =  rp_h + max(1,me)
      rp_gradf =  rp_g + mi
      rp_gradh =  rp_gradf + n
      rp_gradg =  rp_gradh + n*max(1,me)
      rp_hess  =  rp_gradg + n*mi
      rp_xhess =  rp_hess + n*n
      rp_a     =  rp_xhess + n*n
      rp_vt    =  rp_a + n*(me+mi)
      rp_s     =  rp_vt + n*n
      rp_NEXT  =  rp_s + min(n,me+mi)
      rwk_SIZE =  dimrwk - rp_NEXT + 1
      
      if (rwk_SIZE.lt.0) then
         write(*,*)'OUT OF REAL WORKSPACE IN SECSUFF'
         stop
      endif

c -------------------------------------------------------------------------
c     COMPUTE FUNCTIONS, GRADIENTS AND THE HESSIAN OF f AT x
c -------------------------------------------------------------------------

      call funceval(f,rwk(rp_h),rwk(rp_g),x,iparam,rparam,
     $     rwk(rp_NEXT),rwk_SIZE,iwk,dimiwk)

      call gradeval(rwk(rp_gradf),rwk(rp_gradh),rwk(rp_gradg),x,
     $     iparam,rparam,rwk(rp_NEXT),rwk_SIZE,iwk,dimiwk)

      call hleval(rwk(rp_hess),x,y,z,iparam,rparam,rwk(rp_NEXT),
     $     rwk_SIZE,iwk,dimiwk)

c -------------------------------------------------------------------------
c     DETERMINE LIST OF ACTIVE INEQUALITY CONSTRAINTS
c -------------------------------------------------------------------------

      nbind = 0

      do i = 1,mi
         if (dabs(rwk(rp_g-1+i)).lt.gtol) then
            nbind = nbind + 1
            bindlist(i) = 1
         else
            bindlist(i) = 0              
         endif
      enddo

      ncActive = me + nbind


      if (ncActive.eq.0) then
c ---------------------------------------------------------------------     
c     QUICK RETURN IF NO ACTIVE CONSTRAINTS - NO PROJECTIONS NECESSARY
c ---------------------------------------------------------------------     
         call dsyev('N','U',n,rwk(rp_hess),n,eigval,
     $        rwk(rp_NEXT),rwk_SIZE,info)

         if (info.ne.0) then 
            write(*,*)'Failed in eigenvalue-routine (LAPACK)'
            stop
         endif
         RETURN
      endif


c -------------------------------------------------------------------------
c     FIND DIMENSIONALITY (ncRank), AND A BASIS (vt), FOR 
c     THE SET OF ACTIVE CONSTRAINT-GRADIENTS
c -------------------------------------------------------------------------

      call getBasis(rwk(rp_a),rwk(rp_gradh),rwk(rp_gradg),bindlist,
     $     nbind,rwk(rp_vt),rwk(rp_s),ncRank,rwk(rp_NEXT),rwk_SIZE)

c -------------------------------------------------------------------------
c     PROJECT THE HESSIAN ONTO THE SPACE OF CONSTRAINTS
c -------------------------------------------------------------------------

      call projectHess(rwk(rp_hess),rwk(rp_xhess),n,ncRank,rwk(rp_vt))

c -------------------------------------------------------------------------
c     FIND THE EIGENVALUES OF THE PROJECTED HESSIAN
c     using the LAPACK routine DSYEV:
c     
c     SUBROUTINE DSYEV(JOBZ,UPLO,N,A,LDA,W,WORK,LWORK,INFO )
c ---------------------------------------------------------------------     

      call dsyev('N','U',n-ncRank,rwk(rp_xhess),n,eigval,
     $     rwk(rp_NEXT),rwk_SIZE,info)

      if (info.ne.0) then 
         write(*,*)'Failed in eigenvalue-routine (LAPACK)'
         stop
      endif
      
      return
      end



c =======================================================================   
c     AUXILLIARY ROUTINES CALLED BY 'secsuff' FOLLOW.
c =======================================================================   


      subroutine getBasis(a,gradh,gradg,bindlist,nbind,
     $     vt,s,rank,rwk,dimrwk)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam
 
      integer i,j,row,dimrwk,nbind,bindlist(mi),rank,info,ctr
      
      real*8 a(me+mi,n),vt(n,n),u,s(*),gradh(n,me),
     $     gradg(n,mi),rwk(dimrwk),eps

      parameter (eps = 1.0d-12)

c -----------------------------------------------------------------------   
c     STORE ACTIVE CONSTRAINTS ROW-WISE IN A
c -----------------------------------------------------------------------   

      do i = 1,me
         do j = 1,n
            a(i,j) = gradh(j,i)
         enddo
      enddo

      row = me
      do i = 1,mi
         if (bindlist(i).eq.1) then
            row = row + 1
            do j = 1,n
               a(row,j) = gradg(j,i)
            enddo
         endif
      enddo

c -----------------------------------------------------------------------   
c     PERFORM A SINGULAR-VALUE DECOMPOSITION OF A THROUGH CALL TO LAPACK
c -----------------------------------------------------------------------   
c     
c     SUBROUTINE DGESVD(JOBU,JOBVT,M,N,A,LDA,S,U,LDU,
c     VT,LDVT,WORK,LWORK,INFO )
c     
c     
c     
c     Purpose
c     =======
c     
c     DGESVD computes the singular value decomposition (SVD) of a real
c     M-by-N matrix A,optionally computing the left and/or right singular
c     vectors. The SVD is written
c     
c     A = U * SIGMA * transpose(V)
c     
c     where SIGMA is an M-by-N matrix which is zero except for its
c     min(m,n) diagonal elements,U is an M-by-M orthogonal matrix, and
c     V is an N-by-N orthogonal matrix.  The diagonal elements of SIGMA
c     are the singular values of A; they are real and non-negative, and
c     are returned in descending order.  The first min(m,n) columns of
c     U and V are the left and right singular vectors of A.
c     
c     Note that the routine returns V**T, not V.
c     
c     Arguments
c     =========
c     
c     JOBU    (input) CHARACTER*1
c     = 'N':  no columns of U (no left singular vectors) are
c     computed.
c     
c     JOBVT   (input) CHARACTER*1
c     Specifies options for computing all or part of the matrix
c     V**T:
c     = 'A':  all N rows of V**T are returned in the array VT;

c     
c     M       (input) INTEGER
c     The number of rows of the input matrix A.  M >= 0.
c     
c     
c     N       (input) INTEGER
c     The number of columns of the input matrix A.  N >= 0.
c     
c     
c     A       (input/output) DOUBLE PRECISION array, dimension (LDA,N)
c     On entry, the M-by-N matrix A.
c     
c     LDA     (input) INTEGER
c     The leading dimension of the array A.  LDA >= max(1,M).
c     
c     
c     S       (output) DOUBLE PRECISION array, dimension (min(M,N))
c     The singular values of A, sorted so that S(i) >= S(i+1).
c     
c     
c     U       (output) DOUBLE PRECISION array, dimension (LDU,UCOL)
c     (LDU,M) if JOBU = 'A' or (LDU,min(M,N)) if JOBU = 'S'.
c     If JOBU = 'A', U contains the M-by-M orthogonal matrix U;
c     if JOBU = 'S', U contains the first min(m,n) columns of U
c     (the left singular vectors, stored columnwise);
c     if JOBU = 'N' or 'O', U is not referenced.
c     
c     LDU     (input) INTEGER
c     The leading dimension of the array U.  LDU >= 1; if
c     JOBU = 'S' or 'A', LDU >= M.
c     
c     
c     VT      (output) DOUBLE PRECISION array, dimension (LDVT,N)
c     If JOBVT = 'A', VT contains the N-by-N orthogonal matrix
c     V**T;
c     if JOBVT = 'S', VT contains the first min(m,n) rows of
c     V**T (the right singular vectors, stored rowwise);
c     if JOBVT = 'N' or 'O', VT is not referenced.
c     
c     
c     
c     LDVT    (input) INTEGER
c     The leading dimension of the array VT.  LDVT >= 1; if
c     JOBVT = 'A', LDVT >= N; if JOBVT = 'S', LDVT >= min(M,N).
c     
c     
c     
c     WORK    (workspace/output) DOUBLE PRECISION array, dimension (LWORK)
c     
c     LWORK   (input) INTEGER
c     The dimension of the array WORK. LWORK >= 1.
c     
c     INFO    (output) INTEGER
c     = 0:  successful exit.
c     < 0:  if INFO = -i, the i-th argument had an illegal value.
c     > 0:  if DBDSQR did not converge, INFO specifies how many
c     superdiagonals of an intermediate bidiagonal form B
c     did not converge to zero. See the description of WORK
c     above for details.
c     
c     =====================================================================
c     

      call dgesvd('N','A',me+nbind,n,a,me+mi,s,u,1,vt,n,rwk,dimrwk,info)
        
      if (info.ne.0) then 
         write(*,*)'Failed in SVD-routine (LAPACK)'
         stop
      endif

c -------------------------------------------------------------------------
c     DETERMINE THE NUMERICAL RANK OF A FROM ITS SINGULAR VALUES
c -------------------------------------------------------------------------

      ctr = 0
      do i = 1, min(me + nbind,n)
         if (s(i).gt.eps) ctr = ctr + 1
      enddo	

      rank = ctr

      return
      end


      subroutine projectHess(hess,xhess,n,rank,vt)
      IMPLICIT NONE

      integer i,j,k,l,n,rank
      real*8 hess(n,n),xhess(n,n),vt(n,n),sum

      do i = rank+1,n
         do j = rank+1,n
            sum = 0.0d0
            do k = 1,n
               do l = 1,n
                  sum = sum + vt(i,k)*hess(k,l)*vt(j,l)
               enddo
            enddo
            xhess(i-rank,j-rank) = sum
         enddo	
      enddo	

      
      return
      end
