c--------------------------------------------------------------------
c damping.f
c
c Perform damping of the Perturbed Newton Step to insure feasibility
c--------------------------------------------------------------------
c INPUTS:
c
c z(mi)                     Lagrange multipliers, inequalities
c g(mi)                     Inequality constraints
c
c INPUTS/OUTPUT:
c
c step(neq)                 Perturbed Newton step on input
c                           adjusted-length ditto on output
c
c dg(mi)                    Work vector 
c
c OUTPUTS:
c
c alpha_hat                 Damping for FEASIBILITY
c
c EXTERNAL ROUTINES:
c
c ddot()                    BLAS-routine for dot-product
c dscal()                   BLAS-routine for scaling a vector
c
c ALGORITHM:
c
c FEASIBILITY
c 
c Let g+(i) = g_i[x(i)+ alpha*dx(i)]
c and z+(i) = z(i) + alpha*dz(i)
c
c Ensure that the conditions g+(i) >= 0 for i = 1,...,mi
c                            z+(i) >= 0 for i = 1,...,mi
c
c In this version (STOICHIOMETRIC) we have the special requirement that
c g(x) is an affine function of the form g(x) = n0 + V*x.
c
c Thus,
c
c
c  g+ = g(x+) = n0 + V*(x+) = 
c               n0 + V*(x + alpha*dx) = 
c               n0 + V*x + alpha*V*dx
c
c Noting that  g(x)  = n0 + V*x by definition,
c
c and setting     dg < -- V*dx,
c
c the problem is on the standard form:
c
c Find alpha such that,
c
c  g + alpha*dg >= 0
c  z + alpha*dx >=0
c
c
c The largest alpha satisfying the above conditions is denoted alpha_hat
c Since alpha =< 1, we take  alpha_hat = min(alpha_hat,1)       
c
c --------------------------------------------------------------------
c  

      subroutine damping(step,x,z,g,dg,iparam,rparam,tauMin,alpha_hat)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam
      
      integer i,j,ind,iparam(niparam)
      integer nRxn,nSpecies,nComp,rptr_V 
      real*8 sum

      real*8 step(neq),g(mi),dg(mi),x(n),z(mi),rparam(nrparam),
     $     alpha_hat,dotprod,tau,tauMin,minq,q,big,one,zero,eta,ddot

      parameter (
     $     big = 1.0d20,
     $     zero = 0.0d0,
     $     one = 1.0d0,
     $     eta = 1.0d2)

c --------------------------------------------------------------------
c     Unpack a few things needed
c     (This is non-standard)
c
      nRxn             = iparam(1) 
      nSpecies         = iparam(2)
      nComp            = iparam(3)      
      rptr_V = 2

c
c     Compute dg = V*dx (for equilibrium products and components)
c

      do i = 1,nComp
         sum = 0.0d0
         do j = 1,nRxn
            sum = sum + rparam(rptr_V-1+i+nSpecies*(j-1))*step(j)
         enddo
         dg(i) = sum
      enddo

      do i = 1,nRxn
         dg(i+nComp) = step(i)
      enddo

c --------------------------------------------------------------------
c     INITIALIZE
c --------------------------------------------------------------------

c     Compute "approach to the boundary" parameter tau

      dotprod = ddot(mi,g,1,z,1)
      tau = dmax1(tauMin, 1.0d0 - eta*dotprod)

c --------------------------------------------------------------------
c     FEASIBILITY
c --------------------------------------------------------------------

      alpha_hat = one
      minq = big

c     Constraint on primary variables

      do i = 1,mi
         q = dg(i)/g(i)
         minq = dmin1(minq,q)
      enddo

c     Constraints on inequality multipliers

      do i = 1,mi
         ind = n+me+i
         if (z(i).le.zero) then
            if (step(ind).le.zero) step(ind) = zero
         else
            q = step(ind)/z(i)
            minq = dmin1(minq,q)
         endif
      enddo

      if (minq.ge.zero) then
         alpha_hat = one
      else
         alpha_hat = dmin1(-tau/minq,one)
      endif

c --------------------------------------------------------------------
c     EFFECT SCALING OF THE STEP
c --------------------------------------------------------------------

 100  call dscal(neq,alpha_hat,step,1)


      return
      end	




