c--------------------------------------------------------------------
c func.f
c
c Evaluate function and constraint for the interior-point method
c applied to the geochemistry problem (STOICHIOMETRIC VERSION).
c
c --------------------------------------------------------------------
c INPUTS:
c
c x(n)                        Extents of reaction
c n0(nSpecies)                Reference mole-numbers
c V(nSpecies,nRxn)            Stoichiometric matrix
c mu0(nSpecies)               Reference chemical potentials
c phaseIdentity(nSpecies)     Integer array of phase-identity for 
c                              all species
c reactionType(nSpecies-nComp)
c
c nPhase(pi)                  Number of species in each phase
c nVec(nSpecies)              Work array
c nTot(pi)                    Work array
c
c OUTPUTS:
c
c f                         Gibbs Free Energy
c g(mi)                     Non-negativity constraints 
c     
c EXTERNAL ROUTINES:
c
c chempot                   Evaluate chemical potential
c
c --------------------------------------------------------------------

      subroutine func(f,h,g,x,nSpecies,nRxn,nComp,pi,epsConc,
     $     nPhase,phaseIdentity,reactionType,V,n0,mu0,nVec,nTot)   

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam 
    
      integer i,iptr,nSpecies,nComp,nRxn,pi,nPhase(pi),
     $     phaseIdentity(nSpecies),reactionType(nSpecies-nComp)

      real*8 f,x(n),h(*),g(mi),mu0(nSpecies),V(nSpecies,nRxn),epsConc,
     $     n0(nSpecies),nVec(nSpecies),nTot(pi),sum,zero,chempot

      parameter(zero = 0.0d0)

c --------------------------------------------------------------------
c     FUNCTION EVALUATION
c --------------------------------------------------------------------

c     Compute mole vector from extents-of-reaction

      call xsiToMoles(nVec,n0,x,V,nSpecies,nRxn)

c     Compute total moles for multi-species phases

      call phaseMoles(nTot,nVec,pi,nSpecies,phaseIdentity)

c     Compute the Gibbs function for the mixture 

      sum = zero
      do i = 1, nSpecies
         sum = sum + nVec(i)*chempot(mu0,nVec,nTot,i,phaseIdentity,
     $        nPhase,nSpecies,pi)
      enddo

      f = sum

c --------------------------------------------------------------------
c     INEQUALITY CONSTRAINTS (NON-NEGATIVITY)
c
c     Note 1: 
c
c     Species ordering is always nComp components followed by
c     (nSpecies-nComp) product species. We need only worry
c     about the non-negativity of the product species 
c     resulting from equilibrium reactions (i.e. reactionType = 0)
c     since kinetic species cannot be changed in the minimization
c     step. The number of inequalities is therefore
c
c            mi  = nComp + nRxn 
c
c     Note 2:
c
c     We ensure more than non-negativity - the constraint enforced is:
c
c     n(i) > epsConc
c
c     epsConc is a user-specified tolerance that we never fall below.
c --------------------------------------------------------------------

      do i = 1, nComp
         g(i) = nVec(i) - epsConc
      enddo

      iptr = nComp
      do i = nComp+1,nSpecies
         if (reactionType(i-nComp).eq.0) then
            iptr = iptr + 1
            g(iptr) = nVec(i) - epsConc
         endif
      enddo


      return
      end	

