c--------------------------------------------------------------------
c grad.f
c
c Evaluate gradients of function and constraints for the 
c interior-point method applied to the geochemistry problem.
c This is the STOICHIOMETRIC VERSION.
c --------------------------------------------------------------------
c
c INPUTS:
c
c x(n)                        Extents of reaction
c n0(nSpecies)                Reference mole-numbers
c V(nSpecies,nRxn)            Stoichiometric matrix
c mu0(nSpecies)               Reference chemical potentials
c phaseIdentity(nSpecies)     Integer array of phase-identity for 
c                             all species
c
c nPhase(pi)                  Number of species in each phase
c nVec(nSpecies)              Work array
c nTot(pi)                    Work array
c
c
c OUTPUTS:
c
c gradf(n)                    Gradient of Gibbs Free energy
c gradh(n,me)                 Gradient of Equality constraints
c gradg(n,mi)                 Gradient of Inequality constraints
c     
c
c EXTERNAL ROUTINES:
c
c chempot                     Evaluate chemical potential
c
c --------------------------------------------------------------------

      subroutine grad(gradf,gradh,gradg,x,nSpecies,nRxn,nComp,pi,
     $     nPhase,phaseIdentity,reactionType,V,n0,mu0,nVec,nTot)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam 
          
      integer i,j,colptr,nSpecies,nRxn,nComp,pi,
     $     nPhase(pi),phaseIdentity(nSpecies),
     $     reactionType(nSpecies-nComp)
     
      real*8 gradf(n),gradh(n,*),gradg(n,mi),x(n),
     $     V(nSpecies,nRxn),n0(nSpecies),nVec(nSpecies),
     $     mu0(nSpecies),chempot,nTot(pi),zero,one,sum,mu_j

      parameter(zero = 0.0d0,one = 1.0d0)

c --------------------------------------------------------------------
c     GRADIENT OF THE GIBBS FUNCTION
c --------------------------------------------------------------------

c     Compute mole vector from extents-of-reaction

      call xsiToMoles(nVec,n0,x,V,nSpecies,nRxn)

c     Compute total moles for multi-species phases

      call phaseMoles(nTot,nVec,pi,nSpecies,phaseIdentity)

c     Gradient of the Gibbs function, gradf = V'*mu 

      do i = 1,nRxn
         sum = zero
         do j = 1,nSpecies
            mu_j = chempot(mu0,nVec,nTot,j,phaseIdentity,
     $           nPhase,nSpecies,pi)
            sum = sum + V(j,i)*mu_j
         enddo
         gradf(i) = sum
      enddo

c --------------------------------------------------------------------
c     GRADIENT OF INEQUALITY CONSTRAINTS
c
c     gradg = V' , except for column corresponding 
c                  to non-equilibrium species.
c
c     Note: Species ordering is always nComp components followed by
c         (nSpecies-nComp) product species. We only need to worry
c          about the non-negativity of the product species 
c          resulting from equilibrium reactions (i.e. reactionType = 0)
c          since kinetic species cannot be changed in the 
c          minimization step. 
c --------------------------------------------------------------------

      do j = 1,nComp
         do i = 1,nRxn
            gradg(i,j) = V(j,i)
         enddo
      enddo


      colptr = nComp
      do j = nComp+1,nSpecies
         if (reactionType(j-nComp).eq.0) then
            colptr = colptr + 1 
            do i = 1,nRxn
               gradg(i,colptr) = V(j,i)
            enddo
         endif
      enddo

      return
      end	
