c--------------------------------------------------------------------
c hessian.f
c 
c Analytic evaluation of the hessian of the lagrangian.
c Stoichiometric case.
c --------------------------------------------------------------------
c
c INPUTS:
c
c x(n)                       Extents of reaction
c nSpecies
c nComp
c dimRxn (=nSpecies-nComp)
c nRxn
c pi
c nPhase(pi)                 Number of species in each phase
c phaseIdentity(nSpecies)    Integer array of phase-identity for species
c reactionType(dimRxn)       Integer list of reaction types
c V(nSpecies,nRxn)           Stoichiometric matrix
c n0(nSpecies)               Reference mole-numbers
c     
c OUTPUTS:
c
c hessl(n,n)                 Analytic Hessian of the Lagrangian function
c
c EXTERNAL ROUTINES:
c
c xsiToMoles
c phaseMoles
c chempotDeriv
c
c WORK ARRAYS:
c
c nVec(nSpecies)     
c nTot(pi)           
c A(nComp,nComp)
c B(dimRxn,dimRxn)
c C(nComp,dimRxn)              
c L(nRxn,dimRxn)
c temp(nComp,nRxn)
c
c --------------------------------------------------------------------
c
c ALGORITHM:
c
c Note first that it holds that
c
c   hessl = f_xx,
c
c (where the notation f_xx denotes to derivatives in the x variables)
c since the inequality constraints are linear in x.
c
c If the Hessian in terms of MOLE-NUMBER VARIABLES is denoted hessl_nn,
c it also holds that
c
c   hessl = V^{T}*hessl_nn*V.
c
c However, it is too costly to perform this computation directly
c (approximately O(nSpecies^3)), so we break up the computation 
c into the appropriate sub-matrices and use their special properties.
c
c Let 
c                [ A    C]
c     hessl_nn = [       ]
c                [ C^T  B]
c
c where:
c
c  A - Symmetric nComp*nComp matrix
c  B - Symmetric dimRxn*dimRxn matrix
c  C - Rectangular nComp*dimRxn matrix
c
c The special structure of V is used, so we write
c
c      [ V_hat ]
c V =  [       ]
c      [  D    ]
c
c where:
c
c  V_hat - nComp*nRxn 
c  D     - dimRxn*nRxn
c
c In the case where the equilibrium reactions span the entire 
c stoichiometric space, dimRxn = nRxn, and D = I_nRxn.
c In general though, the nRxn columns of D have all zero elements
c except for one '1' corresponding to the appropriate equilibrium 
c product species.
c
c We now have an expression for hessl as follows:
c
c
c hessl = V_hat^T*A*V_hat + L*D +  (D*L)^T + D^T*B*D
c           (T.1)          (T.2)   (T.3)      (T.4)
c where
c      L = V_hat^T*C
c
c Only T.1 is computed using a dense matrix multiplication.
c T.2 & T.3 represent permutations of rows/columns L (which is computed),
c and T.4 is assembled by accessing the appropriate elements of B
c
c --------------------------------------------------------------------
      
      subroutine hessian(hessl,x,nSpecies,nComp,dimRxn,nRxn,pi,
     $     nPhase,phaseIdentity,reactionType,V,n0,nVec,nTot,
     $     A,B,C,L,temp,globalRxnInd)   


      IMPLICIT NONE
 
      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam 

      real*8 gasconst,abstemp,pressure
      logical ideal
      common /thermoparam/ gasconst,abstemp,pressure,ideal

      integer i,j,pi,pj,ind,nSpecies,nComp,dimRxn,nRxn,nPhase(pi),
     $     phaseIdentity(nSpecies),reactionType(dimRxn),
     $     globalRxnInd(nRxn)

      real*8 hessl(n,n),x(n),V(nSpecies,nRxn),n0(nSpecies),
     $     nVec(nSpecies),nTot(pi),zero,chempotDeriv,
     $     A(nComp,nComp),B(dimRxn,dimRxn),C(nComp,dimRxn),
     $     L(nRxn,dimRxn),temp(nComp,nRxn)

      parameter (zero = 0.0d0)

c --------------------------------------------------------------------
c     ERROR-TRAP
c --------------------------------------------------------------------

      if (.not.ideal) then
         write(*,10)
 10      format(/,'ERROR in Hessian',/
     $        'Non-ideal case is not implemented ---',/,
     $        'STOPPING')

         STOP
      endif

c --------------------------------------------------------------------
c     SET UP
c --------------------------------------------------------------------

c     Compute mole vector from extents-of-reaction 

      call xsiToMoles(nVec,n0,x,V,nSpecies,nRxn)

c     Compute total moles for multi-species phases 

      call phaseMoles(nTot,nVec,pi,nSpecies,phaseIdentity)

c     Calculate sub-matrices A, B and C of the Hessian 
c     in mole-number space.

c     /A - symmetric, nComp*nComp/

      do i = 1,nComp
         do j = i,nComp
            A(i,j) = chempotDeriv(nVec,nTot,phaseIdentity,
     $           nSpecies,pi,nPhase,i,j)
         enddo
      enddo

c     /B - symmetric, (dimRxn)*(dimRxn)/ 

      do i = 1,dimRxn
         do j = i,dimRxn
            B(i,j) = chempotDeriv(nVec,nTot,phaseIdentity,
     $           nSpecies,pi,nPhase,i+nComp,j+nComp)
         enddo
      enddo

c     /C - rectangular nComp*(dimRxn)/

      do i = 1,nComp
         do j = 1,dimRxn
            C(i,j) = chempotDeriv(nVec,nTot,phaseIdentity,
     $           nSpecies,pi,nPhase,i,j+nComp)
         enddo
      enddo

c     Initialize array of mapping from list of equilibrium reactions
c     to global list of reactions

      ind = 0
      do i = 1,dimRxn
         if (reactionType(i).eq.0) then
            ind = ind + 1
            globalRxnInd(ind) = i
         endif
      enddo

c --------------------------------------------------------------------
c     COMPUTE THE nRxn*nRxn HESSIAN IN REACTION-COORDINATE SPACE
c --------------------------------------------------------------------

c     
c     hessl <--- V_hat^T * A * V_hat
c

c     First, temp < --- A*V_hat (A is symmetric)

      call dsymm('L','U', nComp, nRxn, 1.0d0, A, nComp,
     $     V, nSpecies, 0.0d0, temp, nComp)

c     Then hessl <--- V_hat^T * temp

      call dgemm('T', 'N', nRxn, nRxn, nComp, 1.0d0, V, nSpecies,
     $     temp, nComp, 0.0d0, hessl, nRxn)


c     Compute L < --- V_hat^T * C

      call dgemm('T', 'N', nRxn, dimRxn, nComp, 1.0d0, V, 
     $     nSpecies, C, nComp, 0.0d0, L, nRxn)

c
c     hessl(i,j) <--- hessl(i,j) + L(i,p(j)) + L(j,p(i)) + B(p(i),p(j))
c

      do i = 1,nRxn
         do j = i,nRxn
            pI = globalRxnInd(i)
            pJ = globalRxnInd(j)

            hessl(i,j) = hessl(i,j) + L(i,pJ) + L(j,pI) + B(pI,pJ)

         enddo
      enddo

      do i = 2,nRxn
         do j = 1,i-1
            hessl(i,j) = hessl(j,i)
         enddo
      enddo


      return
      end	


