c--------------------------------------------------------------------
c damping.f
c
c Perform damping of the Perturbed Newton Step to insure feasibility
c--------------------------------------------------------------------     
c
c INPUTS:
c
c z(mi)
c g(mi)
c
c INPUTS/OUTPUT:
c
c step(neq)                 Perturbed Newton step on input
c                           adjusted-length ditto on output
c
c gstep(mi)                 Work vector
c
c OUTPUTS:
c
c alpha_hat                 Damping for FEASIBILITY
c
c EXTERNAL ROUTINES:
c
c ddot()                    BLAS-routine for dot-product
c dscal()                   BLAS-routine for scaling a vector
c
c ALGORITHM:
c
c FEASIBILITY
c 
c Let g+(i) = g_i[x(i)+ alpha*step_x(i)]
c and z+(i) = z(i) + alpha*step_z(i)
c
c Ensure that the conditions g+(i) >= 0 for i = 1,...,mi
c                            z+(i) >= 0 for i = 1,...,mi
c
c
c The largest alpha satisfying the above conditions is denoted alpha_hat
c Since alpha =< 1, we take  alpha_hat = min(alpha_hat,1)       
c
c --------------------------------------------------------------------
c  

      subroutine damping(step,x,z,g,gstep,iparam,rparam,
     $     tauMin,alpha_hat)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam
      
      integer i,ind,iparam(niparam)

      real*8 step(neq),g(mi),gstep(mi),x(n),z(mi),rparam(nrparam),
     $     alpha_hat,dotprod,tau,tauMin,minq,q,big,one,zero,eta,ddot

      parameter (
     $     big = 1.0d20,
     $     zero = 0.0d0,
     $     one = 1.0d0,
     $     eta = 1.0d2)


c --------------------------------------------------------------------
c     INITIALIZE
c --------------------------------------------------------------------

c     Compute "approach to the boundary" parameter tau

      dotprod = ddot(mi,g,1,z,1)
      tau = dmax1(tauMin, 1.0d0 - eta*dotprod)

c --------------------------------------------------------------------     
c     FEASIBILITY
c -------------------------------------------------------------------- 
    
      alpha_hat = one
      minq = big

c     Constraint on primary variables

      do i = 1,n
         if (g(i).le.zero) then
            if (step(i).le.zero) step(i) = zero
         else
            q = step(i)/g(i)
            minq = dmin1(minq,q)
         endif
      enddo

c     Constraints on inequality multipliers

      do i = 1,mi
         ind = n+me+i
         if (z(i).le.zero) then
            if (step(ind).le.zero) step(ind) = zero
         else
            q = step(ind)/z(i)
            minq = dmin1(minq,q)
         endif
      enddo

      if (minq.ge.zero) then
         alpha_hat = one
      else
         alpha_hat = dmin1(-tau/minq,one)
      endif

c --------------------------------------------------------------------
c     EFFECT SCALING OF THE STEP
c --------------------------------------------------------------------

 100  call dscal(neq,alpha_hat,step,1)


      return
      end	




