c--------------------------------------------------------------------
c func.f
c
c Evaluate function and constraint for the interior-point method
c applied to the geochemistry problem. 
c
c --------------------------------------------------------------------
c INPUTS:
c
c x(n)                      molenumbers
c
c aa(nComp,nSpecies)        Formula matrix for the system
c bb(nComp)                 Component totals
c mu0(nSpecies)             Reference potentials
c phaseIdentity(nSpecies)   Integer array of phase-identity for 
c                           all species
c
c nPhase(pi)                Number of species in each phase
c   
c
c OUTPUTS:
c
c f                         Gibbs Free Energy
c h(me)                     Mass-balance & constancy constraints
c g(mi)                     Non-negativity constraints (simple bounds)
c     
c EXTERNAL ROUTINES:
c
c chempot                   Evaluate chemical potential
c
c --------------------------------------------------------------------
c

      subroutine func(f,h,g,x,nComp,nSpecies,pi,epsConc,
     $     nPhase,phaseIdentity,aa,bb,mu0,nTot)   

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam 
    
      integer i,j,nComp,nSpecies,pi,nPhase(pi),phaseIdentity(nSpecies)

      real*8 f,x(n),h(me),g(mi),mu0(nSpecies),aa(nComp,nSpecies),
     $     bb(nComp),nTot(pi),sum,zero,chempot,epsConc

      parameter(zero = 0.0d0)

c --------------------------------------------------------------------
c     FUNCTION EVALUATION
c --------------------------------------------------------------------

c     Total phase moles

      call phaseMoles(nTot,x,pi,nSpecies,phaseIdentity)

c      Compute the gibbs function for the mixture 

      sum = zero
      do i = 1, nSpecies
         sum = sum +  x(i)*chempot(mu0,x,nTot,i,
     $        phaseIdentity,nPhase,nSpecies,pi)
      enddo

      f = sum

c --------------------------------------------------------------------
c     EQUALITY CONSTRAINTS (MASS-BALANCE)
c --------------------------------------------------------------------

      do j = 1, nComp
         h(j) = -bb(j)
         do i = 1, nSpecies
            h(j) = h(j) + x(i)*aa(j,i)
         enddo
      enddo

c --------------------------------------------------------------------
c     INEQUALITY CONSTRAINTS (NON-NEGATIVITY)
c     
c     Note:
c
c     We require that the mole-numbers satisfy
c     
c      x(i) >=epsConc
c
c     where epsConc is a user-specified tolerance-bound.
c --------------------------------------------------------------------

      do i = 1, nSpecies
         g(i) = x(i) - epsConc
      enddo


      return
      end	

