c--------------------------------------------------------------------
c grad.f
c
c Evaluate gradients of function and constraints for the 
c interior-point method applied to the geochemistry problem.
c     
c --------------------------------------------------------------------
c INPUTS:
c
c x(n)                      Concentration unknowns organized as mole-
c                               fractions, followed by total moles.
c
c aa(nComp,nSpecies)        Formula matrix for the system
c b(nComp)                  Component totals
c mu0(nSpecies)             Reference potentials
c phaseIdentity(nSpecies)   Integer array of phase-identity for 
c                               all species
c
c nPhase(pi)                Number of species in each phase
c
c
c OUTPUTS:
c
c gradf(n)                  Gradient of Gibbs Free energy
c gradh(n,me)               Gradient of Equality constraints
c gradg(n,mi)               Gradient of Inequality constraints
c     
c
c EXTERNAL ROUTINES:
c
c chempot                   Evaluate chemical potential
c
c --------------------------------------------------------------------

      subroutine grad(gradf,gradh,gradg,x,nComp,nSpecies,
     $     pi,nPhase,phaseIdentity,aa,mu0,nTot)

      IMPLICIT NONE

      integer n,me,mi,neq,niparam,nrparam
      common /nips_sizes/n,me,mi,neq,niparam,nrparam 
          
      integer i,j,nComp,nSpecies,pi,nPhase(pi),phaseIdentity(nSpecies)
     

      real*8 gradf(n),gradh(n,me),gradg(n,mi),x(n),aa(nComp,nSpecies),
     $     mu0(nSpecies),chempot,nTot(pi),zero,one

      parameter(zero = 0.0d0,one = 1.0d0)

c
c     Compute total moles for multi-species phases
c
      call phaseMoles(nTot,x,pi,nSpecies,phaseIdentity)

c --------------------------------------------------------------------
c     GRADIENT OF THE GIBBS FUNCTION
c --------------------------------------------------------------------

      do i = 1,nSpecies
         gradf(i) = chempot(mu0,x,nTot,i,phaseIdentity,
     $        nPhase,nSpecies,pi)
      enddo

c --------------------------------------------------------------------
c     GRADIENT OF EQUALITY CONSTRAINTS
c --------------------------------------------------------------------

      do i = 1,nSpecies
         do j = 1,nComp
            gradh(i,j) = aa(j,i)
         enddo
      enddo

c --------------------------------------------------------------------
c     GRADIENT OF INEQUALITY CONSTRAINTS
c --------------------------------------------------------------------

      do i = 1,n
         do j = 1,mi
            gradg(i,j) = zero
         enddo
      enddo

      do i = 1,nSpecies
         gradg(i,i) = one
      enddo


      return
      end	
