/*   Load-balancing algorithm for chemistry calculations. 

     Beta version Jan 98  Joe Eaton
     refinements and debugs, May 98 Steven Bryant and Joe Eaton
     
     Idea is to minimize communication by only exchanging cells with 
     a large work load associated. The current work load is estimated
     by assuming it is similar to the previous time step's load, and 
     an appropriate distribution of work is calculated.  
*/

#include <stdio.h>
#include <stdlib.h>
#include "../../Main/msg.h" 
#include "reaction.h"

#define loc_offset_(i,j,k) (i)+ nx*((j) + ny*(k))

#define offset_(i,j,k,specie) \
  i+padx+1 + (nx+2*(padx+1))*(j+pady+1 + \
  (ny+2*(pady+1))*(k+padz+1 + (nz+2* (padz+1))*(specie)))

#define load_(i,p) load[(i-1) + 2*(p)]

/* Implement parallel load balancing using Heavy/light split algorithm
   Calculate load average, and fill load array with over/under distance
   and total number of cells over. 

   float locIter  -> input work loads from previous time step
   int  send_list -> output list of processors and cells to send away
   int  recv_list -> output list of processors and cells to recv from
   int  cell_list -> output list of local cells current proc works on,
                     and if this proc is overworked then the last _myh_
                     cells are the cells to send away. 
                     cell_list[n]= offset into locIter array
*/

int loads(int *nmsg)
{ 

float mywork, avgwork, avgcell;
float heavy_limit;
int cells,local_cells;  
int totwork;
int i,j,k,n,numheavy;
int myn,myh;
int stats[2];
int *load;
int *heavy_list; 
int *heavy_list_from; 
int *displ, *rcounts;

int *dest;
MPI_Request *Req;
MPI_Status *Status;


mywork=0.0;
for(i= 0; i < nx; i++) 
   for(j=0; j< ny; j++)
      for(k=0; k< nz; k++){ 
        mywork+= locIter_(i,j,k);
}

local_cells = nx*ny*nz;
cells = local_cells;
totwork = (int)mywork;
stats[0]= cells;
stats[1]= totwork;
/* gisumX( 145, &stats, 2); */
MSG_Gisum(stats, 2, comm);
cells = stats[0];
totwork = stats[1];

avgwork = (float)totwork/nProcs;
avgcell  = (float)totwork/cells; 

/* if ( myProc == 0 ) printf("%d: Global work=%d   avgwork = %f \n",myProc,totwork,avgwork);  */

/*  300000 is chosen for a 64 node mixed waste contamination example  */

if(totwork < 1000  ){
  for(i =0 ;i < local_cells; i++)
     cell_list[i]=i;
  return 0; 
}
/* || (mywork>avgwork && mywork/avgwork < 1.2)) return 0; */ /* quick return */

/* Store relative work load */
stats[0] = (int)avgwork - mywork;

/*
if(myProc == 0)
  printf("Totwork= %10d Avgwork= %8.3f Mywork= %8.3f Avgcell= %4.2f\n",
	 totwork,avgwork,mywork,avgcell);
*/

/* Split heavy/light */
myh = 0;
myn = local_cells;
heavy_limit = 1.5*avgcell;
if(mywork> avgwork){
  myn=0;
  for(k=0; k< nz; k++)
    for(j=0; j< ny; j++)
      for(i=0; i< nx; i++) {
	if (locIter_(i,j,k) >= heavy_limit) /* heavy cell found */
	  {
	    myh+=1;
	    cell_list[local_cells - myh] = loc_offset_(i,j,k);
	    /*
	    printf("%d: cell %d %d %d is heavy %6.4f\n",
                   myProc, i,j,k,locIter_(i,j,k));
	    */
	  }
	else
	  {
	    cell_list[myn] = loc_offset_(i,j,k); 
	    myn+=1;
	  }
      }
  if(mywork > avgwork){
    int i=0;
    float tempwork = avgwork;
    while(tempwork>0){
      tempwork -= locIter[cell_list[i]];
      i++;
    }
    myn=i;
    myh = local_cells - myn;
  }
}

/*
printf("%d:Heavy: %d  light: %d  total: %d \n",myProc,myh,myn,local_cells);
*/
stats[1] = myh;

load = (int *)malloc(2*nProcs*sizeof(int));

MPI_Allgather(stats, 2, MPI_INT, load, 2, MPI_INT, comm);

/* Now everyone knows the following info:
 1) which processors are overworked, and by how much
 2) the size of the "heavy" array of cells to be redistributed
*/


dest = (int*)malloc(nProcs*sizeof(int));
n=1;
dest[0]=0;
for(i=0;i<nProcs;i++) if(load_(1,i) > 0){dest[0]+=1; dest[n++] = i; }
/*
printf("%d: dest = ",myProc);
for(i=0;i<nProcs;i++) printf(" %d ",dest[i]);
printf("\n");
*/

Req = (MPI_Request *)malloc((nProcs-dest[0])*sizeof(MPI_Request));
Status = (MPI_Status *)malloc((nProcs-dest[0])*sizeof(MPI_Status));

numheavy=0;
for(i=0;i<nProcs;i++) numheavy += load_(2,i);

/*
printf("%d:Global #heavy cells= %d \n",myProc,numheavy);
*/

heavy_list_from =(int*)malloc(numheavy*sizeof(int));
heavy_list =(int*)malloc(numheavy*sizeof(int));

for(i=0; i < myh; i++) {
  heavy_list_from[i]= locIter[cell_list[local_cells-1-i]];
}
    /*  printf("%d:Local heavy list= ",myProc); 
        for(i=0; i < myh ; i++) printf("%d ",heavy_list_from[i]);
        printf("\n");      */

for(i=0; i < myh; i++) 
     cell_list[local_cells-i] = cell_list[local_cells-1-i];
cell_list[myn] = -1;

/*
printf("cell_list:"); for(i=0;i<local_cells+1;i++) 
printf("%d ",cell_list[i]); printf("\n");
*/
displ = (int *)malloc(nProcs*sizeof(int));
rcounts = (int *)malloc(nProcs*sizeof(int));
k = 0;
for(i=0; i<nProcs; ++i) { 
   rcounts[i] = load_(2,i);
   displ[i] = k;
   k+= load_(2,i);
}

/*  printf("rcts:"); for(i=0;i<nProcs;i++) printf("%d ",rcounts[i]);printf("\n");
    printf("dipl:"); for(i=0;i<nProcs;i++) printf("%d ",displ[i]);printf("\n");    */

MPI_Allgatherv(heavy_list_from, myh, MPI_INT, 
               heavy_list, rcounts, displ, MPI_INT, comm);

/*
 Now we also know exactly which cells are heavy and in what order. 
 all that is left is to assign cells from heavy to light processors and
 calculate the sends and receives needed to move the data.
*/

/*
printf("%d:heavy list= ",myProc); 
for(i=0; i < numheavy ; i++) printf("%d ",heavy_list[i]);
printf("\n");
*/

if( mywork > avgwork) { /* I am overworked */
  /*
    run a prefix operation over available processor to determine
    first one with available space
  */
  int nextp=0;
  int ndest=0;
  int sourcecell=0;
  int p = 0;

/*     printf("%d:heavy list= ",myProc); 
       for(i=0; i < numheavy ; i++) printf("%d ",heavy_list[i]);
       printf("\n");      */

  for(p=0; p< myProc; p++){
    if(load_(1,p) < 0 && load_(2,p) > 0){  /* if this proc is heavy */
      while( load_(1,p) < 0 && load_(2,p) > 0){ /* unload */
        while(load_(1,nextp)>0 && nextp<nProcs) {/* while this proc is light */
	   load_(1,nextp) -= heavy_list[sourcecell];
           load_(1,p) += heavy_list[sourcecell];
              sourcecell += 1;
              load_(2,p) -= 1; 
              if( load_(2,p) <=0 ) break;
	}
        while(load_(1,nextp) <=0) nextp += 1;
      }
    }
  }
  
  for(p=0;p< 2*nProcs; p++) send_list[p]=0;

  if(nextp == myProc) nextp +=1;
  /*
  distribute my excess work to processors that are not overworked
  */
  while( load_(1,myProc) < 0 && myh > 0) { /* until I am balanced */
    while(load_(1,nextp)>0 && nextp<nProcs) {/* if this proc has room left */
          send_list_(0,ndest) = nextp;
          send_list_(1,ndest) += 1; 
	  /*
	    printf("%d: Sending c[%d]%d to %d\n",
	           myProc, sourcecell, heavy_list[sourcecell], nextp); 
	  */
          load_(1,nextp) -= heavy_list[sourcecell];
          load_(1,myProc) += heavy_list[sourcecell];
          sourcecell += 1;
          myh -= 1;
          if(sourcecell >= numheavy) break;
          if(myh <=  0) break;
    } 
    if( send_list_(1,ndest) > 0){
     ndest +=1;
     *nmsg -=1;
    }
    nextp += 1;
  } 

 if (abs(*nmsg) > 0 )  {
   printf("%d:Send_list:",myProc);
   for(i=0; i< abs(*nmsg); i++){
    printf("%d %d ",send_list_(0,i),send_list_(1,i));
   }
   printf("\n");
 }

/* Send the send list to all receiving procs */
for(i=0;i<dest[0];i++){ 
  MPI_Send(send_list,  2*nProcs,  MPI_INT, dest[i+1], dest[i+1], comm );  
}

return 0;

}
else /* I need to find out who is sending me work and how much. */
{
  int nlist,npend,this=0,p;
  int *list_buffer;
#define list_buf_(i,j,k) list_buffer[2*nProcs*i + j + 2*k]
   nlist  = nProcs - dest[0];
   list_buffer = (int *)malloc(2*nProcs*nlist*sizeof(int));
   
   for(p=0;p< 2*nProcs; p++) recv_list[p]=0;

   for(i=0;i<nlist; i++){
      MPI_Irecv(&list_buf_(i,0,0), 2*nProcs, MPI_INT, MPI_ANY_SOURCE,myProc,
                comm, &Req[i]);
   }
   npend= nlist;
   i =0;
   while( npend >0){
      MPI_Waitany(nlist, Req, &this, Status);
      k=0;
      while( list_buf_(this,1,k) !=0 ){
	if(list_buf_(this,0,k) == myProc) {
            recv_list_(0,i) = Status[0].MPI_SOURCE;
            recv_list_(1,i) = list_buf_(this,1,k);
            *nmsg += 1;
            i++;
	}
        k++;
      }
      npend--;
   }

   free(list_buffer);

if ( abs(*nmsg) > 0 )  {	
   printf("%d:Recv_list:",myProc);
   for(i=0; i< abs(*nmsg); i++)
     printf("%d %d ",recv_list_(0,i),recv_list_(1,i));
   printf("\n");
   }
}

free(displ);
free(rcounts);
free(heavy_list);
free(load);
free(dest);
free(Req);
free(Status);
return 0;
}
