c-----------------------------------------------------------------------
c getRxnRates.f
c
c Evaluate the reaction-rates corresponding to kinetic species.
c Mass-action type and Monod type expressions are currently implemented.
c
c-----------------------------------------------------------------------
cc
c Revised 8 jan 97 by Steven Bryant
c
c Purpose of Revision: incorporate Monod style kinetics
c	Monod kinetic expressions are of the form
c
c		d[X]/dt = k [X] [S]/(K + [S]) 
c	where
c	X is typically a microbe, S is a substrate (e.g. a hydrocarbon)
c	k is a rate constant and K is the so-called half saturation constant
c
c Revisions made
c     array of half saturation constants passed into routine and declared
c     local variables conc_to_use and Monod_term declared
c     loop to compute Monod-style terms in rate expression installed
c
c-----------------------------------------------------------------------
c INPUTS:  
c 
c nvec(nSpecies)        concentration vector at which rates are evaluated
c V(nSpecies,nR)        stoichiometric matrix
c
c reactionType(nR)      Current classification of reactions 
c                        (0=equilibrium, 1=kinetic, 2= Monod type expressions)
c     
c k(nR,2)               array of rate-constants:
c                        k(i,1) = forward-rate of ith reaction
c                        k(i,2) = backward-rate of ith reaction **NOT USED**
c
c beta(nComp,nR)        exponents to the components used in rate-laws
c
c SLB 8 Jan 97
c HalfSatConst(nComp,nR)  half saturation constants for the components which
c                         appear in the Monod denominators.  NB that a zero
c                         entry is taken to mean that that component does not
c                         participate in the rate expression.
c SLB end
c
c epsConc               lower-bound for mole-numbers
c     
c OUTPUT:
c
c rate(nR)              resulting rate vector
c                       (equilibrium rates are padded with zeros)
c
c
c ALGORITHM:
c
c We use as the main composition variable the molar 
c concentration of species,
c
c c_i = n_i/volWater
c
c where volWater denotes the volume of water. 
c
c (Recall that we have normalized the system to UNIT 
c  aqueous phase volume!)
c
c A typical rate-expression is the mass-action law (reactionType = 1): 
c
c r_i = Kfwd_i*[c_1^(beta(1,i)*....*c_nComp^(beta(nComp,i))]  - Kbwd*c_(nComp+i)
c For mineral products, it is instead
c
c r_i = Kfwd_i*[c_1^(beta(1,i)*....*c_nComp^(beta(nComp,i))]  - Kbwd,
c
c i.e., the mineral itself does not directly affect the rate.
c
c
c SLB 8 Jan 97
c
c A generalised Monod-style rate expression (reactionType=2) is also allowed: 
c
c r_i = Kfwd_i*       c_1    *       c_2    *...*         c_nComp
c              -------------------------------------------------------- 
c              [K(1,i) + c_1]*[K(2,i) + c_2]*...*[K(nComp,i) + c_nComp]
c
c             * [c_1^(beta(1,i)*....*c_nComp^(beta(nComp,i))]
c             
c SLB end
c
c Note that we have made these rates continous (cf. KEMOD!).
c
c-----------------------------------------------------------------------
c     
      subroutine getRxnRates(rate,nvec,V,reactionType,phaseIdentity,
     $     nPhase,k,beta,K_halfsat,epsConc)

      IMPLICIT NONE
      INTRINSIC log, exp

      integer nSpecies,nComp,nR
      common /chemSizes/nSpecies,nComp,nR

      real*8 rate(nR),nvec(nSpecies),V(nSpecies,nR),beta(nComp,nR),
     $     k(nR,2),epsConc,zero,psum

c
c SLB 8 Jan 97
      real*8 K_halfsat(nComp,nR)
      real*8 conc_to_use , Monod_term
c SLB end
c
      integer reactionType(nR),phaseIdentity(nSpecies),nPhase(*),
     $     prodInd,i,j
 
      parameter(zero = 0.0d0)

      do i = 1,nR

         if (reactionType(i).eq.0) then

c -------------------------------------------------------------------     
c     EQUILIBRIUM REACTION - SET TO ZERO FOR CONSISTENCY
c -------------------------------------------------------------------     

            rate(i) = zero

         else

c -------------------------------------------------------------------
c     KINETIC REACTION (SAFEGUARD AGAINST NUMERICAL ZERO)
c -------------------------------------------------------------------

            psum = zero
            do j = 1,nComp
               psum = psum + beta(j,i)*log( dmax1(nvec(j),epsConc) )
            enddo
c
c SLB 8 Jan 97
c               The preceding loop applies both to mass-action (type=1)
c               and to Monod-style (type=2) kinetics.
c               Loop again over components if Monod-style kinetics
c            
            if (reactionType(i) .eq. 2) then
              do j = 1, nComp
                if (K_halfsat(j,i) .gt. zero ) then
                   conc_to_use = dmax1(nvec(j),epsConc) 
                   Monod_term = conc_to_use / 
     $                         (K_halfsat(j,i) + conc_to_use)
                   psum = psum + log(Monod_term)
                endif
              enddo
            endif
c SLB end
c

            prodInd = nComp + i

            if (nPhase(phaseIdentity(prodInd)).gt.1) then

c     REGULAR MASS-ACTION or MONOD-STYLE TYPE

               rate(i) = k(i,1)*exp(psum) - 
     $              k(i,2)*dmax1(nvec(prodInd),epsConc)

            else

c     MINERAL TYPE 

               rate(i) =  k(i,1)*exp(psum) - 
     $              k(i,2)

            endif

         endif

      enddo


      return
      end

