c-----------------------------------------------------------------------
c maxTimeStep.f
c
c Find the largest time-step that preserves lower bound on moles.
c --------------------------------------------------------------------
c
c Revised 8 Jan 97 by Steven Bryant
c
c Purpose of revision: incorporate Monod style kinetics 
c
c Revisions made
c     allow for new reaction classification, reactionType=2
c --------------------------------------------------------------------
c
c INPUTS: 
c
c rate(nR)                    vector of reaction rates
c nvec(nSpecies)              vector of molenumbers
c V(nSpecies,nR)              stoichiometric matrix 
c reactionType(nR)            Current classification of reactions 
c                             (0=equilibrium, 1=kinetic, 2=Monod kinetics)
c epsConc                     Lowest bound for any concentration
c     
c OUTPUT:
c
c maxTimeStep
c
c ALGORITHM:
c
c maxTimeStep is the largest gamma > 0 for which
c
c nvec + gamma*V*rate > epsConc,
c
c if only the kinetic rates are taken into account.
c
c-----------------------------------------------------------------------
c     
      function maxTimeStep(rate,nvec,V,reactionType,epsConc)

      IMPLICIT NONE

      integer nSpecies,nComp,nR
      common /chemSizes/nSpecies,nComp,nR

      integer i,j,reactionType(nR),prodInd
      real*8 maxTimeStep,rate(nR),nvec(nSpecies),V(nSpecies,nR),
     $     epsConc,distToBdry,big,zero,one,eps,dni,gamma


      parameter(zero = 0.0d0,one = 1.0d0,
     $     eps = 1.0d-4,big = 1.0d20)
      
      gamma = big

c-----------------------------------------------------------------------
c     PERFORM A RATE-LIMITING STEP BEFORE COMPUTING MAXTIMESTEP.
c     THE PURPOSE IS TO ENSURE THAT WE WILL BE ABLE TO FIND 
c     SOME (STRICTLY) POSITIVE TIME-STEP IN THE SECTIONS BELOW.
c-----------------------------------------------------------------------

      call ratelim(rate,nvec,V,reactionType,epsConc)

c-----------------------------------------------------------------------
c     COMPONENTS: CHECK ALL REACTIONS AND SAFEGUARD
c-----------------------------------------------------------------------

      do i = 1,nComp
         dni = zero
         do j = 1,nR
cccc            if (reactionType(j).eq.1) dni = dni + V(i,j)*rate(j)
c SLB 8 Jan 97           
            if (reactionType(j).eq.1.or.reactionType(j).eq.2) 
     $                                    dni = dni + V(i,j)*rate(j)
c SLB end
c
         enddo

         if (dni.lt.zero)  then
            distToBdry = nvec(i) - epsConc
c
c...always have distToBdry > 0 if dni < 0 due to ratelim.
c
            gamma = dmin1(gamma,-distToBdry/dni)
         endif
      enddo

c-----------------------------------------------------------------------
c     PRODUCT SPECIES: EACH IS AFFECTED BY ONE REACTION ONLY
c-----------------------------------------------------------------------

      do i = 1, nR
cccc         if (reactionType(i).eq.1) then
c SLB 8 Jan 97
          if (reactionType(i).eq.1.or.reactionType(i).eq.2) then
c SLB end
c
            prodInd = nComp + i
            dni = rate(i)
            if (dni.lt.zero)  then
               distToBdry = nvec(prodInd) - epsConc
c
c...always have distToBdry > 0 if dni < 0 due to ratelim.
c
               gamma = dmin1(gamma,-distToBdry/dni)
            endif
         endif
      enddo


c-----------------------------------------------------------------------
c     CONSERVATIVE TIME-STEP (SLIGHTLY SMALLER)
c-----------------------------------------------------------------------

      maxTimeStep = (one-eps)*gamma


      return
      end

