c ----------------------------------------------------------------------    
c odeDriver.f
c
c Advance the solution to the equilibrium/kinetic system from time 
c t = t0 to t = t0 + dtTrans.
c 
c ----------------------------------------------------------------------    
c
c Revised 8 jan 97 by Steven Bryant
c
c Purpose of Revision: incorporate Monod style kinetics
c	Monod kinetic expressions are of the form
c
c		d[X]/dt = k [X] [S]/(K + [S]) 
c	where
c	X is typically a microbe, S is a substrate (e.g. a hydrocarbon)
c	k is a rate constant and K is the so-called half saturation constant
c
c Revisions made
c     array of half saturation constants passed into routine, declared
c     and passed to routine getRxnRates
c
c-----------------------------------------------------------------------
c INPUTS:    
c
c dtTrans                        the transport time-step
c t0                             beginning of step
c moles(nSpecies)                vector of moles to be reacted
c epsConc                        lower bound for mole-numbers
c piMulti                        number of multi-species phases
c piPure                         number of single-species phases
c nPhase(piMulti+piPure)         number of species within each phase
c phaseIdentity(nSpecies)        the phase identity of each species
c reactionType(nR)               reaction classification
c formulaMatrix(nComp,nSpecies)  formula-matrix
c V(nSpecies,nR)                 stoichiometric matrix
c mu(nSpecies)                   chemical reference potentials
c rateconst(nR,2)                forward and backward rate-constants
c ratepow(nComp,nR)              exponents on componets for rate-laws
c SLB 8 Jan 97
c HalfSatConst(nComp,nR)         constants in Monod rate-law denominators
c SLB end
c idebug                         Control level of output from odeDriver
c                                (0 = off, 1 = 0n)
c idebugEq                       Control level of output from eqStep
c                                (0 = off, 1 = On, 2 = Really on)
c
c OUTPUT:
c 
c newmoles(nSpecies)            vector of reacted molenumbers
c                               at the advanced time-level (t0+dtCum)
c
c ntTaken                       number of time-steps taken.
c
c peakIter                      max number of iterations taken
c                               in any equilibration.
c
c cumIter                       total number of iterations for 
c                               all equilibrations over the 
c                               entire integration.
c WORK SPACE:
c
c rates(nR,4)                   rate computations
c refRate(nR)                   used if switching is enabled
c avgRate(nR)                   used to hold the weighted rate 
c
c-----------------------------------------------------------------------
c GENERAL
c
c The mixture is reacted from t = tc to t = tc + dtTrans.
c The integration-strategy is determined by the user-supplied
c parameter 'odeAlgType'. The following choices are implemented:
c
c     odeAlgType  = 0        Forward Euler
c     odeAlgType  = 1        Explicit second-order Runge-Kutta
c     odeAlgType  = 2        Explicit fourth-order Runge-Kutta
c-----------------------------------------------------------------------
c
c First, a reaction time-step is defined based on user-specified
c parameter ntReact
c
c dtRef = dtTrans/ntReact
c
c This is the target time-step to be used.
c However, it may have to be reduced further due to positivity-violations.
c
c Let m denote an iteration counter within the time-stepping.
c Designate equilibration by the symbol 'E'. Furthermore, let 
c n^[m], n^[m,*] denote unequilibrated and equilibrated mixtures
c at the same iteration-level (i.e. n^[m,*]  <---- E(n^[m]))
c
c ALGORITHM:
c
c Receive unreacted composition n(t^n) 
c Set n^[0] =  n(t^n)
c
c  (0) TIME-LOOP (for m = 0,1,...)
c
c       (a) Equilibrate:
c           n^[m,*] <--- E(n^[m])
c
c  (1.a) Compute rates r_0( n^[m,*])
c
c  (1.b) Compute largest acceptable time-step based on r_0
c
c  (2) LOOP TO COMPUTE TANGENTS: l = 2,nlevels (RK2 and RK4 only)
c
c    (2.a) Predict
c          n_(l)^[m]  <--- n^[m,*] + dt*V*r_(l-1)
c
c    (2.b) Equilibrate predicted mixture
c          n_(l)^[m,*] <--- E(n_(l)^[m])
c
c    (2.c) Compute new rate
c          r_l  = r(n_(l)^[m,*])
c
c    (2.d) Check for positivity-violations with current time-step.
c          If a violation occurs, compute new dt and GOTO 2
c
c   end loop
c
c
c  (3) COMPUTE FINAL UPDATE
c
c  (3.a) Compute the average rates based on result of (2)
c        Note: average rate DOES NOT generate a positivity-
c        violation, becauase of the conditions imposed on all
c        the rates from which the average is formed.
c
c  (3.b) Update using prescription for FE,RK2 or RK4:
c
c         n_c^[m]  <--- n^[m,*] + F(r_1,...,r_4)
c
c  (3.c) Equilibration of the corrected mixture:
c  
c        n_c^[m,*]  <--- E(n_c^[m])
c 
c
c  (4) UPDATE
c
c       n^[m+1] <--- n_c^[m,*]
c       m       <--- m + 1
c       dtCum = dtCum + dt
c
c       Check if done
c
c  (5) SPECIES-SWITCHING STEP (only if 'switchEnabled' is set)
c
c      check if reactions can be switched from kinetic to equilibrium
c      description, based on heuristic rules.
c
c  (6) GOTO 0.a OR 1.a depending on STEP 5
c
c ----------------------------------------------------------------------    
c  

c SLB 8 Jan 97  inserted HalfSatConst array into call
c SLB 14/10/97  receive initial guess for equilbration

      subroutine odeDriver(moles,
     $     InitGuess,newmoles,odeAlgType,rates,avgRate,
     $     refRate,epsConc,t0,dtTrans,ntReact,switchEnabled,piMulti,
     $     piPure,nRQ,nRK,nProducts,nPhase,nS,nC,phaseIdentity,
     $     reactionType,reactionTypeSave,nipsIparm,formulaMatrix,V,
     $     mu,rateconst,ratepow,HalfSatConst,compCharge,nipsRparm,
     $     rwk,dimrwk,iwk,dimiwk,idebugEq,idebug,debugunit,
     $     ntTaken,peakIter,cumIter,ierror,nipsCparm)

      IMPLICIT NONE

      integer nProducts, nC, nS
      integer nSpecies,nComp,nR
      common /chemSizes/nSpecies,nComp,nR

c     integer i,l,nTangents,odeAlgType,piMulti,piPure,pi,nRQ,nRK,
c    $     nPhase(piMulti+piPure),phaseIdentity(nSpecies),
c    $     reactionType(nR),reactionTypeSave(nR),idebug,idebugEq,
c    $     debugunit,ierror,ntTaken,ntReact,iter,cumIter,peakIter,
c    $     nipsIparm(*),dimiwk,dimrwk,iwk(dimiwk),nRKSAVE,nRQSAVE
c
c     real*8 moles(nSpecies),newmoles(nSpecies),epsConc,V(nSpecies,nR),
c    $     t0,dtTrans,dtRef,dtCum,dtCalc,dtTol,dt,maxTimeStep,
c    $     formulaMatrix(nComp,nSpecies),mu(nSpecies),rates(nR,4),
c    $     avgRate(nR),refRate(nR),rateconst(2*nR),ratepow(nComp*nR),
c    $     compCharge(nComp),nipsRparm(*),wcoef(1:4),rwk(dimrwk)
c
c SLB 8 Jan 97
c     real*8 HalfSatConst(nComp*nR)
c SLB end
c
c SLB 14/10/97 declare initial guess
c     real*8 InitGuess(nSpecies)


ccc SLB 3 Feb 98  Declare arrays only with parameters passed in the
ccc               call to the routine.  This eliminates problems
ccc               with zero-dimension arrays on the first call to the
ccc               routine, at which point the variables nR, nComp and
ccc               nSpecies are zero.

      integer i,l,nTangents,odeAlgType,piMulti,piPure,pi,nRQ,nRK,
     $     nPhase(piMulti+piPure),phaseIdentity(nS),
     $     reactionType(nProducts),reactionTypeSave(nProducts),
     $     idebug,idebugEq,
     $     debugunit,ierror,ntTaken,ntReact,iter,cumIter,peakIter,
     $     nipsIparm(*),dimiwk,dimrwk,iwk(dimiwk),nRKSAVE,nRQSAVE

      real*8 moles(nS),newmoles(nS),epsConc,V(nS,nProducts),
     $     t0,dtTrans,dtRef,dtCum,dtCalc,dtTol,dt,maxTimeStep,
     $     formulaMatrix(nC,nS),mu(nS),rates(nProducts,4),
     $     avgRate(nProducts),refRate(nProducts),
     $     rateconst(2*nProducts),ratepow(nC*nProducts),
     $     compCharge(nC),nipsRparm(*),wcoef(1:4),rwk(dimrwk)
      real*8 HalfSatConst(nC*nProducts)
      real*8 InitGuess(nS)

      logical switchEnabled,switchOccurred,ALL_KINETIC

      character*15 nipsCparm(*)

      parameter(dtTol = 1.0d-4)

c-----------------------------------------------------------------------
c SET-UP
c-----------------------------------------------------------------------

      nSpecies = nS
      nComp = nC
      nR = nProducts
      dtRef    = dtTrans/dble(ntReact)
      dtCum    = 0.0d0
      ntTaken  = 0
      cumIter  = 0
      peakIter = 0
      pi = piMulti + piPure

      if (switchEnabled) then
         nRKSAVE = nRK
         nRQSAVE = nRQ
         do i = 1,nR
            reactionTypeSave(i) = reactionType(i)
         enddo
      endif

      if (odeAlgType.eq.0) then
         nTangents = 1
         wcoef(1) = 0.0d0
      elseif (odeAlgType.eq.1) then 
         nTangents = 2
         wcoef(1) = 0.0d0
         wcoef(2) = 1.0d0
      elseif (odeAlgType.eq.2) then 
         nTangents = 4
         wcoef(1) = 0.0d0
         wcoef(2) = 0.5d0
         wcoef(3) = 0.5d0
         wcoef(4) = 1.0d0
      endif

c-----------------------------------------------------------------------
c     EQUILIBRATE THE INCOMING COMPOSITION 
c-----------------------------------------------------------------------

 1    ALL_KINETIC = (nRK.gt.0).and.(nRQ.eq.0)

      if (.not. ALL_KINETIC) then
         call eqStep(moles,InitGuess,epsConc,piMulti,piPure,
     $        nRQ,nProducts,nPhase,nSpecies,nComp,
     $        phaseIdentity,reactionType,nipsIparm,formulaMatrix,V,mu,
     $        compCharge,nipsRparm,rwk(1),dimrwk,
     $        iwk(1),dimiwk,idebugEq,debugunit,iter,ierror,nipsCparm)

         if (ierror.ne.0) GOTO 999
         if (idebug.ne.0) then
            write(debugunit,100)iter
            call debugOutput('Initial EQ moles:',moles,
     $           nipsCparm(pi+1),nSpecies,debugunit,1)
         endif
         cumIter = cumIter + iter
         peakIter = max(iter,peakIter)
      endif

c-----------------------------------------------------------------------
c     COMPUTE INITIAL REACTION-RATES FOR THE PURPOSE 
c     OF POSSIBLE SPECIES-SWITCHING LATER.
c-----------------------------------------------------------------------

      if (switchEnabled) then
c
c SLB 8 Jan 97   insert half satn constants in the call
c
      call getRxnRates(refRate,moles,V,reactionType,phaseIdentity,
     $        nPhase,rateconst,ratepow,HalfSatConst,epsConc)
      endif

c-----------------------------------------------------------------------
c     TOP OF TIME-STEPPING
c-----------------------------------------------------------------------

 10   if (idebug.ne.0) then
         write(debugunit,200) dtCum
         do i = 1,nR
            write(debugunit,210)i,reactionType(i)
         enddo
      endif

c-----------------------------------------------------------------------
c     COMPUTE RATES AT:  n^[m,*]
c     AND ESTABLISH A LARGEST TIME-STEP BASED ON THIS RATE
c-----------------------------------------------------------------------

c
c SLB 8 Jan 97   insert half satn constants in the call
c
      call getRxnRates(rates(1,1),moles,V,reactionType,
     $     phaseIdentity,nPhase,rateconst,ratepow,HalfSatConst,
     $     epsConc)

      if (idebug.ne.0) call debugOutput('Rates',rates(1,1),
     $     nipsCparm(pi+nComp+1), nR,debugunit,2)

      dtCalc = maxTimeStep(rates(1,1),moles,V,reactionType,epsConc)

      dt = dmin1(dtCalc, dmin1(dtRef,dtTrans-dtCum))

 15   if (idebug.ne.0) write(debugunit,250) dt

c-----------------------------------------------------------------------
c     LOOP TO COMPUTE REMAINING TANGENTS (RK2 AND RK4 ONLY)
c-----------------------------------------------------------------------

      do l = 2,nTangents
         
c-----------------------------------------------------------------------
c     PREDICT NEW POINT:  n_l^[m]  <--- n^[m,*] + wcoef(l)*dt*V*r_l
c-----------------------------------------------------------------------

         call applyRates(moles,newmoles,rates(1,l-1),V,
     $        reactionType,wcoef(l)*dt)

         if (idebug.ne.0) call debugOutput('Predict',newmoles,
     $        nipsCparm(pi+1),nSpecies,debugunit,1)

c-----------------------------------------------------------------------
c     EQUILIBRATE PREDICTED POINT:   n_l^[m,*] <--- E(n_l^[m])
c-----------------------------------------------------------------------

         if (.not.ALL_KINETIC) then
         call eqStep(newmoles,InitGuess,epsConc,piMulti,piPure,
     $        nRQ,nProducts,nPhase,nSpecies,nComp,
     $        phaseIdentity,reactionType,nipsIparm,formulaMatrix,V,mu,
     $        compCharge,nipsRparm,rwk(1),dimrwk,
     $        iwk(1),dimiwk,idebugEq,debugunit,iter,ierror,nipsCparm)

            if (ierror.ne.0) GOTO 999
            if (idebug.ne.0) then
               write(debugunit,100)iter
               call debugOutput('Equilibrated predicted moles:',
     $              newmoles,nipsCparm(pi+1),nSpecies,debugunit,1)
            endif
            cumIter = cumIter + iter
            peakIter = max(iter,peakIter)
         endif

c-----------------------------------------------------------------------
c     COMPUTE NEW RATES :   r(n_l^[m,*])
c-----------------------------------------------------------------------
C
c SLB 8 Jan 97   insert half satn constants in the call
c
         call getRxnRates(rates(1,l),newmoles,V,reactionType,
     $        phaseIdentity,nPhase,rateconst,ratepow,HalfSatConst,
     $        epsConc)

         if (idebug.ne.0) call debugOutput('Rates',rates(1,l),
     $        nipsCparm(pi+nComp+1),nR,debugunit,2)


c-----------------------------------------------------------------------
c     CHECK IF NEW RATES WILL PRODUCE A POSITIVITY VIOLATION
c-----------------------------------------------------------------------

         dtCalc = maxTimeStep(rates(1,l),moles,V,reactionType,epsConc)

         if (idebug.ne.0) write(debugunit,255) l,dt

         if (dtCalc.lt.dt) then
            dt = dtCalc
            if (idebug.ne.0) write(debugunit,300)
            GOTO 15
         endif

      enddo

c-----------------------------------------------------------------------
c     COMPUTE AVERAGE RATE FOR THIS TIME-STEP 
c-----------------------------------------------------------------------
      
      call avgRates(rates,avgRate,odeAlgType)

      if (idebug.ne.0) call debugOutput('Average rate:',avgRate,
     $     nipsCparm(pi+nComp+1),nR,debugunit,2)

c-----------------------------------------------------------------------
c     UPDATE USING AVERAGE RATE: 
c-----------------------------------------------------------------------

      call applyRates(moles,newmoles,avgRate,V,reactionType,dt)

      if (idebug.ne.0) call debugOutput('Corrected moles:',
     $     newmoles,nipsCparm(pi+1),nSpecies,debugunit,1)

c-----------------------------------------------------------------------
c     FINAL EQUILIBRATION: 
c-----------------------------------------------------------------------

      if (.not.ALL_KINETIC) then
         call eqStep(newmoles,InitGuess,epsConc,piMulti,piPure,
     $        nRQ,nProducts,nPhase,nSpecies,nComp,
     $        phaseIdentity,reactionType,nipsIparm,formulaMatrix,V,mu,
     $        compCharge,nipsRparm,rwk(1),dimrwk,
     $        iwk(1),dimiwk,idebugEq,debugunit,iter,ierror,nipsCparm)
         if (ierror.ne.0) GOTO 999
         if (idebug.ne.0) then
            write(debugunit,*)'Final equilibration:'
            write(debugunit,100)iter
         endif
         cumIter = cumIter + iter
         peakIter = max(iter,peakIter)
      endif

      if (idebug.ne.0) then
         write(debugunit,150)t0+dtCum+dt
         call debugOutput('Moles',newmoles,
     $        nipsCparm(pi+1),nSpecies,debugunit,1)
      endif

c-----------------------------------------------------------------------
c     END OF TIME-STEP LOOP
c-----------------------------------------------------------------------

      ntTaken = ntTaken + 1
      dtCum = dtCum + dt

      if (dtCum.ge.dtTrans*(1.0-dtTol)) GOTO 999

c-----------------------------------------------------------------------
c     DECIDE ON STRATEGY FOR NEXT TIME-STEP
c-----------------------------------------------------------------------

      if (switchEnabled) call switchModule(newmoles,avgRate,refRate,
     $     reactionType,nRQ,nRK,switchOccurred,idebug,debugunit)

c-----------------------------------------------------------------------
c     UPDATE SOLUTION FOR THIS ITERATION LEVEL
c-----------------------------------------------------------------------

      call dcopy(nSpecies,newmoles,1,moles,1)

      if (switchEnabled) then
         if (nRK.eq.0) GOTO 999
         if (switchOccurred) GOTO 1
      endif

      GOTO 10 

c-----------------------------------------------------------------------
c     EXIT TO CALLING ROUTINE AFTER RE-SETTING ARRAYS
c-----------------------------------------------------------------------

 999  if (switchEnabled) then
         nRK = nRKSAVE
         nRQ = nRQSAVE
         do i = 1,nR
            reactionType(i) = reactionTypeSave(i)
         enddo
      endif
      RETURN


c-----------------------------------------------------------------------
c     FORMAT STATEMENTS
c

 100  format(/,'Nonlinear iterations taken=',1x,i3)
 150  format(1x,'time=', f8.2,/)
 200  format('Top of time-stepping loop',/,'dtCum =',e10.4,/,
     $     'List of reactions for this time-step:',/,
     $     'reaction number',10x,'active?')
 210  format(i2,25x,i1)
 250  format('Time-step chosen at begining of step',1x,'dt=',1x,e10.4)
 255  format('Time-step chosen for level:',1x,i1,1x,'dt=',1x,e10.4)
 260  format('Overall time-step',1x,'dt=',1x,e10.4)
 300  format('A positivity violation occurred')

      END

