c -------------------------------------------------------------------     
c ratelim.f
c 
c Limit the reaction-rates.
c -------------------------------------------------------------------     
c
c Revised 8 Jan 97 by Steven Bryant
c
c Purpose of revision: incorporate Monod style kinetics 
c
c Revisions made
c     allow for new reaction classification, reactionType=2
c --------------------------------------------------------------------
c
c INPUTS:
c 
c nvec(nSpecies)
c
c V(nSpecies,nR)        stoichiometric matrix
c
c reactionType(nR)      Current classification of reactions 
c                        (0=equilibrium, 1=kinetic, 2=Monod kinetics)
c epsConc               Lower bound on mole-numbers
c
c OUTPUT:
c 
c rate(nR)             Limited rates
c
c ------------------------------------------------------------------- 
c    
c     THE RATE-LIMITING STEP
c
c     On physical grounds we "limit" the rate computed above to ensure 
c     that it will not proceed in such a direction as would lead 
c     to the complete depletion of either reactants or products.
c     By "complete depletion", we mean molenumbers that fall below
c     the user-established value for "approximate zero".
c
c     - A reaction can not proceed in the RIGHT (forward) direction
c     if any of its reactants have fallen below the threshold value, 
c     'lower'. Note that only those reactants that have a negative
c     stoichiometric coefficient are depleted in this direction.
c
c     - A reaction can not proceed in the LEFT (reverse) direction
c     if (1) the product species has been consumed, i.e., if the 
c     molenumber of the product has fallen below the smallest
c     allowed value, or (2) if components with positive 
c     stoichiometric coefficient have fallen below the threshold
c     value.
c
c -------------------------------------------------------------------     

      subroutine ratelim(rate,nvec,V,reactionType,epsConc)
      IMPLICIT NONE

      integer nSpecies,nComp,nR
      common /chemSizes/nSpecies,nComp,nR

      integer i,j,prodInd,reactionType(nR)

      real*8  rate(nR),compJ,coeffJ,nvec(nSpecies),V(nSpecies,nR),
     $     epsConc,lower,zero

      logical proceedLeft,proceedRight

      parameter(zero = 0.0d0)

      lower = 2.0*epsConc

      do i = 1, nR
cccc         if (reactionType(i).eq.1) then 
c SLB 8 Jan 97
            if (reactionType(i).eq.1.or.reactionType(i).eq.2) then
c SLB end
            prodInd = nComp + i
            proceedRight = .true.
            proceedLeft  = .true.

c -------------------------------------------------------------------     
c     COMPONENTS
c -------------------------------------------------------------------     

            do j = 1,nComp
               compJ = nvec(j)
               coeffJ = V(j,i)
               if (compJ.le.lower) then 
                  if (coeffJ.lt.zero) proceedRight = .false.
                  if (coeffJ.gt.zero) proceedLeft = .false.
               endif
            enddo

c -------------------------------------------------------------------     
c     PRODUCT SPECIES
c -------------------------------------------------------------------     

            proceedLeft = proceedLeft.and.(nvec(prodInd).gt.lower)

c -------------------------------------------------------------------     
c     LIMIT THE RATE
c -------------------------------------------------------------------     

            if (.not.proceedRight) rate(i) = dmin1( zero, rate(i) )
            if (.not.proceedLeft)  rate(i) = dmax1( zero, rate(i) )

         endif
      enddo



      return
      end
