/*-----------------------------------------------------------------------
  reactOneCell.c
 
  React the mixture in a single cell from t=tc to t=tc+dt.
 
  For an EQUILIBRIUM system,  perform a flash-calculation.
 
  For MIXED kinetic/equilibrium do a series of ODE solves/
  flash-calculations.
 
  For a purely KINETIC case, perform an ODE solve.

  INPUTS:     
 
  dt            Transport time-step employed
 
  INPUT/OUTPUT:
 
  cellmoles         On call: transported, unreacted MOLENUMBERS
                              at tc+dt.
                    On return: reacted MOLENUMBERS at tc+dt.
  cellInitGuesss    On call: initial guess for eq calculation.
                    On return: trashed
 
  OUTPUT:
 
  rwk(dimrwk)       temporary real*8 work-space
  iwk(dimiwk)       temporary integer work-space
 
  Error            Error flag for the reactionStep. 
                    The following values are defined:
                                  
                    Error = 0        Normal return
                    Error = 1        NIPS error; exceeded maximum 
                                       number of iterations
                    Error = 2        NIPS error; exceeded number of 
                                       back-tracks in linesearch.
                    Error = 3        NIPS error; ill-conditioning 
                                       of linear system.
                    Error = 4        NIPS error; infeasible 
                                       initial point
                    Error = 5        Time-integration error 
                                       (NOT IMPLEMENTED!)
 
      
 ----------------------------------------------------------------------- */ 
#include <stdio.h>
#include "reaction.h"

int reactOneCell( double dt, 
                  double* cellmoles, double* cellInitGuess,
                  int* Iter, int* ntTaken,
                  int debugEq, int debugOde,
                  int kinetic_mode, int equilibrium_mode,
                  int initial_conditions)
          
{ 

int  species;
int  Error=0;
int  peakIter;
double lower ;

lower = 1.1 * chmEpsConc ;

  /* impose concentration threshold on the cellmoles */

   for (species = 0 ; species < nSpecies ; species++) {
       cellmoles[species] = max_(cellmoles[species], lower) ; }

  /* impose concentration threshold on the initial guess */

   for (species = 0 ; species < nSpecies ; species++) {
       cellInitGuess[species] = max_(cellInitGuess[species], lower) ; }

  /*
               if (idebug.ge.1) then
                  if (initial_conditions) then
                     write(debugunit,1050) i,j,k,xc(i),yc(j),zc(k)
                  else
                     write(debugunit,1000) i,j,k,xc(i),yc(j),zc(k),time
                  endif
                  call debugOutput('Initial molenumbers',
     $                 rwk(rptr_moles),Cparm(pi+1),nSpecies,debugunit,1)
               endif

  */

*ntTaken =0;
*Iter = 0;

if(debugEq >=2 || debugOde >0 ){
  printf("Initial Molenumbers:\n");
  printf("-------------------------\n");
for(species =0; species < nSpecies; species++){
   printf("%16.12g  \t %s \n",cellmoles[species],specieName[species]);
}
printf("\n");
}

  if (equilibrium_mode||initial_conditions) {

  /*c-----------------------------------------------------------------------
      PERFORM THE EQUILIBRIUM CALCULATION
      Note: this is only done if (A) the system is of purely
      equilibrium type, in which case no time-integration is 
      needed, or (B) if this is the equilibration of the initial
      conditions, in which case kinetic product species 
      are held constant.
  -----------------------------------------------------------------------*/
    EqStep(cellmoles, cellInitGuess, &chmEpsConc,&chemPhaseMulti,
	   &chemPhasePure, &nRQ, &nProducts, nSpeciesInPhase, 
	   &nSpecies, &nComps, chmPhaseIdentity,
	   reactionType, iRxnpar, chemAmatrix, chemVmatrix, mu0,
	   compCharge, rRxnpar, dwk, &dimDwk,
	   iwk,&dimIwk,&debugEq,&debugUnit,Iter,&Error, cRxnpar);

  } else {
    
/*-----------------------------------------------------------------------
      SOLVE A DIFFERENTIAL OR MIXED ALGEBRAIC-DIFFERENTIAL SYSTEM
 -----------------------------------------------------------------------*/

/*c SLB 8 Jan 97
      Send the half saturation constants after the rate powers and before the 
      species charges
 
  SLB 14/10/97 send initial guess to odeDriver thence to eqStep 
  SLB 04/03/98 OdeDriver receives the current concentrations, i.e.
               the unreacted composition, in the first array passed; it
               receives an initial guess for the reacted composition in 
               the second array passed (this facilitates equilibration).
               OdeDriver writes the new concentrations, i.e. 
               reacted and equilibrated, to the array it receives in the
               third parameter passed.  
               --NB--  Within OdeDriver this third array gets overwritten
               during evaluation of the kinetic reactions.  By passing
               cellInitGuess as the third array, THE INITIAL GUESS USED
               IN CALLS TO EQSTEP FROM WITHIN ODEDRIVER GETS ALTERED.
               This does not appear to be a bad thing, actually; in effect
               it automatically updates the initial guess.  But it WILL
               cause slightly different behaviour from the earlier versions
               of the code, wherein a different location in memory was 
               passed as the third array.               */

    OdeDriver(cellmoles, cellInitGuess, cellInitGuess,
	      &odeAlgType, rates, avgrate, refrate, 
	      &chmEpsConc, &time_t, &dt, &ntReact,
	      &switchFlag, &chemPhaseMulti,&chemPhasePure,
	      &nRQ,&nRK,&nProducts, nSpeciesInPhase, 
	      &nSpecies, &nComps, chmPhaseIdentity,
	      reactionType,iwk,
	      iRxnpar, chemAmatrix, chemVmatrix, mu0,
	      kval, ratepow , halfSatConst, compCharge , rRxnpar,
	      dwk, &dimDwk,iwk,&dimIwk,
	      &debugEq,&debugOde,&debugUnit,
	      ntTaken,&peakIter,Iter,&Error,cRxnpar);

/*    STORE SOLUTION */
    for (species =0; species < nSpecies; species++){
      cellmoles[species] = cellInitGuess[species];
    }
  }

  return Error;
}

