/*-----------------------------------------------------------------------
  reactionStep.c
 
  React the mixture in a list of gridblocks from t=tc to t=tc+deltat.
 
  For an EQUILIBRIUM system,  perform a flash-calculation.
 
  For MIXED kinetic/equilibrium do a series of ODE solves/
  flash-calculations.
 
  For a purely KINETIC case, perform an ODE solve.

 -----------------------------------------------------------------------
  Translated to C Jan 98 by Joe Eaton 

   Added dynamic memory allocation, load balancing and communication
   procedures to reactionStep algorithm. Simplified calling sequence,
   allow for more general ODE packages. 

 -----------------------------------------------------------------------
 
  Revised 1 Dec 97 by Steven Bryant
 
  Purpose of revision: incorporate measure of computational work in the 
                       chemistry routines, as per Joe Eaton developmental work
 
 ------------------------------------------------------------------------
 
  Revised 8 jan 97 by Steven Bryant
 
  Purpose of Revision: incorporate Monod style kinetics
 	Monod kinetic expressions are of the form
 
 		d[X]/dt = k [X] [S]/(K + [S]) 
 	where
 	X is typically a microbe, S is a substrate (e.g. a hydrocarbon)
 	k is a rate constant and K is the so-called half saturation constant
 
  Revisions made
      declare and define new pointer to the HalfSaturationConstants 
      amend pointer to species charges accordingly
      add HalfSaturationConstants array to the call to odeDriver
 
  ------------------------------------------------------------------
      
  INPUTS:     
 
  dt            Transport time-step employed
 
  debug            Debugging enabled if > 0 
                    The following levels are used:
 
                    debug = 0        No debugging output
                    debug = 1        Results and statistics ONLY
                                       from reactionStep
                    debug = 2        debug = 1  AND  information
                                       about the convergence in eqStep.
                    debug = 3        debug = 1 AND information 
                                       about the progress in odeDriver.
                    debug = 4        All of the above 
                                       (some of it more verbose)
  
 
  INPUT/OUTPUT:
 
  cbar              On call: transported, unreacted MOLENUMBERS
                              at tc+deltat.
                    On return: reacted MOLENUMBERS at tc+deltat.
 
  OUTPUT:
 
 
  Error            Error flag for the reactionStep. 
                    The following values are defined:
                                  
                    Error = 0        Normal return
                    Error = 1        NIPS error; exceeded maximum 
                                       number of iterations
                    Error = 2        NIPS error; exceeded number of 
                                       back-tracks in linesearch.
                    Error = 3        NIPS error; ill-conditioning 
                                       of linear system.
                    Error = 4        NIPS error; infeasible 
                                       initial point
                    Error = 5        Time-integration error 
                                       (NOT IMPLEMENTED!)
                    Error = 6        Memory error in buffer allocation. 
 
  cumNtTaken        The total number of time-steps taken for 
                    all gridblocks during the transport time-step. 
 
  cumIter           The total number of nonlinear iterations 
                    for all gridblocks (and all reaction time-steps, 
                    if applicable) during the transport time-step. 
  
  peakIter          The largest number of iterations needed for
                    any call to the equilibrium module.
  
  locIter           chemistry work load at each space point
                
  
  exceptionLoc(3)   if Error = 0:
                     The (i,j,k) coordinates of the gridblock reported 
                     as having the highest iteration count.
  
                    if Error > 0:
                     The (i,j,k) coordinates of the gridblock where
                     an error was first detected.
      
  REQUIRED WORK-SPACE:

  rwk(dimrwk)       temporary real*8 work-space
  iwk(dimiwk)       temporary integer work-space

  If only equilibrium reactions occur, we need at least
      
   sizeRwk = nSpecies
      
  For a mixed kinetic/equilibrium system, we need
      
   sizeRwk = 2*nSpecies + 2*nR
      
  if  a second-order integrator (RK-2) is used,  and
      
   sizeRwk = 2*nSpecies + 4*nR
      
   if a fourth-order integrator (RK-4) is used.
      
 
 ----------------------------------------------------------------------- */ 
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "../../Main/msg.h"
#include "reaction.h"

#define loc_offset_(i,j,k) (i)+ nx*((j) + ny*(k))

#define offset_(i,j,k,specie) \
  i+padx+1 + (nx+2*(padx+1))*(j+pady+1 + \
  (ny+2*(pady+1))*(k+padz+1 + (nz+2* (padz+1))*(specie)))

int reactionStep( double dt, int loadlim, int EqOnly,
                  int* cumNtTaken, int* cumIter, int* peakIter,
                  int* exceptionLoc) {


int  debugEq,debugOde;
int  debugLoadBal;
int  i=0,j=0,k=0,species,icell;
int  Error=0,Iter,ntTaken;
int  ir=0,is,nmsg;
int  cellptr;
int  cellcuts[256];
/* int  floats_per_cell; */
int  doubles_per_cell;
int  actCumIter=0;
int  actntTaken=0;

MPI_Request *Req=NULL;
MPI_Status *Status=NULL;
/*  float* Buffer[21];
    float* this;
    float* that;  */
double* Buffer[257];
double* this;
double* that;
double ElapsedTime;

int  kinetic_mode,equilibrium_mode,initial_conditions;

/*-----------------------------------------------------------------------

      SET-UP flags and loops

----------------------------------------------------------------------- */     

ElapsedTime = 0;

kinetic_mode = (nRK!=0)&&(nRQ==0);
equilibrium_mode = (nRK==0)&&(nRQ!=0);
initial_conditions = (EqOnly==1);

switch (outFlags_t[OF_REACTIONS]){

case 0:   
         debugLoadBal = 0;
         debugEq = 0;
         debugOde = 0;
         break;
case 1:
         debugLoadBal = 1;
         debugEq = 0;
         debugOde = 0;
         break;
case 2:
         debugLoadBal = 2;
         debugEq = 1;
         debugOde = 0;
         break;
case 3: 
         debugLoadBal = 2;
         debugEq = 0;
         debugOde = 1;
         break;
case 4: 
         debugLoadBal = 2;
         debugEq = 2;
         debugOde = 1;
         break; 
default:
        
        printf("Invalid type for reaction output /n");
        exit(0);
}

/*-----------------------------------------------------------------------
      IF THERE ARE NO REACTIONS OR
      IF THIS IS THE INITIAL EQUILIBRATION STEP
      AND THE SYSTEM IS OF ALL-KINETIC TYPE: RETURN.
-----------------------------------------------------------------------*/
 
if((nRK==0)&&(nRQ==0)) return 0;
if(initial_conditions&&kinetic_mode)  return 0;


/*----------------------------------------------------------------------
      Calculate work distribution for this time step 
  ---------------------------------------------------------------------*/

nmsg=0;
cell_list[nx*ny*nz]= -1;
if(loadlim > 0 && !EqOnly) {
  if(debugLoadBal && myProc == 0) printf("Calling load balancing...");
  Error = loads( &nmsg);
  if(Error) return Error;
/*  wait for everybody to find out what cell lists to expect  */
  MPI_Barrier(comm); 

  if (debugLoadBal) printf("Proc:%d needs %d messages\n",myProc,nmsg);
  Req = (MPI_Request *)malloc(abs(nmsg)*sizeof(MPI_Request)); 
  Status = (MPI_Status *)malloc(abs(nmsg)*sizeof(MPI_Status));
}
 
/*  floats_per_cell = 2*nSpecies;  */
    doubles_per_cell = 2*nSpecies;

ElapsedTime = MPI_Wtime();

if(nmsg>0){  /* if I am receiving extra cells */
  
  for (ir = 0; ir < nmsg; ir++){
    /*
    Buffer[ir]= (float*)malloc(recv_list_(1,ir)*floats_per_cell*sizeof(float));
    */
    Buffer[ir]= (double*)malloc(recv_list_(1,ir)*doubles_per_cell*sizeof(double));
    if(Buffer[ir]==NULL){ 
      printf("%d:Can't alloc Buffer[%d]\n",myProc,ir);
      return 6;
    }
  }

  /* Post receive buffers for incoming messages */
  for ( ir = 0; ir < nmsg; ir++){
    MPI_Irecv(Buffer[ir],recv_list_(1,ir)*doubles_per_cell,MPI_DOUBLE,
	      recv_list_(0,ir),myProc,comm, &Req[ir]);  
    if (debugLoadBal>1) printf("%d: posting Rec for %d\n",myProc, recv_list_(0,ir));
  }
} 
else if(nmsg < 0){   /* I am sending cells to helpers */
  /* Allocate send buffer */
  for (is = 0; is < -nmsg; is++){
    Buffer[is]= (double*)malloc(send_list_(1,is)*doubles_per_cell*sizeof(double));
    if(Buffer[is]==NULL){
     printf("%d:Can't alloc Buffer[%d]\n",myProc, is);
     return 6;
    }
  }
  
  /* Pack data and Send out work to helpers according to plan */
  cellptr=nx*ny*nz; 
  for ( is =0; is < -nmsg; is ++){  /* for each msg */
    cellcuts[is] = cellptr;
    this = Buffer[is];
    if (debugLoadBal>1)
       printf("%d:packmsg%d c[%d]x%d send to %d\n",myProc,is,
	      cell_list[cellptr],send_list_(1,is),send_list_(0,is));
    for(icell = 0;  icell < send_list_(1,is); icell++){
      
      k = floor(cell_list[cellptr]/(nx*ny));
      j = floor((cell_list[cellptr]%(nx*ny))/nx);
      i = cell_list[cellptr]%nx;
      cellptr -= 1;
      
      for(species = 0; species<nSpecies; species++)
	(*this++) = cbar_(i,j,k,species);
      
      if(chmGuessType==0){
	for(species = 0; species < nSpecies; species++)
	  (*this++) = cbar_(i,j,k,species);
      }
      else
	for(species = 0; species < nSpecies; species++)
	  (*this++) = (double)conc_(i,j,k,species);
    }	

    if (debugLoadBal>1) printf("%d: Packed %d cells for msg %d\n ",
			       myProc,icell,is);

    MPI_Send(Buffer[is],send_list_(1,is)*doubles_per_cell,MPI_DOUBLE,
	     send_list_(0,is), send_list_(0,is), comm);  
    if (debugLoadBal>1) printf("%d: Sent msg %d.\n",myProc,is);
  }
}

/*  ElapsedTime = MPI_Wtime() - ElapsedTime;
    printf("%d: Time in Send/Receive = %g \n",myProc,ElapsedTime);
    ElapsedTime = MPI_Wtime();    */

/*-----------------------------------------------------------------------
      LOOP OVER ALL CELLS THIS PROCESSOR HAS BEEN ASSIGNED
  -----------------------------------------------------------------------*/

*cumIter    = 0;
*cumNtTaken = 0;
*peakIter   = 0;
actCumIter  = 0;
actntTaken  = 0;
icell=0;
if (debugLoadBal>1) printf("%d:Beginning local cell loop..\n",myProc);
while(cell_list[icell] >= 0 ){ /* Until end of cell list */

/*  INITIALIZE SPECIES MOLE VECTOR FOR THIS GRIDBLOCK */ 

  k = floor(cell_list[icell]/(nx*ny));
  j = floor((cell_list[icell]%(nx*ny))/nx);
  i = cell_list[icell]%nx;
  if (debugLoadBal > 1) printf("cell %d %d %d = %d\n",i,j,k,cell_list[icell]);
  
  for(species = 0; species<nSpecies; species++){
    moles[species] = (double)cbar_(i,j,k,species);
  } 
  if(chmGuessType==0){
    for(species = 0; species < nSpecies; species++){
      InitGuess[species] = (double)cbar_(i,j,k,species);
    }
  }
  else
    for(species = 0; species < nSpecies; species++){
      InitGuess[species] = (double)conc_(i,j,k,species);
    }


  /* Call general chemistry package to solve for new state */

  Error = reactOneCell(dt, moles, InitGuess, &Iter, &ntTaken, 
		       debugEq, debugOde, kinetic_mode, equilibrium_mode,
		       initial_conditions ); 

if(debugEq >=2 || debugOde >0 ){
  printf("cell %d %d %d \n", i,j,k);
  printf("Reacted Composition:\n");
  printf("-------------------------\n");
for(species =0; species < nSpecies; species++){
   printf("%16.12g  \t %s \n",moles[species],specieName[species]);
}
printf("\n");
}
 
/*-----------------------------------------------------------------------
      CHECK FOR ERRORS AND UPDATE STATISTICS.
 -----------------------------------------------------------------------*/

  if (Error!=0){
    exceptionLoc[0] = i;
    exceptionLoc[1] = j;
    exceptionLoc[2] = k;
    
    printf("ERROR: Proc: %d Cell = (%d, %d, %d), time = %lf %d\n",
	   myProc,i,j,k,time_t,Error); 
    return Error;
  }
  
  /*-----------------------------------------------------------------------
    STORE RESULT FOR THIS GRID-BLOCK IN GLOBAL ARRAY
    AND PROCEED WITH NEXT GRIDBLOCK (SCALE BACK TO 
    ACTUAL SIZE OF GRIDBLOCK)
    -----------------------------------------------------------------------*/
  for( species = 0; species < nSpecies; species++){
    cbar_(i,j,k,species) = moles[species];
  }
  if(!EqOnly) locIter_(i,j,k) = Iter + ntTaken;
  actCumIter += Iter;
  actntTaken  += ntTaken;
  *cumNtTaken += ntTaken;
  *cumIter += Iter;
  if (Iter >= *peakIter) {
    exceptionLoc[0] = i;
    exceptionLoc[1] = j;
    exceptionLoc[2] = k;
    *peakIter = Iter;
  }
  icell++;

} 
if (debugLoadBal>1) printf("%d:Local cells done.\n",myProc);
if (debugLoadBal) printf("%d: Local Act CumIter = %d Act ntTaken = %d\n",
            myProc,actCumIter,actntTaken);

ElapsedTime = MPI_Wtime() - ElapsedTime;
printf("%d: Time in Local Cells Chemistry = %g \n",myProc,ElapsedTime);
ElapsedTime = MPI_Wtime();

/* Done with local calculation, proceed to shared cells */

if(nmsg > 0 ) { /* Receive and unpack and process cells */
  is = nmsg;
  while(nmsg>0){
    if (debugLoadBal>1) printf("%d: Waiting to recv... ",myProc); 
    MPI_Waitany(is, Req, &ir, Status); 
    if (debugLoadBal>1) printf("%d:Recvd msg %d from %d\n",myProc, ir,Status[0].MPI_SOURCE);

    this = Buffer[ir];
    /* Unpack and process this message */
    for(icell = 0; icell < recv_list_(1,ir); icell++){
    /* Mark cell point in buffer for return data */
      that = this;
      
      for(species = 0; species<nSpecies; species++){
	moles[species] = *this++;
      } 
      for(species = 0; species < nSpecies; species++){
	InitGuess[species] = *this++;
      }
      Error = reactOneCell(dt, moles, InitGuess, &Iter, &ntTaken, 
			   debugEq, debugOde, kinetic_mode,equilibrium_mode,
			   initial_conditions ); 
      if (Error!=0){
	printf("ERROR: Proc: %d Cell = (%d, %d, %d), time = %lf %d\n",
	       recv_list_(0,ir),i,j,k,time_t,Error); 
	return Error;
      }
      
      for(species = 0; species<nSpecies; species++){
	*that++ = moles[species];
      } 
      *that++ = Iter;
      *that++ = ntTaken;
      actCumIter += Iter;
      actntTaken += ntTaken;
    } 
    nmsg--; 
  }
  ElapsedTime = MPI_Wtime() - ElapsedTime;
  printf("%d: Time in NonLocal Cells Chemistry = %g \n",myProc,ElapsedTime);
  ElapsedTime = MPI_Wtime();

  for(ir = 0; ir < is; ir++){
    /* Return processed cell(s) to sender */
     if (debugLoadBal>1)  printf("%d: Returning %d cells to %d\n",myProc,
   	   recv_list_(1,ir), recv_list_(0,ir));
    MPI_Send(Buffer[ir],recv_list_(1,ir)*doubles_per_cell,MPI_DOUBLE,
	      recv_list_(0,ir), recv_list_(0,ir), comm); 
  } 
/*   MPI_Waitall(is,Req,Status);  */
  for(i=0; i < is; i++) free(Buffer[i]);

/*    ElapsedTime = MPI_Wtime() - ElapsedTime;
      printf("%d: Time returning NonLocal Cells = %g \n",myProc,ElapsedTime);
      ElapsedTime = MPI_Wtime();    */

}
else if(nmsg < 0){ /* Wait for returned data to finalize the step */
  /* Post receive buffers for incoming messages */
  for ( is = 0; is < -nmsg; is++){
    MPI_Irecv(Buffer[is],send_list_(1,is)*doubles_per_cell,MPI_DOUBLE,
	      send_list_(0,is),myProc,comm, &Req[is]);  
  }
  ir = -nmsg;
  while(-nmsg > 0){
    if (debugLoadBal>1) printf("%d: Waiting for reply... ",myProc); 
    MPI_Waitany(ir, Req, &is, Status); 
    if (debugLoadBal>1) printf("%d:Msg %d returned from %d\n",myProc, is, send_list_(0,is));
    /* Gather statistics and store data in standard locations */
    cellptr = cellcuts[is];

    for(icell=0; icell < send_list_(1,is); icell++ ){
      this = Buffer[is];
      this+= icell*doubles_per_cell;

      k = floor(cell_list[cellptr]/(nx*ny));
      j = floor((cell_list[cellptr]%(nx*ny))/nx);
      i = cell_list[cellptr]%nx; 

      cellptr -= 1;
      for( species = 0; species < nSpecies; species++){
	cbar_(i,j,k,species) = *this++;
      } 
      Iter = *this++;
      ntTaken = *this;
      *cumIter += Iter;
      *cumNtTaken += ntTaken;
/*        if(!EqOnly) locIter_(i,j,k) = Iter + ntTaken;   */
/*  use a history-weighted average of the work in order to 
    distribute cells on the next iteration;  the weight is arbitrary  */
      if(!EqOnly) locIter_(i,j,k) = 0.5*(Iter + ntTaken + locIter_(i,j,k)) ;   
      if (Iter >= *peakIter) *peakIter = Iter;
    }
    if (debugLoadBal>1) printf("%d:Unpacked msg %d from %d\n",myProc,is,send_list_(0,is));
    nmsg++;
  } 
  for(i =0 ; i< ir; i++) free(Buffer[i]);
/*    ElapsedTime = MPI_Wtime() - ElapsedTime;
      printf("%d: Time receiving NonLocal Cells = %g \n",myProc,ElapsedTime);
      ElapsedTime = MPI_Wtime();    */
}

  if(outFlags_t[OF_REACTIONS]== 1){
    printf("%d: Local+NonLocal Act CumIter = %d Act ntTaken = %d\n",
            myProc,actCumIter,actntTaken);
  }

/*  printf("%d:all reactions done\n",myProc);
    printf("%d: Act CumIter = %d Act ntTaken = %d\n",myProc,actCumIter,actntTaken);  */
if(Req!=NULL) free(Req);
if(Status!=NULL) free(Status);
return Error;
}
