/* This file has the routines needed to obtain interpolated
boundary condition and well and leak values */

#include "ftn.h"
#include "debug.h"
#include <stdio.h>

#include "driverGlobals.h"

#define NPHASES 1

static double interpValue(double t_in, int item, int set, float *data,
			  int *nInterp, int maxnInterp, int nItems);
static double avgValue(double a, double b,
		       int item, int set, float *data,
		       int *nInterp, int maxnInterp, int nItems);

/* Return the pressure boundary condition values */
double pBCValue(double tp_in, int phase, int region) {
  return interpValue(tp_in,phase,region-1,bc_f,
		     nBCInterp_f,maxnBCInterp_f,NPHASES);
}
double PBCValue(double *tp_in, int *phase, int *region) {
  return pBCValue(*tp_in,*phase-1,*region);
}

/* Return the concentration boundary condition values */
double cBCValue(double tc_in, int specie, int region) {
  return interpValue(tc_in,specie,region-1,bc_t,
		     nBCInterp_t,maxnBCInterp_t,nSpecies);
}
double CBCValue(double *tc_in, int *specie, int *region) {
  return cBCValue(*tc_in,*specie-1,*region);
}

/* Return the well injection rates */
double pWellValue(double tp_in, int phase, int well) {
  return interpValue(tp_in,phase,well,wellInj_f,
		     nWellInterp_f,maxnWellInterp_f,NPHASES);
}
double PWellValue(double *tp_in, int *phase, int *well) {
  return pWellValue(*tp_in,*phase-1,*well-1);
}

/* Return the well injection concentrations */
double cWellValue(double tc_in, int specie, int well) {
  return interpValue(tc_in,specie,well,wellInj_t,
		     nWellInterp_t,maxnWellInterp_t,nSpecies);
}
double CWellValue(double *tc_in, int *specie, int *well) {
  return cWellValue(*tc_in,*specie-1,*well-1);
}

/* Return the average leakage rate */
double avgLeakRate(double tc0, double tc1, int specie, int leak) {
  return avgValue(tc0,tc1,specie,leak,leakRate,
		  nLeakInterp,maxnLeakInterp,nSpecies);
}
double AvgLeakRate(double *tc0, double *tc1, int *specie, int *leak) {
  return avgLeakRate(*tc0, *tc1, *specie-1, *leak-1);
}

/* Return the interpolated or averaged values */
/* Note: The item and set indexes start at 0
         Pressure      has item=phase
         Concentration has item=specie
	 Well          has set=well
         BC            has set=region (converted to 0-based!)
*/

#define TDATA(i) (data[2*((i)+maxnInterp*((item)+nItems*(set)))])
#define VDATA(i) (data[1+2*((i)+maxnInterp*((item)+nItems*(set)))])

static double interpValue(double t_in, int item, int set, float *data,
			  int *nInterp, int maxnInterp, int nItems) {
  int i;
  int n = nInterp[item + nItems*set];
  double t0 = (double) TDATA(0);

  if ( n == 1 || t_in <= t0 ) return (double) VDATA(0);
  for( i = 1; i < n ; i++) {
    double t1 = (double) TDATA(i);

    if ( t_in <= t1 ) {
      return ( (t1 - t_in)*VDATA(i-1) + (t_in - t0)*VDATA(i) ) / (t1 - t0);
    }
    t0 = t1;
  }
  return (double) VDATA(n-1);
}

static double avgValue(double a, double b,
		       int item, int set, float *data,
		       int *nInterp, int maxnInterp, int nItems) {
  int i, ia, ib;
  double sum;
  int n = nInterp[item + nItems*set];

  /* Check that a < b */
  if(a >= b) {
    if(a == b) {
      return interpValue(a,item,set,data,nInterp,maxnInterp,nItems);
    } else {
      double c = a; a = b; b = c;
    }
  }

  /* Check degenerate cases */
  if(n == 1 || b <= TDATA(0)) return (double) VDATA(0);
  if(a >= TDATA(n-1)) return (double) VDATA(n-1);

  /* Find interval ia containing a */ 
  for(ia=0; ia<n-1; ia++) {
    if(a<TDATA(ia)) break;
  }

  /* Check if a and b in same interval (then ia>0) */
  if(b<=TDATA(ia)) {
    double midpt = (a+b)/2;
    return ( VDATA(ia-1)*(TDATA(ia) - midpt)
	     + VDATA(ia)*(midpt - TDATA(ia-1)) ) / (TDATA(ia) - TDATA(ia-1));
  }

  /* Find interval ib containing b */ 
  for(ib=ia+1; ib<n; ib++) {
    if(b<TDATA(ib)) break;
  }

  /* integral over ia */
  if(ia==0) {
    sum = (TDATA(0) - a)*VDATA(0);
  } else {
    double midpt = (a+TDATA(ia))/2;
    sum = (TDATA(ia) - a)*( VDATA(ia-1)*(TDATA(ia) - midpt)
            + VDATA(ia)*(midpt - TDATA(ia-1)) ) / (TDATA(ia) - TDATA(ia-1));
  }

  /* integral over full intervals between a and b */
  for(i=ia+1; i<ib; i++) {
    sum += ( VDATA(i-1) + VDATA(i) )*( TDATA(i) - TDATA(i-1) ) / 2;
  }

  /* integral over ib */
  if(ib==n) {
    sum += (b - TDATA(n-1))*VDATA(n-1);
  } else {
    double midpt = (b+TDATA(ib-1))/2;
    sum += (b - TDATA(ib-1))*( VDATA(ib-1)*(TDATA(ib) - midpt)
	     + VDATA(ib)*(midpt - TDATA(ib-1)) ) / (TDATA(ib) - TDATA(ib-1));
  }

  return sum/(b-a);
}
