/******************************************************************************

                    THE CENTER FOR SUBSURFACE MODELING
                  PARALLEL SUBSURFACE SIMULATOR (PARSSIM)

Flow:
  single-phase
  general geometry
  pressure equation (Mixed Method)
  wells
  BCs

Transport:
  ParTrans (Characteristics or Higher Order Godunov Mixed Method)
    multiphase
    wells
    general geometry
    BCs
  Biochemistry
    PNNL CCl_4 Biodegradation
  Radionuclide decay

-------------------------------------------------------------------------------
PRINCIPLE INVESTIGATOR AUTHORS

Todd Arbogast (arbogast@ticam.utexas.edu)
Clint N. Dawson  (clint@ticam.utexas.edu)
Mary F. Wheeler  (mfw@ticam.utexas.edu)

Center for Subsurface Modeling
Mary F. Wheeler, Director
Texas Institute for Computational and Applied Mathematics
TAY 2.400; C0200
The University of Texas at Austin
Austin, Texas  78712

-------------------------------------------------------------------------------
PRIMARY DEVELOPMENTAL AUTHORS
  Driver:    Todd Arbogast
  Flow:      Lawrence Cowsar
  Transport: Todd Arbogast, Clint Dawson, Ashokkumar Chilakapati,
             Douglas Moore, and Ivan Yotov

-------------------------------------------------------------------------------
ACKNOWLEDGMENTS
  The development of this code has been supported primarily by
  the U.S. Department of Energy (DOE) through (1) the Partnership in
  Computational Science (PICS) "Grand Challenge Problems in Environmental
  Modeling and Remediation: Groundwater Contaminant Transport",
  and through (2) the Advanced Computational Technology Initiative (ACTI),
  "Research in New Generation Framework for Petroleum Reservoir Simulation".

  Partial support has also come from other grants from the DOE,
  the U.S. National Science Foundation, and
  the State of Texas Governer`s Energy Office.

COPYRIGHT
  No portion of this document or program may be reproduced,
  transmitted or otherwise copied without prior written permission of
  the authors, except for the internal use of the Center for Subsurface
  Modeling (CSM), The University of Texas at Austin. The authors make
  no representations or warranties about the correctness of any portion
  of the program code, supplementary codes, or associated documentation,
  nor about the suitability of this program for any purpose, and is in
  no way liable for any damages resulting from its use or misuse.

-------------------------------------------------------------------------------
DIFFERENTIAL EQUATIONS

Multi-phase flow and transport:
  See J.C. Parker, "Multiphase flow and transport in porous media,"
  Reviews of Geophysics, 27 (1989), pp. 311-328.

******************************************************************************/

/*
Todd Arbogast
The University of Texas at Austin
August 1996
*/
#include <stdio.h>
#include "msgs.h"
#include "version.h"

#define PSS_ALLOCATE_MEMORY
#include "driverGlobals.h"
#include "typeNumbers.h"

int pssInitialize(char* inFileName0, char* echoFileName,
		  char* echoChemFileName, int echo, int echoChem);
int pssDriver(double* time_flow, double* time_trans, double* time_out, double* time_chem);
void writeTypeInfo();
void expr_listUnits(FILE* fp);

/******************************************************************************
Title: checkExtraMsg
Description: This routine should be called in debugging mode at a place where
             all messages have been received.  If all messages have not
             been received, an error message is printed.
Input: A section heading identifier
*/
#define WAIT_TIME .05
void checkExtraMsg(char* string) {
  /* 
     double time;
     gsyncX();
     time = dclockX() + WAIT_TIME;
     while(dclockX() < time) { continue; };
     if(iprobeX(-1)) {
     printf("EXTRA MSG IN %s: node %4.4d, type %d, src %d, bytes %d\n",
            string,myProc,typeinfoX(),nodeinfoX(),lengthinfoX());
     time = dclockX() + WAIT_TIME;
     while(dclockX() < time) { continue; };
     gsyncX();
  */
  double time;
  int flag;
  MPI_Status status;

  MPI_Barrier(comm);
  time = MPI_Wtime() + WAIT_TIME;
  while(MPI_Wtime() < time) { continue; };

  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, comm, &flag, &status); 
  if(flag) {
    int len;
    MPI_Get_count(&status,MPI_BYTE,&len);
    printf("EXTRA MSG IN %s: node %4.4d, type %d, src %d, bytes %d\n",
	   string,myProc,status.MPI_TAG,status.MPI_SOURCE,len);
  }

  time = MPI_Wtime() + WAIT_TIME;
  while(MPI_Wtime() < time) { continue; };
  MPI_Barrier(comm);
}

/******************************************************************************
Title: timeSince
Description: Returns the time since the last call.
             On first call, returns the current time.
*/
double timeSince() {
  static double myTime1=0;
  double myTime0 = myTime1;

  /* myTime1 = dclockX(); */
  myTime1 = MPI_Wtime();

  return (myTime1 - myTime0);
}

/******************************************************************************
Title: timeClock
Description: Returns the time in hours, minutes, and seconds,
             given a time in seconds.
*/
void timeClock(double time, int* hours, int* minutes, double* seconds) {
  double timeLeft=time;

  *hours = timeLeft/3600;
  timeLeft -= 3600* (*hours);
  *minutes = timeLeft/60;
  timeLeft -= 60* (*minutes);
  *seconds = timeLeft;
}

/******************************************************************************
Title: main
Description: Parses command line arguments, and initializes parallel
             communications and timings.
Inputs: Command line arguments (see the code below for available arguments,
        or try "-h" to list the options).
Return: Error code.
*/

/* MACROS */
#define ABORT(s0) \
  do{ if(!myProc) fprintf(stderr,s0); error = 1; goto exit; }while(0)

int main(int argc, char* argv[]) {
  char inFileNameDefault[] = IN_FILE_NAME_DEFAULT;
  char echoFileNameDefault[] = ECHO_FILE_NAME_DEFAULT;
  char echoChemFileNameDefault[] = ECHO_CHEM_FILE_NAME_DEFAULT;
  char* inFileName0 = inFileNameDefault;
  char* echoFileName = echoFileNameDefault;
  char* echoChemFileName = echoChemFileNameDefault;
  int echo = 0;
  int echoChem = 0;
  int versionOnly = 0;

  double time_flow=0,time_trans=0,time_out=0,time_run,time_chem=0;

  int error = 0;
  int i;

  /* INITIALIZE PARALLEL COMMUNICATIONS, TIMINGS, ETC. */
  /*
    initmsgX(nProcs,argv[0],0);
    nProcs = nnodesX();
    if(nProcs > 10000) ABORT("Too many nodes (10000 maximum)\n");
    time_run = timeSince();
    myProc = mynodeX();
  */
  MPI_Init(&argc, &argv);
  MPI_Comm_size(comm,&nProcs);
  if(nProcs > 10000) ABORT("Too many nodes (10000 maximum)\n");
  time_run = timeSince();
  MPI_Comm_rank(comm,&myProc);

  initDebug(myProc);

  /* PARSE OTHER COMMAND LINE ARGUMENTS */
  for(i = 1; i < argc; i++) {
    if(argv[i][0] == '-') {
      switch(argv[i][1]) {
      case 'i':	i++;
	if(i < argc && argv[i][0] != '-') {
	  inFileName0 = argv[i];
	} else {
	  i--; inFileName0 = inFileNameDefault;
	}
	break;
      case 'e':	echo = 1; i++;
	if(i < argc && argv[i][0] != '-') {
	  echoFileName = argv[i];
	} else {
	  i--; echoFileName = echoFileNameDefault;
	}
	break;
      case 'E':	echo = 1;
	echoFileName = '\0';
	break;
      case 'c':	echoChem = 1; i++;
	if(i < argc && argv[i][0] != '-') {
	  echoChemFileName = argv[i];
	} else {
	  i--; echoChemFileName = echoChemFileNameDefault;
	}
	break;
      case 't':
	if(!myProc) writeTypeInfo();
	return 0;
      case 'u':
	if(!myProc) expr_listUnits(stdout);
	return 0;
      case 'v': i++;
	versionOnly = 1;
	break;
      case 'h':
	if(!myProc) {
	  printf("Usage: parssim [command-options]\n");
	  printf("  Where the optional [command-options] are:\n");
	  printf("    -i [<file name>]   The initial general input <file name>, or \"%s\" if omitted\n",
		 inFileNameDefault);
	  printf("    -e [<file name>]   Echo input to <file name>, or \"%s\" if omitted\n",
		 echoFileNameDefault);
	  printf("    -E                 Echo input to standard out\n");
	  printf("    -c [<file name>]   Echo chemistry input to <file name>, or \"%s\" if omitted\n",
echoChemFileNameDefault);
	  printf("    -t                 Just display PARSSIM input type information\n");
	  printf("    -u                 Just display PARSSIM input physical units information\n");
	  printf("    -v                 Just display PARSSIM version information\n");
	  printf("    -h                 Print this usage information only\n");
	}
	return 0;
      default:
	if(!myProc)
	  fprintf(stderr,"Unrecognized command line option: %s\n",argv[i]);
      }
    } else if(!myProc) {
      fprintf(stderr,"Unrecognized item on command line: %s\n",argv[i]);
      fprintf(stderr,"For command line arguments, try: %s -h\n",argv[0]);
      return 0;
    }
  }

  /* OUTPUT HEADER */
  if(!myProc) {
    printf("\n");
    printf("    THE CENTER FOR SUBSURFACE MODELING\n");
    printf("       PARALLEL SUBSURFACE SIMULATOR\n\n");

    printf("           Code:    PARSSIM%d\n",N_FLOWING_PHASES);
    printf("           Version: %s\n",VERSION);
    printf("           Tag:     %s\n",TAG_VERSION);
    printf("           Date:    %s\n",DATE);
    printf("           Phases:  %d (flowing)\n",N_FLOWING_PHASES);
  }
  if(versionOnly) return 0;

  /* INITIALIZE PARSSIM */
  if(!myProc) printf("\nINITIALIZING PARSSIM\n");
  error = pssInitialize(inFileName0, echoFileName,
			echoChemFileName, echo, echoChem);
  if(debug >= DEBUG_INPUT) dumpDriverGlobals(1,1);
  if(error) goto exit;
  if(debug >= DEBUG_PARALLEL) checkExtraMsg("INITIALIZING PARSSIM");

  /* CALL PARSSIM */
  if(computeFlow || callTransport) {
 
/* debug */
    if(!myProc) printf("\nCALLING PARSSIM DRIVER\n");
    error = pssDriver(&time_flow, &time_trans, &time_out, &time_chem);
    if(error) goto exit;
    if(debug >= DEBUG_PARALLEL) checkExtraMsg("PARSSIM DRIVER");
  }

  /* CONCLUDING BUSINESS */
  if(debug >= DEBUG_MISC) closeDebug();
  /* time_run = dclockX() - time_run; */
  time_run = MPI_Wtime() - time_run;
  {
    double dbuf[5];
    double dbuf_wk[5];
    dbuf[0] = time_flow;
    dbuf[1] = time_trans;
    dbuf[2] = time_out;
    dbuf[3] = time_run;
    dbuf[4] = time_chem;
    /* gdmaxX(MSG_MAXTIME,dbuf,5); */
    MSG_Gdmax2(dbuf,dbuf_wk,5,comm);
    time_flow  = dbuf[0];
    time_trans = dbuf[1];
    time_out   = dbuf[2];
    time_run   = dbuf[3];
    time_chem  = dbuf[4];
  }

  if(!myProc) {
    int hours,minutes;
    double seconds;

    printf("\n");
    if(computeFlow) {
      timeClock(time_flow, &hours, &minutes, &seconds);
      printf("CPU time in flow:        %3.2d:%2.2d:%05.2f\n",
	     hours, minutes, seconds);
    }
    if(callTransport) {
      timeClock(time_trans-time_chem, &hours, &minutes, &seconds);
      printf("CPU time in transport:   %3.2d:%2.2d:%05.2f\n",
	     hours, minutes, seconds);
    }
    if(computeChemistry) {
      timeClock(time_chem, &hours, &minutes, &seconds);
      printf("CPU time in chemistry:   %3.2d:%2.2d:%05.2f\n",
             hours, minutes, seconds);
    }
    if(computeFlow || callTransport) {
      timeClock(time_out, &hours, &minutes, &seconds);
      printf("CPU time writing output: %3.2d:%2.2d:%05.2f\n",
	     hours, minutes, seconds);
      timeClock(time_run, &hours, &minutes, &seconds);
      printf("TOTAL CPU TIME:          %3.2d:%2.2d:%05.2f\n\n",
	     hours, minutes, seconds);
    }
  }
  /* gsyncX(); */
  MPI_Barrier(comm);

 exit:
  printf("NODE %4.4d EXIT (Error code %d)\n", myProc, error);
  /* exitmsgX(); */
  MPI_Finalize();
  return error;
}
