/*
Output routines for 1-D concentration history at a single grid cell.
We assume Tecplot format.
*/

#include <stdlib.h>
#include <stdio.h>

#include "msgs.h"
#include "driverGlobals.h"
#include "typeNumbers.h"
#include "debug.h"

#define ERR_MEMORY 1
#define ERR_OPENING_FILE 2

#define min(a,b) ((a)<(b) ? (a) : (b))
#define max(a,b) ((a)>(b) ? (a) : (b))

FILE* openFile(char* base, char* extension, int count);

static FILE** fp = NULL;

static int* stepOut_hist_t;
static double* tOut_hist_t;
static int firstCall = 1;
static double* sumConc;

/************************************************************************
Utility routines for history output
*/
static int outhist_openFiles() {
  char base[121];
  char extension[] = ".plt";
  int hist,specie,m,iLo,jLo,kLo,iHi,jHi,kHi;
  char typeLetters[4][3] = {"CP","MP","CT","MT"};

  if(myProc) return 0;

  for(hist = 0; hist < nHistories_t; hist++) {
    int typeNo = 0;

    /* Unlocalize */
    iLo = histIndexLo_t_(0,hist) + xOffset_g_(0);
    jLo = histIndexLo_t_(1,hist) + yOffset_g_(0);
    kLo = histIndexLo_t_(2,hist) + zOffset_g_(0);
    iHi = histIndexHi_t_(0,hist) + xOffset_g_(0);
    jHi = histIndexHi_t_(1,hist) + yOffset_g_(0);
    kHi = histIndexHi_t_(2,hist) + zOffset_g_(0);

    /* File name */
    switch(histType_t[hist]) {
    case HIST_CONC_PHASE: typeNo = 0; break;
    case HIST_MOLES_PHASE: typeNo = 1; break;
    case HIST_CONC_TOTAL: typeNo = 2; break;
    case HIST_MOLES_TOTAL: typeNo = 3; break;
    }

    if(iLo == iHi && jLo == jHi && kLo == kHi) {
      sprintf(&base[0],"concHist%d%s%d,%d,%d",
	      hist+1,typeLetters[typeNo],iLo,jLo,kLo);
    } else {
      sprintf(&base[0],"concHist%d%s%d,%d,%d-%d,%d,%d",
	      hist+1,typeLetters[typeNo],iLo,jLo,kLo,iHi,jHi,kHi);
    }
    fp[hist] = openFile(base,extension,-1);
    if(fp[hist] == NULL) return ERR_OPENING_FILE;
    
    /* PREAMBLE */
    fprintf(fp[hist],"TITLE= \"");
    switch(histType_t[hist]) {
    case HIST_MOLES_PHASE:
      fprintf(fp[hist],"Total phase moles");
      break;
    case HIST_MOLES_TOTAL:
      fprintf(fp[hist],"Total moles");
      break;
    case HIST_CONC_PHASE:
      fprintf(fp[hist],"Pore volume average of phase concentration");
      break;
    case HIST_CONC_TOTAL:
      fprintf(fp[hist],"Pore volume average of total moles");
      break;
    }
    if(iLo == iHi && jLo == jHi && kLo == kHi) {
      fprintf(fp[hist]," at");
    } else {
      fprintf(fp[hist]," from (%d,%d,%d) to",iLo,jLo,kLo);
    }
    fprintf(fp[hist]," (%d,%d,%d)\"\n",iHi,jHi,kHi);
    
    fprintf(fp[hist],"VARIABLES= \"TIME\", \"STEP\"");
    for(m = 0; m < nHistSpecies_t[hist]; m++) {
      specie = histSpecie_t_(m,hist);
      fprintf(fp[hist],", \"%d=%s\"",specie,specieName[specie-1]);
    }
    fprintf(fp[hist],"\n");
    fprintf(fp[hist],"ZONE T=\"Rectangular zone\"\n");
    fprintf(fp[hist],"DT=(SINGLE SINGLE");
    for(m = 0; m < nHistSpecies_t[hist]; m++)
      fprintf(fp[hist]," SINGLE");
    fprintf(fp[hist],")\n");
  }

  return 0;
}

void outhist_closeFiles() {
  int hist;

  if(!myProc) {
    for(hist = 0; hist < nHistories_t; hist++) {
      if(fp[hist] != NULL) fclose(fp[hist]);
    }
  }
  return;
}

/************************************************************************
Title: setHistVolFactor
Description: Determine reciprocal of region pore volume
*/
static void setHistVolFactor() {
  int hist,i,j,k,iLo,jLo,kLo,iHi,jHi,kHi;
  double vol;

  for(hist = 0; hist < nHistories_t; hist++) {
    if(histVolFactor_t[hist] == 1) continue;

    iLo = max(1,histIndexLo_t_(0,hist)) - 1;
    jLo = max(1,histIndexLo_t_(1,hist)) - 1;
    kLo = max(1,histIndexLo_t_(2,hist)) - 1;
    iHi = min(nx,histIndexHi_t_(0,hist)) - 1;
    jHi = min(ny,histIndexHi_t_(1,hist)) - 1;
    kHi = min(nz,histIndexHi_t_(2,hist)) - 1;

    vol = 0;
    for(k=kLo; k<=kHi; k++)
    for(j=jLo; j<=jHi; j++)
    for(i=iLo; i<=iHi; i++) {
      vol += effPorosity_(i,j,k)*(xpComp_(i+1)-xpComp_(i))
	*(ypComp_(j+1)-ypComp_(j))*(zpComp_(k+1)-zpComp_(k));
    }
    MSG_Gdsum1(&vol,comm);
    
    histVolFactor_t[hist] = 1/vol;
  }
}

/************************************************************************
Title: outputConcHist_init
Description: Initialize outputConcHist
*/
static int outputConcHist_init() {
  int error;
  int hist,m,i;
  double max;

  /* DEFINE VOLUME FACTORS */

  setHistVolFactor();

  /* MEMMORY ALLOCATION */

  stepOut_hist_t = (int*) malloc(nHistories_t*sizeof(int));
  tOut_hist_t = (double*) malloc(nHistories_t*sizeof(double));
  if(stepOut_hist_t == NULL || tOut_hist_t == NULL) return ERR_MEMORY;

  m = nHistSpecies_t[0];
  for(hist=1; hist<nHistories_t; hist++) {
    if(nHistSpecies_t[hist] > m) m = nHistSpecies_t[hist];
  }
  sumConc = (double*) malloc(m*sizeof(double));
  if(sumConc == NULL) return ERR_MEMORY;

  /* OPEN HISTORY FILES */

  if(!myProc) {
    fp = (FILE**) malloc(nHistories_t*sizeof(FILE*));
    if(fp == NULL) return ERR_MEMORY;
    error = outhist_openFiles();
    if(error) return error;
  }

  /* INITIALIZE TIMING */

  max = dtMax_t[0];
  for(i=1; i<ndtMax_t; i++) { if(max < dtMax_t[i]) max = dtMax_t[i]; }

  for(hist = 0; hist < nHistories_t; hist++) {
    if(dStepOut_hist_t[hist] == 0) dStepOut_hist_t[hist] = 1;
    if(dStepOut_hist_t[hist] < 0) dStepOut_hist_t[hist] = nStepsMax_t + 1;
    if(dtOut_hist_t[hist] == 0) dStepOut_hist_t[hist] = 1;
    if(dtOut_hist_t[hist] <= 0) dtOut_hist_t[hist] = tFinal + max + 1;
    stepOut_hist_t[hist] = step_t;
    tOut_hist_t[hist] = time_t;
  }

  return 0;
}

/************************************************************************
Title: outputConcHist
Description: Write concentration histories at single grid cells
*/
int outputConcHist() {
  int error;
  int hist,specie,m,iLo,jLo,kLo,iHi,jHi,kHi,i,j,k;

  /* OPEN HISTORY FILES AND INITIALIZE */
  if(firstCall) {
    firstCall = 0;
    error = outputConcHist_init();
    if(error) return error;
  }

  /* LOOP OVER HISTORIES AND WRITE CONCENTRATIONS */
  for(hist = 0; hist < nHistories_t; hist++) {
    if(step_t == stepOut_hist_t[hist] || time_t >= tOut_hist_t[hist]) {
      if(step_t == stepOut_hist_t[hist])
	stepOut_hist_t[hist] += dStepOut_hist_t[hist];
      while(time_t >= tOut_hist_t[hist])
	tOut_hist_t[hist] += dtOut_hist_t[hist];
    } else {
      continue;
    }

    iLo = max(1,histIndexLo_t_(0,hist)) - 1;
    jLo = max(1,histIndexLo_t_(1,hist)) - 1;
    kLo = max(1,histIndexLo_t_(2,hist)) - 1;
    iHi = min(nx,histIndexHi_t_(0,hist)) - 1;
    jHi = min(ny,histIndexHi_t_(1,hist)) - 1;
    kHi = min(nz,histIndexHi_t_(2,hist)) - 1;

    for(m = 0; m < nHistSpecies_t[hist]; m++) {
      specie = histSpecie_t_(m,hist)-1;
      sumConc[m] = 0;
      if(histType_t[hist]==HIST_CONC_PHASE ||
	 histType_t[hist]==HIST_MOLES_PHASE) {
	for(k=kLo; k<=kHi; k++)
        for(j=jLo; j<=jHi; j++)
	for(i=iLo; i<=iHi; i++) {
	  double vol = (xpComp_(i+1)-xpComp_(i))
	    *(ypComp_(j+1)-ypComp_(j))*(zpComp_(k+1)-zpComp_(k));
	  sumConc[m] += vol*effPorosity_(i,j,k)*conc_(i,j,k,specie);
	}
      } else {
	for(k=kLo; k<=kHi; k++)
        for(j=jLo; j<=jHi; j++)
	for(i=iLo; i<=iHi; i++) {
	  double vol = (xpComp_(i+1)-xpComp_(i))
	    *(ypComp_(j+1)-ypComp_(j))*(zpComp_(k+1)-zpComp_(k));
	  int tt = traceType[specie]-1;
	  if(tt < 0) {
	    sumConc[m] += vol*conc_(i,j,k,specie)/jacInv_(i,j,k);
	  } else {
	    sumConc[m] += vol*totCoef_(i,j,k,tt)*conc_(i,j,k,specie);
	  }
	}
      }
    }
    MSG_Gdsum(sumConc,nHistSpecies_t[hist],comm);

    if(!myProc) {
      fprintf(fp[hist],"%g %d  ",time_t,step_t);
      for(m = 0; m < nHistSpecies_t[hist]; m++) {
	fprintf(fp[hist]," %g",sumConc[m]*histVolFactor_t[hist]);
      }
      fprintf(fp[hist],"\n");
      fflush(fp[hist]);
    }
  }

  return 0;
}
