/*
Output routines for the pressure, velocities and concentration.
Also output routine for chemistry computational workload.
We assume a single grid for all data.  The grid is the cell centers
and the boundary cell face and edge centers, and the 8 vertices
of the domain.
*/

/* Robert McLay, 3/26/97, and Todd Arbogast, 8/97 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "driverGlobals.h"
#include "../Transport/transportGlobals.h"
#include "typeNumbers.h"

#define ERR_MEMORY 1
#define ERR_OPENING_FILE 2

#define ALWAYS_SPREAD_BUFFER 1

void collectFltArray(float* array, int _nDim, int* _pad, int* shared,
                     int* lwrBnd, int* upBnd, int* arrSizeG, int *gblOrigin,
                     int* _nDom, int* _myDom, FILE* fp,
                     int (*writeBuf)(void *buf, int sizeper,
                                     int number, FILE* fp2));
/* The output buffer (OB) contains one or more data items
   (and possibly the grid point values themselves) at each grid point.
   It is a 4-D array, the first index selects the data (and possibly
   grid point values) and the last 3 indices select the grid point */
static int outbufSizeGrid = -1;  /* Size of the OB grid */
static int nGridItems = 0;       /* Number of Grid items in OB */
static int nDataItems = 0;       /* Number of Data items in OB */

static int haveBndy = 0;         /* Have a boundary flag */
static int lw[4];                /* OB grid lower index */
static int up[4];                /* OB grid upper index */
static int upBnd[4];             /* OB grid upper index for collectFltArray */
static int sz[4];                /* OB grid size=up-lw+1 */
static int arrSizeG[4];          /* Global array size */
static int gblOrigin[4];         /* The global coordinates of Lower left bottom
                                    grid location */
static int zeros[4]={0,0,0,0};

static int nToWriteOnLine;       /* Number of items to write per line */

/* assume lw[0] = 0 */
#define outbuf_(m,i,j,k) \
  outbuf[(m) + sz[0]*((i) - lw[1] + sz[1]*((j)-lw[2] + sz[2]*((k)-lw[3])))]
#define outbuf_D_(m,i,j,k) outbuf[nGridItems + (m) \
             + sz[0]*((i) - lw[1] + sz[1]*((j)-lw[2] + sz[2]*((k)-lw[3])))]

/************************************************************************
Fill in corner points in the data
*/
void spreadOutBuf(float *outbuf)
{
  float f = 1.0/3.0;
  int m,i,j,k;

  if (lw[1] < 0 && lw[2] < 0)
    for (k = 0; k < nz; k++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,-1,-1,k)
          = 0.5*(outbuf_D_(m,-1,0,k) + outbuf_D_(m,0,-1,k));
  if (lw[1] < 0 && up[2] >= ny)
    for (k = 0; k < nz; k++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,-1,ny,k)
          = 0.5*(outbuf_D_(m,-1,ny-1,k) + outbuf_D_(m,0,ny,k));
  if (lw[2] < 0 && up[1] >= nx)
    for (k = 0; k < nz; k++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,nx,-1,k)
          = 0.5*(outbuf_D_(m,nx-1,-1,k) + outbuf_D_(m,nx,0,k));
  if (up[1] >= nx && up[2] >= ny)
    for (k = 0; k < nz; k++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,nx,ny,k)
          = 0.5*(outbuf_D_(m,nx-1,ny,k) + outbuf_D_(m,nx,ny-1,k));

  if (lw[1] < 0 && lw[3] < 0)
    for (j = 0; j < ny; j++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,-1,j,-1)
          = 0.5*(outbuf_D_(m,-1,j,0) + outbuf_D_(m,0,j,-1));
  if (lw[1] < 0 && up[3] >= nz)
    for (j = 0; j < ny; j++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,-1,j,nz)
          = 0.5*(outbuf_D_(m,-1,j,nz-1) + outbuf_D_(m,0,j,nz));
  if (lw[3] < 0 && up[1] >= nx)
    for (j = 0; j < ny; j++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,nx,j,-1)
          = 0.5*(outbuf_D_(m,nx-1,j,-1) + outbuf_D_(m,nx,j,0));
  if (up[1] >= nx && up[3] >= nz)
    for (j = 0; j < ny; j++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,nx,j,nz)
          = 0.5*(outbuf_D_(m,nx-1,j,nz) + outbuf_D_(m,nx,j,nz-1));

  if (lw[2] < 0 && lw[3] < 0)
    for (i = 0; i < nx; i++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,i,-1,-1)
          = 0.5*(outbuf_D_(m,i,0,-1) + outbuf_D_(m,i,-1,0));
  if (lw[2] < 0 && up[3] >= nz)
    for (i = 0; i < nx; i++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,i,-1,nz)
          = 0.5*(outbuf_D_(m,i,-1,nz-1) + outbuf_D_(m,i,0,nz));
  if (lw[3] < 0 && up[2] >= ny)
    for (i = 0; i < nx; i++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,i,ny,-1)
          = 0.5*(outbuf_D_(m,i,ny-1,-1) + outbuf_D_(m,i,ny,0));
  if (up[2] >= ny && up[3] >= nz)
    for (i = 0; i < nx; i++)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,i,ny,nz)
          = 0.5*(outbuf_D_(m,i,ny-1,nz) + outbuf_D_(m,i,ny,nz-1));

  if (lw[1] < 0 && lw[2] < 0 && lw[3] < 0)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,-1,-1,-1)
          = f*(outbuf_D_(m,0,-1,-1)
               + outbuf_D_(m,-1,0,-1) + outbuf_D_(m,-1,-1,0));
  if (up[1] >= nx && lw[2] < 0 && lw[3] < 0)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,nx,-1,-1)
          = f*(outbuf_D_(m,nx-1,-1,-1)
               + outbuf_D_(m,nx,0,-1) + outbuf_D_(m,nx,-1,0));
  if (up[1] >= nx && up[2] >= ny && lw[3] < 0)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,nx,ny,-1)
          = f*(outbuf_D_(m,nx,ny-1,-1)
               + outbuf_D_(m,nx-1,ny,-1) + outbuf_D_(m,nx,ny,0));
  if (lw[1] < 0 && up[2] >= ny && lw[3] < 0)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,-1,ny,-1)
          = f*(outbuf_D_(m,-1,ny-1,-1)
               + outbuf_D_(m,0,ny,-1) + outbuf_D_(m,-1,ny,0));
  if (lw[1] < 0 && lw[2] < 0 && up[3] >= nz)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,-1,-1,nz)
          = f*(outbuf_D_(m,-1,-1,nz-1)
               + outbuf_D_(m,0,-1,nz) + outbuf_D_(m,-1,0,nz));
  if (up[1] >= nx && lw[2] < 0 && up[3] >= nz)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,nx,-1,nz)
          = f*(outbuf_D_(m,nx,-1,nz-1)
               + outbuf_D_(m,nx-1,-1,nz) + outbuf_D_(m,nx,0,nz));
  if (up[1] >= nx && up[2] >= ny && up[3] >= nz)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,nx,ny,nz)
          = f*(outbuf_D_(m,nx,ny,nz-1)
               + outbuf_D_(m,nx,ny-1,nz) + outbuf_D_(m,nx-1,ny,nz));
  if (lw[1] < 0 && up[2] >= ny && up[3] >= nz)
      for (m = 0; m < nDataItems; m++)
        outbuf_D_(m,-1,ny,nz)
          = f*(outbuf_D_(m,-1,ny,nz-1)
               + outbuf_D_(m,-1,ny-1,nz) + outbuf_D_(m,0,ny,nz));

}

/************************************************************************
Form the grid cell centers and boundary points
*/
void formGrid(float* outbuf)
{
  int i,j,k;

  for (k = 0; k < nz; k++)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        {
          outbuf_(0,i,j,k) = 0.125*(xp_(i,j,k)           + xp_(i+1,j,k)+
                                    xp_(i,j+1,k)         + xp_(i+1,j+1,k) +
                                    xp_(i,j,k+1)         + xp_(i+1,j,k+1) +
                                    xp_(i,j+1,k+1) + xp_(i+1,j+1,k+1));
          outbuf_(1,i,j,k) = 0.125*(yp_(i,j,k)           + yp_(i+1,j,k)+
                                    yp_(i,j+1,k)         + yp_(i+1,j+1,k) +
                                    yp_(i,j,k+1)         + yp_(i+1,j,k+1) +
                                    yp_(i,j+1,k+1) + yp_(i+1,j+1,k+1));
          outbuf_(2,i,j,k) = 0.125*(zp_(i,j,k)           + zp_(i+1,j,k)+
                                    zp_(i,j+1,k)         + zp_(i+1,j+1,k) +
                                    zp_(i,j,k+1)         + zp_(i+1,j,k+1) +
                                    zp_(i,j+1,k+1) + zp_(i+1,j+1,k+1));
        }

  if (lw[1] < 0)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        {
          outbuf_(0,-1,j,k) = 0.25*(xp_(0,j,k) + xp_(0,j+1,k)
                                    + xp_(0,j+1,k+1) + xp_(0,j,k+1));
          outbuf_(1,-1,j,k) = 0.25*(yp_(0,j,k) + yp_(0,j+1,k)
                                    + yp_(0,j+1,k+1) + yp_(0,j,k+1));
          outbuf_(2,-1,j,k) = 0.25*(zp_(0,j,k) + zp_(0,j+1,k)
                                    + zp_(0,j+1,k+1) + zp_(0,j,k+1));
        }

  if (up[1] >= nx)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        {
          outbuf_(0,nx,j,k) = 0.25*(xp_(nx,j,k) + xp_(nx,j+1,k)
                                    + xp_(nx,j+1,k+1) + xp_(nx,j,k+1));
          outbuf_(1,nx,j,k) = 0.25*(yp_(nx,j,k) + yp_(nx,j+1,k)
                                    + yp_(nx,j+1,k+1) + yp_(nx,j,k+1));
          outbuf_(2,nx,j,k) = 0.25*(zp_(nx,j,k) + zp_(nx,j+1,k)
                                    + zp_(nx,j+1,k+1) + zp_(nx,j,k+1));
        }

  if (lw[2] < 0)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        {
          outbuf_(0,i,-1,k) = 0.25*(xp_(i,0,k) + xp_(i+1,0,k)
                                    + xp_(i+1,0,k+1) + xp_(i,0,k+1));
          outbuf_(1,i,-1,k) = 0.25*(yp_(i,0,k) + yp_(i+1,0,k)
                                    + yp_(i+1,0,k+1) + yp_(i,0,k+1));
          outbuf_(2,i,-1,k) = 0.25*(zp_(i,0,k) + zp_(i+1,0,k)
                                    + zp_(i+1,0,k+1) + zp_(i,0,k+1));
        }

  if (up[2] >= ny)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        {
          outbuf_(0,i,ny,k) = 0.25*(xp_(i,ny,k) + xp_(i+1,ny,k)
                                    + xp_(i+1,ny,k+1) + xp_(i,ny,k+1));
          outbuf_(1,i,ny,k) = 0.25*(yp_(i,ny,k) + yp_(i+1,ny,k)
                                    + yp_(i+1,ny,k+1) + yp_(i,ny,k+1));
          outbuf_(2,i,ny,k) = 0.25*(zp_(i,ny,k) + zp_(i+1,ny,k)
                                    + zp_(i+1,ny,k+1) + zp_(i,ny,k+1));
        }

  if (lw[3] < 0)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        {
          outbuf_(0,i,j,-1) = 0.25*(xp_(i,j,0) + xp_(i+1,j,0)
                                    + xp_(i+1,j+1,0) + xp_(i,j+1,0));
          outbuf_(1,i,j,-1) = 0.25*(yp_(i,j,0) + yp_(i+1,j,0)
                                    + yp_(i+1,j+1,0) + yp_(i,j+1,0));
          outbuf_(2,i,j,-1) = 0.25*(zp_(i,j,0) + zp_(i+1,j,0)
                                    + zp_(i+1,j+1,0) + zp_(i,j+1,0));
        }

  if (up[3] >= nz)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        {
          outbuf_(0,i,j,nz) = 0.25*(xp_(i,j,nz) + xp_(i+1,j,nz)
                                    + xp_(i+1,j+1,nz) + xp_(i,j+1,nz));
          outbuf_(1,i,j,nz) = 0.25*(yp_(i,j,nz) + yp_(i+1,j,nz)
                                    + yp_(i+1,j+1,nz) + yp_(i,j+1,nz));
          outbuf_(2,i,j,nz) = 0.25*(zp_(i,j,nz) + zp_(i+1,j,nz)
                                    + zp_(i+1,j+1,nz) + zp_(i,j+1,nz));
        }

  if (lw[1] < 0 && lw[2] < 0)
    for (k = 0; k < nz; k++)
      {
        outbuf_(0,-1,-1,k) = 0.5*(xp_(0,0,k) + xp_(0,0,k+1));
        outbuf_(1,-1,-1,k) = 0.5*(yp_(0,0,k) + yp_(0,0,k+1));
        outbuf_(2,-1,-1,k) = 0.5*(zp_(0,0,k) + zp_(0,0,k+1));
      }
  if (lw[1] < 0 && up[2] >= ny)
    for (k = 0; k < nz; k++)
      {
        outbuf_(0,-1,ny,k) = 0.5*(xp_(0,ny,k) + xp_(0,ny,k+1));
        outbuf_(1,-1,ny,k) = 0.5*(yp_(0,ny,k) + yp_(0,ny,k+1));
        outbuf_(2,-1,ny,k) = 0.5*(zp_(0,ny,k) + zp_(0,ny,k+1));
      }
  if (lw[2] < 0 && up[1] >= nx)
    for (k = 0; k < nz; k++)
      {
        outbuf_(0,nx,-1,k) = 0.5*(xp_(nx,0,k) + xp_(nx,0,k+1));
        outbuf_(1,nx,-1,k) = 0.5*(yp_(nx,0,k) + yp_(nx,0,k+1));
        outbuf_(2,nx,-1,k) = 0.5*(zp_(nx,0,k) + zp_(nx,0,k+1));
      }
  if (up[1] >= nx && up[2] >= ny)
    for (k = 0; k < nz; k++)
      {
        outbuf_(0,nx,ny,k) = 0.5*(xp_(nx,ny,k) + xp_(nx,ny,k+1));
        outbuf_(1,nx,ny,k) = 0.5*(yp_(nx,ny,k) + yp_(nx,ny,k+1));
        outbuf_(2,nx,ny,k) = 0.5*(zp_(nx,ny,k) + zp_(nx,ny,k+1));
      }

  if (lw[1] < 0 && lw[3] < 0)
    for (j = 0; j < ny; j++)
      {
        outbuf_(0,-1,j,-1) = 0.5*(xp_(0,j,0) + xp_(0,j+1,0));
        outbuf_(1,-1,j,-1) = 0.5*(yp_(0,j,0) + yp_(0,j+1,0));
        outbuf_(2,-1,j,-1) = 0.5*(zp_(0,j,0) + zp_(0,j+1,0));
      }
  if (lw[1] < 0 && up[3] >= nz)
    for (j = 0; j < ny; j++)
      {
        outbuf_(0,-1,j,nz) = 0.5*(xp_(0,j,nz) + xp_(0,j+1,nz));
        outbuf_(1,-1,j,nz) = 0.5*(yp_(0,j,nz) + yp_(0,j+1,nz));
        outbuf_(2,-1,j,nz) = 0.5*(zp_(0,j,nz) + zp_(0,j+1,nz));
      }
  if (lw[3] < 0 && up[1] >= nx)
    for (j = 0; j < ny; j++)
      {
        outbuf_(0,nx,j,-1) = 0.5*(xp_(nx,j,0) + xp_(nx,j+1,0));
        outbuf_(1,nx,j,-1) = 0.5*(yp_(nx,j,0) + yp_(nx,j+1,0));
        outbuf_(2,nx,j,-1) = 0.5*(zp_(nx,j,0) + zp_(nx,j+1,0));
      }
  if (up[1] >= nx && up[3] >= nz)
    for (j = 0; j < ny; j++)
      {
        outbuf_(0,nx,j,nz) = 0.5*(xp_(nx,j,nz) + xp_(nx,j+1,nz));
        outbuf_(1,nx,j,nz) = 0.5*(yp_(nx,j,nz) + yp_(nx,j+1,nz));
        outbuf_(2,nx,j,nz) = 0.5*(zp_(nx,j,nz) + zp_(nx,j+1,nz));
      }

  if (lw[2] < 0 && lw[3] < 0)
    for (i = 0; i < nx; i++)
      {
        outbuf_(0,i,-1,-1) = 0.5*(xp_(i,0,0) + xp_(i+1,0,0));
        outbuf_(1,i,-1,-1) = 0.5*(yp_(i,0,0) + yp_(i+1,0,0));
        outbuf_(2,i,-1,-1) = 0.5*(zp_(i,0,0) + zp_(i+1,0,0));
      }
  if (lw[2] < 0 && up[3] >= nz)
    for (i = 0; i < nx; i++)
      {
        outbuf_(0,i,-1,nz) = 0.5*(xp_(i,0,nz) + xp_(i+1,0,nz));
        outbuf_(1,i,-1,nz) = 0.5*(yp_(i,0,nz) + yp_(i+1,0,nz));
        outbuf_(2,i,-1,nz) = 0.5*(zp_(i,0,nz) + zp_(i+1,0,nz));
      }
  if (lw[3] < 0 && up[2] >= ny)
    for (i = 0; i < nx; i++)
      {
        outbuf_(0,i,ny,-1) = 0.5*(xp_(i,ny,0) + xp_(i+1,ny,0));
        outbuf_(1,i,ny,-1) = 0.5*(yp_(i,ny,0) + yp_(i+1,ny,0));
        outbuf_(2,i,ny,-1) = 0.5*(zp_(i,ny,0) + zp_(i+1,ny,0));
      }
  if (up[2] >= ny && up[3] >= nz)
    for (i = 0; i < nx; i++)
      {
        outbuf_(0,i,ny,nz) = 0.5*(xp_(i,ny,nz) + xp_(i+1,ny,nz));
        outbuf_(1,i,ny,nz) = 0.5*(yp_(i,ny,nz) + yp_(i+1,ny,nz));
        outbuf_(2,i,ny,nz) = 0.5*(zp_(i,ny,nz) + zp_(i+1,ny,nz));
      }

  if (lw[1] < 0 && lw[2] < 0 && lw[3] < 0)
    {
      outbuf_(0,-1,-1,-1) = xp_(0,0,0);
      outbuf_(1,-1,-1,-1) = yp_(0,0,0);
      outbuf_(2,-1,-1,-1) = zp_(0,0,0);
    }
  if (up[1] >= nx && lw[2] < 0 && lw[3] < 0)
    {
      outbuf_(0,nx,-1,-1) = xp_(nx,0,0);
      outbuf_(1,nx,-1,-1) = yp_(nx,0,0);
      outbuf_(2,nx,-1,-1) = zp_(nx,0,0);
    }
  if (up[1] >= nx && up[2] >= ny && lw[3] < 0)
    {
      outbuf_(0,nx,ny,-1) = xp_(nx,ny,0);
      outbuf_(1,nx,ny,-1) = yp_(nx,ny,0);
      outbuf_(2,nx,ny,-1) = zp_(nx,ny,0);
    }
  if (lw[1] < 0 && up[2] >= ny && lw[3] < 0)
    {
      outbuf_(0,-1,ny,-1) = xp_(0,ny,0);
      outbuf_(1,-1,ny,-1) = yp_(0,ny,0);
      outbuf_(2,-1,ny,-1) = zp_(0,ny,0);
    }
  if (lw[1] < 0 && lw[2] < 0 && up[3] >= nz)
    {
      outbuf_(0,-1,-1,nz) = xp_(0,0,nz);
      outbuf_(1,-1,-1,nz) = yp_(0,0,nz);
      outbuf_(2,-1,-1,nz) = zp_(0,0,nz);
    }
  if (up[1] >= nx && lw[2] < 0 && up[3] >= nz)
    {
      outbuf_(0,nx,-1,nz) = xp_(nx,0,nz);
      outbuf_(1,nx,-1,nz) = yp_(nx,0,nz);
      outbuf_(2,nx,-1,nz) = zp_(nx,0,nz);
    }
  if (up[1] >= nx && up[2] >= ny && up[3] >= nz)
    {
      outbuf_(0,nx,ny,nz) = xp_(nx,ny,nz);
      outbuf_(1,nx,ny,nz) = yp_(nx,ny,nz);
      outbuf_(2,nx,ny,nz) = zp_(nx,ny,nz);
    }
  if (lw[1] < 0 && up[2] >= ny && up[3] >= nz)
    {
      outbuf_(0,-1,ny,nz) = xp_(0,ny,nz);
      outbuf_(1,-1,ny,nz) = yp_(0,ny,nz);
      outbuf_(2,-1,ny,nz) = zp_(0,ny,nz);
    }
  return;
}

/************************************************************************
*/
void internalPresType() {
  int i,j,k;

  switch(presType_out) {
  case PRES_TYPE_POTENTIAL:
    break;
  case PRES_TYPE_PRESSURE:
    for(k=0; k<nz; k++)
    for(j=0; j<ny; j++)
    for(i=0; i<nx; i++) {
      pressure_(i,j,k) -= phaseDensity*gravity*zp_(i,j,k);
    }
    break;
  case PRES_TYPE_HYDRAULIC_HEAD:
    for(k=0; k<nz; k++)
    for(j=0; j<ny; j++)
    for(i=0; i<nx; i++) {
      pressure_(i,j,k) *= phaseDensity*gravity;
    }
    break;
  case PRES_TYPE_HEAD:
    for(k=0; k<nz; k++)
    for(j=0; j<ny; j++)
    for(i=0; i<nx; i++) {
      pressure_(i,j,k) -= zp_(i,j,k);
      pressure_(i,j,k) *= phaseDensity*gravity;
    }
    break;
  }
}

/************************************************************************
*/
int externalPresType(float* outbuf)
{
  int i,j,k;
  int size0 = 0;
  float *grid = 0;

#define grid_(m,i,j,k) \
  grid[(m) + size0*((i) - lw[1] + sz[1]*((j)-lw[2] + sz[2]*((k)-lw[3])))]

  switch(presType_out) {
  case PRES_TYPE_HEAD:
  case PRES_TYPE_PRESSURE:
    if(nGridItems > 0)
      {
	size0 = sz[0];
	grid = outbuf;
      }
    else
      {
	size0 = 3;
	grid = (float *) malloc(sizeof(float)*3*outbufSizeGrid);
	if (grid == NULL)
	  return ERR_MEMORY;
	formGrid(grid);
      }
    break;
  }

  switch(presType_out) {
  default:
  case PRES_TYPE_POTENTIAL:
    break;
  case PRES_TYPE_PRESSURE:
    for(k = lw[3]; k <= up[3]; k++)
      for(j = lw[2]; j <= up[2]; j++)
        for(i = lw[1]; i <= up[1]; i++)
          outbuf_D_(0,i,j,k) += phaseDensity*gravity*grid_(2,i,j,k);
    break;
  case PRES_TYPE_HYDRAULIC_HEAD:
    for(k = lw[3]; k <= up[3]; k++)
      for(j = lw[2]; j <= up[2]; j++)
        for(i = lw[1]; i <= up[1]; i++)
          outbuf_D_(0,i,j,k) /= phaseDensity*gravity;
    break;
  case PRES_TYPE_HEAD:
    for(k = lw[3]; k <= up[3]; k++)
      for(j = lw[2]; j <= up[2]; j++)
        for(i = lw[1]; i <= up[1]; i++)
          {
            outbuf_D_(0,i,j,k) /= phaseDensity*gravity;
            outbuf_D_(0,i,j,k) += grid_(2,i,j,k);
          }
    break;
  }

  switch(presType_out) {
  case PRES_TYPE_HEAD:
  case PRES_TYPE_PRESSURE:
    if(nGridItems <= 0)
      free(grid);
    break;
  }

#undef grid_

  return 0;
}

/************************************************************************
*/
static int iFltArrayCntr = 0;
void resetFltArrayCntr()
{
  iFltArrayCntr = 0;  
}

int writeFltArray(void * array, int sizeper, int size, FILE *fp)
{
  int icnt = iFltArrayCntr;
  float *a = array;
  int ja;

  for (ja = 0; ja < size; ja++)
    {
      icnt++;
      fprintf(fp,"%g ",a[ja]);
      if (icnt >= nToWriteOnLine)
        {
          icnt = 0;
          fprintf(fp,"\n");
        }
    }
  iFltArrayCntr = icnt;
  return 0;
}

void endWriteFltArray(FILE *fp)
{
  if(iFltArrayCntr != 0) fprintf(fp,"\n");
}

/************************************************************************
Utility routines for output
*/
FILE* openFile(char* base, char* extension, int count)
{
  int len;                      /* len of outdir name */
  char fn[256];

  len = strlen(&outDir[0]);
  if (len == 0)
    {
      sprintf(&outDir[0],".%c",SLASH);
      len = 2;
    }
  if (outDir[len-1] == SLASH)
    len--;
  if(count >= 0)
    sprintf(fn,"%.*s%c%s.%03d%s",len,&outDir[0],SLASH,base,count,extension);
  else
    sprintf(fn,"%.*s%c%s%s",len,&outDir[0],SLASH,base,extension);

  return fopen(&fn[0],"w");
}

static void output_setParams(int nItems) {

  if (outbufSizeGrid < 0)
    {
      int ja;
      int nSize[3];

      nSize[0] = nx;
      nSize[1] = ny;
      nSize[2] = nz;

      outbufSizeGrid = 1;
      switch(outFormat3D) {
      case OUT_3D_FORMAT_TECPLOT:
      case OUT_3D_FORMAT_EYE:
	{
	  for (ja = 0; ja < 3; ja++)
	    {
	      if (myDom[ja] == 0)
		{
		  haveBndy = 1;
		  lw[ja+1] = -1;
		}
	      else
		lw[ja+1] = 0;
	      
	      if (myDom[ja] == (nDom[ja]-1))
		{
		  haveBndy = 1;
		  up[ja+1] = nSize[ja];
		}
	      else
		up[ja+1] = nSize[ja]-1;
	      
	      sz[ja+1] = up[ja+1] - lw[ja+1] + 1;
	      outbufSizeGrid *= sz[ja+1];
	      upBnd[ja+1] = up[ja+1] - lw[ja+1];
	    }
	  lw[0] = 0;

	  arrSizeG[1] = nx_g + 2;
	  arrSizeG[2] = ny_g + 2;
	  arrSizeG[3] = nz_g + 2;
	}
	break;
      case OUT_3D_FORMAT_RAW:
      default:
	{
	  for (ja = 0; ja < 3; ja++)
	    {
	      lw[ja+1] = 0;
	      up[ja+1] = nSize[ja]-1;

	      sz[ja+1] = up[ja+1] - lw[ja+1] + 1;
	      outbufSizeGrid *= (sz[ja+1] + 1);
	      upBnd[ja+1] = up[ja+1] - lw[ja+1];
	    }
	  lw[0] = 0;

	  arrSizeG[1] = nx_g;
	  arrSizeG[2] = ny_g;
	  arrSizeG[3] = nz_g;
	  break;
	}
      }

      gblOrigin[0] = 0;
      gblOrigin[1] = xOffset_g_(0) + lw[1] + 1;
      gblOrigin[2] = yOffset_g_(0) + lw[2] + 1;
      gblOrigin[3] = zOffset_g_(0) + lw[3] + 1;

      switch(outFormat3D) {
      case OUT_3D_FORMAT_TECPLOT:
        nGridItems = 3;
        break;
      case OUT_3D_FORMAT_RAW:
      case OUT_3D_FORMAT_EYE:
      default:
        nGridItems = 0;
        break;
      }
    }

  nDataItems = nItems;
  sz[0] = nGridItems + nDataItems;
  up[0] = sz[0]-1;
  upBnd[0] = up[0];
  arrSizeG[0] = sz[0];

  switch(outFormat3D) {
  case OUT_3D_FORMAT_TECPLOT:
    nToWriteOnLine = sz[0];
    break;
  case OUT_3D_FORMAT_RAW:
  case OUT_3D_FORMAT_EYE:
  default:
    nToWriteOnLine = 5;
    break;
  }
}

static int output_time(FILE** fp_time, char* timeFileName,
                       int step, double time) {
  if (myProc != 0) return 0;

  if(*fp_time == NULL)
    {
      *fp_time = openFile(timeFileName, "", -1);
      if (*fp_time == NULL)
	return ERR_OPENING_FILE;
    }
  fprintf(*fp_time,"%d %g\n",step,time);
  fflush(*fp_time);

  return 0;
}

static int output_openFile(FILE** fp, char* fileName, int icnt,
			   int noHeading, char* title,
			   char (*itemNames)[SIZE_WORD]) {
  char extension[8];
  int i;

  if (myProc != 0) return 0;

  switch(outFormat3D) {
  case OUT_3D_FORMAT_TECPLOT:
    strcpy(extension,".plt");
    break;
  case OUT_3D_FORMAT_RAW:
  case OUT_3D_FORMAT_EYE:
  default:
    strcpy(extension,"");
    break;
  }

  *fp = openFile(fileName, extension, icnt);
  if (*fp == NULL)
    return ERR_OPENING_FILE;

  if(noHeading) return 0;

  switch(outFormat3D) {
  case OUT_3D_FORMAT_TECPLOT:
    fprintf(*fp,"TITLE= \"%s\"\n",title);
    fprintf(*fp,"VARIABLES= \"X\", \"Y\", \"Z\"");
    for(i=0; i<nDataItems; i++)
      fprintf(*fp,", \"%s\"",itemNames[i]);
    fprintf(*fp,"\n");

    fprintf(*fp,"ZONE T=\"Rectangular zone\"\n");
    fprintf(*fp," I=%d, J=%d, K=%d, F=POINT\n",
	    arrSizeG[1],arrSizeG[2],arrSizeG[3]);
    fprintf(*fp,"DT=(SINGLE SINGLE SINGLE");
    for(i=0; i<nDataItems; i++)
      fprintf(*fp," SINGLE");
     fprintf(*fp,")\n");
    break;
  case OUT_3D_FORMAT_RAW:
    break;
  case OUT_3D_FORMAT_EYE:
  default:
    fprintf(*fp,"f grid.000\n");
  }

  return 0;
}

static void myCollectFltArray(float* array, int* pad, int* shared, FILE* fp) {
  resetFltArrayCntr();
  if(lw[0] == up[0])
    collectFltArray(array, 3, &pad[1], &shared[1],
		    &zeros[1], &upBnd[1], &arrSizeG[1], &gblOrigin[1],
		    nDom, myDom, fp, writeFltArray);
  else
    collectFltArray(array, 4, pad, shared,
		    zeros, upBnd, arrSizeG, gblOrigin,
		    nDom, myDom, fp, writeFltArray);
  endWriteFltArray(fp);
}

/************************************************************************
Title: outputPres
Description: Write distributed pressure to a single file
*/
int outputPres()
{
  FILE* fp;                     /* file pointer */
  static int icnt = 0;          /* counter over output file number */
  int i,j,k;                    /* counters over space indices */
  int error;                    /* return flag */
  float *outbuf;                /* output buffer */
  static FILE * fp_time = NULL; /* file pointer for simulation times */

  output_setParams(1);

  error = output_time(&fp_time, "presTime", step_f, time_f);
  if(error) return error;

  {
    char (title[])[SIZE_WORD] = {"pressure"};
    error = output_openFile(&fp, "pres", icnt, 0, title[0], title);
  }
  if(error) return error;

  outbuf = (float *) malloc(sizeof(float)*sz[0]*outbufSizeGrid);
  if (outbuf == NULL)
    return ERR_MEMORY;

  /* Fill output buffer with data */
  for (k = 0; k < nz; k++)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(0,i,j,k) = pressure_(i,j,k);

  if (lw[1] < 0)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        outbuf_D_(0,-1,j,k)
          = pressure_(0,j,k); /* = lagrangePresX_(j,k,0); */

  if (up[1] >= nx)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        outbuf_D_(0,nx,j,k)
          = pressure_(nx-1,j,k); /* = lagrangePresX_(j,k,1); */

  if (lw[2] < 0)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        outbuf_D_(0,i,-1,k)
          = pressure_(i,0,k); /* = lagrangePresY_(i,k,0); */

  if (up[2] >= ny)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        outbuf_D_(0,i,ny,k)
          = pressure_(i,ny-1,k); /* = lagrangePresY_(i,k,1); */

  if (lw[3] < 0)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(0,i,j,-1)
          = pressure_(i,j,0); /* = lagrangePresZ_(i,j,0); */

  if (up[3] >= nz)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(0,i,j,nz)
          = pressure_(i,j,nz-1); /* = lagrangePresZ_(i,j,1); */

  spreadOutBuf(outbuf);

  error = externalPresType(outbuf);
  if (error) return error;

  if(nGridItems > 0) formGrid(outbuf);

  myCollectFltArray(outbuf, zeros, zeros, fp);
  if (myProc == 0)
    fclose(fp);

  free(outbuf);

  icnt++;
  return 0;
}

/************************************************************************
Title: outputAvgVel
Description: Write distributed velocity to a single file.  Here we average the
             velocity to the centroid.
*/
static int outputAvgVel()
{
  FILE* fp;                     /* file pointer */
  static int icnt = 0;          /* counter over output file number */
  int i,j,k;                    /* space counters */
  int error;                    /* return flag */
  float *outbuf;                /* output buffer */
  static FILE * fp_time = NULL; /* file pointer for simulation times */
  int n;
  char (itemNames[])[SIZE_WORD] = {"x-velocity","y-velocity","z-velocity"};

  /* no padding and no overlap for Velocity at centroid */
  switch(outFormat3D) {
  case OUT_3D_FORMAT_TECPLOT:
    output_setParams(3);
    break;
  case OUT_3D_FORMAT_EYE:
  default:
    output_setParams(1);
    break;
  }

  error = output_time(&fp_time, "velTime", step_f, time_f);
  if(error) return error;

  outbuf = (float *) malloc(sizeof(float)*sz[0]*outbufSizeGrid);
  if (outbuf == NULL)
    return ERR_MEMORY;

  if(nGridItems > 0) formGrid(outbuf);
  n = 0;

  /* Fill output buffer with data, i.e., averaged Velocity */

  /* Vel X*/
  for (k = 0; k < nz; k++)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,k) = 0.5*(velX_(i,j,k) + velX_(i+1,j,k));

  if (lw[1] < 0)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        outbuf_D_(n,-1,j,k) = velX_(0,j,k);

  if (up[1] >= nx)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        outbuf_D_(n,nx,j,k) = velX_(nx,j,k);

  if (lw[2] < 0)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,-1,k) = outbuf_D_(n,i,0,k);

  if (up[2] >= ny)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,ny,k) = outbuf_D_(n,i,ny-1,k);

  if (lw[3] < 0)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,-1) = outbuf_D_(n,i,j,0);

  if (up[3] >= nz)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,nz) = outbuf_D_(n,i,j,nz-1);

  if(nDataItems == 1)
    {
      spreadOutBuf(outbuf);

      error = output_openFile(&fp, "velx", icnt, 0,
			      itemNames[0], &itemNames[0]);
      if(error) return error;

      myCollectFltArray(outbuf, zeros, zeros, fp);
      if (myProc == 0)
	fclose(fp);

    }
  else
    n++;

  /* Vel Y*/
  for (k = 0; k < nz; k++)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,k) = 0.5*(velY_(i,j,k) + velY_(i,j+1,k));

  if (lw[1] < 0)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        outbuf_D_(n,-1,j,k) = outbuf_D_(n,0,j,k);

  if (up[1] >= nx)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        outbuf_D_(n,nx,j,k) = outbuf_D_(n,nx-1,j,k);

  if (lw[2] < 0)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,-1,k) = velY_(i,0,k);

  if (up[2] >= ny)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,ny,k) = velY_(i,ny,k);

  if (lw[3] < 0)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,-1) = outbuf_D_(n,i,j,0);

  if (up[3] >= nz)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,nz) = outbuf_D_(n,i,j,nz-1);

  if(nDataItems == 1)
    {
      spreadOutBuf(outbuf);

      error = output_openFile(&fp, "vely", icnt, 0,
			      itemNames[1], &itemNames[1]);
      if(error) return error;

      myCollectFltArray(outbuf, zeros, zeros, fp);
      if (myProc == 0)
	fclose(fp);

    }
  else
    n++;

  /* Vel Z*/
  for (k = 0; k < nz; k++)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,k) = 0.5*(velZ_(i,j,k) + velZ_(i,j,k+1));

  if (lw[1] < 0)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        outbuf_D_(n,-1,j,k) = outbuf_D_(n,0,j,k);

  if (up[1] >= nx)
    for (k = 0; k < nz; k++)
      for (j = 0; j < ny; j++)
        outbuf_D_(n,nx,j,k) = outbuf_D_(n,nx-1,j,k);

  if (lw[2] < 0)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,-1,k) = outbuf_D_(n,i,0,k);

  if (up[2] >= ny)
    for (k = 0; k < nz; k++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,ny,k) = outbuf_D_(n,i,ny-1,k);

  if (lw[3] < 0)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,-1) = velZ_(i,j,0);

  if (up[3] >= nz)
    for (j = 0; j < ny; j++)
      for (i = 0; i < nx; i++)
        outbuf_D_(n,i,j,nz) = velZ_(i,j,nz);

  spreadOutBuf(outbuf);

  if(nDataItems == 1)
    error = output_openFile(&fp, "velz", icnt, 0,
			      itemNames[2], &itemNames[2]);
  else
    error = output_openFile(&fp, "vel", icnt, 0, "velocity", itemNames);
  if(error) return error;

  myCollectFltArray(outbuf, zeros, zeros, fp);
  if (myProc == 0)
    fclose(fp);

  free(outbuf);

  icnt++;
  return 0;
}

/************************************************************************
Title: outputRawVel
Description: Write distributed velocity to a single file.  Here we use the
             raw velocities at face centers
*/
static int outputRawVel()
{
  FILE* fp;                     /* file pointer */
  static int icnt = 0;          /* counter over output file number */
  int i,j,k;                    /* space counters */
  int error;                    /* return flag */
  float *outbuf;                /* output buffer */
  static FILE * fp_time = NULL; /* file pointer for simulation times */
  char (itemNames[])[SIZE_WORD] = {"x-velocity","y-velocity","z-velocity"};

  output_setParams(1);

  error = output_time(&fp_time, "velTime", step_f, time_f);
  if(error) return error;

  outbuf = (float *) malloc(sizeof(float)*sz[0]*outbufSizeGrid);
  if (outbuf == NULL)
    return ERR_MEMORY;

  /* Fill output buffer with data */

  /* Vel X*/
  if (myDom[0] == nDom[0]-1) {
    sz[1] = nx + 1;
    up[1] = nx;
    upBnd[1] = nx;
    arrSizeG[1] = nx_g + 1;
  }
  for (k = 0; k < sz[3]; k++)
    for (j = 0; j < sz[2]; j++)
      for (i = 0; i < sz[1]; i++)
        outbuf_D_(0,i,j,k) = velX_(i,j,k);

  error = output_openFile(&fp, "velx", icnt, 0, itemNames[0], &itemNames[0]);
  if(error) return error;

  myCollectFltArray(outbuf, zeros, zeros, fp);
  if (myProc == 0)
    fclose(fp);
  if (myDom[0] == nDom[0]-1) {
    sz[1] = nx;
    up[1] = nx - 1;
    upBnd[1] = nx - 1;
    arrSizeG[1] = nx_g;
  }

  /* Vel Y*/
  if (myDom[1] == nDom[1]-1) {
    sz[2] = ny + 1;
    up[2] = ny;
    upBnd[2] = ny;
    arrSizeG[2] = ny_g + 1;
  }
  for (k = 0; k < sz[3]; k++)
    for (j = 0; j < sz[2]; j++)
      for (i = 0; i < sz[1]; i++)
        outbuf_D_(0,i,j,k) = velY_(i,j,k);

  error = output_openFile(&fp, "vely", icnt, 0, itemNames[1], &itemNames[1]);
  if(error) return error;

  myCollectFltArray(outbuf, zeros, zeros, fp);
  if (myProc == 0)
    fclose(fp);
  if (myDom[1] == nDom[1]-1) {
    sz[2] = ny;
    up[2] = ny - 1;
    upBnd[2] = ny - 1;
    arrSizeG[2] = ny_g;
  }

  /* Vel Z*/
  if (myDom[2] == nDom[2]-1) {
    sz[3] = nz + 1;
    up[3] = nz;
    upBnd[3] = nz;
    arrSizeG[3] = nz_g + 1;
  }
  for (k = 0; k < sz[3]; k++)
    for (j = 0; j < sz[2]; j++)
      for (i = 0; i < sz[1]; i++)
        outbuf_D_(0,i,j,k) = velZ_(i,j,k);

  error = output_openFile(&fp, "velz", icnt, 0, itemNames[2], &itemNames[2]);
  if(error) return error;

  myCollectFltArray(outbuf, zeros, zeros, fp);
  if (myProc == 0)
    fclose(fp);
  if (myDom[2] == nDom[2]-1) {
    sz[3] = nz;
    up[3] = nz - 1;
    upBnd[3] = nz - 1;
    arrSizeG[3] = nz_g;
  }

  free(outbuf);

  icnt++;
  return 0;
}

/************************************************************************
Title: outputVel
Description: Write distributed velocity to a single file.
*/
int outputVel()
{
  switch(outFormat3D) {
  case OUT_3D_FORMAT_TECPLOT:
  case OUT_3D_FORMAT_EYE:
    return outputAvgVel();
  case OUT_3D_FORMAT_RAW:
  default:
    return outputRawVel();
  }
}

/************************************************************************
Title: outputConc
Description: Write distributed concentrations to a single file
*/
int outputConc()
{
  FILE* fp;                     /* file pointer */
  static int icnt = 0;          /* counter over output file number */
  char base[26];
  int i,j,k;                    /* counters over space indices */
  int error;                    /* return flag */
  int concPad[4];               /* the padding */
  float *outbuf;                /* output buffer */
  int needOutbuf=0;             /* need an output buffer flag */
  static FILE * fp_time = NULL; /* file pointer for simulation times */

  int m,specie0,specie1;
  int nPerFile = 1;

  switch(outFormat3D) {
  case OUT_3D_FORMAT_TECPLOT:
    nPerFile = (nSpeciesPerOutfile > nSpecies) ? nSpecies : nSpeciesPerOutfile;
    break;
  case OUT_3D_FORMAT_RAW:
  case OUT_3D_FORMAT_EYE:
  default:
    break;
  }

  output_setParams(nPerFile);

  needOutbuf = haveBndy || (sz[0] > 1) || ALWAYS_SPREAD_BUFFER;
  concPad[0] = 0;

  error = output_time(&fp_time, "concTime", step_t, time_t);
  if(error) return error;

  if (needOutbuf)
    {
      outbuf = (float *) malloc(sizeof(float)*sz[0]*outbufSizeGrid);
      if (outbuf == NULL)
        return ERR_MEMORY;
      concPad[1] = 0;
      concPad[2] = 0;
      concPad[3] = 0;

      if(nGridItems > 0) formGrid(outbuf);
    }
  else
    {
      concPad[1] = padx+1;
      concPad[2] = pady+1;
      concPad[3] = padz+1;
    }

  /* Fill output buffer with data and open output file(s) */

  specie0 = 0;
  while (specie0 < nSpecies)
    {
      specie1 = specie0 + nPerFile;
      if(specie1 > nSpecies) {
	nPerFile = nSpecies - specie0;
	specie1 = nSpecies;
	output_setParams(nPerFile);
      }

      if(nPerFile > 1)
	sprintf(&base[0],"conc%d-%d",specie0+1,specie1);
      else
	sprintf(&base[0],"conc%d",specie0+1);
      error = output_openFile(&fp, base, icnt, 0,
                              "concentration", &specieName[specie0]);
      if(error) return error;

      if (needOutbuf)
        {
	  if(nGridItems > 0) formGrid(outbuf);

          for (k=lw[3]; k <= up[3]; k++)
            for (j=lw[2]; j <= up[2]; j++)
              for (i=lw[1]; i <= up[1]; i++)
		for (m=0; m<nPerFile; m++)
		  outbuf_D_(m,i,j,k) = conc_(i,j,k,specie0+m);
          spreadOutBuf(outbuf);
        }
      else
        outbuf = &conc_array_(specie0);

      myCollectFltArray(outbuf, concPad, zeros, fp);
      if (myProc == 0)
        fclose(fp);

      specie0 = specie1;
    }

  if (needOutbuf)
    free(outbuf);

  icnt++;
  return 0;
}

/************************************************************************
Title: outputWork
Description: Write workload array to a single file
*/
int outputWork()
{
  FILE* fp;                     /* file pointer */
  static int icnt = 0;          /* counter over output file number */
  int i,j,k;                    /* counters over space indices */
  int error;                    /* return flag */
  float *outbuf;                /* output buffer */
  static FILE * fp_time = NULL; /* file pointer for simulation times */

  output_setParams(1);  

  error = output_time(&fp_time, "workTime", step_t, time_t);
  if(error) return error;

  outbuf = (float *) malloc(sizeof(float)*sz[0]*outbufSizeGrid);
  if (outbuf == NULL)
    return ERR_MEMORY;
  
  /* Fill output buffer with data and open output file(s) */

  {
    char (title[])[SIZE_WORD] = {"workload"};
    error = output_openFile(&fp, "work", icnt, 0, title[0], title);
  }
  if(error) return error;


/* just fill outbuf_D with the unvarnished work load */
    for (k=0; k < nz; k++)
      for (j=0; j < ny; j++)
	for (i=0; i < nx; i++)
	  outbuf_D_(0,i,j,k) = locIter_(i,j,k);

  if(nGridItems > 0) formGrid(outbuf); 

  myCollectFltArray(outbuf, zeros, zeros, fp);
  if (myProc == 0)
    fclose(fp);
  
  free(outbuf);

  icnt++;
  return 0;
}

/************************************************************************
Title: outputGrid
Description: Write distributed grid to a single file.  Here we average the
             coordinates to the centroid.  Return immediately if the grid
             is included in the data files.
*/
int outputGrid()
{
  if (outFormat3D == OUT_3D_FORMAT_RAW) return 0;

  FILE* fp;                     /* file pointer */
  static int icnt = 0;          /* counter over output file number */
  int i;
  int error;                    /* return flag */
  float* outbuf;                /* averaged grid */

  /* no padding and no overlap for Grid at centroid*/
  output_setParams(3);
  if(nGridItems > 0) return 0; /* grid in data files, not separate */
  outbuf = (float *) malloc(sizeof(float)*3*outbufSizeGrid);
  if (outbuf == NULL)
    return ERR_MEMORY;

  /* Fill output buffer with data, i.e., averaged Grid */

  if(!zIsDepth) for(i=0; i<(nx+1)*(ny+1)*(nz+1); i++) zp[i] = -zp[i];
  formGrid(outbuf);
  if(!zIsDepth) for(i=0; i<(nx+1)*(ny+1)*(nz+1); i++) zp[i] = -zp[i];

  {
    char (title[])[SIZE_WORD] = {"grid"};
    error = output_openFile(&fp, title[0], icnt, 1, title[0], title);
  }
  if(error) return error;
  switch(outFormat3D) {
  case OUT_3D_FORMAT_RAW:
    break;
  case OUT_3D_FORMAT_EYE:
  default:
    fprintf(fp,"i\n%d %d %d\n",arrSizeG[1],arrSizeG[2],arrSizeG[3]);
    break;
  }

  myCollectFltArray(outbuf, zeros, zeros, fp);
  if (myProc == 0)
    fclose(fp);

  free(outbuf);

  icnt++;
  return 0;
}

#undef outbuf_
#undef outbuf_D_
