#include <stdio.h>
#include <math.h>
#include <string.h>
#include "assert.h"
#include "read.h"

#include "driverGlobals.h"
#include "typeNumbers.h"

/*****************************************************************************
Title: packChemData
Description: Pack the Chem data in arrays for Fortran reaction routines
Return: error
*/

int packChemData()
{
  int i,j;
  int iptr=0;

  iRxnparSZ = 9;
  iRxnpar = (int*)malloc(9*sizeof(int));
  if(iRxnpar == NULL) return ERR_MEMORY;

  iRxnpar[iptr++] = locH;
  iRxnpar[iptr++] = locWater;
  iRxnpar[iptr++] = chmMaxIter;
  iRxnpar[iptr++] = nViol;
  iRxnpar[iptr++] = ideal;
  iRxnpar[iptr++] = chmScaleFlag;
  iRxnpar[iptr++] = chmTestSolFlag;
  iRxnpar[iptr++] = chmInterpFlag;
  iRxnpar[iptr++] = hessFlag;
    
  /** Real Arrays**/

  /* The formula matrix A(nComps,nSpecies):
     Note that A = [ I_nComps stoich^{T} ]    */

  chemAmatrix = (double*)malloc(nComps*nSpecies*sizeof(double));
  if(chemAmatrix == NULL) return ERR_MEMORY;

  iptr = 0;

  for (j = 0; j < nComps; j++)
    for (i = 0; i < nComps; i++)
      chemAmatrix[iptr++] = (i == j) ? 1.0 : 0.0;

  for (j = 0; j < nProducts; j++)
    for (i = 0; i < nComps; i++)
      chemAmatrix[iptr++] = stoich_(j,i);

  /* The stoichiometric matrix V(nSpecies,nProducts)

     Note that V = [ -stoich        ]
                   [   I_nProducts  ]
   */

  chemVmatrix = (double*)malloc(nProducts*nSpecies*sizeof(double));
  if(chemVmatrix == NULL) return ERR_MEMORY ;

  iptr = 0;
  for (j = 0; j < nProducts; j++)
    for (i = 0; i < nSpecies; i++)
      {
	if (i < nComps)
	  chemVmatrix[iptr++] = -stoich_(j,i);
	else
	  chemVmatrix[iptr++] = (i == (nComps + j)) ? 1.0 : 0.0;
      }

  /* kVal[nProducts*2]

     We store (reaction by reaction) either:
     keq,zero (if reactionType = CHEM_RXN_EQ)
     OR
     kfwd,kbwd (if reactionType is kinetic)
   */

  kval = (double*)malloc(nProducts*2*sizeof(double));
  if(kval == NULL) return ERR_MEMORY;

  iptr = 0;
  for (i = 0; i < nProducts; i++)
    kval[iptr++] =
      (reactionType[i] == CHEM_RXN_EQ) ? pow(10.0,peqk[i]) : pow(10.0,pkf[i]);

  for (i = 0; i < nProducts; i++)
    kval[iptr++]
      = (reactionType[i] == CHEM_RXN_EQ) ? 0.0 : pow(10.0,pkb[i]);


  /* ratepow(nComps,nProducts)

     We store the appropriate powers to be used in rate-laws 
     (if reactionType(i) = CHEM_RXN_EQ), otherwise nComps zeros.
   */

  ratepow = (double*)malloc(nComps*nProducts*sizeof(double));
  if(ratepow == NULL) return ERR_MEMORY;

  iptr = 0;
  for (i = 0; i < nProducts; i++)
    for (j = 0; j < nComps; j++)
      ratepow[iptr++] = (reactionType[i] == CHEM_RXN_EQ) ? 0.0 : stoichK_(i,j);


  rRxnparSZ = 6;
  rRxnpar = (double*)malloc(6*sizeof(double));
  if(rRxnpar == NULL) return ERR_MEMORY;

  iptr = 0; 

  rRxnpar[iptr++] = absTemp;
  rRxnpar[iptr++] = chemPres;
  rRxnpar[iptr++] = chmAbsTol;
  rRxnpar[iptr++] = chmAlpha;
  rRxnpar[iptr++] = tauMin;
  rRxnpar[iptr++] = chmRho;

  /* Character Array */


  cRxnparSZ = chemPhases + nSpecies;
  cRxnpar   = (char *) malloc((cRxnparSZ*cSIZE+1)*sizeof(char));
  if (cRxnpar == NULL) return ERR_MEMORY;

  iptr = 0;
  for (i = 0; i < chemPhases; i++)
    strncpy(&cRxnpar_(0,iptr++),&phaseNames_(0,i),cSIZE);

  for (i = 0; i < nSpecies; i++)
    strncpy(&cRxnpar_(0,iptr++),&chmSpecieName_(0,i),cSIZE);
  
  cRxnpar[cRxnparSZ*cSIZE] = '\0';

  return 0;
}

/*****************************************************************************
Title: echoChemData
Description: Echo the Chem data in processed form
Return: error
*/

#define SMALL (1.0e-10)

char* cstr(char* str) {
  static char ss[cSIZE+1];
  int i = 0;
  while(str[i] != ' ' && i < cSIZE) {
    ss[i] = str[i];
    i++;
  }
  ss[i] = '\0';
  return ss;
}

int echoChemData(char* echoChemFileName) {
  FILE* fp;
  int i,j,specie;
  int plus;
  int error = 0;

  if(myProc) return 0;

  fp = fopen(echoChemFileName,"w");
  if(fp == NULL) ABORT(-1);

  fprintf(fp,"GENERAL CHEMISTRY INPUT DATA\n");
  fprintf(fp,"\nRun %d, description:\n%s\n",runNumber,runDescription);

  /* ECHO GENERAL DATA */

  fprintf(fp,"\n\n<GENERAL INFORMATION>\n\n");
  fprintf(fp,"Absolute temperature.........................  %g\n",absTemp);
  fprintf(fp,"Pressure.....................................  %g\n",chemPres);
  fprintf(fp,"Phase-behavior...............................  %3d\n",ideal);
  fprintf(fp,"Location of hydrogen in the component list...  %3d\n",locH);
  fprintf(fp,"Location of water in the component list......  %3d\n",locWater);
  fprintf(fp,"Total number of components employed..........  %3d\n",nComps);
  fprintf(fp,"Total number of species in the system........  %3d\n",nSpecies);
  fprintf(fp,"Number of multi-species phases...............  %3d\n",
	  chemPhaseMulti);
  fprintf(fp,"Maximum number of single-species phases......  %3d\n",
	  chemPhasePure);

  /* ECHO PHASES */

  fprintf(fp,"\n\n<LIST OF PHASES>\n\n");
  fprintf(fp,"Number      Name         Number of      Molar Density\n");
  fprintf(fp,"                          Species    moles/(aq. phase vol.)\n");
  fprintf(fp,"-----------------------------------------------------------\n");
  for(i=0; i<chemPhases; i++) {
    fprintf(fp," %3d    %-15s    %3d",
	    i+1,cstr(&phaseNames_(0,i)),nSpeciesInPhase[i]);
    if(nSpeciesInPhase[i] > 1)
      fprintf(fp,"           %g\n",chmPhaseDensity[i]);
    else
      fprintf(fp,"\n");
  }

  /* ECHO COMPONENTS */

  fprintf(fp,"\n\n<LIST OF COMPONENTS>\n\n");
  fprintf(fp,"Number      Name         Charge\n");
  fprintf(fp,"-------------------------------\n");
  for(i=0; i<nComps; i++) {
    fprintf(fp," %3d    %-15s    %g",
	    i+1,cstr(&chmSpecieName_(0,i)),compCharge[i]);
    if(i+1 == locH) fprintf(fp,"     (Hydrogen)");
    if(i+1 == locWater) fprintf(fp,"     (Water)");
    fprintf(fp,"\n");
  }

  /* ECHO PRODUCT-SPECIES */

  fprintf(fp,"\n\n<LIST OF PRODUCT SPECIES>\n\n");
  for(i=0; i<nProducts; i++) {
    double speciesCharge = 0;
    specie = nComps + i;
    for(j = 0; j < nComps; j++)
      speciesCharge += compCharge[j]*stoich_(i,j);
    fprintf(fp,"Product %d (Specie %d) = %s\n",
	    i+1,specie+1,cstr(&chmSpecieName_(0,specie)));
    fprintf(fp,"  Phase = %s\n",
	    cstr(&phaseNames_(0,chmPhaseIdentity[specie]-1)));
    fprintf(fp,"  Charge = %g\n",speciesCharge);
    switch(reactionType[i]) {
    case CHEM_RXN_EQ:
      fprintf(fp,"  Type = EQUILIBRIUM PRODUCT\n");
      fprintf(fp,"  Equilibrium reaction constant, log10(K) = %g\n",peqk[i]);
      fprintf(fp,"  Consistent reference potential, mu0     = %g\n",
	      mu0[specie]);
      break;
    case CHEM_RXN_K_MASS_ACTION:
    case CHEM_RXN_K_MONOD:
      switch(reactionType[i]) {
      case CHEM_RXN_K_MASS_ACTION:
	fprintf(fp,"  Type = KINETIC MASS ACTION PRODUCT\n");
	break;
      case CHEM_RXN_K_MONOD:
	fprintf(fp,"  Type = KINETIC MONOD PRODUCT\n");
	break;
      }
      fprintf(fp,"  Consistent reference potential, mu0 = %g\n",mu0[specie]);
      fprintf(fp,"    (for switch to equilibrium)\n");
      fprintf(fp,"  Forward rate-constant, log10(K_f)   = %g\n",pkf[i]);
      fprintf(fp,"  Backward rate-constant, log10(K_b)  = %g\n",pkb[i]);
    }
    
    /* ...  WRITE REACTION */

    fprintf(fp,"\n  Reaction:\n");
    plus = 0;
    for(j = 0; j < nComps; j++) {
      if(stoich_(i,j) < -SMALL || stoich_(i,j) > SMALL) {
	if(plus)
	  fprintf(fp," +\n");
	else
	  plus = 1;
	fprintf(fp,"    %g * [%s]",stoich_(i,j),cstr(&chmSpecieName_(0,j)));
      }
    }
    fprintf(fp,"\n      ---> [%s]\n",
	    cstr(&chmSpecieName_(0,specie)));

    /* ...  WRITE RATE-LAW EXPONENTS FOR KINETIC PRODUCTS */

    switch(reactionType[i]) {
    case CHEM_RXN_K_MASS_ACTION:
      fprintf(fp,"\n  Rate-law:\n");
      fprintf(fp,"    r = K_f * {\n");
      plus = 0;
      for(j = 0; j < nComps; j++) {
	if(stoichK_(i,j) < -SMALL || stoichK_(i,j) > SMALL) {
	  if(plus)
	    fprintf(fp," *\n");
	  else
	    plus = 1;
	  fprintf(fp,"                [%s]^(%g)",
		  cstr(&chmSpecieName_(0,j)),stoichK_(i,j));
	}
      }
      fprintf(fp,"\n              } - K_b");
      if(nSpeciesInPhase[chmPhaseIdentity[specie]-1] > 1)
	fprintf(fp," * [%s]\n",cstr(&chmSpecieName_(0,specie)));
      else
	fprintf(fp,"\n");
      break;
    case CHEM_RXN_K_MONOD:
      fprintf(fp,"\n  Rate-law:\n");
      fprintf(fp,"    r = K_f * {\n");
      plus = 0;
      for(j = 0; j < nComps; j++) {
	if(stoichK_(i,j) < -SMALL || stoichK_(i,j) > SMALL) {
	  if(plus)
	    fprintf(fp," *\n");
	  else
	    plus = 1;
	  fprintf(fp,"                [%s]^(%g)",
		  cstr(&chmSpecieName_(0,j)),stoichK_(i,j));
	}
      }
      fprintf(fp,"\n              } *\n");
      fprintf(fp,"              {\n");
      plus = 0;
      for(j = 0; j < nComps; j++) {
	if(halfSatConst_(j,i) != 0) {
	  if(plus)
	    fprintf(fp," *\n");
	  else
	    plus = 1;
	  fprintf(fp,"                [%s] / ( %g + [%s] )",
		  cstr(&chmSpecieName_(0,j)),halfSatConst_(j,i),
		  cstr(&chmSpecieName_(0,j)));
	}
      }
      fprintf(fp,"\n              } - K_b");
      if(nSpeciesInPhase[chmPhaseIdentity[specie]-1] > 1)
	fprintf(fp," * [%s]\n",cstr(&chmSpecieName_(0,specie)));
      else
	fprintf(fp,"\n");
      break;
    }
  }

  /* ECHO SIMULATION DATA */

  fprintf(fp,"\n\n<SIMULATION PARAMETERS>\n\n");
  fprintf(fp,"Maximum number of iterations.................  %d\n",chmMaxIter);
  fprintf(fp,"Absolute error tolerance.....................  %g\n",chmAbsTol);
  fprintf(fp,"Lower bound for concentrations...............  %g\n",chmEpsConc);
  if (chmScaleFlag == 0)
    fprintf(fp,"Scaling is not used\n");
  else
    fprintf(fp,"Scaling is enabled\n");
  if (chmTestSolFlag == 0)
    fprintf(fp,"Second-order sufficiency is not checked\n");
  else
    fprintf(fp,"Second-order sufficiency is checked\n");
  if (chmInterpFlag == 0)
    fprintf(fp,"Simple reduction scheme in line-search\n");
  else
    fprintf(fp,"Cubic interpolation in line-search\n");
  fprintf(fp,"Number of non-monotonic line-searches........  %3d\n",nViol);
  fprintf(fp,"Alpha parameter..............................  %g\n",chmAlpha);
  if (hessFlag == 0)
    fprintf(fp,"Finite-difference Hessian is used\n");
  else
    fprintf(fp,"Analytic Hessian is used\n");
  fprintf(fp,"Percentage of approach to boundary...........  %g\n",tauMin);
  fprintf(fp,"Reduction of perturbation parameter..........  %g\n",chmRho);
  fprintf(fp,"Ratio of reaction/transport time-steps.......  %3d\n",ntReact);
  fprintf(fp,"ODE integration (0 = EF, 1 = RK2, 2 = RK4)...  %3d\n",
	  odeAlgType);
  if (switchFlag == 0)
    fprintf(fp,"Regular time-integration is used\n");
  else
    fprintf(fp,"Species switching enabled\n");

  fclose(fp);

exit:
  switch(error) {
  case -1:
    fprintf(stderr,"Error opening chem echo file %s\n",echoChemFileName);
    break;
  }
  return error;
}
