/*
Todd Arbogast
The University of Texas at Austin
August 1996 and March 1997
*/

/******************************************************************************
Title: read3D<TYPE>
Description: Read and distribute a 3D parallel <TYPE> array of cell- or
             point-centered data.  The array is assumed to be dimensioned
	     in Fortran as (n, 1-px:nnx+px, 1-py:nny+py, 1-pz:nnz+pz),
	     where nnx = nx (cell == !pointx) or nx+1 (pointx), etc.
Inputs: units     For double or float data, the units conversion factor.
        point?    Boolean for whether the array is point-centered, as
                  opposed to being cell-centered.
        n         Number of items per grid location.
	p?        The padding or shadow width.
Output: array     The array to be read.
*/
int CAT2(read3D,TYPE_CAP)(TYPE* array,
#ifdef TYPE_UNITS
			  double units,
#endif
			  int pointx, int pointy, int pointz,
			  int n, int px, int py, int pz) {
  int error = 0;
  int i_g,j_g,k_g, i_l,j_l,k_l, myI,myJ,myK;
  int l,m = 0;
  int nnx = nx, nny = ny, nnz = nz, nnx_g = nx_g, nny_g = ny_g, nnz_g = nz_g;
  TYPE* buffer = 0;
  int nBuf, size = 2*SIZE_BUF;

  if(pointx) { nnx++; nnx_g++; }
  if(pointy) { nny++; nny_g++; }
  if(pointz) { nnz++; nnz_g++; }
  do {
    size /= 2; if(size <= 1) { error = ERR_MEMORY; goto exit; }
    buffer = (TYPE*) malloc(size*sizeof(TYPE));
  } while(buffer == NULL);
  /* giminX(MSG_RD_MINSIZE,&size,1); */
  MSG_Gimin1(&size,comm);

  nBuf = size; m = nBuf;
  for(k_g = 0 ; k_g < nnz_g; k_g++) {
    k_l = kLocal(k_g); myK = ( -pz <= k_l && k_l < nnz+pz );
    for(j_g = 0 ; j_g < nny_g; j_g++) {
      j_l = jLocal(j_g); myJ = ( -py <= j_l && j_l < nny+py );
      for(i_g = 0 ; i_g < nnx_g; i_g++) {
	i_l = iLocal(i_g); myI = ( -px <= i_l && i_l < nnx+px );

	for(l = 0; l < n ; l++) {
	  if(m == nBuf) {
#ifdef TYPE_UNITS
	    CAT3(read,TYPE_CAP,ArrayUnits)(buffer,size,&nBuf,
					   n*nnx_g*nny_g*nnz_g,units);
#else
	    CAT3(read,TYPE_CAP,Array)(buffer,size,&nBuf,n*nnx_g*nny_g*nnz_g);
#endif
	    m = 0;
	  }
	  if(myI && myJ && myK) array_(l,i_l,j_l,k_l) = buffer[m];
	  m++;
	}
      }
    }
  }

 exit:
  free(buffer);
  return error;
}

/******************************************************************************
Title: setup3D<TYPE>
Description: Set up a 3D parallel <TYPE> array of cell- or
             point-centered data.  There may be a fixed number of numbers
             per cell or point.  The array is assumed to be dimensioned
	     in Fortran as (n, 1-px:nnx+px, 1-py:nny+py, 1-pz:nnz+pz),
	     where nn? = n? (cell == !point?) or n?+1 (point?).

	     Possible options are:
	       1. constant <value(s)>
	       2. nearly constant <value(s)>
	          <numNot>
		  {
		  <x0> <x1> <y0> <y1> <z0> <z1> <value(s)>
                  ...
                  }
	       3. { <value(s)> ... } (no units allowed)
	       4. { [<units>] <value(s)> ... } (units apply to all numbers)
	       5. ??? (return ERR_IN_GRID_ARRAY for possible further
                       processing)
	       6. nearly ??? (return  ERR_IN_GRID_ARRAY for possible
                              further processing; this routine
	                      can be called to process the nearly patches)

	     We assume the indices given for patches (in the "nearly" case)
	     are in the range 1:n? for cell arrays and 0:n? for vertex arrays.

	     The string records the option desired, and the nearly flag
	     is set according to whether it was read.

	     On call, nearly must be set to 0, unless only a nearly case is
	     to be processed.
Inputs:      dimen     For double or float data, the dimensions of the data.
             point?    is a boolean for whether the array is point-centered,
                       as opposed to being cell-centered.
             n         is the number of items per grid location.
	     p?        is the padding or shadow width.
             charSize  is the size of the string variable.
	     nearly    is a boolean for the nearly option, if it was read.
Outputs:     array     is the array to be set up.
             string    is the setup command read.
	     nearly    is a boolean for processing only the nearly option.
*/
int CAT2(setup3D,TYPE_CAP)(TYPE* array,
#ifdef TYPE_UNITS
			   int* dimen,
#endif
			   int pointx, int pointy, int pointz,
			   int n, int px, int py, int pz,
			   char* string, int charSize, int* nearly) {
  int error = 0;
  int nnx = nx, nny = ny, nnz = nz;
  int i,j,k,l,m;
#ifdef TYPE_UNITS
  double units = 1;
#endif
  if(pointx) nnx++; if(pointy) nny++; if(pointz) nnz++;

  if(charSize < 21) ABORT(ERR_IN_CODE);

  if(!*nearly) {
    error = getString(string,charSize); if(error) goto exit;
    if(!strcmp(string,"nearly")) {
      *nearly = 1;
      readString(string,charSize);
    };

    if(!strcmp(string,"constant")) {
      int nPoints = (nnx + 2*px)*(nny+2*py)*(nnz+2*pz);
      TYPE* value = (TYPE*) malloc(n*sizeof(TYPE));
      if(value == NULL) ABORT(ERR_MEMORY);
      for(l=0; l<n; l++) {
#ifdef TYPE_UNITS
	error = CAT3(prs_pack,TYPE_CAP,Units)(&value[l],dimen);
	if(error) goto exit;
#else
	CAT2(pack,TYPE_CAP)(value[l]);
#endif
      }
      CAT2(unPack,TYPE_CAP)();
      m = 0;
      for(i=0; i<nPoints; i++)
      for(l=0; l<n; l++) {
	array[m] = value[l];
	m++;
      }
      free(value);
    } else if(string[0] == BEGIN_LIST && string[1] == '\0') {
      if(*nearly) ABORT(ERR_ON_LINE);
#ifdef TYPE_UNITS
      error = getUnits(&units,dimen); if(error) goto exit;
      error = CAT2(read3D,TYPE_CAP)(array, units,
				    pointx,pointy,pointz, n, px,py,pz);
#else
      error = CAT2(read3D,TYPE_CAP)(array, pointx,pointy,pointz, n, px,py,pz);
#endif
      if(error) goto exit;
      readString(string,charSize);
      if(string[0] != END_LIST || string[1] != '\0')
	ABORT(ERR_END_LIST_NOT_READ);
    } else ABORT(ERR_IN_GRID_ARRAY);
  }

  if(*nearly) {
    int numNot,nn, x0,x1, y0,y1, z0,z1;
    TYPE* value = (TYPE*) malloc(n*sizeof(TYPE));
    if(value == NULL) ABORT(ERR_MEMORY);

    packInt(numNot);
    unPackInt();
    beginList();
    for(nn=0; nn<numNot; nn++) {
      packInt(x0); packInt(x1);
      packInt(y0); packInt(y1);
      packInt(z0); packInt(z1);
      for(l=0; l<n; l++) {
#ifdef TYPE_UNITS
	error = CAT3(prs_pack,TYPE_CAP,Units)(&value[l],dimen);
	if(error) goto exit;
#else
	CAT2(pack,TYPE_CAP)(value[l]);
#endif
      }
      unPackInt();
      CAT2(unPack,TYPE_CAP)();
      if(!pointx) {x0--; x1--;}
      if(!pointy) {y0--; y1--;}
      if(!pointz) {z0--; z1--;}
      x0 = max(-px,x0-xOffset_g_(0)); x1 = min(nnx+px-1,x1-xOffset_g_(0));
      y0 = max(-py,y0-yOffset_g_(0)); y1 = min(nny+py-1,y1-yOffset_g_(0));
      z0 = max(-pz,z0-zOffset_g_(0)); z1 = min(nnz+pz-1,z1-zOffset_g_(0));
      for(k=z0; k<=z1; k++)
      for(j=y0; j<=y1; j++)
      for(i=x0; i<=x1; i++)
      for(m=0; m<n; m++)
	array_(m,i,j,k) = value[m];
    }
    endList();
    free(value);
  }

 exit:
  return error;
}

/******************************************************************************
Title: read3DBndry<TYPE>
Description: Read and distribute a 3D parallel <TYPE> array of cell- or
             point-centered data on the exterior boundary.  The array is
             assumed to be dimensioned in Fortran as
             (n, 1-p1:nn1+p1, 1-p2:nn2+p2),
	     where nn? = n? (cell == !point?) or n?+1 (point?)
             and the indices 1 and 2 are x, y, or z, depending on the
	     side number.
Inputs: units     For double or float data, the units conversion factor.
        side      is the side number, one of
                    SIDE_X, SIDE_Y, SIDE_Z
	point?    Boolean for whether the array is point-centered, as
                  opposed to being cell-centered.
        n         Number of items per grid location.
	p?        The padding or shadow width.
Output: arrayBndry     The array to be read.
*/
int CAT2(read3DBndry,TYPE_CAP)
  (TYPE* arrayBndry,
#ifdef TYPE_UNITS
   double units,
#endif
   int side, int point1,int point2, int n, int p1,int p2) {
  int error = 0;
  int i_g,j_g, i_l,j_l, myI,myJ;
  int l,m = 0;
  int nn1, nn2, nn1_g, nn2_g;
  TYPE* buffer = 0;
  int nBuf, size = 2*SIZE_BUF;

  switch(side) {
  case SIDE_X:
    nn1 = ny; nn2 = nz;
    nn1_g = ny_g; nn2_g = nz_g;
    break;
  case SIDE_Y:
    nn1 = nx; nn2 = nz;
    nn1_g = nx_g; nn2_g = nz_g;
    break;
  case SIDE_Z:
    nn1 = nx; nn2 = ny;
    nn1_g = nx_g; nn2_g = ny_g;
    break;
  default:
    ABORT(ERR_IN_CODE);
  }
  if(point1) { nn1++; nn1_g++; }
  if(point2) { nn2++; nn2_g++; }

  do {
    size /= 2; if(size <= 1) { error = ERR_MEMORY; goto exit; }
    buffer = (TYPE*) malloc(size*sizeof(TYPE));
  } while(buffer == NULL);
  /* giminX(MSG_RD_MINSIZE,&size,1); */
  MSG_Gimin1(&size,comm);

  nBuf = size; m = nBuf;
  for(j_g = 0 ; j_g < nn2_g; j_g++) {
    j_l = (side == SIDE_Z) ? jLocal(j_g) : kLocal(j_g);
    myJ = ( -p2 <= j_l && j_l < nn2+p2 );
    for(i_g = 0 ; i_g < nn1_g; i_g++) {
      i_l = (side == SIDE_X) ? jLocal(i_g) : iLocal(i_g);
      myI = ( -p1 <= i_l && i_l < nn1+p1 );

      for(l = 0; l < n ; l++) {
	if(m == nBuf) {
#ifdef TYPE_UNITS
	  CAT3(read,TYPE_CAP,ArrayUnits)(buffer,size,&nBuf,
					 n*nn1_g*nn2_g,units);
#else
	  CAT3(read,TYPE_CAP,Array)(buffer,size,&nBuf,n*nn1_g*nn2_g);
#endif
	  m = 0;
	}
	if(myI && myJ) arrayBndry_(l,i_l,j_l) = buffer[m];
	m++;
      }
    }
  }

 exit:
  free(buffer);
  return error;
}

/******************************************************************************
Title: setup3DBndry<TYPE>
Description: Set up a 3D parallel <TYPE> array of cell- or
             point-centered data on the external boundary.  There may be a
	     fixed number of numbers per cell or point.   The array is
             assumed to be dimensioned in Fortran as
             (n, 1-p1:nn1+p1, 1-p2:nn2+p2),
	     where nn? = n? (cell == !point?) or n?+1 (point?)
             and the indices 1 and 2 are x, y, or z, depending on the
	     side number.

	     Possible options are:
	       1. constant <value(s)>
	       2. nearly constant <value(s)>
	          <numNot>
		  {
		  <numNot_lo_1> <numNot_hi_1>
		          <numNot_lo_2> <numNot_hi_2> <value(s)> ...
		  }
	       3. { <value(s)> ... } (no units allowed)
	       4. { [<units>] <value(s)> ... } (units apply to all numbers)
	       5. ??? (return ERR_IN_GRID_ARRAY for possible further
                       processing)
	       6. nearly ??? (return  ERR_IN_GRID_ARRAY for possible
                              further processing; this routine can be
			      called again to process the nearly patches)

	     We assume the indices given for patches (in the "nearly" case)
	     are in the range 1:n? for cell arrays and 0:n? for vertex arrays.

	     The string records the option desired, and the nearly flag
	     is set according to whether it was read.

	     On call, nearly must be set to 0, unless only a nearly case is
	     to be processed.
Inputs:      dimen       For double or float data, the dimensions of the data.
             side        is the side number, one of
                           SIDE_X, SIDE_Y, SIDE_Z
             point?      is a boolean for whether the array is point-centered,
                         as opposed to being cell-centered.
             n           is the number of items per grid location.
	     p?          is the padding or shadow width.
             charSize    is the size of the string variable.
	     nearly      is a boolean for the nearly option, if it was read.
Outputs:     arrayBndry  is the array to be set up.
             string      is the setup command read.
	     nearly      is a boolean for processing only the nearly option.
*/
int CAT2(setup3DBndry,TYPE_CAP)
  (TYPE* arrayBndry,
#ifdef TYPE_UNITS
   int* dimen,
#endif
   int side, int point1, int point2, int n, int p1, int p2,
   char* string, int charSize, int* nearly) {
  int error = 0;
  int nn1,nn2;
  int i,j,l,m;
#ifdef TYPE_UNITS
  double units = 1;
#endif

  if(charSize < 21) ABORT(ERR_IN_CODE);

  switch(side) {
  case SIDE_X:
    nn1 = ny; nn2 = nz;
    break;
  case SIDE_Y:
    nn1 = nx; nn2 = nz;
    break;
  case SIDE_Z:
    nn1 = nx; nn2 = ny;
    break;
  default:
    ABORT(ERR_IN_CODE);
  }
  if(point1) nn1++; if(point2) nn2++;

  if(!*nearly) {
    error = getString(string,charSize); if(error) goto exit;
    if(!strcmp(string,"nearly")) {
      *nearly = 1;
      readString(string,charSize);
    };

    if(!strcmp(string,"constant")) {
      int nPoints = (nn1 + 2*p1)*(nn2+2*p2);
      TYPE* value = (TYPE*) malloc(n*sizeof(TYPE));
      if(value == NULL) ABORT(ERR_MEMORY);
      for(l=0; l<n; l++) {
#ifdef TYPE_UNITS
	error = CAT3(prs_pack,TYPE_CAP,Units)(&value[l],dimen);
	if(error) goto exit;
#else
	CAT2(pack,TYPE_CAP)(value[l]);
#endif
      }
      CAT2(unPack,TYPE_CAP)();
      m = 0;
      for(i=0; i<nPoints; i++)
      for(l=0; l<n; l++) {
	arrayBndry[m] = value[l];
	m++;
      }
      free(value);
    } else if(string[0] == BEGIN_LIST && string[1] == '\0') {
      if(*nearly) ABORT(ERR_ON_LINE);
#ifdef TYPE_UNITS
      error = getUnits(&units,dimen); if(error) goto exit;
      error = CAT2(read3DBndry,TYPE_CAP)(arrayBndry, units,
					 side, point1,point2, n, p1,p2);
#else
      error = CAT2(read3DBndry,TYPE_CAP)(arrayBndry,
					 side, point1,point2, n, p1,p2);
#endif
      if(error) goto exit;
      readString(string,charSize);
      if(string[0] != END_LIST || string[1] != '\0')
	ABORT(ERR_END_LIST_NOT_READ);
    } else ABORT(ERR_IN_GRID_ARRAY);
  }

  if(*nearly) {
    int numNot, numNot_lo_1, numNot_hi_1, numNot_lo_2, numNot_hi_2;
    TYPE* value = (TYPE*) malloc(n*sizeof(TYPE));
    if(value == NULL) ABORT(ERR_MEMORY);

    packInt(numNot);
    unPackInt();
    beginList();
    for(l=0; l<numNot; l++) {
      packInt(numNot_lo_1); packInt(numNot_hi_1);
      packInt(numNot_lo_2); packInt(numNot_hi_2);
      for(l=0; l<n; l++) {
#ifdef TYPE_UNITS
	error = CAT3(prs_pack,TYPE_CAP,Units)(&value[l],dimen);
	if(error) goto exit;
#else
	CAT2(pack,TYPE_CAP)(value[l]);
#endif
      }
      unPackInt();
      CAT2(unPack,TYPE_CAP)();
      if(!point1) {numNot_lo_1--; numNot_hi_1--;}
      if(!point2) {numNot_lo_2--; numNot_hi_2--;}

      switch(side) {
      case SIDE_X:
	numNot_lo_1 = max(-p1,numNot_lo_1-yOffset_g_(0));
	numNot_hi_1 = min(nn1+p1-1,numNot_hi_1-yOffset_g_(0));
	numNot_lo_2 = max(-p2,numNot_lo_2-zOffset_g_(0));
	numNot_hi_2 = min(nn2+p2-1,numNot_hi_2-zOffset_g_(0));
	break;
      case SIDE_Y:
	numNot_lo_1 = max(-p1,numNot_lo_1-xOffset_g_(0));
	numNot_hi_1 = min(nn1+p1-1,numNot_hi_1-xOffset_g_(0));
	numNot_lo_2 = max(-p2,numNot_lo_2-zOffset_g_(0));
	numNot_hi_2 = min(nn2+p2-1,numNot_hi_2-zOffset_g_(0));
	break;
      case SIDE_Z:
	numNot_lo_1 = max(-p1,numNot_lo_1-xOffset_g_(0));
	numNot_hi_1 = min(nn1+p1-1,numNot_hi_1-xOffset_g_(0));
	numNot_lo_2 = max(-p2,numNot_lo_2-yOffset_g_(0));
	numNot_hi_2 = min(nn2+p2-1,numNot_hi_2-yOffset_g_(0));
	break;
      }
      for(m=0; m<n; m++)
      for(j=numNot_lo_2; j<=numNot_hi_2; j++)
      for(i=numNot_lo_1; i<=numNot_hi_1; i++)
	arrayBndry_(m,i,j) = value[m];
    }
    endList();
    free(value);
  }

 exit:
  return error;
}
