#include <stdio.h>
#include <stdlib.h>
#include "msgs.h"
#include "read.h"
#include "typeNumbers.h"
#include <string.h>
 
#include "driverGlobals.h"

#define STRING_IS(str) !strcmp(string,str)

/******************************************************************************
The variables that are set up have the following form.

nBCRegions
          gives the number of user specified regions of the boundary
          that have the same boundary condition.
bcRegion?_(i,j,m)
          gives the region index number (FORTRAN-STYLE) of the appropriate
          face of the boundary.  These numbers are -1 (if there is no
          outer boundary, i.e., the "boundary" is internal to the domain
          in the domain decomposition), 0 (not used), and 1, ... , nBCRegions.
          The face is specified by the coordinate normal direction
          "?"=1, 2, or 3 (x, y, or z) and "m"=0 or 1 (face of low or high
          indices). "i" and "j" loop over the 2-D indices on the face.
bcType_f_(region)
bcType_t_(region)
          gives the type of boundary condition found on the corresponding
          region of the boundary (see bc,h for symbolic type numbers).
          The "region" is defined by the "bcRegion?" arrays, and goes from
          1 to nBCRegions for flow and from -1:nBCRegions for transport.
          The default type is homogeneous Neumann (no-flow).
bcPresType_in[region]
          gives the interpretation of the input pressure as potential,
          pressure, hydraulic head, or head.
maxnBCInterp_f
maxnBCInterp_t
          gives the maximum number of interpolation points used below.
nBCInterp_f_(region)
nBCInterp_t_(specie,region)
          gives the number of interpolation points that are used to define
	  the function associated with the boundary condition
	  "region"=1,...,nBCRegions (for each "specie"). (There is no
	  need for points for the default no-flow condition.)
bc_f_(i,j,region)
bc_t_(i,j,specie,region)
          defines the piece-wise linear function associated with the
	  boundary condition "region" (for each "specie"). "i"=0,1 defines
	  whether the quantity looked at is time or the function value,
	  respectively.  "j"=0,1,...,nBCInterp_?_([specie,]region)-1
          indexes the interpolation points.
******************************************************************************/


/******************************************************************************
Title: readBCS
Description: Read and preprocess the boundary condition input data.
Return: Error code (see reader.h).
*/
int readBCS() {
  static char BOUNDARY_CONDITIONS[] = {"BOUNDARY CONDITIONS"};
  static char Region[] = {"Region"};
  static char Flow[] = {"Flow"};
  static char Transport[] = {"Transport"};

  int error = 0;
  int i,j,k,n;
  char string[21];
  int nearly = 0;

  bcRegion1 = (int*) malloc(2*ny*nz*sizeof(int));
  bcRegion2 = (int*) malloc(2*nx*nz*sizeof(int));
  bcRegion3 = (int*) malloc(2*nx*ny*sizeof(int));
  if(bcRegion1 == NULL || bcRegion2 == NULL
     || bcRegion3 == NULL) ABORT(ERR_MEMORY);

  for(n=0; n<2*ny*nz; n++) bcRegion1[n] = 1;
  for(n=0; n<2*nx*nz; n++) bcRegion2[n] = 1;
  for(n=0; n<2*nx*ny; n++) bcRegion3[n] = 1;
      
  skipToSection(BOUNDARY_CONDITIONS);
  packInt(nBCRegions);
  unPackInt();
  n = nBCRegions;
  if(nBCRegions <= 0) nBCRegions = 1;
  if(computeFlow) {
    bcType_f = (int*) malloc(nBCRegions*sizeof(int));
    bcPresType_in = (int*) malloc(nBCRegions*sizeof(int));
    if(bcType_f == NULL || bcPresType_in == NULL) ABORT(ERR_MEMORY);

    for(i=1; i<=nBCRegions; i++) {
      bcType_f_(i) = BC_NOFLOW_F;
      bcPresType_in_(i) = PRES_TYPE_POTENTIAL;
    }
  }
  if(callTransport) {
/*  SLB 13 Feb 98 malloc an extra spot for bcType corresponding to
      nBCRegions+1, in case Parcel has altered bcRegionX in that fashion
    bcType_t  = (int*) malloc((2+nBCRegions)*sizeof(int));
*/
    bcType_t  = (int*) malloc((3+nBCRegions)*sizeof(int));
    if(bcType_t == NULL) ABORT(ERR_MEMORY);

    bcType_t_(-1) = BC_INTERNAL_T;
    for(i=0; i<=nBCRegions; i++) bcType_t_(i) = BC_NOFLOW_T;
/* SLB 13 Feb 98 set element to correspond to internal boundary as
      labeled in Parcel */
    bcType_t_(nBCRegions+1) = BC_INTERNAL_T;
  }
  if(n <= 0) goto done; /* assume no-flow on the entire boundary */

  packInt(maxnBCInterp_f);
  packInt(maxnBCInterp_t);
  unPackInt();

  /* read boundary side maps */

  if(!periodicBC[0]) {
    nearly = 0;
    error = setup3DBndryInt(&bcRegion1_(0,0,0),SIDE_X,
			  0,0,1,0,0,string,21,&nearly);
    if(error) goto exit;
    nearly = 0;
    error = setup3DBndryInt(&bcRegion1_(0,0,1),SIDE_X,
			    0,0,1,0,0,string,21,&nearly);
    if(error) goto exit;
  }
  if(!periodicBC[1]) {
    nearly = 0;
    error = setup3DBndryInt(&bcRegion2_(0,0,0),SIDE_Y,
			    0,0,1,0,0,string,21,&nearly);
    if(error) goto exit;
    nearly = 0;
    error = setup3DBndryInt(&bcRegion2_(0,0,1),SIDE_Y,
			    0,0,1,0,0,string,21,&nearly);
    if(error) goto exit;
  }
  if(!periodicBC[2]) {
    nearly = 0;
    error = setup3DBndryInt(&bcRegion3_(0,0,0),SIDE_Z,
			    0,0,1,0,0,string,21,&nearly);
    if(error) goto exit;
    nearly = 0;
    error = setup3DBndryInt(&bcRegion3_(0,0,1),SIDE_Z,
			    0,0,1,0,0,string,21,&nearly);
    if(error) goto exit;
  }

  /* read BC information for each region number */

  if(computeFlow) {
    nBCInterp_f = (int*) calloc(nBCRegions,sizeof(int));
    if(nBCInterp_f == NULL) ABORT(ERR_MEMORY);
    if(maxnBCInterp_f > 0) {
      bc_f = (float*)
	malloc(2*maxnBCInterp_f*nBCRegions*sizeof(float));
      if(bc_f == NULL) ABORT(ERR_MEMORY);
    }
  }
  if(computeTransport) {
    if(nSpecies > 0) {
      nBCInterp_t = (int*) calloc(nSpecies*nBCRegions,sizeof(int));
      if(nBCInterp_t == NULL) ABORT(ERR_MEMORY);
      if(maxnBCInterp_t > 0) {
	bc_t = (float*)
	  malloc(2*maxnBCInterp_t*nSpecies*nBCRegions*sizeof(float));
	if(bc_t == NULL) ABORT(ERR_MEMORY);
      }
    }
  }

  for(n=1; n<=nBCRegions; n++) {
    readSubSection(0);
    skipToSubSection(Region);

    if(computeFlow) {
      readSubSubSection(0);
      skipToSubSubSection(Flow);
      packInt(bcType_f_(n));
      unPackInt();

      switch(bcType_f_(n)) {
      case BC_NOFLOW_F_INPUT:
	bcType_f_(n) = BC_NOFLOW_F;
	break;
      case BC_NEUMANN_F_INPUT:
	bcType_f_(n) = BC_NEUMANN_F;
	break;
      case BC_DIRICHLET_F_INPUT:
	bcType_f_(n) = BC_DIRICHLET_F;
	break;
      default:
	ABORT(-5);
      }
      {
	int dimen_v[] = {DIMEN_VELOCITY};
	int dimen_p[] = {DIMEN_PRESSURE};
	int dimen_h[] = {DIMEN_LENGTH};
	int* dimen1 = dimen_v;

	switch(bcType_f_(n)) {
	case BC_DIRICHLET_F:
	  readString(string,21);
	  if(STRING_IS("potential")) {
	    bcPresType_in_(n) = PRES_TYPE_POTENTIAL;
	    dimen1 = dimen_p;
	  } else if(STRING_IS("pressure")) {
	    bcPresType_in_(n) = PRES_TYPE_PRESSURE;
	    dimen1 = dimen_p;
	  } else if(STRING_IS("hydraulicHead")) {
	    bcPresType_in_(n) = PRES_TYPE_HYDRAULIC_HEAD;
	    dimen1 = dimen_h;
	  } else if(STRING_IS("head")) {
	    bcPresType_in_(n) = PRES_TYPE_HEAD;
	    dimen1 = dimen_h;
	  } else ABORT(-1);
	case BC_NEUMANN_F:
	  packInt(nBCInterp_f_(n));
	  unPackInt();
	  if(nBCInterp_f_(n) > maxnBCInterp_f) ABORT(-4);
	  beginList();
	  for(j=0; j<nBCInterp_f_(n); j++) {
	    packFloatUnits(bc_f_(0,j,n),DIMEN_TIME);
	    error = prs_packFloatUnits(&bc_f_(1,j,n),dimen1);
	    if(error) goto exit;
	  }
	  endList();
	  break;
	default:
	  break;
	}
      }
      for(i=0; i<nBCRegions; i++) {
	if(nBCInterp_f[i] > 1) {
	  steadyFlow = 0;
	  break;
	}
      }
    }

    if(computeTransport) {
      readSubSubSection(0);
      skipToSubSubSection(Transport);
      packInt(bcType_t_(n));   
      unPackInt();

      switch(bcType_t_(n)) {
      case BC_NOFLOW_T:
	break;
      case BC_FLOW_T:
      case BC_DIRICHLET_T:
	beginList();
	for(i=0; i<nSpecies; i++) {
	  packInt(nBCInterp_t_(i,n));
	  unPackInt();
	  if(nBCInterp_t_(i,n) > maxnBCInterp_t) ABORT(-4);
	  beginList();
	  for(j=0; j<nBCInterp_t_(i,n); j++) {
	    packFloatUnits(bc_t_(0,j,i,n),DIMEN_TIME);
	    packFloatUnits(bc_t_(1,j,i,n),DIMEN_MOLAR_CONC);
	  }
	  endList();
	}
	endList();
	break;
      default:
	ABORT(-5);
      }
    }
  }

 done:
  if(myNbr_(-1,0,0) > -1) {
    for(j=0; j<ny; j++)
    for(k=0; k<nz; k++)
      bcRegion1_(j,k,0) = -1;
  } else {
    for(j=0; j<ny; j++)
    for(k=0; k<nz; k++)
      if(bcRegion1_(j,k,0) < 1 || bcRegion1_(j,k,0) > nBCRegions) ABORT(-3);
  }
  if(myNbr_(1,0,0)>-1) {
    for(j=0; j<ny; j++)
    for(k=0; k<nz; k++)
      bcRegion1_(j,k,1) = -1;
  } else {
    for(j=0; j<ny; j++)
    for(k=0; k<nz; k++)
      if(bcRegion1_(j,k,1) < 1 || bcRegion1_(j,k,1) > nBCRegions) ABORT(-3);
  }
  if(myNbr_(0,-1,0)>-1) {
    for(i=0; i<nx; i++)
    for(k=0; k<nz; k++)
      bcRegion2_(i,k,0) = -1;
  } else {
    for(i=0; i<nx; i++)
    for(k=0; k<nz; k++)
      if(bcRegion2_(i,k,0) < 1 || bcRegion2_(i,k,0) > nBCRegions) ABORT(-3);
  }
  if(myNbr_(0,1,0)>-1) {
    for(i=0; i<nx; i++)
    for(k=0; k<nz; k++)
      bcRegion2_(i,k,1) = -1;
  } else {
    for(i=0; i<nx; i++)
    for(k=0; k<nz; k++)
      if(bcRegion2_(i,k,1) < 1 || bcRegion2_(i,k,1) > nBCRegions) ABORT(-3);
  }
  if(myNbr_(0,0,-1)>-1) {
    for(i=0; i<nx; i++)
    for(j=0; j<ny; j++)
      bcRegion3_(i,j,0) = -1;
  } else {
    for(i=0; i<nx; i++)
    for(j=0; j<ny; j++)
      if(bcRegion3_(i,j,0) < 1 || bcRegion3_(i,j,0) > nBCRegions) ABORT(-3);
  }
  if(myNbr_(0,0,1)>-1) {
    for(i=0; i<nx; i++)
    for(j=0; j<ny; j++)
      bcRegion3_(i,j,1) = -1;
  } else {
    for(i=0; i<nx; i++)
    for(j=0; j<ny; j++)
      if(bcRegion3_(i,j,1) < 1 || bcRegion3_(i,j,1) > nBCRegions) ABORT(-3);
  }

 exit:
  switch(error) {
  case -1:
    if(!myProc) fprintf(stderr,"Unsupported pressure type\n");
    error = ERR_ON_LINE;
    break;
  case -3:
    fprintf(stderr,"BC Region number out of range on processor %d\n",myProc);
    error = ERR_BEFORE_LINE;
  case -4:
    if(!myProc) fprintf(stderr,"Too many interpolation times\n");
    error = ERR_ON_LINE;
    break;
  case -5:
    if(!myProc) fprintf(stderr,"Unsupported boundary condition type\n");
    error = ERR_ON_LINE;
    break;
  }
  return error;
}
