#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "msgs.h"
#include "read.h"

#include "driverGlobals.h"
#include "typeNumbers.h"

/******************************************************************************
Title: pingNbrs
Description: Sends a ping message to each neighbor processor and waits
             for a response from each neighbor node
*/
void pingNbrs() {
  int i,j,k;
  char null[1] = {""};
  MPI_Request req = MPI_REQUEST_NULL ;
  MPI_Status status;

  /* For each neighbor processor, send ping */
  for(k=-1; k<=1; k++)
  for(j=-1; j<=1; j++)
  for(i=-1; i<=1; i++) {
    /* if(myNbr_(i,j,k) != -1) asendX(MSG_PING+myProc,null,1,myNbr_(i,j,k)); */
    if(myNbr_(i,j,k) != -1) {
      MPI_Isend(null,1,MPI_BYTE,myNbr_(i,j,k),MSG_PING+myProc,comm,&req);
      MPI_Request_free(&req);
    }
  }

  /* For each neighbor processor, await reply */
  for(k=-1; k<=1; k++)
  for(j=-1; j<=1; j++)
  for(i=-1; i<=1; i++) {
    /* if(myNbr_(i,j,k) != -1) recvX(MSG_PING+myNbr_(i,j,k),null,1); */
    if(myNbr_(i,j,k) != -1) {
      MPI_Recv(null,1,MPI_BYTE,MPI_ANY_SOURCE,MSG_PING+myNbr_(i,j,k),
	       comm,&status);
    }
  }
}

/******************************************************************************
Title: distributeIndices
Description: Distributes 1-D element indices uniformly over the processors.
             Sets the global indices of the lowest element in
               i0_g[0] for myDom[dir]-1
               i0_g[1] for myDom[dir]
               i0_g[2] for myDom[dir]+1
               i0_g[3] for myDom[dir]+2
Inputs: n_g   The global number of elements in the coordinate direction
        dir   The cartesian direction
Output: i0_g  The global indices array
Return: The size of the subdomain.
*/
int distributeIndices(int* i0_g, int n_g, int dir) {
  int n,nn0,m;

  n = n_g/nDom[dir];           /* number of elements is n or n+1 */
  nn0 = nDom[dir]*(n+1) - n_g; /* number with n elements */

  m = myDom[dir]-1; i0_g[0] = (m < nn0) ? m*n : m*(n+1) - nn0;
  m = myDom[dir];   i0_g[1] = (m < nn0) ? m*n : m*(n+1) - nn0;
  m = myDom[dir]+1; i0_g[2] = (m < nn0) ? m*n : m*(n+1) - nn0;
  m = myDom[dir]+2; i0_g[3] = (m < nn0) ? m*n : m*(n+1) - nn0;
  if(i0_g[0] < 0) i0_g[0] = 0;
  if(i0_g[3] > n_g) i0_g[3] = n_g;

  return (i0_g[2] - i0_g[1]);
}

/******************************************************************************
Title: defineDomainDecomp
Description: Define the domain decomposition.
*/

void defineDomainDecomp() {
  int i,j,k;

  /* SET PARALLEL STRUCTURE */
  myDom[2] = myProc/(nDom[0]*nDom[1]);
  myDom[1] = (myProc - nDom[0]*nDom[1]*myDom[2])/nDom[0];
  myDom[0] = myProc - nDom[0]*(myDom[1] + nDom[1]*myDom[2]);

  /* Define myNbr (account for periodicity) */
  for(k=-1; k<=1; k++) {
    int kk = myDom[2]+k;
    if(kk < 0) kk = nDom[2]-1;
    if(kk > nDom[2]-1) kk = 0;
    for(j=-1; j<=1; j++) {
      int jj = myDom[1]+j;
      if(jj < 0) jj = nDom[1]-1;
      if(jj > nDom[1]-1) jj = 0;
      for(i=-1; i<=1; i++) {
	int ii = myDom[0]+i;
	if(ii < 0) ii = nDom[0]-1;
	if(ii > nDom[0]-1) ii = 0;

	myNbr_(i,j,k) = ii + nDom[0]*(jj + nDom[1]*kk);
      }
    }
  }
  if(!periodicBC[0]) {
    if(myDom[0] == 0) {
      for(k=-1; k<=1; k++)
      for(j=-1; j<=1; j++) {
	myNbr_(-1,j,k) = -1;
      }
    }
    if(myDom[0] == nDom[0]-1) {
      for(k=-1; k<=1; k++)
      for(j=-1; j<=1; j++) {
	myNbr_(1,j,k) = -1;
      }
    }
  }
  if(!periodicBC[1]) {
    if(myDom[1] == 0) {
      for(k=-1; k<=1; k++)
      for(i=-1; i<=1; i++) {
        myNbr_(i,-1,k) = -1;
      }
    }
    if(myDom[1] == nDom[1]-1) {
      for(k=-1; k<=1; k++)
      for(i=-1; i<=1; i++) {
        myNbr_(i,1,k) = -1;
      }
    }
  }
  if(!periodicBC[2]) {
    if(myDom[2] == 0) {
      for(j=-1; j<=1; j++)
      for(i=-1; i<=1; i++) {
        myNbr_(i,j,-1) = -1;
      }
    }
    if(myDom[2] == nDom[2]-1) {
      for(j=-1; j<=1; j++)
      for(i=-1; i<=1; i++) {
        myNbr_(i,j,1) = -1;
      }
    }
  }

  /* Ping neighbors */
  if (debug >= DEBUG_PARALLEL) {
    if(!myProc) printf("Pinging neighbors\n");
    /* gsyncX(); */
    MPI_Barrier(comm);
    pingNbrs();
    printf("  Ping successful for NODE %4.4d\n",myProc);
    /* gsyncX(); */
    MPI_Barrier(comm);
  }

  /* DEFINE LOCAL SIZES AND INDICES */
  if(nx_g < 1) nx_g = 1;
  if(ny_g < 1) ny_g = 1;
  if(nz_g < 1) nz_g = 1;

  nx = distributeIndices(&xOffset_g_(-1), nx_g, (int)0);
  ny = distributeIndices(&yOffset_g_(-1), ny_g, (int)1);
  nz = distributeIndices(&zOffset_g_(-1), nz_g, (int)2);
}

/******************************************************************************
Title: ReadGrid1
Description: Reads and initialized type 1 grids (uniform rectangular).
Return: Error code (see reader.h).
*/
int ReadGrid1() {
  int error = 0;
  int i,j,k;
  double x0, xMax;
  double y0, yMax;
  double z0, zMax;
  double dx,dy,dz;
  double x00,y00,z00;

  packDoubleUnits(x0,DIMEN_LENGTH); packDoubleUnits(xMax,DIMEN_LENGTH);
  packDoubleUnits(y0,DIMEN_LENGTH); packDoubleUnits(yMax,DIMEN_LENGTH);
  packDoubleUnits(z0,DIMEN_LENGTH); packDoubleUnits(zMax,DIMEN_LENGTH);
  unPackDouble();

  dx = (xMax - x0)/nx_g;
  dy = (yMax - y0)/ny_g;
  dz = (zMax - z0)/nz_g;

  x00 = xOffset_g_(0)*dx;
  y00 = yOffset_g_(0)*dy;
  z00 = zOffset_g_(0)*dz;

  for(k=0; k<=nz; k++)
  for(j=0; j<=ny; j++)
  for(i=0; i<=nx; i++) {
    xp_(i,j,k) = x00 + i*dx;
    yp_(i,j,k) = y00 + j*dy;
    zp_(i,j,k) = z00 + k*dz;
  }

 exit:
  return error;
}

/******************************************************************************
Title: ReadGrid2
Description: Reads and initialized type 2 grids (nonuniform rectangular).
Return: Error code (see reader.h).
*/
int ReadGrid2() {
  int error = 0;
  int i,j,k,m,nMax;
  double* point;
  int increasing, decreasing;

  nMax = (nx_g>ny_g) ? nx_g : ny_g;
  if(nMax < nz_g) nMax = nz_g;
  point = (double*) malloc((nMax+1)*sizeof(double));
  if(point == NULL) ABORT(ERR_MEMORY);

  beginList();
  for(i=0; i<=nx_g; i++)
    packDoubleUnits(point[i],DIMEN_LENGTH);
  endList();
  unPackDouble();
  m = 0;
  for(k=0; k<=nz; k++)  
  for(j=0; j<=ny; j++) 
  for(i=0; i<=nx; i++) {
    xp[m] = point[xOffset_g_(0) + i]; m++;
  }
  increasing = 1;
  for(m=0; m<nx; m += 1) {
    if(xp[m] >= xp[m + 1]) {
      increasing = 0;
      break;
    }
  }
  decreasing = 1;
  for(m=0; m<nx; m += 1) {
    if(xp[m] <= xp[m + 1]) {
      decreasing = 0;
      break;
    }
  }
  if(!increasing && !decreasing) ABORT(-4);

  beginList();
  for(j=0; j<=ny_g; j++)
    packDoubleUnits(point[j],DIMEN_LENGTH);
  endList();
  unPackDouble();
  m = 0;
  for(k=0; k<=nz; k++)  
  for(j=0; j<=ny; j++) 
  for(i=0; i<=nx; i++) {
    yp[m] = point[yOffset_g_(0) + j]; m++;
  }
  increasing = 1;
  for(m=0; m<(nx+1)*ny; m += (nx+1)) {
    if(yp[m] >= yp[m + nx + 1]) {
      increasing = 0;
      break;
    }
  }
  decreasing = 1;
  for(m=0; m<(nx+1)*ny; m += (nx+1)) {
    if(yp[m] <= yp[m + nx + 1]) {
      decreasing = 0;
      break;
    }
  }
  if(!increasing && !decreasing) ABORT(-4);

  beginList();
  for(k=0; k<=nz_g; k++)
    packDoubleUnits(point[k],DIMEN_LENGTH);
  endList();
  unPackDouble();
  m = 0;
  for(k=0; k<=nz; k++)  
  for(j=0; j<=ny; j++) 
  for(i=0; i<=nx; i++) {
    zp[m] = point[zOffset_g_(0) + k]; m++;
  }
  increasing = 1;
  for(m=0; m<(nx+1)*(ny+1)*nz; m += (nx+1)*(ny+1)) {
    if(zp[m] >= zp[m + (nx+1)*(ny+1)]) {
      increasing = 0;
      break;
    }
  }
  decreasing = 1;
  for(m=0; m<(nx+1)*(ny+1)*nz; m += (nx+1)*(ny+1)) {
    if(zp[m] <= zp[m + (nx+1)*(ny+1)]) {
      decreasing = 0;
      break;
    }
  }
  if(!increasing && !decreasing) ABORT(-4);

 exit:
  free(point);
  return(error);
}

/******************************************************************************
Title: ReadGrid3
Description: Reads and initialized type 3 grids (nonuniform rectangular
             in x-y, variable in z).
Return: Error code (see reader.h).
*/
int ReadGrid3() {
  int error = 0;
  int i,j,k,m,nMax;
  double* point;
  int increasing, decreasing;

  nMax = (nx_g>ny_g) ? nx_g : ny_g;
  point = (double*) malloc((nMax+1)*sizeof(double));
  if(point == NULL) ABORT(ERR_MEMORY);

  beginList();
  for(i=0; i<=nx_g; i++)
    packDoubleUnits(point[i],DIMEN_LENGTH);
  endList();
  unPackDouble(); m = 0;
  for(k=0; k<=nz; k++)  
  for(j=0; j<=ny; j++) 
  for(i=0; i<=nx; i++) {
    xp[m] = point[xOffset_g_(0) + i]; m++;
  }
  increasing = 1;
  for(m=0; m<nx; m += 1) {
    if(xp[m] >= xp[m + 1]) {
      increasing = 0;
      break;
    }
  }
  decreasing = 1;
  for(m=0; m<nx; m += 1) {
    if(xp[m] <= xp[m + 1]) {
      decreasing = 0;
      break;
    }
  }
  if(!increasing && !decreasing) ABORT(-4);

  beginList();
  for(j=0; j<=ny_g; j++)
    packDoubleUnits(point[j],DIMEN_LENGTH);
  endList();
  unPackDouble(); m = 0;
  for(k=0; k<=nz; k++)  
  for(j=0; j<=ny; j++) 
  for(i=0; i<=nx; i++) {
    yp[m] = point[yOffset_g_(0) + j]; m++;
  }
  increasing = 1;
  for(m=0; m<(nx+1)*ny; m += (nx+1)) {
    if(yp[m] >= yp[m + nx + 1]) {
      increasing = 0;
      break;
    }
  }
  decreasing = 1;
  for(m=0; m<(nx+1)*ny; m += (nx+1)) {
    if(yp[m] <= yp[m + nx + 1]) {
      decreasing = 0;
      break;
    }
  }
  if(!increasing && !decreasing) ABORT(-4);

  free(point);
  {
    double units;
    int dimen[] = {DIMEN_LENGTH};
    beginList();
    error = getUnits(&units,dimen); if(error) goto exit;
    error = read3DDouble(zp, units, 1,1,1, 1, 0,0,0); if(error) goto exit;
    endList();
  }

 exit:
  return(error);
}

/******************************************************************************
Title: ReadGrid4
Description: Reads and initialized type 4 grids (fully logically rectangular).
Return: Error code (see reader.h).
*/
int ReadGrid4() {
  int error = 0;
  double units;
  int dimen[] = {DIMEN_LENGTH};

  error = getUnits(&units,dimen); if(error) goto exit;
  error = read3DDouble(xp, units, 1,1,1, 1, 0,0,0); if(error) goto exit;
  endList();

  beginList();
  error = getUnits(&units,dimen); if(error) goto exit;
  error = read3DDouble(yp, units, 1,1,1, 1, 0,0,0); if(error) goto exit;
  endList();

  beginList();
  error = getUnits(&units,dimen); if(error) goto exit;
  error = read3DDouble(zp, units, 1,1,1, 1, 0,0,0); if(error) goto exit;
  endList();

 exit:
  return error;
}

/******************************************************************************
Title: readGrid
Description: Read and preprocess the grid input data
Return: Error code (see reader.h).
*/
int readGrid() {
  static char GRID[] = {"GRID"};

  int error = 0;
  char string[21];
  int i;

  skipToSection(GRID);

  packInt(nx_g);
  packInt(ny_g);
  packInt(nz_g);
  packInt(periodicBC[0]);
  packInt(periodicBC[1]);
  packInt(periodicBC[2]);
  unPackInt();
  for(i=0; i<3; i++) {
    if(periodicBC[i] && nDom[i] <= 1) ABORT(-5);
  }

  defineDomainDecomp();

  packInt(zIsDepth);

  if(callTransport) {
    if( nx < padx+1 && (myNbr_(-1,0,0) >= 0 || myNbr_(1,0,0) >= 0) ) ABORT(-1);
    if( ny < pady+1 && (myNbr_(0,-1,0) >= 0 || myNbr_(0,1,0) >= 0) ) ABORT(-2);
    if( nz < padz+1 && (myNbr_(0,0,-1) >= 0 || myNbr_(0,0,1) >= 0) ) ABORT(-3);
  }

  xp = (double*) malloc((nx+1)*(ny+1)*(nz+1)*sizeof(double));
  yp = (double*) malloc((nx+1)*(ny+1)*(nz+1)*sizeof(double));
  zp = (double*) malloc((nx+1)*(ny+1)*(nz+1)*sizeof(double));
  if(xp == NULL || yp  == NULL || zp == NULL) ABORT(ERR_MEMORY);

  readString(string,21);
  if(!strcmp(string,"uniform")) {
    identityMap = 1;
    error = ReadGrid1(); if(error) goto exit;
  }
  else if(!strcmp(string,"rectangular")) {
    identityMap = 1;
    error = ReadGrid2(); if(error) goto exit;
  }
  else if(!strcmp(string,"xy-rectangular")) {
    identityMap = 0;
    diagMapPerm = 0;
    error = ReadGrid3(); if(error) goto exit;
  }
  else if(!strcmp(string,"{")) {
    identityMap = 0;
    diagMapPerm = 0;
    error = ReadGrid4(); if(error) goto exit;
  }
  else ABORT(ERR_UNSUPPORTED_OPTION);

  unPackInt();
  if(!zIsDepth) for(i=0; i<(nx+1)*(ny+1)*(nz+1); i++) zp[i] = -zp[i];

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case -1:
      fprintf(stderr,"Padx too big\n");
      error = ERR_BEFORE_LINE;
      break;
    case -2:
      fprintf(stderr,"Pady too big\n");
      error = ERR_BEFORE_LINE;
      break;
    case -3:
      fprintf(stderr,"Padz too big\n");
      error = ERR_BEFORE_LINE;
      break;
    case -4:
      fprintf(stderr,"Improper grid (non-monotone)\n");
      error = ERR_BEFORE_LINE;
      break;
    case -5:
      fprintf(stderr,
	"Must use >= 2 processors in each direction with periodic BC's\n");
      error = ERR_ON_LINE;
      break;
    }
  }
  return error;
}
