#include <stdio.h>
#include <stdlib.h>
#include "read.h"
#include "msgs.h"
#include "ftn.h"
#include "debug.h"

#include "driverGlobals.h"

#define ERR_LEAK_OUTSIDE_DOMAIN -1
static int incorrectLeak = -1;

#define min(a,b) ((a)<(b) ? (a) : (b))
#define max(a,b) ((a)>(b) ? (a) : (b))

double ElementVol(float*,float*,float*,float*,float*,float*,float*,float*);

/******************************************************************************
Title: localizeLeaks
Description: Localizes the leaks (turn global info into local info)
*/

static int localizeLeaks() {
  int error = 0;
  int leak,m;

  /* localize indices (and order lo to hi if necessary) */

  for(leak=0; leak<nLeaks; leak++) {
    for(m=0; m<3; m++) {
      if(leakIndexLo_(m,leak) > leakIndexHi_(m,leak)) {
	int ind = leakIndexLo_(m,leak);
	leakIndexLo_(m,leak) = leakIndexHi_(m,leak);
	leakIndexHi_(m,leak) = ind;
      }
    }

    if( leakIndexLo_(0,leak) < 1 || leakIndexHi_(0,leak) > nx_g ||
	leakIndexLo_(1,leak) < 1 || leakIndexHi_(1,leak) > ny_g ||
	leakIndexLo_(2,leak) < 1 || leakIndexHi_(2,leak) > nz_g ) {
      incorrectLeak = leak + 1;
      ABORT(ERR_LEAK_OUTSIDE_DOMAIN);
    }

    leakIndexLo_(0,leak) = leakIndexLo_(0,leak) - xOffset_g_(0);
    leakIndexLo_(1,leak) = leakIndexLo_(1,leak) - yOffset_g_(0);
    leakIndexLo_(2,leak) = leakIndexLo_(2,leak) - zOffset_g_(0);
    leakIndexHi_(0,leak) = leakIndexHi_(0,leak) - xOffset_g_(0);
    leakIndexHi_(1,leak) = leakIndexHi_(1,leak) - yOffset_g_(0);
    leakIndexHi_(2,leak) = leakIndexHi_(2,leak) - zOffset_g_(0);
  }

 exit:
  return error;
}

/******************************************************************************
Title: normalizeLeakRates
Description: Divide Leak rates by total volume
*/

static int normalizeLeakRates() {
  int error = 0;
  int n,leak,specie,i,j,k;

  for(leak=0; leak<nLeaks; leak++) {
    int iLo = max(0,leakIndexLo_(0,leak)-1);
    int jLo = max(0,leakIndexLo_(1,leak)-1);
    int kLo = max(0,leakIndexLo_(2,leak)-1);
    int iHi = min(nx,leakIndexHi_(0,leak));
    int jHi = min(ny,leakIndexHi_(1,leak));
    int kHi = min(nz,leakIndexHi_(2,leak));
    double vol=0;

    for(k=kLo; k<kHi; k++)
    for(j=jLo; j<jHi; j++)
    for(i=iLo; i<iHi; i++) {
      float pt0[3], pt1[3], pt2[3], pt3[3], pt4[3], pt5[3], pt6[3], pt7[3];

      pt0[0] = xp_(i,j,k);
      pt0[1] = yp_(i,j,k);
      pt0[2] = zp_(i,j,k);
      pt1[0] = xp_(i+1,j,k);
      pt1[1] = yp_(i+1,j,k);
      pt1[2] = zp_(i+1,j,k);
      pt2[0] = xp_(i+1,j+1,k);
      pt2[1] = yp_(i+1,j+1,k);
      pt2[2] = zp_(i+1,j+1,k);
      pt3[0] = xp_(i,j+1,k);
      pt3[1] = yp_(i,j+1,k);
      pt3[2] = zp_(i,j+1,k);
      pt4[0] = xp_(i,j,k+1);
      pt4[1] = yp_(i,j,k+1);
      pt4[2] = zp_(i,j,k+1);
      pt5[0] = xp_(i+1,j,k+1);
      pt5[1] = yp_(i+1,j,k+1);
      pt5[2] = zp_(i+1,j,k+1);
      pt6[0] = xp_(i+1,j+1,k+1);
      pt6[1] = yp_(i+1,j+1,k+1);
      pt6[2] = zp_(i+1,j+1,k+1);
      pt7[0] = xp_(i,j+1,k+1);
      pt7[1] = yp_(i,j+1,k+1);
      pt7[2] = zp_(i,j+1,k+1);

      vol += ElementVol(pt0,pt1,pt2,pt3,pt4,pt5,pt6,pt7);
    }
    MSG_Gdsum1(&vol,comm);

    for(specie=0; specie<nSpecies; specie++) {
      for(n=0; n<maxnLeakInterp; n++) {
	leakRate_(1,n,specie,leak) /= vol;
      }
    }
  }

  return error;
}

/******************************************************************************
Title: readLeaks
Description: Read and preprocess the leaking contaminant source data
*/

int readLeaks() {
  static char LEAKS[] = {"LEAKS"};
  static char Leak[] = {"Leak"};

  int i,j,k;
  int error = 0;

  skipToSection(LEAKS);
  packInt(nLeaks);
  packInt(maxnLeakInterp);
  unPackInt();
  if(nLeaks <= 0) { nLeaks = 0; return 0; }

  if(nLeaks) {
    if(maxnLeakInterp < 0) maxnLeakInterp = 0;

    leakIndexLo = (int*) malloc(3*nLeaks*sizeof(int));
    leakIndexHi = (int*) malloc(3*nLeaks*sizeof(int));
    if(leakIndexLo == NULL || leakIndexHi == NULL) ABORT(ERR_MEMORY);

    if(nSpecies) {
      nLeakInterp = (int*) malloc(nSpecies*nLeaks*sizeof(int));
      if(nLeakInterp == NULL) ABORT(ERR_MEMORY);
      for(i=0; i<nSpecies*nLeaks; i++) nLeakInterp[i] = 0;
      if(maxnLeakInterp) {
	leakRate = (float*)
	  malloc(2*maxnLeakInterp*nSpecies*nLeaks*sizeof(float));
	if(leakRate == NULL) ABORT(ERR_MEMORY);
	for(i=0; i<2*maxnLeakInterp*nSpecies*nLeaks; i++) leakRate[i] = 0;
      }
    }
  }

  for(i=0; i<nLeaks; i++) {
    readSubSection(0);
    skipToSubSection(Leak);

    packInt(leakIndexLo_(0,i));
    packInt(leakIndexLo_(1,i));
    packInt(leakIndexLo_(2,i));
    packInt(leakIndexHi_(0,i));
    packInt(leakIndexHi_(1,i));
    packInt(leakIndexHi_(2,i));

    beginList();
    for(j=0; j<nSpecies; j++) {
      packInt(nLeakInterp[j + nSpecies*i]);
      unPackInt();
      if(nLeakInterp[j + nSpecies*i] > maxnLeakInterp) ABORT(ERR_ON_LINE);
      beginList();
      for(k=0; k< nLeakInterp[j + nSpecies*i]; k++) {
	packFloatUnits(leakRate_(0,k,j,i),DIMEN_TIME);
	packFloatUnits(leakRate_(1,k,j,i),DIMEN_PER_TIME); /* moles/time */
      }
      endList();
    }
    endList();
    unPackFloat();
    for(j=0; j<nSpecies; j++)
    for(k=0; k<nLeakInterp[i]; k++) {
      if(leakRate_(1,k,j,i) < 0) leakRate_(1,k,j,i) = -leakRate_(1,k,j,i);
    }
  }
  unPackInt();

  error = localizeLeaks(); if(error) goto exit;
  error = normalizeLeakRates(); if(error) goto exit;

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case ERR_LEAK_OUTSIDE_DOMAIN:
      fprintf(stderr,"Leak %d outside domain\n",incorrectLeak);
      error = ERR_BEFORE_LINE;
      break;
    default:
      break;
    }
  }
  return error;
}
