#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "msgs.h"
#include "read.h"
 
#include "driverGlobals.h"
#include "typeNumbers.h"

#define max(x,y) ((x)>(y)) ? (x) : (y)
#define min(x,y) ((x)<(y)) ? (x) : (y)

#define STRING_IS(str) !strcmp(string,str)

/******************************************************************************
Title: readMaterials
Description: Read and preprocess the porous medium material input data.
Return: Error code (see reader.h).

The permeability is stored as a symmetric tensor
  On the (i,j,k) cell: perm_(PERM_XX,i,j,k) is perm_xx, etc.
*/
int readMaterials(int *sorpTypePorosity, int *nSorpTypes_in) {
  static char MATERIAL_PROPERTIES[] = {"MATERIAL PROPERTIES"};
  static char* PermHeading[] = {"Permeability","Conductivity"};
  static char Dispersion[] = {"Dispersion"};
  static char Sorption[] = {"Sorption"};

  int dimen_perm[] = {DIMEN_AREA};
  int dimen_cond[] = {DIMEN_VELOCITY};
  int* dimenPerm = dimen_perm;

  int error = 0;
  int nCells = nx*ny*nz;
  int i,j,k,m,n,nearly,keyRead;
  char string[21];

  int byComponents = 0;
  int numberComponents;
  int order[6] = {0,0,0,0,0,0};

  permType = 0;
  *sorpTypePorosity = 0;

  skipToSection(MATERIAL_PROPERTIES);
  packDoubleUnits(gravity,DIMEN_ACCEL);

  porosity = (double*) malloc(nCells*sizeof(double));
  perm = (double*) malloc(6*nCells*sizeof(double));
  if(porosity == NULL || perm == NULL) ABORT(ERR_MEMORY);
  for(i=0; i<6*nCells; i++) perm[i] = 0.0;

  /* POROSITY */
  nearly = 0;
  {
    int dimen[] = {DIMEN_NUMBER};
    error = setup3DDouble(porosity, dimen,
			  0,0,0, 1, 0,0,0, string, 21, &nearly);
  }
  if(error) goto exit;

  /* PERMEABILITY */
  if(computeFlow) {
    readSubSection(0);
    skipToSubSections(PermHeading,2,&keyRead);
    if(keyRead == 0) {
      permType_in = PERM_TYPE_PERMEABILITY;
    } else if(keyRead == 1) {
      permType_in = PERM_TYPE_CONDUCTIVITY;
      dimenPerm = dimen_cond;
    } else ABORT(-2);

    readString(string,21);
    if(STRING_IS("scalar")) {
      permType = PERM_SCALAR;
      numberComponents = 1;
    } else if(STRING_IS("diagonal")) {
      permType = PERM_DIAGONAL;
      numberComponents = 3;
    } else if(STRING_IS("symmetric")) {
      permType = PERM_SYMMETRIC;
      diagMapPerm = 0;
      numberComponents = 6;
    } else if(STRING_IS("face")) {
      permType = PERM_FACE;
      numberComponents = 3;
      facePerm = (double*) malloc( ((nx+1)*ny*nz + nx*(ny+1)*nz + nx*ny*(nz+1))
				   *sizeof(double) );
      if(facePerm == NULL) ABORT(ERR_MEMORY);
    } else ABORT(ERR_UNSUPPORTED_OPTION);

    /* determine data layout */
    if(permType == PERM_SCALAR || permType == PERM_FACE) {
      byComponents = 1;
    } else {
      readString(string,21);
      if(STRING_IS("by")) readString(string,21);
      if(STRING_IS("components")) {
	byComponents = 1;
      } else if(STRING_IS("cells")) {
      } else ABORT(ERR_UNSUPPORTED_OPTION);

      for(n=0; n<numberComponents; n++) {
	readString(string,21);
	if(STRING_IS("xx")) { order[n] = PERM_XX;
	} else if(STRING_IS("yy")) { order[n] = PERM_YY;
	} else if(STRING_IS("zz")) { order[n] = PERM_ZZ;
	} else if(numberComponents == 3) { ABORT(ERR_UNSUPPORTED_OPTION);
	} else if(STRING_IS("xy")) { order[n] = PERM_XY;
	} else if(STRING_IS("xz")) { order[n] = PERM_XZ;
	} else if(STRING_IS("yz")) { order[n] = PERM_YZ;
	} else { ABORT(ERR_UNSUPPORTED_OPTION);
	}
      }
      for(n=0; n<numberComponents; n++)
	for(m=n+1; m<numberComponents; m++) {
	  if(order[n] == order[m]) ABORT(-1);
	}
    }

    /* read perm */
    if(permType == PERM_FACE) {
      nearly = 0;
      error = setup3DDouble(facePerm, dimenPerm,
			    1,0,0, 1, 0,0,0, string, 21, &nearly);
      if(error) goto exit;
      nearly = 0;
      error = setup3DDouble(&facePerm[(nx+1)*ny*nz], dimenPerm,
			    0,1,0, 1, 0,0,0, string, 21, &nearly);
      if(error) goto exit;
      nearly = 0;
      error = setup3DDouble(&facePerm[(nx+1)*ny*nz + nx*(ny+1)*nz], dimenPerm,
			    0,0,1, 1, 0,0,0, string, 21, &nearly);
      if(error) goto exit;
      for(k=0; k<nz; k++)
      for(j=0; j<ny; j++)
      for(i=0; i<nx; i++) {
	perm_(PERM_XX,i,j,k) = (facePermX_(i+1,j,k) + facePermX_(i,j,k))/2;
	perm_(PERM_YY,i,j,k) = (facePermY_(i,j+1,k) + facePermY_(i,j,k))/2;
	perm_(PERM_ZZ,i,j,k) = (facePermZ_(i,j,k+1) + facePermZ_(i,j,k))/2;
      }
    } else if(byComponents) {
      double* permComponent = (double*) malloc(nCells*sizeof(double));
      if(permComponent == NULL) ABORT(ERR_MEMORY); 

      for(n=0; n<numberComponents; n++) {
	nearly = 0;
	error = setup3DDouble(permComponent, dimenPerm,
			      0,0,0, 1, 0,0,0, string, 21, &nearly);
	if(error) goto exit;

	if(permType == PERM_SCALAR) {
	  m = 0;
	  for(k=0; k<nz; k++)
	  for(j=0; j<ny; j++)
	  for(i=0; i<nx; i++) {
	    perm_(PERM_XX,i,j,k) = permComponent[m];
	    perm_(PERM_YY,i,j,k) = permComponent[m];
	    perm_(PERM_ZZ,i,j,k) = permComponent[m];
	    m++;
	  }
	} else {
	  m = 0;
	  for(k=0; k<nz; k++)
	  for(j=0; j<ny; j++)
	  for(i=0; i<nx; i++) {
	    perm_(order[n],i,j,k) = permComponent[m];
	    m++;
	  }
	}
      }
      free(permComponent);
    } else {
      double* permCells =
	(double*) malloc(nCells*numberComponents*sizeof(double));
      if(permCells == NULL) ABORT(ERR_MEMORY); 

      nearly = 0;
      error = setup3DDouble(permCells, dimenPerm,
			    0,0,0, numberComponents, 0,0,0,
			    string, 21, &nearly);
      if(error) goto exit;

      m = 0;
      for(k=0; k<nz; k++)
      for(j=0; j<ny; j++)
      for(i=0; i<nx; i++) {
	for(n=0; n<numberComponents; n++) {
	  perm_(order[n],i,j,k) = permCells[n + numberComponents*m];
	}
	m++;
      }
      free(permCells);
    }

    if(permType == PERM_SCALAR || permType == PERM_DIAGONAL) {
      for(k=0; k<nz; k++)
      for(j=0; j<ny; j++)
      for(i=0; i<nx; i++) {
	perm_(PERM_XY,i,j,k) = 0;
	perm_(PERM_XZ,i,j,k) = 0;
	perm_(PERM_YZ,i,j,k) = 0;
      }
    }
  }

  /* Dispersion */
  if(computeTransport) {
    readSubSection(0);
    skipToSubSection(Dispersion);
    packInt(uniformDispersion);
    unPackInt();
    if(uniformDispersion) {
      packDoubleUnits(molDiff,DIMEN_DIFFUSION);
      packDoubleUnits(longDisp,DIMEN_LENGTH);
      packDoubleUnits(transDisp,DIMEN_LENGTH);
    } else {
      /* change molDiff, longDisp, transDisp in order to foul the "if 
         statement" which will skip dispersion calualation if 
         molDiff = 0, longDisp = 0 and transDisp = 0;  */
      molDiff = 1;
      longDisp = 1;
      transDisp = 1;

      molDiffAry = (double*) malloc(nCells*sizeof(double));
      longDispAry = (double*) malloc(nCells*sizeof(double));
      transDispAry = (double*) malloc(nCells*sizeof(double));
      if(molDiffAry == NULL || longDispAry == NULL || transDispAry == NULL) 
        ABORT(ERR_MEMORY);

      {
      int dimen[] = {DIMEN_DIFFUSION};
      nearly = 0;
      error = setup3DDouble(molDiffAry, dimen,
                            0,0,0, 1, 0,0,0, string, 21, &nearly);
      if(error) goto exit;
      }
      {
      int dimen[] = {DIMEN_LENGTH};
      nearly = 0;
      error = setup3DDouble(longDispAry, dimen,
                            0,0,0, 1, 0,0,0, string, 21, &nearly);
      if(error) goto exit;
      nearly = 0;
      error = setup3DDouble(transDispAry, dimen,
                            0,0,0, 1, 0,0,0, string, 21, &nearly);
      if(error) goto exit;
      }
    }
  }

  /* Sorption */
  if(callTransport) {
    readSubSection(0);
    skipToSubSection(Sorption);

     /* SORPTION TYPES
       Type 0 = proportional to porosity or not subject to sorption
       1 = type1
       2 = type2, etc.
       nSorpTypes does not include type 0 internal to the code
    */
    packInt(nSorpTypes);
    unPackInt();
    if(nSorpTypes <= 0) ABORT(ERR_ON_LINE);
    *nSorpTypes_in = nSorpTypes;
    sorp = (double*) malloc(nx*ny*nz*nSorpTypes*sizeof(double));
    if(sorp == NULL) ABORT(ERR_MEMORY);
    j = 0;
    beginList();
    for(i=0; i<nSorpTypes; i++) {
      nearly = 0;
      {
	int dimen[] = {DIMEN_NUMBER};
	error = setup3DDouble(&sorp[j*nx*ny*nz], dimen, 0,0,0, 1, 0,0,0,
			      string, 21, &nearly);
      }
      if(!error) {
	j++;
      } else if(error == ERR_IN_GRID_ARRAY
		&& !strcmp(string,"porosity") && !nearly) {
	if(*sorpTypePorosity) ABORT(-11);
	*sorpTypePorosity = j+1;
	error = 0;
      } else goto exit;
    }
    endList();
    nSorpTypes = j;
    if(nSorpTypes) {
      sorp = (double*) realloc((void*)sorp,nx*ny*nz*nSorpTypes*sizeof(double));
    } else {
      sorp = (double*) realloc((void*)sorp,1*sizeof(double));
    }
    if(sorp == NULL) ABORT(ERR_MEMORY);
  }

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case -1:
      fprintf(stderr,"Duplicate option not allowed\n");
      error = ERR_BEFORE_LINE;
      break;
    case -2:
      fprintf(stderr,"Unsupported permeability type\n");
      error = ERR_ON_LINE;
      break;
    case -11:
      fprintf(stderr,"Sorption \"porosity\" appears more than once\n");
      error = ERR_ON_LINE;
      break;
    default:
      error = ERR_BEFORE_LINE;
    }
  }
  return error;
}

void internalPerm() {
  if(permType_in != PERM_TYPE_PERMEABILITY) {
    int n;
    double constant = phaseViscosity/(phaseDensity*gravity);
    for(n=0; n<6*nx*ny*nz; n++) perm[n] *= constant;
    if(permType == PERM_FACE) {
      for(n=0; n<(nx+1)*ny*nz+nx*(ny+1)*nz+nx*ny*(nz+1); n++)
	facePerm[n] *= constant;
    }
  }
}
