#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "read.h"
#include "msg.h"
#include "expr.h"

#include "driverGlobals.h"

/******************************************************************************
Title: readRND
Description: Read and preprocess the radio-nuclide decay data.
Return: Error code (see reader.h).
*/
int readRND() {
  static char RADIONUCLIDE_DECAY[]
    = {"RADIONUCLIDE DECAY"};
  static char Chain[] = {"Chain"};

  int i,j,k;
  int error = 0;
  double ln2 = log(2);
  int chain,nChains, chainLen, specie;
  int iOffset = 1, rOffset = 0;
  char chainType[7];
  int branching;
  int* specieAppeared;
  double gramsPerMass = 0;

  /* READ RND
     Format of iParmRND:
     (1) number of chains
     (2) for each chain:
         number of Species in chain
         the Species in the chain (in generational order), counting from 1
     Format of rParmRND:
       For each chain:
       (1) atomic numbers for each specie
       (2) half-lives for each specie (converted to negative ODE
           decay constants below)
       (3) branching ratios (default values if linear)
           for each specie in the chain (in generational order)
           in a [number of species in chain]^2 matrix
           (these with the half-lives are converted to the ODE decay
           coefficients below)
  */

  skipToSection(RADIONUCLIDE_DECAY);

  packInt(nChains);
  unPackInt();
  computeRND = (nChains > 0);
  niParmRND = (computeRND) ? 1 : 0;
  nrParmRND = 0;

  if(!myProc) {
    gramsPerMass = expr_gramsPerMass();
  }
  MPI_Bcast((void*)&gramsPerMass,1,MPI_DOUBLE,0,comm);

  for(i=0; i<nChains; i++) {
    readSubSection(0);
    skipToSubSection(Chain);
    packInt(chainLen);
    readString(chainType,7);
    unPackInt();

    niParmRND += chainLen + 1;
    nrParmRND += chainLen*(2 + chainLen);
    iParmRND = (int*) realloc(iParmRND,niParmRND*sizeof(int));
    rParmRND = (double*) realloc(rParmRND,nrParmRND*sizeof(double));
    if(rParmRND == NULL || iParmRND == NULL) ABORT(ERR_MEMORY);
    iParmRND[iOffset] = chainLen;
    iOffset++;
    for(k=0; k<chainLen*chainLen; k++) {
      rParmRND[rOffset + 2*chainLen + k] = 0;
    }

    branching = !strcmp(chainType,"branch");
    beginList();
    for(j=0; j<chainLen; j++) {
      packInt(iParmRND[iOffset + j]); /* specie number */
      unPackInt();
      specie = iParmRND[iOffset + j];
      if( specie > nSpecies || specie <= 0 ) ABORT(-1);
      rParmRND[rOffset + j]
	= molecularWeight[specie-1]*gramsPerMass; /* atomic number */
      packDoubleUnits(rParmRND[rOffset + chainLen + j],DIMEN_TIME);
                                        /* half life */
      if(branching) { /* branch ratios of previous species into this one */
	for(k=0; k<j; k++) {
	  packDoubleUnits(rParmRND[rOffset + chainLen*(2+k) + j],DIMEN_NUMBER);
	}
      } else if(j > 0) rParmRND[rOffset + chainLen*(1+j) + j] = 1;
    }
    endList();
    iOffset += chainLen;
    rOffset += chainLen*(2 + chainLen);
    unPackDouble();
  }
  iParmRND[0] = nChains;

  unPackInt();

  /* PROCESS RND DATA
     Check consistency and change to decay/creation constants
     Require legal specie numbers, in one chain only, and positive
       atomic numbers, half-lives, and branching ratios (that sum to <= 1)
  */
  iOffset = 1;
  rOffset = 0;
  specieAppeared = (int*) malloc((nSpecies+1)*sizeof(int));
  for(i=0; i<=nSpecies; i++) specieAppeared[i] = 0;

  if(specieAppeared == NULL) ABORT(ERR_MEMORY);
  for(chain=0; chain<nChains; chain++) {
    chainLen = iParmRND[iOffset];
    iOffset++;
    /* legal species, in one chain only */
    for(i=0; i<chainLen; i++) {
      specie = iParmRND[iOffset+i];
      if( specie > nSpecies || specie <= 0 ) ABORT(-1);
      if(specieAppeared[specie]) ABORT(-2);
      specieAppeared[specie] = 1;
    }
    /* positive atomic numbers */
    for(i=0; i<chainLen; i++) {
      if(rParmRND[rOffset+i] <= 0) ABORT(-1);
    }
    rOffset += 2*chainLen;
    /* positive branching ratios that sum to <= 1 */
    for(i=0; i<chainLen-1; i++) {
      double temp = 0;
      for(j=i+1; j<chainLen; j++) {
	if(rParmRND[rOffset+j+chainLen*i] < 0) ABORT(-1);
	temp += rParmRND[rOffset+j+chainLen*i];
      }
      if(temp > 1) ABORT(-1);
    }
    rOffset += chainLen*chainLen;
    iOffset += chainLen;
  }
  free(specieAppeared);

  /* switch to decay constants (must be unique unless stable) */
  iOffset = 1;
  rOffset = 0;
  for(chain=0; chain<nChains; chain++) {
    chainLen = iParmRND[iOffset];
    rOffset += chainLen;
    for(i=0; i<chainLen; i++) {
      rParmRND[rOffset+i]
	= (rParmRND[rOffset+i] > 0) ? -ln2/rParmRND[rOffset+i] : 0;
      for(j=0; j<i-1; j++) {
	if ( rParmRND[rOffset+i]
	    && rParmRND[rOffset+i] == rParmRND[rOffset+j] ) ABORT(-3);
      }
    }
    rOffset += chainLen;
    for(i=0; i<chainLen; i++) {
      rParmRND[rOffset+i+chainLen*i] = rParmRND[rOffset-chainLen+i];
      for(j=i+1; j<chainLen; j++) {
	rParmRND[rOffset+j+chainLen*i]
	  = -rParmRND[rOffset+j+chainLen*i]*rParmRND[rOffset+i+chainLen*i];
      }
    }
    iOffset += chainLen + 1;
    rOffset += chainLen*chainLen;
  }

 exit:
  if(error < 0 && !myProc) {
    char msg_inconsistent[] = {"Inconsistent RND data"};

    switch(error) {
    case -1:
      fprintf(stderr,"%s\n",msg_inconsistent);
      error = ERR_BEFORE_LINE;
      break;
    case -2:
      fprintf(stderr,"%s: A specie can appear only once\n",msg_inconsistent);
      error = ERR_BEFORE_LINE;
      break;
    case -3:
      fprintf(stderr,"%s: Each specie must have a unique half-life\n",
	      msg_inconsistent);
      error = ERR_BEFORE_LINE;
      break;
    case -4:
      fprintf(stderr,"Incorrect number of Chains\n");
      error = ERR_ON_LINE;
      break;
    }
  }
  return error;
}
