#include <stdio.h>
#include "read.h"
#include "typeNumbers.h"

#include "driverGlobals.h"

/******************************************************************************
Title: distributeDomains
Description: Define the distribution of the domains in the three
             coordinate directions.  There is flexibility if
             nDom[?] <= 0.  Use nDom[2] = 1 if flexible.
Return: Error code (-1 if the requested decomposition is not possible).
*/
int distributeDomains() {
  int i,j;

  if(nDom[2] <= 0) nDom[2] = 1;

  if(nDom[0] <= 0 && nDom[1] <= 0) {
    for(i=1; i<=nProcs; i++)
    for(j=1; j<=i; j++) {
      if(i*j*nDom[2] == nProcs) {
	nDom[0] = i;
	nDom[1] = j;
	return 0;
      }
    }
    return -1;
  }
  if(nDom[0] <= 0) {
    for(i=1; i<=nProcs; i++) {
      if(i*nDom[1]*nDom[2] == nProcs) {
	nDom[0] = i;
	return 0;
      }
    }
    return -1;
  }
  if(nDom[1] <= 0) {
    for(j=1; j<=nProcs; j++) {
      if(nDom[0]*j*nDom[2] == nProcs) {
	nDom[1] = j;
	return 0;
      }
    }
    return -1;
  }

  if(nDom[0]*nDom[1]*nDom[2] == nProcs) return 0;
  return -1;
}

/******************************************************************************
Title: readSolutionParams
Description: Read and preprocess the solution parameter input data.
Return: Error code (see reader.h).
*/
int readSolutionParams() {
  static char SOLUTION_PARAMETERS[] = {"SOLUTION PARAMETERS"};
  static char Flow[] = {"Flow"};
  static char Transport[] = {"Transport"};
  static char Chemistry[] = {"Chemistry"};
  static char CMM[] = {"CMM"};

  int error = 0;
  int i;

  skipToSection(SOLUTION_PARAMETERS);

  dWorkspace_f = 0;
  dWorkspace_t = 0;
  dWorkspace_c = 0;
  iWorkspace_c = 0;

  packInt(nDom[0]);
  packInt(nDom[1]);
  packInt(nDom[2]);
  unPackInt();
  error = distributeDomains(); if(error) goto exit;

  if(computeFlow) {
    readSubSection(0);
    skipToSubSection(Flow);
    packInt(maxIterIF_f);
    packDoubleUnits(relTolIF_f,DIMEN_NUMBER);
    packDoubleUnits(absTolIF_f,DIMEN_NUMBER);
    packInt(pcTypeIF_f);
    unPackInt();
    switch(pcTypeIF_f) {
    case PC_IF_NOPRECOND_F:
    case PC_IF_DIAGONAL_F:
    case PC_IF_BALANCESTYLE_F:
      break;
    default:
      ABORT(ERR_UNSUPPORTED_OPTION);
    }
    packInt(dWorkspace_f);
    unPackInt();
    if(dWorkspace_f < 0) dWorkspace_f = 0;
  }

  if(computeTransport) {
    readSubSection(0);
    skipToSubSection(Transport);
    packInt(maxIter_t);
    packDoubleUnits(relTol_t,DIMEN_NUMBER);
    packInt(dWorkspace_t);
    unPackInt();
    if(dWorkspace_t < 0) dWorkspace_t = 0;
    packDoubleUnits(cflFactor_t,DIMEN_NUMBER);
    if(computeTransport == ADV_CMM) {
      readSubSubSection(0);
      skipToSubSubSection(CMM);
      packInt(ntCutMax_t);
      packInt(padx);
      packInt(pady);
      packInt(padz);
      packDoubleUnits(volTol_t,DIMEN_NUMBER);
      packInt(volRefine_t[0]);
      packInt(volRefine_t[1]);
      packInt(volRefine_t[2]);
      unPackInt();
      for(i=0; i<3; i++) if(volRefine_t[i] <= 0) volRefine_t[i] = 1;
    }
    switch(computeTransport) {
    case 0:
      break;
    case ADV_CMM:
      if(padx < 1) padx = 1;
      if(pady < 1) pady = 1;
      if(padz < 1) padz = 1;
      break;
    case ADV_HOG:
    case ADV_GODUNOV:
      padx = 2;
      pady = 2;
      padz = 2;
      break;
    case ADV_DIFF_ONLY:
      padx = 1;
      pady = 1;
      padz = 1;
      break;
    default:
      ABORT(ERR_IN_CODE);
    }
  }

  if(computeChemistry) {
    readSubSection(0);
    skipToSubSection(Chemistry);
    packInt(chmLoadBalFlag);
    packInt(chmMaxIter);
    packDoubleUnits(chmAbsTol,DIMEN_NUMBER);
    packDoubleUnits(chmEpsConc,DIMEN_MOLAR_CONC);
    packInt(chmScaleFlag);
    packInt(chmTestSolFlag);
    packInt(chmGuessType);
    packInt(chmInterpFlag);
    packInt(nViol);
    packDoubleUnits(chmAlpha,DIMEN_NUMBER);
    packInt(hessFlag);
    packDoubleUnits(tauMin,DIMEN_NUMBER);
    packDoubleUnits(chmRho,DIMEN_NUMBER);
    packInt(ntReact);
    packInt(odeAlgType);
    packInt(switchFlag);
    packInt(dWorkspace_c);
    packInt(iWorkspace_c);

    unPackInt();
    unPackDouble();
    if(nProcs == 1) chmLoadBalFlag = 0;
    if(chmMaxIter <= 0   || chmAbsTol <= 0.0 || chmEpsConc <= 0.0 ||
       chmLoadBalFlag < 0 || chmLoadBalFlag > 1 ||
       chmScaleFlag < 0  || chmScaleFlag > 1 || 
       chmInterpFlag < 0 || chmInterpFlag > 1 ||
       hessFlag < 0      || hessFlag > 1 ||
       odeAlgType < 0    || odeAlgType > 2 ||
       switchFlag < 0    || switchFlag > 1 ||
       chmTestSolFlag < 0|| chmTestSolFlag > 1 ||
       ntReact < 1       || nViol < 0 ||
       chmAlpha > 0.5    || chmRho >= 1.0 ||
       tauMin <= 0.0     || tauMin >= 1.0)  ABORT(-2);
    if(chmGuessType < 0) {
      chmGuessType = -1;
    } else if(chmGuessType > 0) {
      chmGuessType = 1;
    }
    if(dWorkspace_c < 0) dWorkspace_c = 0;
    if(iWorkspace_c < 0) iWorkspace_c = 0;
  }

 exit:
  switch(error) {
  case -1:
    fprintf(stderr,
	    "Cannot allocate %d processor(s) in a (%d,%d,%d) rectangular array\n",
	    nProcs,nDom[0],nDom[1],nDom[2]);
    error = ERR_ON_LINE;
    break;
  case -2:
    fprintf(stderr,"Invalid chemistry simulation parameter");
    error = ERR_BEFORE_LINE;
    break;
  }
  return error;
}
