#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "read.h"

#include "driverGlobals.h"
#include "typeNumbers.h"

#define MIN(i,j)             ((i<j) ? (i) : (j))

void blankPad(char* string,int size);

/******************************************************************************
Title: readSpecies
Description: Read and preprocess the chemical species data.
Return: Error code (see reader.h).
*/
int readSpecies(int sorpTypePorosity, int nSorpTypes_in) {
  static char CHEMICAL_SPECIES[] = {"CHEMICAL SPECIES"};
  static char Chemistry[] = {"Chemistry"};
  static char * componentTypes[]
    = {"Component", "H Component", "H2O Component"};
  static char Product[] = {"Product"};
  enum ComponentEnum {Component=0, HComponent, H2OComponent};

  int len;
  int error = 0;
  int i,comp, prod, keyRead, specie;
  int badPhase = 0;

  if(sorpTypePorosity == 0) sorpTypePorosity = nSorpTypes_in + 1;

  skipToSection(CHEMICAL_SPECIES);

  locH = 0;
  locWater = 0;

  packInt(nComps);
  packInt(nProducts);
  unPackInt();
  if(nComps < 0 || nProducts < 0) ABORT(-6);
  nSpecies = nComps + nProducts;
  if(nSpecies <= 0) ABORT(-8);

/* SLB 10 Feb 98   check for no products but general chemistry; 
                   catching it here avoids obscure input errors later */
  if(computeChemistry && nProducts < 1) ABORT(-14);

  specieName = (char(*)[SIZE_WORD])malloc(nSpecies*sizeof(char [SIZE_WORD]));
  molecularWeight  = (double*) malloc(nSpecies*sizeof(double));
  sorpType         = (int*) malloc(nSpecies*sizeof(int));
  chmSpecieName    = (char *) malloc((nSpecies*SpecieNameSZ+1)*sizeof(char));
  primaryPhase     = (int*) malloc(nSpecies*sizeof(int));
  mu0              = (double*) malloc(nSpecies*sizeof(double));
  chmPhaseIdentity = (int*) malloc(nSpecies*sizeof(int));
  phaseDist        = (double*) malloc((N_FLOWING_PHASES+1)*nSpecies*sizeof(double));
  nSpeciesInPhase  = (int *) malloc(chemPhases*sizeof(int));
  
  if(specieName == NULL || molecularWeight == NULL || sorpType == NULL ||
     primaryPhase == NULL || chmSpecieName == NULL ||
     phaseDist == NULL || nSpeciesInPhase == NULL || mu0 == NULL)
    ABORT(ERR_MEMORY);
  if(nComps > 0) {
    compCharge = (double*) malloc(nComps*sizeof(double));
    if(compCharge == NULL) ABORT(ERR_MEMORY);
  }
  if(nProducts > 0) {
    stoich = (double *) malloc(nProducts*nComps*sizeof(double));
    reactionType = (int *) malloc(nProducts*sizeof(int));
    peqk         = (double *) malloc(nProducts*sizeof(double));
    peqkX        = (double *) malloc(nProducts*sizeof(double));
    pkf          = (double *) malloc(nProducts*sizeof(double));
    pkb          = (double *) malloc(nProducts*sizeof(double));
    stoichK      = (double *) malloc(nProducts*nComps*sizeof(double));
    halfSatConst = (double *) malloc(nComps*nProducts*sizeof(double));
    if(stoich == NULL || reactionType == NULL || peqk == NULL ||
       peqkX == NULL || pkf == NULL || pkb == NULL ||
       stoichK == NULL || halfSatConst == NULL) ABORT(ERR_MEMORY);
    for(i = 0; i < nComps*nProducts; i++) halfSatConst[i] = 0;
  }

  nRQ = nRK = 0;
  for(i = 0; i < chemPhases; nSpeciesInPhase[i++] = 0);

  for(specie = 0; specie < nSpecies; specie++) {
    readSubSection(0);
    if(specie < nComps) {
      skipToSubSections(componentTypes,3,&keyRead);
    } else {
      skipToSubSection(Product);
    }

    readString(specieName[specie],SIZE_WORD);
    len = strlen(specieName[specie]);
    len = MIN(len,SpecieNameSZ);
    blankPad(&chmSpecieName_(0,specie),SpecieNameSZ);
    strncpy(&chmSpecieName_(0,specie),specieName[specie],len);

    packDoubleUnits(molecularWeight[specie],DIMEN_MASS);

    packDoubleUnits(phaseDist_(0,specie),DIMEN_NUMBER);
    unPackDouble();
    if(phaseDist_(0,specie) < 0) {
      primaryPhase[specie] = 0; /* immobile */
      phaseDist_(0,specie) = 0;
      phaseDist_(1,specie) = 0;
    } else {
      primaryPhase[specie] = 1; /* mobile */
      phaseDist_(1,specie) = 1;
    }

    if(phaseDist_(0,specie) == 0) {
      sorpType[specie] = 0;
    } else {
      packInt(sorpType[specie]);
      unPackInt();
      if(sorpType[specie] < 1 || sorpType[specie] > nSorpTypes_in) ABORT(-13);
      if(sorpType[specie] == sorpTypePorosity) {
	sorpType[specie] = 0;
      } else if(sorpType[specie] > sorpTypePorosity) {
	sorpType[specie]--;
      }
    }

    if(computeChemistry) {
      readSubSubSection(0);
      skipToSubSubSection(Chemistry);

      packInt(chmPhaseIdentity[specie]);
      unPackInt();
      if(chmPhaseIdentity[specie] < 1 || chmPhaseIdentity[specie] > chemPhases)
	ABORT(ERR_ON_LINE);
      nSpeciesInPhase[chmPhaseIdentity[specie]-1]++;

      if(specie < nComps) {
	comp = specie;
	packDoubleUnits(compCharge[specie],DIMEN_NUMBER);
	if(keyRead == HComponent) {
	  if(locH) ABORT(-7);
	  locH = comp+1;
	} else if(keyRead == H2OComponent) {
	  if(locWater) ABORT(-8);
	  locWater = comp+1;
	}
      } else {
	prod = specie - nComps;
	beginList();
	for(comp = 0; comp < nComps; comp++)
	  packDoubleUnits(stoich_(prod,comp),DIMEN_NUMBER);
	endList();
	packInt(reactionType[prod]);
	unPackInt();
	switch(reactionType[prod]) {
	case CHEM_RXN_EQ:
	  packDoubleUnits(peqk[prod],DIMEN_NUMBER);
	  /* Equil Prod: LOG10 of EQ Const.*/
	  nRQ++;
	  for(comp = 0; comp < nComps; comp++) stoichK_(prod,comp) = 0;
	  break;
	case CHEM_RXN_K_MASS_ACTION:
	  packDoubleUnits(pkf[prod],DIMEN_NUMBER);
	  packDoubleUnits(pkb[prod],DIMEN_NUMBER);
	  beginList();
	  for(comp = 0; comp < nComps; comp++)
	    packDoubleUnits(stoichK_(prod,comp),DIMEN_NUMBER);
	  endList();
	  unPackInt();
	  nRK++;
	  break;
	case CHEM_RXN_K_MONOD:
	  packDoubleUnits(pkf[prod],DIMEN_NUMBER);
	  packDoubleUnits(pkb[prod],DIMEN_NUMBER);
	  beginList();
	  for(comp = 0; comp < nComps; comp++)
	    packDoubleUnits(stoichK_(prod,comp),DIMEN_NUMBER);
	  endList();
	  unPackInt();
	  beginList();
	  for(comp = 0; comp < nComps; comp++)
	    packDoubleUnits(halfSatConst_(comp,prod),DIMEN_MOLAR_CONC);
	  endList();
	  unPackInt();
	  nRK++;
	  break;
	default:
	  ABORT(-2);
	}
      }
    }
  }

  /* check data */
  if(computeChemistry) {
    chemPhasePure = 0;

    for(i = 0; i < chemPhases; i++) {
      if(nSpeciesInPhase[i] == 0) {
        badPhase = i;
        ABORT(-12);
      }
    }
    
    for(i = 0; i < chemPhases; i++)
      if(nSpeciesInPhase[i] == 1) chemPhasePure++;
  
    chemPhaseMulti = chemPhases - chemPhasePure;

#if UNSF
    if(nRK != 0) ABORT(-3);
#endif
#if UNSF || SF
    if(locWater <= 0) ABORT(-4);
#endif
#if RNSF 
    if(chmTestSolFlag != 0) ABORT(-5);
#endif
  }

  /*************************************************************************

    POST-PROCESSING OF THE READ DATA

    Convert read equilibrium constants into a consistent 
    set of chemical reference potentials to be used in 
    minimization algorithms. 
     
    (1a) For equilibrium reactions, convert the given log10 
         equilibrium constants, peqk, from molar units to mole-fraction 
         based units , peqkX, (i.e. dimensionless). 
         Note that this conversion is not exact, unless the 
         molar density specified stays constant.

    (1b) For kinetic reactions, compute the corresponding 
         log10 backward and forward rates into the equilibrium
         constant that would result if the reaction was allowed
         to go to completion. This is simply peqk = pkf - pkb.
         Then proceed as in 1a.

    (2)  For all species, compute chemical reference-
         potentials COMPATIBLE with the equilibrium constants 
         calculated in 1a-1b. This procedure is descibed in 
         "Chemical reaction equilibrium analysis", 
         Smith&Missen,Wiley-Interscience,1982, pp. 214-215
 
         The chemical reference-potentials corresponding to
         components are set to zero. For each product species,
         we then set:

         mu0(nComps+i) = -peqkX(i)*ln(10)

  ****************************************************************************/

  if(computeChemistry) {
    unPackDouble();
    unPackInt();

    /* CONVERT FROM  peqk --> peqkX */

#if SF || UNSF
    for(prod = 0; prod < nProducts; prod++) {
      peqkX[prod] = (reactionType[prod] == 0) ? peqk[prod] : pkf[prod];
      peqkX[prod] -= log10(chmPhaseDensity[chmPhaseIdentity[nComps+prod]-1]);
    }
    for(comp = 0; comp < nComps; comp++) {
      double logrhoj;

      if(comp == locWater-1) continue;

      logrhoj = log10(chmPhaseDensity[chmPhaseIdentity[comp]-1]);
      for(prod = 0; prod < nProducts; prod++) {
	peqkX[prod] += stoich_(prod,comp)*logrhoj;
      }
    }
#elif RNSF
    /*  Uses molar variables; modify only if water is a component */

    for(prod = 0; prod < nProducts; prod++) {
      peqkX[prod] = (reactionType[prod] == 0) ? peqk[prod] : pkf[prod];
    }
    if(locWater > 0) {
      double logrhoj;
      comp = locWater - 1;
      logrhoj = log10(chmPhaseDensity[chmPhaseIdentity[comp]-1]);
      for(prod = 0; prod < nProducts; prod++) 
        peqkX[prod] -= stoich_(prod,comp)*logrhoj;
    }
#endif

    /* COMPUTE mu0 FROM peqkX  */
    for(comp = 0; comp < nComps; comp++)
      mu0[comp] = 0.0;
    for(prod = 0; prod < nProducts; prod++)
      mu0[prod+nComps] = -log(10.0)*peqkX[prod];
  }

  /* auto selection of initial guess */
  if(chmGuessType == CHEM_INIT_GUESS_AUTO) {
    chmGuessType = CHEM_INIT_GUESS_TRAN;
    for(prod = 0; prod < nProducts; prod++) {
      if(reactionType == CHEM_RXN_EQ) {
	chmGuessType = CHEM_INIT_GUESS_PREV;
	break;
      }
    }
  }

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case -1:
      fprintf(stderr,"Inconsistent chemical species/phase data\n");
      error = ERR_BEFORE_LINE;
      break;
    case -2:
      fprintf(stderr,"Illegal reaction type\n");
      error = ERR_ON_LINE;
      break;
    case -3:
      fprintf(stderr,"Kinetic reactions NOT supported with UNSF\n");
      error = ERR_ON_LINE;
      break;
    case -4:
      fprintf(stderr,"Include H2O Component with this chemistry version\n");
      error = ERR_ON_LINE;
      break;
    case -5:
      fprintf(stderr,"RNSF does NOT support test Solution\n");
      error = ERR_ON_LINE;
      break;
    case -6:
      fprintf(stderr,
              "Negative number of components or products not allowed\n");
      error = ERR_BEFORE_LINE;
      break;
    case -7:
      fprintf(stderr,"Location of H+ specified more than once\n");
      error = ERR_BEFORE_LINE;
      break;
    case -8:
      fprintf(stderr,"Location of H2O specified more than once\n");
      error = ERR_BEFORE_LINE;
      break;
    case -9:
      fprintf(stderr,"Number of species must be positive\n");
      error = ERR_ON_LINE;
      break;
    case -12:
      {
        char str[PhaseNameSZ+1];
        int i;
        strncpy(str,&phaseNames_(0,badPhase),PhaseNameSZ);
        for(i = PhaseNameSZ-1; i > 0; i--)
          if(str[i] != ' ') break;
        str[i+1] = '\0';
        fprintf(stderr,"Phase %d (%s) does not have any species\n",
                badPhase+1,str);
        error = ERR_BEFORE_LINE;
      }
      break;
    case -13:
      fprintf(stderr,"Sorption type out of range\n");
      error = ERR_BEFORE_LINE;
      break;
    case -14:
      fprintf(stderr,"General Chemistry selected, but no Product species found.\n");
      error = ERR_BEFORE_LINE;
      break;
    }
  }
  return error;
}
