#include <stdlib.h>
#include <stdio.h>
#include "read.h"
#include "typeNumbers.h"

#include "driverGlobals.h"

#define ERR_NEED_1_PROC_IN_Z -1
#define ERR_UNSUPPORTED_WELL -2
#define ERR_INVERTED_WELL -3
#define ERR_WELL_OUTSIDE_DOMAIN -4

static int incorrectWell;

/******************************************************************************
Title: localizeWells
Description: Localizes the wells (turn global info into local info)
             and removes inactive wells and wells outside the padded region
*/

#define MYWELL(i,j)  (-padx<(i) && (i)<=nx+padx && -pady<(j) && (j)<=ny+pady)

int localizeWells() {
  int error = 0;
  int well,well_l,i,j,m,specie;

  /* localize indices */

  for(well=0; well<nWells; well++) {
    if(wellType[well] == WELL_INACTIVE) continue;

    if(hWellIndex_(0,well) < 1 || hWellIndex_(0,well) > nx_g
       || hWellIndex_(1,well) < 1 || hWellIndex_(1,well) > ny_g) {
      incorrectWell = well + 1;
      ABORT(ERR_WELL_OUTSIDE_DOMAIN);
    }

    switch(wellType[well]) {
    case WELL_INJECTION:
    case WELL_REINJECTION:
      if(zWellIndexLo_(0,well) < 1 || zWellIndexHi_(0,well) > nz_g) {
	incorrectWell = well + 1;
	ABORT(ERR_WELL_OUTSIDE_DOMAIN);
      }
      if(zWellIndexLo_(0,well) > zWellIndexHi_(0,well)) {
	incorrectWell = well + 1;
	ABORT(ERR_INVERTED_WELL);
      }
      break;
    }
    switch(wellType[well]) {
    case WELL_EXTRACTION:
    case WELL_BOTTOMHOLE_PROD:
    case WELL_REINJECTION:
      if(zWellIndexLo_(1,well) < 1 || zWellIndexHi_(1,well) > nz_g) {
	incorrectWell = well + 1;
	ABORT(ERR_WELL_OUTSIDE_DOMAIN);
      }
      if(zWellIndexLo_(1,well) > zWellIndexHi_(1,well)) {
	incorrectWell = well + 1;
	ABORT(ERR_INVERTED_WELL);
      }
      break;
    }

    /* periodic case not implemented in its entirity
    if(period[0]) {
      if(xOffset_g_(0) = 0) {
	if(hWellIndex_(0,well) >= nx_g-padx)
	  hWellIndex_(0,well) = hWellIndex_(0,well) - nx_g;
      }
      if(xOffset_g_(0)+nx = nx_g) {
	if(hWellIndex_(0,well) <= padx)
	  hWellIndex_(0,well) = hWellIndex_(0,well) + nx_g;
      }
    }
    if(period[1]) {
      if(yOffset_g_(0) = 0) {
	if(hWellIndex_(1,well) >= ny_g-pady)
	  hWellIndex_(1,well) = hWellIndex_(1,well) - ny_g;
      }
      if(yOffset_g_(0)+ny = ny_g) {
	if(hWellIndex_(1,well) <= pady)
	  hWellIndex_(1,well) = hWellIndex_(1,well) + ny_g;
      }
    }
    */

    hWellIndex_(0,well) = hWellIndex_(0,well) - xOffset_g_(0);
    hWellIndex_(1,well) = hWellIndex_(1,well) - yOffset_g_(0);
    zWellIndexLo_(0,well) = zWellIndexLo_(0,well) - zOffset_g_(0);
    zWellIndexLo_(1,well) = zWellIndexLo_(1,well) - zOffset_g_(0);
    zWellIndexHi_(0,well) = zWellIndexHi_(0,well) - zOffset_g_(0);
    zWellIndexHi_(1,well) = zWellIndexHi_(1,well) - zOffset_g_(0);
  }

  /* Remove excess wells */

  well_l = -1;
  for(well=0; well<nWells; well++) {
    i = hWellIndex_(0,well);
    j = hWellIndex_(1,well);
    if( MYWELL(i,j) && (wellType[well] != WELL_INACTIVE) ) {
      well_l++;
      wellType[well_l] = wellType[well];
      hWellIndex_(0,well_l) = i;
      hWellIndex_(1,well_l) = j;
      switch(wellType[well_l]) {
      case WELL_INJECTION:
      case WELL_REINJECTION:
	if(zWellIndexLo_(0,well) <= zWellIndexHi_(0,well)) {
	  zWellIndexLo_(0,well_l) = zWellIndexLo_(0,well);
	  zWellIndexHi_(0,well_l) = zWellIndexHi_(0,well);
	} else {
	  incorrectWell = well + 1;
	  ABORT(ERR_INVERTED_WELL);
	}
	break;
      default:
	zWellIndexLo_(0,well_l) = nz;
	zWellIndexHi_(0,well_l) = 0;
      }
      switch(wellType[well_l]) {
      case WELL_EXTRACTION:
      case WELL_BOTTOMHOLE_PROD:
      case WELL_REINJECTION:
	if(zWellIndexLo_(1,well) <= zWellIndexHi_(1,well)) {
	  zWellIndexLo_(1,well_l) = zWellIndexLo_(1,well);
	  zWellIndexHi_(1,well_l) = zWellIndexHi_(1,well);
	} else {
	  incorrectWell = well + 1;
	  ABORT(ERR_INVERTED_WELL);
	}
	break;
      default:
	zWellIndexLo_(1,well_l) = nz;
	zWellIndexHi_(1,well_l) = 0;
      }
      wellRadius[well_l] = wellRadius[well];
      if(computeFlow) {
	wellPres[well_l] = wellPres[well];
	nWellInterp_f[well_l] = nWellInterp_f[well];
	for(m=0; m<maxnWellInterp_f; m++) {
	  wellInj_f_(0,m,well_l) = wellInj_f_(0,m,well);
	  wellInj_f_(1,m,well_l) = wellInj_f_(1,m,well);
	}
      }
      if(computeTransport) {
	for(specie=0; specie<nSpecies; specie++) {
	  nWellInterp_t_(specie,well_l) = nWellInterp_t_(specie,well);
	  for(m=0; m<maxnWellInterp_t; m++) {
	    wellInj_t_(0,m,specie,well_l) = wellInj_t_(0,m,specie,well);
	    wellInj_t_(1,m,specie,well_l) = wellInj_t_(1,m,specie,well);
	  }
	}
      }
    }
  }
  nWells = well_l + 1;

 exit:
  return error;
}

/******************************************************************************
Title: readWells
Description: Read and preprocess the well data
*/

int readWells() {
  static char WELLS[] = {"WELLS"};
  static char Well[] = {"Well"};
  static char Flow[] = {"Flow"};
  static char Transport[] = {"Transport"};

  int i,j,k;
  int error = 0;

  skipToSection(WELLS);
  packInt(nWells);
  packInt(maxnWellInterp_f);
  packInt(maxnWellInterp_t);
  unPackInt();

  if(nWells < 0) nWells = 0;
  if(nWells) {
    if ( nDom[2] != 1 ) ABORT(ERR_NEED_1_PROC_IN_Z);
    if(maxnWellInterp_f < 0) maxnWellInterp_f = 0;
    if(maxnWellInterp_t < 0) maxnWellInterp_t = 0;

    wellType = (int*) malloc(nWells*sizeof(int));
    hWellIndex = (int*) malloc(2*nWells*sizeof(int));
    zWellIndexLo = (int*) malloc(2*nWells*sizeof(int));
    zWellIndexHi = (int*) malloc(2*nWells*sizeof(int));
    if(wellType == NULL || hWellIndex == NULL
       || zWellIndexLo == NULL || zWellIndexHi == NULL) ABORT(ERR_MEMORY);
    wellRadius = (double*) malloc(nWells*sizeof(double));
    if(wellRadius == NULL) ABORT(ERR_MEMORY);
    for(i=0;i<nWells;i++) wellRadius[i] = 0;
    if(computeFlow) {
      nWellInterp_f = (int*) malloc(nWells*sizeof(int));
      if(nWellInterp_f == NULL) ABORT(ERR_MEMORY);
      for(i=0;i<nWells;i++) nWellInterp_f[i] = 0;
      if(maxnWellInterp_f) {
	wellInj_f = (float*) malloc(2*maxnWellInterp_f*nWells*sizeof(float));
	if(wellInj_f == NULL) ABORT(ERR_MEMORY);
	for(i=0;i<2*maxnWellInterp_f*nWells;i++) wellInj_f[i] = 0;
      }
      wellPres = (double*) malloc(nWells*sizeof(double));
      if(wellPres == NULL) ABORT(ERR_MEMORY);
      for(i=0;i<nWells;i++) wellPres[i] = 0;
    }
    if(computeTransport && nSpecies) {
      nWellInterp_t = (int*) malloc(nSpecies*nWells*sizeof(int));
      if(nWellInterp_t == NULL) ABORT(ERR_MEMORY);
      for(i=0;i<nSpecies*nWells;i++) nWellInterp_t[i] = 0;
      if(maxnWellInterp_t) {
	wellInj_t = (float*)
	  malloc(2*maxnWellInterp_t*nSpecies*nWells*sizeof(float));
	if(wellInj_t == NULL) ABORT(ERR_MEMORY);
	for(i=0;i<2*maxnWellInterp_t*nSpecies*nWells;i++) wellInj_t[i] = 0;
      }
    }
  }

  for(i=0; i<nWells; i++) {
    readSubSection(0);
    skipToSubSection(Well);
    packInt(wellType[i]);
    packDoubleUnits(wellRadius[i],DIMEN_LENGTH);
    packInt(hWellIndex_(0,i));
    packInt(hWellIndex_(1,i));
    packInt(zWellIndexLo_(0,i));
    packInt(zWellIndexHi_(0,i));

    unPackInt();
    if(wellType[i] == WELL_INACTIVE) continue;

    switch(wellType[i]) {
    case WELL_INJECTION:
      zWellIndexLo_(1,i) = 0;
      zWellIndexHi_(1,i) = 0;
      break;
    case WELL_EXTRACTION:
    case WELL_BOTTOMHOLE_PROD:
      unPackInt();
      zWellIndexLo_(1,i) = zWellIndexLo_(0,i);
      zWellIndexLo_(0,i) = 0;
      zWellIndexHi_(1,i) = zWellIndexHi_(0,i);
      zWellIndexHi_(0,i) = 0;
      break;
    case WELL_REINJECTION:
      packInt(zWellIndexLo_(1,i));
      packInt(zWellIndexHi_(1,i));
      break;
    default:
      incorrectWell = i+1;
      ABORT(ERR_UNSUPPORTED_WELL);
    }

    if(computeFlow) {
      switch(wellType[i]) {
      case WELL_BOTTOMHOLE_PROD:
	readSubSubSection(0);
	skipToSubSubSection(Flow);
	packDoubleUnits(wellPres[i],DIMEN_PRESSURE);
	break;
      case WELL_INJECTION:
      case WELL_EXTRACTION:
      case WELL_REINJECTION:
	readSubSubSection(0);
	skipToSubSubSection(Flow);
	packInt(nWellInterp_f[i]);
	unPackInt();
	if(nWellInterp_f[i] > maxnWellInterp_f)
	  ABORT(ERR_ON_LINE);
	beginList();
	for(k=0; k<nWellInterp_f[i]; k++) {
	  packFloatUnits(wellInj_f_(0,k,i),DIMEN_TIME);
	  packFloatUnits(wellInj_f_(1,k,i),DIMEN_VOLUME_PER_TIME);
	}
	endList();
	break;
      default:
	break;
      }
      switch(wellType[i]) {
      case WELL_INJECTION:
      case WELL_EXTRACTION:
      case WELL_REINJECTION:
	unPackFloat();
	for(k=0; k<nWellInterp_f[i]; k++) {
	  if(wellInj_f_(1,k,i) < 0) wellInj_f_(1,k,i) = -wellInj_f_(1,k,i);
	}
	break;
      default:
	break;
      }
    }

    if(computeTransport) {
      switch(wellType[i]) {
      case WELL_INJECTION:
      case WELL_REINJECTION:
	readSubSubSection(0);
	skipToSubSubSection(Transport);
	beginList();
	for(j=0; j<nSpecies; j++) {
	  packInt(nWellInterp_t[j + nSpecies*i]);
	  unPackInt();
	  if(nWellInterp_t[j + nSpecies*i] > maxnWellInterp_t)
	    ABORT(ERR_ON_LINE);
	  beginList();
	  for(k=0; k< nWellInterp_t[j + nSpecies*i]; k++) {
	    packFloatUnits(wellInj_t_(0,k,j,i),DIMEN_TIME);
	    packFloatUnits(wellInj_t_(1,k,j,i),DIMEN_MOLAR_CONC);
	  }
	  endList();
	}
	endList();
	unPackFloat();
	for(j=0; j<nSpecies; j++)
	for(k=0; k< nWellInterp_t[i]; k++) {
	  if(wellInj_t_(1,k,j,i) < 0)
	    wellInj_t_(1,k,j,i) = -wellInj_t_(1,k,j,i);
	}
	break;
      default:
	break;
      }
    }
  }

  unPackInt(); unPackDouble(); unPackFloat();
  if(computeFlow) {
    for(i=0; i<nWells; i++) if(nWellInterp_f[i] > 1) steadyFlow = 0;
  }
  error = localizeWells(); if(error) goto exit;

  if(!computeFlow && nWells) {
    nWellInterp_f = (int*) malloc(nWells*sizeof(int));
    wellInj_f = (float*) malloc(2*nWells*sizeof(float));
    if(nWellInterp_f == NULL || wellInj_f == NULL) ABORT(ERR_MEMORY);
      
    for(i=0; i<nWells; i++) {
      nWellInterp_f[i] = 1;
      wellInj_f[i] = 0;
      wellInj_f[i+nWells] = 0;
    }
  }

 exit:
  if(error < 0 && !myProc) {
    switch(error) {
    case ERR_NEED_1_PROC_IN_Z:
      fprintf(stderr,"Must use 1 domain in the third direction with wells\n");
      error = ERR_ON_LINE;
      break;
    case ERR_UNSUPPORTED_WELL:
      fprintf(stderr,"Well %d type undefined\n",incorrectWell);
      error = ERR_BEFORE_LINE;
      break;
    case ERR_INVERTED_WELL:
      fprintf(stderr,"Well %d depth indices inverted\n",incorrectWell);
      error = ERR_BEFORE_LINE;
      break;
    case ERR_WELL_OUTSIDE_DOMAIN:
      fprintf(stderr,"Well %d outside domain\n",incorrectWell);
      error = ERR_BEFORE_LINE;
      break;
    default:
      break;
    }
  }
  return error;
}
