/******************************************************************************
Parallel Read Scalars (PRS)

This package reads scalar data in parallel.  It consists of the two files
  reader.c and reader.h .

Todd Arbogast
The University of Texas at Austin
August 1996
Modified July 1997

SYNTACTIC STRUCTURE OF GENERAL INPUT FILES

The general input file(s) contain:
  sectional commands
  keywords
  data items
  list delineators
  white space (WS), including proper white space (PWS)
  comments (explicit and implicit)
  commands
Within an input file, data are arranged in sections, subsections, and
  subsubsections and ordered item-by-item, possibly separated by
  list delineators, special commands, comments, and white-space.

Sections, Subsections, and Subsubsections are initiated by a
  SECTION, SUBSECTION, or SUBSUBSECTION command, respectively,
  and a specified string.  See below for command syntax.
Keywords are specified strings of characters.

Data items consist of:
  int
  float
  double
  word of characters (terminated by a SPACE, newline (EOL), or TAB)
  line of characters (including spaces and tabs, terminating before EOL)
List delineators are:
  BEGIN_LIST '{' to indicate the beginning of a list;
  END_LIST '}' to indicate the end of a list.
    Lists are user specified numbers of items.  The deliminators check that
    the user has counted the number of list items correctly.

WS (white space) consists of SPACE ' ', TAB '\t', EOL '\n' (end-of-line),
  COMMA ',', SEMICOLON ';', and COLON ':'.  WS is ignored on input.
PWS (proper white space) consists of SPACE ' ', TAB '\t', EOL '\n'.
  Command names and keywords are terminated by PWS.

Explicit comments appear
  After the COMMENT symbol '#' (the comment runs to the EOL symbol,
    and the '#' cannot be imbedded in a string).
  Between the BEGIN_COMMENT and END_COMMENT commands (see below).
Implicit comments can be allowed:
  Before the sectional commands.  This enables skipping of unnessessary
  sectional units.

Commands alter the way the data file(s) are read.  Commands take the form
  COMMAND symbol '$' followed immediately (no space)
    by a command name.  Note that the COMMAND symbol is ignored
    in any explicit comment (but not in implicit comments).
  Commands are
  $ is a null command
  $$ is a null command
  $INCLUDE, followed by PWS and a new file name.
    This causes subsequent reading to take place in the new file, until
    the EOF (end-of-file) is read.  Reading then returns to the original file.
    A finite levell of nesting is imposed.
  $BEGIN_COMMENT causes input to be ignored until the matching END_COMMENT
    command is read (such commands are nested) or the EOF is reached.
  $END_COMMENT terminates a BEGIN_COMMENT command.
  ${# is equivalent to BEGIN_COMMENT.
  $#} is equivalent to END_COMMENT.
  $LITERAL terminates reading WS and other symbols.  Used to allow special
    symbols in data strings (',', ';', '#', and '$').  (This command is
    not needed before INCLUDE file names.)
  $SECTION begins a section of the data file.  Reading may not pass a
    section command except by explicit action.    
  $S is equivalent to SECTION.
  $SUBSECTION begins a subsection of the data file.  Reading may not pass
    a section or subsection command except by explicit action.    
  $SS is equivalent to SUBSECTION.
  SUBSUBSECTION begins a subsubsection of the data file.  Reading
    may not pass any sectional command except by explicit action.    
  $SSS is equivalent to SUBSUBSECTION.
  $IGNORE_WORDS, followed by PWS and an integer.  Causes that number of
    words to be ignored from the input stream.  Note that the include
    command only is not ignored.  This command can be used to ignore
    preambles in included data files.
  $IGNORE_LINES, followed by PWS and an integer.  Causes that number of
    lines (i.e., up to EOL) to be ignored from the input stream.  The line
    containing the integer is included in the count of lines ignored, and
    ignored from that point on.  Note that the include
    command only is not ignored.  This command can be used to ignore
    preambles in included data files.

REMARKS:
  (1) The readIntArray, readDoubleArray[Units], and readFloatArray[Units]
      commands read a block of data, and it may contain the '@' repetition
      character.  A set of "n" single data items, each with the value "value",
      may be indicated by "n@value".  PWS may precede "value", but only
      spaces may precede '@'.
  (2) Integers may not be followed by ".".
******************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "msgs.h"
#include "reader.h"
#include "expr.h"

#include "debug.h"

#define SIZE_INTBUF           4096
#define SIZE_DOUBLEBUF        4096
#define SIZE_FLOATBUF         4096
#define DEPTH_INCLUDES           5
#define MAX_LENGTH_UNITS_STR   121

/* DEFINE PROPER WHITESPACE */
#define EOL        '\n'  /* end of line character */
#define SPACE      ' '   /* space character */
#define TAB        '\t'  /* tab character */

/* DEFINE IMPROPER WHITESPACE */
#define COMMENT    '#'   /* comment character */
#define COMMA      ','   /* comma separator */
#define SEMICOLON  ';'   /* semicolon separator */
#define COLON      ':'   /* colon separator */

/* DEFINE SPECIAL SYMBOLS */
#define REPETITION '@'
#define BEGIN_LIST '{'
#define END_LIST   '}'
#define BEGIN_UNIT '['
#define END_UNIT   ']'

/* DEFINE PROCESSOR COMMANDS */
#define COMMAND            '$' /* processor commands follow this symbol */
#define SIZE_COMMAND_NAME  21
static char beginCommentCmd[]       = "BEGIN_COMMENT";
static char endCommentCmd[]         = "END_COMMENT";
static char shortBeginCommentCmd[]  = "{#";
static char shortEndCommentCmd[]    = "#}";
static char includeCmd[]            = "INCLUDE";
static char literalCmd[]            = "LITERAL";
static char sectionCmd[]            = "SECTION";
static char shortSectionCmd[]       = "S";
static char subSectionCmd[]         = "SUBSECTION";
static char shortSubSectionCmd[]    = "SS";
static char subSubSectionCmd[]      = "SUBSUBSECTION";
static char shortSubSubSectionCmd[] = "SSS";
static char ignoreWordsCmd[]        = "IGNORE_WORDS";
static char ignoreLinesCmd[]        = "IGNORE_LINES";

static int inFileNo = -1;
static int SIZE_FILENAME;
static char* inFileName[DEPTH_INCLUDES];
static FILE* inFile[DEPTH_INCLUDES];
static int inFileLineNo[DEPTH_INCLUDES];
static FILE* echoFile;

static int skipping = 0;
static int beginSkipFileLineNo;
static char* beginSkipFileName;

/* static int msg = 0; */
static int myProc;
static int echo = 0;

static int   nIntBuf = 0;
static int*  intBuf; /* [SIZE_INTBUF] */
static int** intAdd; /* [SIZE_INTBUF] */
static int      nDoubleBuf = 0;
static double*  doubleBuf; /* [SIZE_DOUBLEBUF] */
static double** doubleAdd; /* [SIZE_DOUBLEBUF] */
static int     nFloatBuf = 0;
static float*  floatBuf; /* [SIZE_FLOATBUF] */
static float** floatAdd; /* [SIZE_FLOATBUF] */

static char* save_keyWord_1;
static char** save_keyWord;
static int save_nKeyWords;

int prs_ignore(int ignoreWords);

/******************************************************************************
Title: prs_getc
Description: Get a character.  Essentially the usual getc C routine.
             This routine is dangerous, simce normal checking for
	     special characters is omitted.
*/
int prs_getc() {
  return getc(inFile[inFileNo]);
}

/******************************************************************************
Title: prs_ungetc
Description: Unget a character.  Use only after prs_getc, as in the usual
             getc and ungetc C routines.
*/
void prs_ungetc(int c) {
  ungetc(c,inFile[inFileNo]);
}

/******************************************************************************
Title: prs_echoString
Description: Echo a string.
*/
void prs_echoString(char* str) {
  if(echo) fprintf(echoFile,"%s",str);
}

/******************************************************************************
Title: prs_echoChar
Description: Echo a char.
*/
void prs_echoChar(int c) {
  if(echo) fprintf(echoFile,"%c",c);
}

/******************************************************************************
Title: prs_beginSkip()
Description: Set the skip line number.
*/
void prs_beginSkip() {
  skipping = 1;
  if(!myProc) {
    beginSkipFileLineNo = inFileLineNo[inFileNo];
    beginSkipFileName = inFileName[inFileNo];
  }
}

/******************************************************************************
Title: prs_endSkip()
Description: Set the skip line number.
*/
void prs_endSkip() {
  skipping = 0;
}

/******************************************************************************
Title: prs_serialReadPWS
Description: Read PWS in serial mode, up to non-punctuation white
             space or one less than the length of the WS, and add
	     the WS terminating character.
Serial.
*/
void prs_serialReadPWS() {
  int c;

  do {
    c = getc(inFile[inFileNo]);
    if(c == EOL) inFileLineNo[inFileNo]++;
  } while(c == SPACE || c == TAB || c == EOL);
  ungetc(c,inFile[inFileNo]);
}

/******************************************************************************
Title: prs_serialReadString
Description: Read a string in serial mode, up to proper WS
             (or one less than the length of the string), and add
	     the string terminating character.  Do not read initial
             proper WS.
Serial.
Input:  len     The length of the string.
Output: string  The string read.
*/
void prs_serialReadString(char* string, int len) {
  int c;
  int i=0;

  for(i=0; i<len-1; i++) {
    c = getc(inFile[inFileNo]);
    if(c == EOL || c == EOF || c == SPACE || c == TAB ) {
      ungetc(c,inFile[inFileNo]);
      break;
    }
    string[i] = c;
  }
  string[i] = '\0';
}

/******************************************************************************
Title: prs_beginFile
Description: Begin reading a file.
Serial.
Return: Error flag (see reader.h).
*/
int prs_beginFile() {
  if(inFileNo >= 0) {
    if(inFileNo >= DEPTH_INCLUDES-1) return ERR_NESTING_FILE;
    prs_serialReadPWS();
    prs_serialReadString(inFileName[inFileNo+1],SIZE_FILENAME);
  }
  inFileNo++;
  inFileLineNo[inFileNo] = 0;

  inFile[inFileNo] = fopen(inFileName[inFileNo],"r");
  if(inFile[inFileNo] == NULL) return ERR_OPENING_FILE;

  ungetc(EOL,inFile[inFileNo]);

  printf("> Reading from file %s\n",inFileName[inFileNo]);
  return 0;
}

/******************************************************************************
Title: prs_finishFile
Description: Finish reading a file.
Serial.
Return: Error flag (see reader.h).
*/
int prs_finishFile() {
  fclose(inFile[inFileNo]);
  inFileNo--; if(inFileNo < 0) return ERR_EOF;
  printf("< Reading from file %s\n",inFileName[inFileNo]);
  return 0;
}

/******************************************************************************
Title: prs_processCmd
Description: Process the commands.  Flag as an error reading the
             SECTIONAL commands.
Serial.
Return: Error or return flag (see reader.h).
*/
int prs_processCmd() {
  int c;
  int error = 0;
  char cmd[SIZE_COMMAND_NAME];

#define CMD_IS(cmdName) !strcmp(cmd,cmdName)

  prs_serialReadString(cmd,SIZE_COMMAND_NAME);
  if(CMD_IS(includeCmd)) {
    error = prs_beginFile(); if(error) return error;
  } else if( CMD_IS(beginCommentCmd) || CMD_IS(shortBeginCommentCmd) ) {
    int nBeginComments = 1;
    while(1) {
      do {
	c = getc(inFile[inFileNo]);
	if(c == EOL) inFileLineNo[inFileNo]++;
      } while(c != COMMAND && c != EOF);
      if(c == EOF) {
	ungetc(c,inFile[inFileNo]);
	return 0;
      }
      prs_serialReadString(cmd,SIZE_COMMAND_NAME);
      if( CMD_IS(endCommentCmd) || CMD_IS(shortEndCommentCmd) ) {
	nBeginComments--;
	if(nBeginComments == 0) return 0;
      } else if( CMD_IS(beginCommentCmd)
		|| CMD_IS(shortBeginCommentCmd) ) nBeginComments++;
    }
  } else if( CMD_IS(endCommentCmd) || CMD_IS(shortEndCommentCmd) ) {
    return ERR_UNMATCHED_END_COMMENT_CMD;
  } else if(CMD_IS(literalCmd)) {
    do {
      c = getc(inFile[inFileNo]);
      if(c == EOL) inFileLineNo[inFileNo]++;
    } while(c == SPACE || c == TAB || c == EOL);
    ungetc(c,inFile[inFileNo]);
    if(echo) fprintf(echoFile,"%c%s ",COMMAND,literalCmd);
    return READ_LITERAL_CMD;
  } else if( CMD_IS(sectionCmd) || CMD_IS(shortSectionCmd) ) {
    if(echo) fprintf(echoFile,"%c%s ",COMMAND,shortSectionCmd);
    return ERR_SECTION_READ;
  } else if( CMD_IS(subSectionCmd) || CMD_IS(shortSubSectionCmd) ) {
    if(echo) fprintf(echoFile,"%c%s ",COMMAND,shortSubSectionCmd);
    return ERR_SUBSECTION_READ;
  } else if( CMD_IS(subSubSectionCmd) || CMD_IS(shortSubSubSectionCmd) ) {
    if(echo) fprintf(echoFile,"%c%s ",COMMAND,shortSubSubSectionCmd);
    return ERR_SUBSUBSECTION_READ;
  } else if( CMD_IS(ignoreWordsCmd) ) {
    error = prs_ignore(1); if(error) return error;
  } else if( CMD_IS(ignoreLinesCmd) ) {
    error = prs_ignore(0); if(error) return error;
  } else {
    return ERR_UNKNOWN_CMD;
  }
  return 0;
}

/******************************************************************************
Title: prs_skipWS
Description: Skip past white space and comments.
             Processes commands along the way.
	     Conditionally stop just after (the last of several) EOL is read.
Input:  stopEOL  Switch to stop just after EOL is read.
Return: Error flag (see reader.h).
*/
int prs_skipWS(int stopEOL) {
  int c;
  int error = 0;

  if(myProc) return 0;
  while(1) {
    c = getc(inFile[inFileNo]);
    switch(c) {
    case EOL:
      if(stopEOL) {
	do {
	  inFileLineNo[inFileNo]++;
	  c = getc(inFile[inFileNo]);
	  if(c == EOF) {
	    error = prs_finishFile(); if(error) return error;
	    c = getc(inFile[inFileNo]);
	  }
	} while(c == EOL);
	ungetc(c,inFile[inFileNo]);
	return 0;
      } else {
	inFileLineNo[inFileNo]++;
      }
    case SPACE:
    case TAB:
    case COMMA:
    case SEMICOLON:
    case COLON:
      break;
    case EOF:
      error = prs_finishFile(); if(error) return error;
      break;
    case COMMENT:
      do { c = getc(inFile[inFileNo]); } while(c != EOL && c != EOF);
      ungetc(c,inFile[inFileNo]);
      break;
    case COMMAND:
      error = prs_processCmd();
      if(error) {
	if(error == READ_LITERAL_CMD) return 0;
	return error;
      }
      break;
    default:
      ungetc(c,inFile[inFileNo]);
      return 0;
    }
  }
}

/******************************************************************************
Title: prs_skipWord
Description: Skip past the next word (of non special characters).
Input:  stopNonPWS       Switch to stop just before non PWS or
                         COMMENT or COMMAND symbols.
        stopListSymbols  Switch to stop just before a BEGIN_LIST
                         or END_LIST symbol.
        echoWord         Switch to echo the word (if echoing is enabled).
Return: Error flag (see reader.h).
*/
void prs_skipWord(int stopNonPWS, int stopListSymbols, int echoWord) {
  int c;
  int echoed = 0;

  if(myProc) return;

  echoWord = echoWord && echo;

  while(1) {
    c = getc(inFile[inFileNo]);
    switch(c) {
    case SPACE:
    case TAB:
    case EOL:
    case EOF:
      ungetc(c,inFile[inFileNo]);
      goto exit;
    case COMMA:
    case SEMICOLON:
    case COLON:
    case COMMENT:
    case COMMAND:
      if(stopNonPWS) {
	ungetc(c,inFile[inFileNo]);
	goto exit;
      }
      break;
    case BEGIN_LIST:
    case END_LIST:
      if(stopListSymbols) {
	ungetc(c,inFile[inFileNo]);
	goto exit;
      }
      break;
    }
    if(echoWord) {
      fprintf(echoFile,"%c",c);
      echoed = 1;
    }
  }
exit:
  if(echoed) fprintf(echoFile,"\n");
}

/******************************************************************************
Title: prs_ignore
Description: Ignore lines or words.  Process the INCLUDE command, but
             no other commands.
Input:  ignoreWords      Switch to ignore words vs. lines.
Return: Error flag (see reader.h).
*/
int prs_ignore(int ignoreWords) {
  int n;
  int c;
  int error = 0;
  char cmd[SIZE_COMMAND_NAME];

  int* nRead;
  int nWordsRead = 0;
  int nLinesRead = 0;

  if(myProc) return 0;

  prs_serialReadPWS();

  if(!fscanf(inFile[inFileNo],"%d",&n)) return ERR_ON_LINE;
  c = getc(inFile[inFileNo]);
  if(c == '.') return ERR_ON_LINE;
  ungetc(c,inFile[inFileNo]);

  nRead = (ignoreWords) ? &nWordsRead : &nLinesRead;

  while(*nRead < n) {
    c = getc(inFile[inFileNo]);
    switch(c) {
    case EOL:
      inFileLineNo[inFileNo]++;
      nLinesRead++;
      break;
    case SPACE:
    case TAB:
      break;
    case EOF:
      error = prs_finishFile(); if(error) return error;
      break;
    case COMMAND:
      prs_serialReadString(cmd,SIZE_COMMAND_NAME);
      if(!strcmp(cmd,includeCmd)) {
	int nOld = inFileLineNo[inFileNo];
	error = prs_beginFile(); if(error) return error;
	nLinesRead += inFileLineNo[inFileNo-1] - nOld - 1;
      } else {
	nWordsRead++;
      }
      break;
    default:
      ungetc(c,inFile[inFileNo]);
      prs_skipWord(0,0,0);
      nWordsRead++;
    }
  }
  return 0;
}

/******************************************************************************
Title: prs_readKeys
Description: Skip or read until one of several key words is read.
             Processes commands along the way.
             Returns the C-number of the key read in "keyRead".
             If reading, not skipping, the key must be read next from
             the input, except commands, WS, and explicit comments.
Input:  keyWord         Pointers to the key words to search for.
                        These key words should be statically defined,
			so that error conditions can be processed properly.
        nKeys           Number of key words.
        skip_echoWord   Flag:  1 = skip to the key (otherwise read to the key)
                              -1 = echo the word read (if echoing is enabled).
Output: keyRead  The C-index of the key read.
Return: Error flag (see reader.h).
*/
int prs_readKeys(char** keyWord, int nKeys, int* keyRead, int skip_echoWord) {
  int k;
  int c,cc;
  int error = 0;
  int next;
  int* possibleKey;
  int stillPossible;

  int skip = (skip_echoWord > 0);
  int echoWord = (skip_echoWord < 0) && echo;
  int echoDone = !echoWord;

  if(nKeys == 1) {
    save_keyWord_1 = *keyWord;
    save_keyWord = &save_keyWord_1;
  } else {
    save_keyWord = keyWord;
  }
  save_nKeyWords = nKeys;
  *keyRead = -1;

  if(myProc) goto exit1;

  for(k=0; k<nKeys; k++) {
    if(keyWord[k][0] == '\0') { *keyRead = k; goto exit1; }
  }
  possibleKey = (int*) malloc(nKeys*sizeof(int));
  if(possibleKey == NULL) { error = ERR_MEMORY; goto exit; }

  do {
    error = prs_skipWS(0);
    if(error) {
      if(error == ERR_EOF || error == ERR_SECTION_READ
	 || error == ERR_SUBSECTION_READ
	 || error == ERR_SUBSUBSECTION_READ) error = ERR_KEY_NOT_READ;
      echoWord = 0;
      goto exit;
    }

    for(k=0; k<nKeys; k++) possibleKey[k] = 1;
    next = -1;

    do {
      c = getc(inFile[inFileNo]);
      if(c == TAB || c == EOL || c == EOF) {
	ungetc(c,inFile[inFileNo]);
        echoDone = 1;
	break;
      }
      if(echoWord && c != SPACE) fprintf(echoFile,"%c",c);

      next++;
      stillPossible = 0;

      for(k=0; k<nKeys; k++) {
	if(possibleKey[k]) {
	  if(c == keyWord[k][next]) {
	    if(keyWord[k][next+1] == '\0') {
	      cc = getc(inFile[inFileNo]);
	      ungetc(cc,inFile[inFileNo]);
	      if(cc == SPACE || cc == TAB || cc == EOL || cc == EOF) {
		if(echo && !echoWord) fprintf(echoFile,"%s\n",keyWord[k]);
		*keyRead = k;
		error = 0;
		goto exit;
	      }
	      possibleKey[k] = 0;
	    }
	  } else {
	    possibleKey[k] = 0;
	  }
	}
	stillPossible += possibleKey[k];
      }

      if(echoWord && c == SPACE) {
	if(stillPossible) {
	  fprintf(echoFile,"%c",c);
	} else {
	  echoDone = 1;
	}
      }
    } while(stillPossible);
  } while(skip);

  error = ERR_KEY_NOT_READ;
  while(!echoDone) {
    c = getc(inFile[inFileNo]);
    if(c == SPACE || c == TAB || c == EOL || c == EOF) {
      ungetc(c,inFile[inFileNo]);
      goto exit;
    }
    fprintf(echoFile,"%c",c);
  }

 exit:
  if(echoWord) fprintf(echoFile,"\n");
  free(possibleKey);
 exit1:
  if(nKeys > 1) {
    /* msg++; gibcastX(msg,keyRead,1,0); */
    MPI_Bcast((void*)keyRead,1,MPI_INT,0,comm);

  } else *keyRead = 0;
  return error;
}

/******************************************************************************
Title: prs_readKey
Description: Skip or read until a key word is read.
             Processes commands along the way.
             Returns the C-number of the key read in "keyRead".
             If reading, not skipping, the key must be read next from
             the input, except commands, WS, and explicit comments.
Input:  keyWord         The key word to search for.
                        This key word should be statically defined,
			so that error conditions can be processed properly.
        skip_echoWord   Flag:  1 = skip to the key (otherwise read to the key)
                              -1 = echo the word read (if echoing is enabled).
Return: Error flag (see reader.h).
*/
int prs_readKey(char* keyWord, int skip_echoWord) {
  int tmp;
  return prs_readKeys(&keyWord,1,&tmp,skip_echoWord);
}

/******************************************************************************
Title: prs_readSection
Description: Skip or read until a SECTION command is read.
             Processes other commands along the way.
Input: skip     Flag: skip to the sectional command (otherwise read to it)
Return: Error flag (see reader.h).
*/
int prs_readSection(int skip) {
  int error = 0;

  if(myProc) return 0;

  while(1) {
    error = prs_skipWS(0);
    switch(error) {
    case ERR_SECTION_READ:
      return 0;
    case ERR_EOF:
      return ERR_SECTION_NOT_READ;
    case ERR_SUBSECTION_READ:
    case ERR_SUBSUBSECTION_READ:
      prs_serialReadPWS();
      prs_skipWord(1,0,1);
      skip = 1;
      break;
    case 0:
      if(!skip) return ERR_SECTION_NOT_READ;
      prs_skipWord(1,0,0);
      break;
    default:
      return error;
    }
  }
}

/******************************************************************************
Title: prs_readSubSection
Description: Skip or read until a SUBSECTION command is read.
             Processes other commands along the way.
Input: skip     Flag: skip to the sectional command (otherwise read to it)
Return: Error flag (see reader.h).
*/
int prs_readSubSection(int skip) {
  int error = 0;

  if(myProc) return 0;

  while(1) {
    error = prs_skipWS(0);
    switch(error) {
    case ERR_SUBSECTION_READ:
      return 0;
    case ERR_EOF:
    case ERR_SECTION_READ:
      return ERR_SUBSECTION_NOT_READ;
    case ERR_SUBSUBSECTION_READ:
      prs_serialReadPWS();
      prs_skipWord(1,0,1);
      skip = 1;
      break;
    case 0:
      if(!skip) return ERR_SUBSECTION_NOT_READ;
      prs_skipWord(1,0,0);
      break;
    default:
      return error;
    }
  }
}

/******************************************************************************
Title: prs_readSubSubSection
Description: Skip or read until a SUBSUBSECTION command is read.
             Processes other commands along the way.
Input: skip     Flag: skip to the sectional command (otherwise read to it)
Return: Error flag (see reader.h).
*/
int prs_readSubSubSection(int skip) {
  int error = 0;

  if(myProc) return 0;

  while(1) {
    error = prs_skipWS(0);
    switch(error) {
    case ERR_SUBSUBSECTION_READ:
      return 0;
    case ERR_EOF:
    case ERR_SECTION_READ:
    case ERR_SUBSECTION_READ:
      return ERR_SUBSUBSECTION_NOT_READ;
    case 0:
      if(!skip) return ERR_SUBSUBSECTION_NOT_READ;
      prs_skipWord(1,0,0);
      break;
    default:
      return error;
    }
  }
}

/******************************************************************************
Title: skipToSections
Description: Skip to one of several new sections.
             We assume we have just read a section command.
*/
int prs_skipToSections(char** keys, int nKeys, int* keyRead) {
  int error = 0;

  prs_beginSkip();

  while(1) {
    error = prs_readKeys(keys,nKeys,keyRead,-1);
    if(error != ERR_KEY_NOT_READ) goto exit;
    error = prs_readSection(1);
    if(error == ERR_SECTION_NOT_READ) error = ERR_SECTION_KEY_NOT_READ;
    if(error) goto exit;
  }

exit:
  if(!error) prs_endSkip();
  return error;
}

/******************************************************************************
Title: skipToSubSections
Description: Skip to one of several new subsections.
             We assume we have just read a subsection command.
*/
int prs_skipToSubSections(char** keys, int nKeys, int* keyRead) {
  int error = 0;

  prs_beginSkip();

  while(1) {
    error = prs_readKeys(keys,nKeys,keyRead,-1);
    if(error != ERR_KEY_NOT_READ) goto exit;
    error = prs_readSubSection(1);
    if(error == ERR_SUBSECTION_NOT_READ) error = ERR_SUBSECTION_KEY_NOT_READ;
    if(error) goto exit;
  }

exit:
  if(!error) prs_endSkip();
  return error;
}

/******************************************************************************
Title: skipToSubSubSections
Description: Skip to one of several new subsubsections.
             We assume we have just read a subsubsection command.
*/
int prs_skipToSubSubSections(char** keys, int nKeys, int* keyRead) {
  int error = 0;

  prs_beginSkip();

  while(1) {
    error = prs_readKeys(keys,nKeys,keyRead,-1);
    if(error != ERR_KEY_NOT_READ) goto exit;
    error = prs_readSubSubSection(1);
    if(error == ERR_SUBSUBSECTION_NOT_READ)
      error = ERR_SUBSUBSECTION_KEY_NOT_READ;
    if(error) goto exit;
  }

exit:
  if(!error) prs_endSkip();
  return error;
}

/******************************************************************************
Title: skipToSection
Description: Skip to a new section.
             We assume we have just read a section command.
*/
int prs_skipToSection(char* keyWord) {
  int tmp;
  return prs_skipToSections(&keyWord,1,&tmp);
}

/******************************************************************************
Title: skipToSubSection
Description: Skip to a new subsection.
             We assume we have just read a subsection command.
*/
int prs_skipToSubSection(char* keyWord) {
  int tmp;
  return prs_skipToSubSections(&keyWord,1,&tmp);
}

/******************************************************************************
Title: skipToSubSubSection
Description: Skip to a new subsubsection.
             We assume we have just read a subsubsection command.
*/
int prs_skipToSubSubSection(char* keyWord) {
  int tmp;
  return prs_skipToSubSubSections(&keyWord,1,&tmp);
}

/******************************************************************************
Title: prs_beginList
Description: Skip WS and then read a BEGIN_LIST symbol.
             Processes commands along the way.
Return: Error flag (see reader.h).
*/
int prs_beginList() {
  int error = 0;
  char c;

  if(myProc) goto exit;

  error = prs_skipWS(0);
  if(error) {
    if(error == ERR_EOF) error = ERR_BEGIN_LIST_NOT_READ;
    goto exit;
  }
  c = getc(inFile[inFileNo]);
  if(c != BEGIN_LIST) {
    error = ERR_BEGIN_LIST_NOT_READ;
    ungetc(c,inFile[inFileNo]);
    goto exit;
  }
  if(echo) fprintf(echoFile,"%c\n",c);

 exit:
  return error;
}

/******************************************************************************
Title: prs_endList
Description: Skip WS and then read an END_LIST symbol.
             Processes commands along the way.
Return: Error flag (see reader.h).
*/
int prs_endList() {
  int error = 0;
  char c;

  if(myProc) goto exit;

  error = prs_skipWS(0);
  if(error) {
    if(error == ERR_EOF || error == ERR_SECTION_READ
       || error == ERR_SUBSECTION_READ
       || error == ERR_SUBSUBSECTION_READ) error = ERR_END_LIST_NOT_READ;
    goto exit;
  }
  c = getc(inFile[inFileNo]);
  if(c != END_LIST) {
    error = ERR_END_LIST_NOT_READ;
    ungetc(c,inFile[inFileNo]);
    goto exit;
  }
  if(echo) fprintf(echoFile,"%c\n",c);

 exit:
  return error;
}

/******************************************************************************
Title: prs_skipToEndList
Description: Skip to just before an unmatched END_LIST symbol.
             Processes commands along the way.
Return: Error flag (see reader.h).
*/
int prs_skipToEndList() {
  int error = 0;
  char c;
  int nBeginLists = 1;

  if(myProc) goto exit;

  while(1) {
    error = prs_skipWS(0);
    if(error) {
      if(error == ERR_EOF || error == ERR_SECTION_READ
	 || error == ERR_SUBSECTION_READ
	 || error == ERR_SUBSUBSECTION_READ)
	error = ERR_END_LIST_NOT_READ;
      goto exit;
    }
    c = getc(inFile[inFileNo]);
    switch(c) {
    case BEGIN_LIST:
      nBeginLists++;
      if(echo) fprintf(echoFile,"%c",c);
      break;
    case END_LIST:
      nBeginLists--;
      if(nBeginLists == 0) {
	ungetc(c,inFile[inFileNo]);
	goto exit;
      }
      if(echo) fprintf(echoFile,"%c",c);
      break;
    default:
      prs_skipWord(1,1,0);
    }
  }

 exit:
  if(echo) fprintf(echoFile,"\n");
  return error;
}

/******************************************************************************
Title: prs_readUnits
Description: Read the units expression of a real number.
Inputs: dimen    The required dimensions of the units expression.
Output: value    The value of the units expression.
Return: Error flag (see reader.h)
*/
int prs_readUnits(double *value, int* dimen) {
  int error = 0;
  int c,i;
  char str[MAX_LENGTH_UNITS_STR];
  int mode = 1;
  int dimenCheck;

  if(myProc) return 0;

  prs_serialReadPWS();
  c = getc(inFile[inFileNo]);
  if(c != BEGIN_UNIT) {
    ungetc(c,inFile[inFileNo]);
    *value = 1;
    return NO_UNITS;
  }
  c = getc(inFile[inFileNo]);
  if(c == BEGIN_UNIT) {
    mode = 2;
  } else {
    c = ungetc(c,inFile[inFileNo]);
  }

  for(i = 0; i < MAX_LENGTH_UNITS_STR-1; i++) {
    prs_serialReadPWS();
    c = getc(inFile[inFileNo]);
    if(c == END_UNIT) break;
    str[i] = c;
  }
  if(i >= MAX_LENGTH_UNITS_STR-1) return ERR_UNITS_TOO_LONG;
  str[i] = '\0';
  if(mode == 2) {
    c = getc(inFile[inFileNo]);
    if(c != END_UNIT) return ERR_ON_LINE;
  }

  if(echo) {
    if(mode == 2) {
      fprintf(echoFile," [[%s]]",str);
    } else {
      fprintf(echoFile," [%s]",str);
    }
  }

  dimenCheck = (mode == 1);
  error = expr_eval(str,value,mode,dimenCheck,dimen);
  switch(error) {
  case 0:                           return 0;
  case EXPR_NEGATIVE:               return ERR_UNITS_NEGATIVE;
  case EXPR_ADDITION_FOUND:         return ERR_UNITS_ADDITION_FOUND;
  case EXPR_DIMENS_FAILED:          return ERR_UNITS_WRONG_DIMEN;
  case EXPR_TOO_BIG:                return ERR_UNITS_TOO_LONG;
  case EXPR_NONNESTED_PARENTHESES:  return ERR_NONNESTED_PARENTHESES;
  case EXPR_ARITHMETIC_OP:          return ERR_ARITHMETIC_OP;
  case EXPR_IMPROPER:               return ERR_UNITS_IMPROPER_EXPR;
  case EXPR_UNKNOWN_UNIT:           return ERR_UNKNOWN_UNIT;
  default:                          return ERR_UNITS;
  }
}

/******************************************************************************
Title: prs_readTempUnits
Description: Read and convert to the absolute temperature expression.
Inputs:   temp            The absolute temperature in the given units.
          echoConversion  Flag to echo the conversion comment.
Outputs:  temp            The absolute temperature in the internal units.
Return: Error flag (see reader.h)
*/
int prs_readTempUnits(double *temp, int echoConversion) {
  int error = 0;
  int c,i;
  double temp_in;
  char* degrees[] = {DEG_K, DEG_C, DEG_F};
  char str[MAX_LENGTH_UNITS_STR];
  int base = expr_baseTempIndex();

  if(base < 0) return ERR_IN_CODE;

  if(myProc) return 0;

  prs_serialReadPWS();
  c = getc(inFile[inFileNo]);
  if(c != BEGIN_UNIT) {
    ungetc(c,inFile[inFileNo]);
    return NO_UNITS;
  }
  temp_in = *temp;

  for(i = 0; i < MAX_LENGTH_UNITS_STR-1; i++) {
    prs_serialReadPWS();
    c = getc(inFile[inFileNo]);
    if(c == END_UNIT) break;
    str[i] = c;
  }
  if(i >= MAX_LENGTH_UNITS_STR-1) return ERR_UNITS_TOO_LONG;
  str[i] = '\0';

  error = expr_convertTempBase(temp,str);
  switch(error) {
  case 0:                  break;
  case EXPR_UNKNOWN_UNIT:  return ERR_UNKNOWN_UNIT;
  default:                 return ERR_UNITS;
  }

  if(echo) {
    fprintf(echoFile," [%s]",str);
    if(echoConversion)
      fprintf(echoFile,
	      "   # %g [%s] = %g [%s]",temp_in,str,*temp,degrees[base]);
  }

  return 0;
}

/******************************************************************************
Title: prs_unPack<TYPE>
Description: In parallel, unpack the <TYPE> buffer.
*/
/* <TYPE == int */
void prs_unPackInt() {
  int i;

  if(nIntBuf) {
    /* msg++; gibcastX(msg,intBuf,nIntBuf,0); */
    MPI_Bcast((void*)intBuf,nIntBuf,MPI_INT,0,comm);
    for(i=0;i<nIntBuf;i++) { *intAdd[i] = intBuf[i]; }
    nIntBuf = 0;
  }
}

/* <TYPE == double */
void prs_unPackDouble() {
  int i;

  if(nDoubleBuf) {
    /* msg++; gdbcastX(msg,doubleBuf,nDoubleBuf,0); */
    MPI_Bcast((void*)doubleBuf,nDoubleBuf,MPI_DOUBLE,0,comm);
    for(i=0;i<nDoubleBuf;i++) { *doubleAdd[i] = doubleBuf[i]; }
    nDoubleBuf = 0;
  }
}

/* <TYPE == float */
void prs_unPackFloat() {
  int i;

  if(nFloatBuf) {
    /* msg++; gsbcastX(msg,floatBuf,nFloatBuf,0); */
    MPI_Bcast((void*)floatBuf,nFloatBuf,MPI_FLOAT,0,comm);
    for(i=0;i<nFloatBuf;i++) { *floatAdd[i] = floatBuf[i]; }
    nFloatBuf = 0;
  }
}

/******************************************************************************
Title: prs_pack<TYPE>
Description: In parallel, pack a <TYPE> into a buffer.
Input:  The address of the variable for which to pack a value.
Return: Error flag (see reader.h).
*/
/* <TYPE == int */
int prs_packInt(int* var) {
  int error = 0;
  int c;

  if(nIntBuf >= SIZE_INTBUF) prs_unPackInt();
  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    if(!fscanf(inFile[inFileNo],"%d",&intBuf[nIntBuf]))
      return ERR_ON_LINE;
    c = getc(inFile[inFileNo]);
    ungetc(c,inFile[inFileNo]);
    if(c == '.') return ERR_ON_LINE;
    if(echo) fprintf(echoFile,"%d\n",intBuf[nIntBuf]);
  }
  intAdd[nIntBuf] = var; nIntBuf++;
  return 0;
}

/* <TYPE == double */
int prs_packDouble(double* var) {
  int error = 0;

  if(nDoubleBuf >= SIZE_DOUBLEBUF) { prs_unPackDouble(); }
  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    if(!fscanf(inFile[inFileNo],"%lg",&doubleBuf[nDoubleBuf]))
      return ERR_ON_LINE;
    if(echo) fprintf(echoFile,"%g\n",doubleBuf[nDoubleBuf]);
  }
  doubleAdd[nDoubleBuf] = var; nDoubleBuf++;
  return 0;
}

/* <TYPE == double (with units) */
int prs_packDoubleUnits(double* var, int* dimen) {
  int error = 0;
  double unit,value;

  if(nDoubleBuf >= SIZE_DOUBLEBUF) { prs_unPackDouble(); }
  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    if(!fscanf(inFile[inFileNo],"%lg",&value)) return ERR_ON_LINE;
    if(echo) fprintf(echoFile,"%g",value);

    error = prs_readUnits(&unit,dimen);
    if(error && error != NO_UNITS) return error;

    doubleBuf[nDoubleBuf] = value*unit;
    if(echo) {
      if(error == NO_UNITS) {
	fprintf(echoFile,"\n");
      } else {
	fprintf(echoFile,"   # %g * %g = %g\n",
		value,unit,doubleBuf[nDoubleBuf]);
      }
    }
  }
  doubleAdd[nDoubleBuf] = var; nDoubleBuf++;
  return 0;
}

/* <TYPE == double (absolute temperature) */
int prs_packDoubleTemp(double* var) {
  int error = 0;

  if(nDoubleBuf >= SIZE_DOUBLEBUF) { prs_unPackDouble(); }
  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    if(!fscanf(inFile[inFileNo],"%lg",&doubleBuf[nDoubleBuf]))
      return ERR_ON_LINE;
    if(echo) fprintf(echoFile,"%g",doubleBuf[nDoubleBuf]);

    error = prs_readTempUnits(&doubleBuf[nDoubleBuf],1);
    if(error && error != NO_UNITS) return error;

    if(echo) fprintf(echoFile,"\n");
  }
  doubleAdd[nDoubleBuf] = var; nDoubleBuf++;
  return 0;
}

/* <TYPE == float */
int prs_packFloat(float* var) {
  int error = 0;

  if(nFloatBuf >= SIZE_FLOATBUF) { prs_unPackFloat(); }
  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    if(!fscanf(inFile[inFileNo],"%g",&floatBuf[nFloatBuf]))
      return ERR_ON_LINE;
    if(echo) fprintf(echoFile,"%g\n",floatBuf[nFloatBuf]);
  }
  floatAdd[nFloatBuf] = var; nFloatBuf++;
  return 0;
}

/* <TYPE == float(with units) */
int prs_packFloatUnits(float* var, int* dimen) {
  int error = 0;
  double unit;
  float value;

  if(nFloatBuf >= SIZE_FLOATBUF) { prs_unPackFloat(); }
  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    if(!fscanf(inFile[inFileNo],"%g",&value)) return ERR_ON_LINE;
    if(echo) fprintf(echoFile,"%g",value);

    error = prs_readUnits(&unit,dimen);
    if(error && error != NO_UNITS) return error;

    floatBuf[nFloatBuf] = value*unit;
    if(echo) {
      if(error == NO_UNITS) {
	fprintf(echoFile,"\n");
      } else {
	fprintf(echoFile," # %g * %g = %g\n",
		value,unit,floatBuf[nFloatBuf]);
      }
    }
  }
  floatAdd[nFloatBuf] = var; nFloatBuf++;
  return 0;
}

/* <TYPE == float (absolute temperature) */
int prs_packFloatTemp(float* var) {
  int error = 0;
  double dTemp;

  if(nFloatBuf >= SIZE_FLOATBUF) { prs_unPackFloat(); }
  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    if(!fscanf(inFile[inFileNo],"%lg",&dTemp)) return ERR_ON_LINE;
    floatBuf[nFloatBuf] = dTemp;
    if(echo) fprintf(echoFile,"%g",floatBuf[nFloatBuf]);

    error = prs_readTempUnits(&dTemp,1);
    if(error && error != NO_UNITS) return error;
    floatBuf[nFloatBuf] = dTemp;

    if(echo) fprintf(echoFile,"\n");
  }
  floatAdd[nFloatBuf] = var; nFloatBuf++;
  return 0;
}

/******************************************************************************
Title: prs_readString
Description: Read a string, up to non-punctuation white space or one less
             than the length of the string, and add the string
             terminating character.
Input:  len     The length of the string.
Output: string  The string read.
Return: Error flag (see reader.h).
*/
int prs_readString(char string[], int len) {
  int c;
  int i;
  int error = 0;

  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    for(i=0; i<len-1; i++) {
      c = getc(inFile[inFileNo]);
      if(c == EOL || c == EOF || c == SPACE || c == TAB ) {
	ungetc(c,inFile[inFileNo]);
	break;
      }
      string[i] = c;
    }
    string[i] = '\0';
    if(echo) fprintf(echoFile,"%s\n",string);
  }

  /* msg++; gabcastX(msg,string,len,0); */
  MPI_Bcast((void*)string,len,MPI_CHAR,0,comm);
  return 0;
}

/******************************************************************************
Title: prs_readLine
Description: Read a line, up to EOL, EOF, or one less than the length of
             the string, and add the string terminating character.
Input:  len     The length of the string.
Output: string  The string read.
Return: Error flag (see reader.h).
*/
int prs_readLine(char string[], int len) {
  int c;
  int i;
  int error = 0;

  if(!myProc) {
    error = prs_skipWS(0); if(error) return error;
    for(i=0; i<len-1; i++) {
      c = getc(inFile[inFileNo]);
      if(c == EOL || c == EOF) {
	ungetc(c,inFile[inFileNo]);
	break;
      }
      string[i] = c;
    }
    string[i] = '\0';
    if(echo) fprintf(echoFile,"%s\n",string);
  }

  /* msg++; gabcastX(msg,string,len,0); */
  MPI_Bcast((void*)string,len,MPI_CHAR,0,comm);
  return 0;
}

/******************************************************************************
Title: prs_read<TYPE>Array
Description: Read a <TYPE> array in parallel up to the size of the
             given array.  Supports multiple calls.  Handles the repetition
	     symbol in the input.
Input:  len            The length of the array.
        nToReadTotal   The total number of items to read through
	               sucessive calls.
Output: array          The array read.
        nRead          The number of items read so far.
Return: Error flag (see reader.h).
*/
/* <TYPE == int */
int prs_readIntArray(int* array, int len, int* nRead, int nToReadTotal) {
  int error = 0;
  char c;
  static int nReadTotal = 0;
  static int nRepeated = 0;
  static int valueRepeated;
  int i=0;

  *nRead = (len >= nToReadTotal-nReadTotal) ? nToReadTotal-nReadTotal : len;
  if(!myProc) {
    while(i<*nRead) {
      if(nRepeated <= 0) {
	error = prs_skipWS(0); if(error) return error;
	if(!fscanf(inFile[inFileNo],"%d",&array[i])) return ERR_IN_DATA_BLOCK;
	do { c = getc(inFile[inFileNo]); } while(c == SPACE);
	if(c != REPETITION) {
	  ungetc(c,inFile[inFileNo]);
	  i++;
	} else {
	  nRepeated = array[i];
	  error = prs_skipWS(0); if(error) return error;
	  if(!fscanf(inFile[inFileNo],"%d",&valueRepeated))
	    return ERR_IN_DATA_BLOCK;
	}
      } else {
	do {
	  array[i] = valueRepeated;
	  i++; nRepeated--;
	} while(i < *nRead && nRepeated > 0);
      }
    }
    nReadTotal += *nRead;
    if(nReadTotal >= nToReadTotal) {
      nReadTotal = 0;
      if(nRepeated > 0) return ERR_IN_DATA_BLOCK;
    }
  }

  /* msg++; gibcastX(msg,&nReadTotal,1,0); */
  /* msg++; gibcastX(msg,array,*nRead,0); */
  MPI_Bcast((void*)&nReadTotal,1,MPI_INT,0,comm);
  MPI_Bcast((void*)array,*nRead,MPI_INT,0,comm);

  if(!myProc && echo) {
    static int j = 0;
    for(i=0; i<*nRead; i++) {
      fprintf(echoFile," %d",array[i]); j++;
      if(j == 5) { fprintf(echoFile,"\n"); j = 0; }
    }
    if(nReadTotal == 0 && j != 5) { fprintf(echoFile,"\n"); j = 0; }
  }
  return 0;
}

/* <TYPE == double */
static void multByUnits(double* array, int n, double units) {
  int i;
  if(units != 1) for(i=0; i<n; i++) array[i] *= units;
}

int prs_readDoubleArrayUnits(double* array, int len, int* nRead,
			     int nToReadTotal, double units) {
  int error = 0;
  char c;
  static int nReadTotal = 0;
  static int nRepeated = 0;
  static double valueRepeated;
  int i=0;

  *nRead = (len >= nToReadTotal-nReadTotal) ? nToReadTotal-nReadTotal : len;
  if(!myProc) {
    while(i<*nRead) {
      if(nRepeated <= 0) {
	error = prs_skipWS(0); if(error) return error;
	if(!fscanf(inFile[inFileNo],"%lg",&array[i])) return ERR_IN_DATA_BLOCK;
	do { c = getc(inFile[inFileNo]); } while(c == SPACE);
	if(c != REPETITION) {
	  ungetc(c,inFile[inFileNo]);
	  i++;
	} else {
	  nRepeated = (int)array[i];
	  error = prs_skipWS(0); if(error) return error;
	  if(!fscanf(inFile[inFileNo],"%lg",&valueRepeated))
	    return ERR_IN_DATA_BLOCK;
	}
      } else {
	do {
	  array[i] = valueRepeated;
	  i++; nRepeated--;
	} while(i < *nRead && nRepeated > 0);
      }
    }
    nReadTotal += *nRead;
    if(nReadTotal >= nToReadTotal) {
      nReadTotal = 0;
      if(nRepeated > 0) return ERR_IN_DATA_BLOCK;
    }
    multByUnits(array,*nRead,units);
  }

  /* msg++; gibcastX(msg,&nReadTotal,1,0); */
  /* msg++; gdbcastX(msg,array,*nRead,0); */
  MPI_Bcast((void*)&nReadTotal,1,MPI_INT,0,comm);
  MPI_Bcast((void*)array,*nRead,MPI_DOUBLE,0,comm);

  if(!myProc && echo) {
    static int j = 0;
    double unitsInv = (units != 0) ? 1/units : 1;
    for(i=0; i<*nRead; i++) {
      fprintf(echoFile," %g",array[i]*unitsInv); j++;
      if(j == 5) { fprintf(echoFile,"\n"); j = 0; }
    }
    if(nReadTotal == 0 && j != 5) { fprintf(echoFile,"\n"); j = 0; }
  }
  return 0;
}

int prs_readDoubleArray(double* array, int len, int* nRead, int nToReadTotal) {
  double one = 1;
  return prs_readDoubleArrayUnits(array, len, nRead, nToReadTotal, one);
}

/* <TYPE == float */
static void fmultByUnits(float* array, int n, double units) {
  int i;
  if(units != 1) for(i=0; i<n; i++) array[i] *= units;
}

int prs_readFloatArrayUnits(float* array, int len, int* nRead,
			    int nToReadTotal, double units) {
  int error = 0;
  char c;
  static int nReadTotal = 0;
  static int nRepeated = 0;
  static float valueRepeated;
  int i=0;

  *nRead = (len >= nToReadTotal-nReadTotal) ? nToReadTotal-nReadTotal : len;
  if(!myProc) {
    while(i<*nRead) {
      if(nRepeated <= 0) {
	error = prs_skipWS(0); if(error) return error;
	if(!fscanf(inFile[inFileNo],"%g",&array[i])) return ERR_IN_DATA_BLOCK;
	do { c = getc(inFile[inFileNo]); } while(c == SPACE);
	if(c != REPETITION) {
	  ungetc(c,inFile[inFileNo]);
	  i++;
	} else {
	  nRepeated = (int)array[i];
	  error = prs_skipWS(0); if(error) return error;
	  if(!fscanf(inFile[inFileNo],"%g",&valueRepeated))
	    return ERR_IN_DATA_BLOCK;
	}
      } else {
	do {
	  array[i] = valueRepeated;
	  i++; nRepeated--;
	} while(i < *nRead && nRepeated > 0);
      }
    }
    nReadTotal += *nRead;
    if(nReadTotal >= nToReadTotal) {
      nReadTotal = 0;
      if(nRepeated > 0) return ERR_IN_DATA_BLOCK;
    }
    fmultByUnits(array,*nRead,units);
  }

  /* msg++; gibcastX(msg,&nReadTotal,1,0); */
  /* msg++; gsbcastX(msg,array,*nRead,0); */
  MPI_Bcast((void*)&nReadTotal,1,MPI_INT,0,comm);
  MPI_Bcast((void*)array,*nRead,MPI_FLOAT,0,comm);

  if(!myProc && echo) {
    static int j = 0;
    float unitsInv = (units != 0) ? 1/units : 1;
    for(i=0; i<*nRead; i++) {
      fprintf(echoFile," %g",array[i]*unitsInv); j++;
      if(j == 5) { fprintf(echoFile,"\n"); j = 0; }
    }
    if(nReadTotal == 0 && j != 5) { fprintf(echoFile,"\n"); j = 0; }
  }
  return 0;
}

int prs_readFloatArray(float* array, int len, int* nRead, int nToReadTotal) {
  double one = 1;
  return prs_readFloatArrayUnits(array, len, nRead, nToReadTotal, one);
}

/******************************************************************************
Title: prs_openReader
Description: Initialize the routines for reading input files.
Inputs: inFileName0      The initial input file name.
        echoFileName     The echo file name.
	length_fileName  The length of the two file names (excluding '\0').
	echo0            Boolean for echoing the input.
Return: Error flag (see reader.h).
*/
int prs_openReader(char* inFileName0, char* echoFileName, int length_fileName,
	       int echo0) {
  int error = 0;
  int i;

  /* myProc = mynodeX(); */
  MPI_Comm_rank(comm,&myProc);

  /* INITIALIZE GENERAL INPUT AND ECHO FILES */
  SIZE_FILENAME = length_fileName + 1;
  if(!myProc) {
    for(i=0; i<DEPTH_INCLUDES; i++) {
      inFileName[i] = (char*) malloc(SIZE_FILENAME);
      if(inFileName[i] == NULL) return ERR_MEMORY;
    }
    strcpy(inFileName[0],inFileName0);
    error = prs_beginFile(); if(error) return error;
  }
  echo = echo0;
  if(!myProc && echo) {
    if(echoFileName == '\0') {
      echoFile = stdout;
    } else {
      echoFile = fopen(echoFileName,"w");
      if(echoFile == NULL) {
	fprintf(stderr,"Error opening echo file %s\n",echoFileName);
	echo = 0;
      }
    }
  }

  /* INITIALIZE BUFFERS */
  intBuf = (int*) malloc(SIZE_INTBUF*sizeof(int));
  intAdd = (int**) malloc(SIZE_INTBUF*sizeof(int*));
  doubleBuf = (double*) malloc(SIZE_DOUBLEBUF*sizeof(double));
  doubleAdd = (double**) malloc(SIZE_DOUBLEBUF*sizeof(double*));
  floatBuf = (float*) malloc(SIZE_FLOATBUF*sizeof(float));
  floatAdd = (float**) malloc(SIZE_FLOATBUF*sizeof(float*));
  if(intBuf == NULL || intAdd == NULL
     || doubleBuf == NULL || doubleAdd == NULL
     || floatBuf == NULL || floatAdd == NULL
     ) return ERR_MEMORY;

  return 0;
}

/******************************************************************************
Title: prs_closeReader
Description: Close the routines for reading input files and process error flag.
Input:  error  The error flag (see reader.h).  Call with error = 0 if no
               error information is desired.
*/
void prs_closeReader(int error) {
  int i;

  if(!myProc) {
    if(!inFile[0]) fclose(inFile[0]);
    if(echo) {
      fprintf(echoFile,"\n");
      if(echoFile != stdout) fclose(echoFile);
    }
  }

  prs_unPackInt();
  prs_unPackDouble();
  prs_unPackFloat();

  free(intBuf);
  free(intAdd);
  free(doubleBuf);
  free(doubleAdd);
  free(floatBuf);
  free(floatAdd);

  prs_processReaderError(error);

  for(i=0; i<DEPTH_INCLUDES; i++) free(inFileName[i]);
}

/******************************************************************************
Title: prs_processReaderError
Description: Process error flag by writing the appropriate message.
Input: error  The error flag (see reader.h).
*/
void prs_processReaderError(int error) {
  if(error) {
    switch(error) {
    case ERR_UNMATCHED_END_COMMENT_CMD:
    case ERR_UNKNOWN_CMD:
    case ERR_IN_DATA_BLOCK:
    case ERR_IN_GRID_ARRAY:
    case ERR_UNSUPPORTED_OPTION:
    case ERR_BEGIN_LIST_NOT_READ:
    case ERR_SECTION_READ:
    case ERR_SUBSECTION_READ:
    case ERR_SUBSUBSECTION_READ:
    case ERR_NONNESTED_PARENTHESES:
    case ERR_ARITHMETIC_OP:
    case ERR_UNITS_NEGATIVE:
    case ERR_UNITS_ADDITION_FOUND:
    case ERR_UNITS_WRONG_DIMEN:
    case ERR_UNITS_TOO_LONG:
    case ERR_UNITS_IMPROPER_EXPR:
    case ERR_UNKNOWN_UNIT:
    case ERR_UNITS:
    case ERR_ON_LINE:
      if(!myProc) {
	switch(error) {
	case ERR_UNMATCHED_END_COMMENT_CMD:
	  fprintf(stderr,"Unmatched end comment command\n");
	  break;
	case ERR_UNKNOWN_CMD:
	  fprintf(stderr,"Unknown %c command\n",COMMAND);
	  break;
	case ERR_IN_DATA_BLOCK:
	  fprintf(stderr,"Error in data block\n");
	  break;
	case ERR_IN_GRID_ARRAY:
	  fprintf(stderr,"Grid array expected\n");
	  break;
	case ERR_UNSUPPORTED_OPTION:
	  fprintf(stderr,"Unsupported input option\n");
	  break;
	case ERR_BEGIN_LIST_NOT_READ:
	  fprintf(stderr,
		  "Beginning of list symbol %c not read\n",BEGIN_LIST);
	  break;
	case ERR_SECTION_READ:
	  fprintf(stderr,"Premature beginning of SECTION command found\n");
	  break;
	case ERR_SUBSECTION_READ:
	  fprintf(stderr,"Premature beginning of SUBSECTION command found\n");
	  break;
	case ERR_SUBSUBSECTION_READ:
	  fprintf(stderr,
		  "Premature beginning of SUBSUBSECTION command found\n");
	  break;
	case ERR_NONNESTED_PARENTHESES:
	  fprintf(stderr,"Expression has nonnested parentheses\n");
	  break;
	case ERR_ARITHMETIC_OP:
	  fprintf(stderr,
		  "Arithmetic operation error in evaluating expression\n");
	  break;
	case ERR_UNITS_NEGATIVE:
	  fprintf(stderr,"Units expression is not positive\n");
	  break;
	case ERR_UNITS_ADDITION_FOUND:
	  fprintf(stderr,
		  "Binary addition or subtraction in units expression\n");
	  break;
	case ERR_UNITS_WRONG_DIMEN:
	  fprintf(stderr,"Unit expression has the wrong dimensions\n");
	  break;
	case ERR_UNITS_TOO_LONG:
	  fprintf(stderr,"Units expression is too long\n");
	  break;
	case ERR_UNITS_IMPROPER_EXPR:
	  fprintf(stderr,
		  "Units expression improperly formed or unknown unit\n");
	  break;
	case ERR_UNKNOWN_UNIT:
	  fprintf(stderr,"Unknown unit\n");
	  break;
	case ERR_UNITS:
	  fprintf(stderr,"Unit expression error\n");
	  break;
	}
	if(inFileNo >= 0)
	  fprintf(stderr,"Error in input file %s on line number %d\n",
		  inFileName[inFileNo],inFileLineNo[inFileNo]);
      }
      break;
    case ERR_BEFORE_LINE:
      if(!myProc) {
	fprintf(stderr,
		"Error in input file %s before line number %d\n",
		inFileName[inFileNo],inFileLineNo[inFileNo]);
      }
      break;
    case ERR_MEMORY:
      fprintf(stderr,"Error allocating memory on processor %d\n",myProc);
      break;
    case ERR_OPENING_FILE:
      fprintf(stderr,"Error opening input file %s\n",inFileName[inFileNo]);
      break;
    case ERR_NESTING_FILE:
      fprintf(stderr,"Too many nested include files: maximum %d allowed\n",
	      DEPTH_INCLUDES);
      break;
    case ERR_EOF:
      break;
    case ERR_IN_CODE:
      fprintf(stderr,"Sorry, there is an error in the computer code\n");
      break;
    case ERR_SECTION_NOT_READ:
    case ERR_SUBSECTION_NOT_READ:
    case ERR_SUBSUBSECTION_NOT_READ:
    case ERR_SECTION_KEY_NOT_READ:
    case ERR_SUBSECTION_KEY_NOT_READ:
    case ERR_SUBSUBSECTION_KEY_NOT_READ:
    case ERR_KEY_NOT_READ:
    case ERR_END_LIST_NOT_READ:
      if(!myProc) {
 	switch(error) {
	case ERR_SECTION_NOT_READ:
	  fprintf(stderr,"Beginning of SECTION command not found\n");
	  break;
	case ERR_SUBSECTION_NOT_READ:
	  fprintf(stderr,"Beginning of SUBSECTION command not found\n");
	  break;
	case ERR_SUBSUBSECTION_NOT_READ:
	  fprintf(stderr,"Beginning of SUBSUBSECTION command not found\n");
	  break;
	case ERR_SECTION_KEY_NOT_READ:
	case ERR_SUBSECTION_KEY_NOT_READ:
	case ERR_SUBSUBSECTION_KEY_NOT_READ:
	case ERR_KEY_NOT_READ:
	  switch(error) {
	  case ERR_SECTION_KEY_NOT_READ:
	    fprintf(stderr,"Section k");
	    break;
	  case ERR_SUBSECTION_KEY_NOT_READ:
	    fprintf(stderr,"Subsection k");
	    break;
	  case ERR_SUBSUBSECTION_KEY_NOT_READ:
	    fprintf(stderr,"Subsubsection k");
	    break;
	  case ERR_KEY_NOT_READ:
	    fprintf(stderr,"K");
	    break;
	  }
	  {
	    int k;
	    fprintf(stderr,"ey not found: %s\n",save_keyWord[0]);
	    for(k=1; k<save_nKeyWords; k++)
	      fprintf(stderr,"               %s\n",save_keyWord[k]);
	  }
	  break;
	case ERR_END_LIST_NOT_READ:
	  fprintf(stderr,"End of list symbol %c not found\n",END_LIST);
	  break;
	}
	if(skipping) {
	  if(inFileNo >= 0) {
	    fprintf(stderr,"Error between input file %s line number %d\n",
		    beginSkipFileName,beginSkipFileLineNo);
	    fprintf(stderr,"          and input file %s line number %d\n",
		    inFileName[inFileNo],inFileLineNo[inFileNo]);
	  } else {
	    fprintf(stderr,"Error after input file %s line number %d\n",
		    beginSkipFileName,beginSkipFileLineNo);
	  }
	} else {
	  if(inFileNo >= 0)
	    fprintf(stderr,"Error in input file %s on line number %d\n",
		    inFileName[inFileNo],inFileLineNo[inFileNo]);
	}
      }
      break;
    case READ_LITERAL_CMD:
      break;
    default:
      fprintf(stderr,"Error on processor %d\n",myProc);
    }
    if(!myProc && (error == ERR_EOF || inFileNo < 0))
      fprintf(stderr,"Premature end of data\n");
  }
}
