static float* g_conc;
static float* g_totcoef;
static float* g_result;
static int* g_compId;
static int g_nComps;
static int* g_pad;

const int dimen = 3;
const int maxSimploidVtxs = (1+dimen)*(3+dimen)/4;

typedef int bool;
const bool FALSE = 0;
const bool TRUE = 1;

extern "C" bool isWell(int k[3]);

class Point
{
  double coord[dimen];
public:
  double & operator[] (const int i) // get a coordinate
    {
      return coord[i];
    }

  friend bool operator == (const Point& pt0, const Point& pt1);
  friend inline Point convex_combine(const Point& vtx_i, const Point& vtx_j, double baryWght);
};

bool operator == (const Point& pt0, const Point& pt1)
{
  for (int i = 0; i < dimen; ++i)
    if (pt0.coord[i] != pt1.coord[i]) return FALSE;
  return TRUE;
}

inline Point convex_combine(const Point& vtx_i, const Point& vtx_j, double baryWght)
{
  Point answer;
  int i = dimen-1;
  do
    answer.coord[i] = (vtx_i.coord[i] +
		       baryWght * (vtx_j.coord[i] - vtx_i.coord[i]));
  while (i--);
  return answer;
}

inline void swap(Point& x, Point& y) {Point t = x; x = y; y = t;}

typedef double *intercept[dimen];

inline void swap(int& x, int& y) {int t = x; x = y; y = t;}

inline double min(double a, double b) {return (a < b)? a: b;}

inline void swap(double& x, double& y) {double t = x; x = y; y = t;}
inline double abs(double x) {return x < 0.0? -x: x;}

double tet_volume(Point tet_vtx[1+dimen])
{
  double mtx[dimen][dimen];
  
  for (int i = 0; i < dimen; ++i)
    for (int j = 0; j < dimen;  ++j)
      mtx[i][j] = tet_vtx[i+1][j] - tet_vtx[0][j];
  
  double prod = 1.0;
  for (i = 0; i < dimen; ++i)
    {
      int maxPos = i;
      for (int j = i+1; j < dimen; ++j)
	if (abs(mtx[maxPos][i]) < abs(mtx[j][i]))
	  maxPos = j;
      
      if (i != maxPos)
	{
	  for (int k = i; k < dimen; ++k)
	    swap(mtx[i][k], mtx[maxPos][k]);
	  prod = -prod;
	}
      
      prod *= mtx[i][i] / (i + 1);
      if (prod == 0.0) return prod;
      for (j = i+1; j < dimen; ++j)
	{
	  double div = mtx[j][i] / mtx[i][i];
	  for (int k = i+1; k < dimen; ++k)
	    mtx[j][k] -= mtx[i][k] * div;
	}
    }
  return prod;
}

static void tetra_pw_integrate(int sign, Point tet_vtx[1+dimen],
			       const int boxSize[dimen],
			       intercept coord, int flat_dim, int lbound[dimen]);

inline bool bit(unsigned mask, unsigned pos)
{ return (mask >>pos) & 1; }

inline unsigned lo1bits(int n1Bits)
{
  return (1 << n1Bits) - 1;
}

inline unsigned hi1bits(int n1Bits, int nBits)
{
  return lo1bits(n1Bits) << (nBits - n1Bits);
}

const unsigned evenBits = ~0U/3;
inline void nextSame1bits(unsigned& mask, unsigned& sign)
{
  unsigned x = mask & -mask;
  mask += x;
  unsigned y = mask & -mask;
  sign ^= !!((x|y) & evenBits);
  y -= x;
  while (!(y & 1)) y >>= 1;
  mask |= y >> 1;
}

static void simploid_pw_integrate(int sign,
				  Point smpl_vtx[maxSimploidVtxs], int s,
				  const int boxSize[dimen],
				  intercept coord, int flat_dim, int lbound[dimen])
     // compute the integral of a function over an (s,dimen-s) simploid
     // where the function is piecewise in dimensions < flat_dim and
     // identified by elements of lbound in dimension >= flat_dim
{
  if (s < 0 || s > dimen) return;
  int jump[2];
  jump[0] = 1;
  jump[1] = 1+dimen-s;
  unsigned lastPerm = hi1bits(s, dimen);
  unsigned perm = lo1bits(s);
  unsigned signBit = 0;
  for(;;)
    {
      Point child_vtx[1+dimen];
      int index = 0;
      child_vtx[0] = smpl_vtx[index];
      
      for (int j = 0; j < dimen; ++j)
	{
	  index += jump[bit(perm,j)];
	  child_vtx[1+j] = smpl_vtx[index];
	}
      
      tetra_pw_integrate(signBit?-sign:sign, child_vtx, boxSize, coord, flat_dim, lbound);
      if (perm < lastPerm)
	nextSame1bits(perm,signBit);
      else break;
    }
}

extern "C" void tetint
(Point tet_vtx[1+dimen], const int sign, int lbound[dimen], int pad[3],
 float* conc, float* totcoef, double concvtx[2][3], const int concsize[3],
 int* compId, float* results, const int nComps);

static void tetra_pw_integrate(int sign, Point tet_vtx[1+dimen],
			       const int boxSize[dimen],
			       intercept coord, int flat_dim, int lbound[dimen])
     // for a piecewise function defined over a cubic grid,
     // compute the integral over a tetrahedron
     // which is known to lie strictly between two parallel planes of
     // the grid in dimensions flat_dim and higher, with that plane
     // identified by lbound[j], j >= flat_dim,
     // and known to lie beyond the lbound[i] plane and
     // before the boxSize[i] plane in dimension i, i < flat_dim
{
  if (flat_dim-- == 0)
    // this tetrahedron lies entirely within one element of the cubic grid
    {
      if (!isWell(lbound))
	{
	  double concvtx[2][3];
	  for (int j = 0; j < 2; ++j)
	    for (int k = 0; k < 3; ++k)
	      concvtx[j][k] = coord[k][j+lbound[k]];
	  tetint(tet_vtx, sign, lbound, g_pad,
		  g_conc, g_totcoef, concvtx, boxSize, g_compId, g_result, g_nComps);
	}
      return;
    }
  
  // sort the tetra vtxs by flat_dim coordinate; keep track of parity
  for (int i = 0; i < dimen; ++i)
    {
      int min_pos = i;
      for (int j = i+1; j <= dimen; ++j)
	if (tet_vtx[min_pos][flat_dim] > tet_vtx[j][flat_dim])
	  min_pos = j;
      if (min_pos != i)
	{
	  swap(tet_vtx[min_pos], tet_vtx[i]);
	  sign = -sign;
	}
    }

  // find the rightmost plane in flat_dim left of the tetra
  double minCoord = tet_vtx[0][flat_dim];
  double* coord_d = coord[flat_dim];
  if (minCoord < coord_d[0])
    lbound[flat_dim] = -1;
  else if (coord_d[boxSize[flat_dim]] <= minCoord)
    lbound[flat_dim] = boxSize[flat_dim];
  else
    {
      int lo = 0;
      int hi = boxSize[flat_dim];
      int nowGap = hi - lo;
      int oldGap = 2*nowGap;
      while (nowGap != 1)
	// coord_d[lo] <= minCoord < coord_d[hi]
	{
	  int mid;
	  if (2*nowGap <= oldGap)
	    {
	      mid = int(((minCoord - coord_d[lo]) * hi +
			 (coord_d[hi] - minCoord) * lo) /
			(coord_d[hi] - coord_d[lo]));
	      if      (mid == lo) ++mid;
	      else if (mid == hi) --mid;
	    }
	  else
	    mid = (lo + hi) / 2;
	  oldGap = nowGap;
	  (minCoord < coord_d[mid] ? hi: lo) = mid;
	  nowGap = hi - lo;
	}
      lbound[flat_dim] = lo;
    }

  double cutPlane[2];
  cutPlane[0] = coord_d[lbound[flat_dim]];
  cutPlane[1] = coord_d[1+lbound[flat_dim]];
  int wayLeft_vtx = 0;
  int wayRight_vtx = 1+dimen;
  while (wayRight_vtx > 0 &&
	 tet_vtx[1+dimen-wayRight_vtx][flat_dim] <= cutPlane[1])
    --wayRight_vtx;
  while (lbound[flat_dim] < boxSize[flat_dim] && wayLeft_vtx <= dimen)
    {
      // Calculate the integral over the next slice between cut planes.
      // The left cut plane has *wayLeft_vtx* vertices strictly to its left,
      // the right cut plane has *wayRight_vtx vertices strictly to its right,
      // and *integrand* is the integral of the function over the part of the
      // tetrahedron left of the left cut plane.

      // There are several ways we could proceed.  We could decompose
      // this region into tetrahedra.  That's hard.  Instead, we could
      // integrate over the entire tetra, then subtract the integrals over
      // the parts to the far left and far right.  This is most attractive
      // when there are few vertices out there.

      // Generalizing the idea, we can compute the integral without decomposing
      // the middle region in essentially two ways:   (number of integrals req)
      // 1. I(tetra) - I(far left) - I(far right): 1 + C(dimen,Rt-1) + C(dimen,Lf-1)
      // 2. I(middle-left) - I(far left)         : C(dimen,Lf) + C(dimen,Rt-1)
      // (and symmetric variant).
      // 1+C(dimen,Lf-1) < C(dimen,Lf) when Lf <= dimen/2.
      // (Here, Rt == wayRight_vtx, Lf = wayLeft_vtx)


      // So, our strategy:
      // for each plane, integrate the tetra over the half-space with fewer
      // vertices, then integrate over the entire tetra if necessary.

      // Integrating over a half-space:
      // When a cut plane partitions the vertices of the tetra into
      // I and K vertices, I+K == 1+dimen, the cross section is topologically
      // a (I-1,K-1) simploid.  The entire tetra is split into a (I,K-1) simploid
      // and a (I-1,K) simploid.  An (I,K) simploid can be triangulated into C(I,K)
      // simplices.

      Point cut_vtx[maxSimploidVtxs];

      // compute the integral of the tetra over one side of the left plane
      bool subtract_left_tip = (2*wayLeft_vtx <= 1+dimen);
      {
	int nVtxs = 0;
	
	for (int i = 0; i < wayLeft_vtx; ++i)
	  {
	    if (subtract_left_tip)
	      cut_vtx[nVtxs++] = tet_vtx[i];
	    double endPoint = tet_vtx[i][flat_dim];
	    double cut_to_i_dist = cutPlane[0] - endPoint;
	    for (int j = wayLeft_vtx; j <= dimen; ++j)
	      {
		double j_to_i_dist = tet_vtx[j][flat_dim]  - endPoint;
		double baryWght = cut_to_i_dist / j_to_i_dist;
		cut_vtx[nVtxs++] = convex_combine(tet_vtx[i], tet_vtx[j], baryWght);
	      }
	  }
	
	if (subtract_left_tip)
	  simploid_pw_integrate(-sign, cut_vtx, wayLeft_vtx-1,
				boxSize, coord, flat_dim, lbound);
	else
	  {
	    for (int j = wayLeft_vtx; j <= dimen; ++j)
	      cut_vtx[nVtxs++] = tet_vtx[j];
	    
	    simploid_pw_integrate(sign, cut_vtx, wayLeft_vtx,
				  boxSize, coord, flat_dim, lbound);
	  }
      }
      
      // compute the integral of the tetra over one side of the right plane
      bool subtract_right_tip = (2*wayRight_vtx < 1+dimen);
      {
	int nVtxs = 0;
	
	for (int i = 0; i < 1+dimen-wayRight_vtx; ++i)
	  {
	    if (!subtract_right_tip)
	      cut_vtx[nVtxs++] = tet_vtx[i];
	    double endPoint = tet_vtx[i][flat_dim];
	    double cut_to_i_dist = cutPlane[1] - endPoint;
	    for (int j = 1+dimen-wayRight_vtx; j <= dimen; ++j)
	      {
		double j_to_i_dist = tet_vtx[j][flat_dim]  - endPoint;
		double baryWght = cut_to_i_dist / j_to_i_dist;
		cut_vtx[nVtxs++] = convex_combine(tet_vtx[i], tet_vtx[j], baryWght);
	      }
	  }
	
	if (!subtract_right_tip)
	  {
	    simploid_pw_integrate(sign, cut_vtx, dimen-wayRight_vtx,
				  boxSize, coord, flat_dim, lbound);
	  }
	else
	  {
	    for (int j = 1+dimen-wayRight_vtx; j <= dimen; ++j)
	      cut_vtx[nVtxs++] = tet_vtx[j];
	    
	    simploid_pw_integrate(-sign, cut_vtx, 1+dimen-wayRight_vtx,
				  boxSize, coord, flat_dim, lbound);
	  }
      }

      if (subtract_left_tip && subtract_right_tip)
	{
	  for (int j = 0; j <= dimen; ++j)
	    cut_vtx[j] = tet_vtx[j];
	  tetra_pw_integrate(sign, cut_vtx, boxSize, coord, flat_dim, lbound);
	}


      // reestablish invariants
      while (wayLeft_vtx <= dimen &&
	     tet_vtx[wayLeft_vtx][flat_dim] < cutPlane[1])
	++wayLeft_vtx;

      if (++lbound[flat_dim] < boxSize[flat_dim])
	{
	  cutPlane[0] = cutPlane[1];
	  cutPlane[1] = coord_d[1+lbound[flat_dim]];
	  
	  while (wayRight_vtx > 0 &&
		 tet_vtx[1+dimen-wayRight_vtx][flat_dim] <= cutPlane[1])
	    --wayRight_vtx;
	}
    }
}

void
cube_pw_integrate(int sign, Point hex_vtx[1<<dimen], intercept coord,
		  const int boxSize[dimen])
     // for a piecewise function defined over a cubic grid,
     // compute the integral over a Kuhn's approximation of
     // a warped cube
{
  Point tet_vtx[1+dimen];

  // Possibly, this code is a specific instance of a general method for dividing
  // a dimen-cube into 2**dimen - dimen simplices, which may be worth exploring

  // These 5 tetrahedra are each positively oriented, which means exactly that
  // det ( 1  1  1  1 ) is positive for each.
  //     ( v0 v1 v2 v3)


  int lbound[dimen];
  lbound[0] = lbound[1] = lbound[2] = 0;
  int mask = (sign == -1)? 7: 0;
  
  tet_vtx[0] = hex_vtx[0^mask];
  tet_vtx[1] = hex_vtx[1^mask];
  tet_vtx[2] = hex_vtx[2^mask];
  tet_vtx[3] = hex_vtx[4^mask];
  tetra_pw_integrate(sign, tet_vtx, boxSize, coord, dimen, lbound);
  
  tet_vtx[0] = hex_vtx[3^mask];
  tet_vtx[1] = hex_vtx[2^mask];
  tet_vtx[2] = hex_vtx[1^mask];
  tet_vtx[3] = hex_vtx[7^mask];
  tetra_pw_integrate(sign, tet_vtx, boxSize, coord, dimen, lbound);
  
  tet_vtx[0] = hex_vtx[5^mask];
  tet_vtx[1] = hex_vtx[4^mask];
  tet_vtx[2] = hex_vtx[7^mask];
  tet_vtx[3] = hex_vtx[1^mask];
  tetra_pw_integrate(sign, tet_vtx, boxSize, coord, dimen, lbound);
  
  tet_vtx[0] = hex_vtx[6^mask];
  tet_vtx[1] = hex_vtx[7^mask];
  tet_vtx[2] = hex_vtx[4^mask];
  tet_vtx[3] = hex_vtx[2^mask];
  tetra_pw_integrate(sign, tet_vtx, boxSize, coord, dimen, lbound);
  
  tet_vtx[0] = hex_vtx[7^mask];
  tet_vtx[1] = hex_vtx[4^mask];
  tet_vtx[2] = hex_vtx[2^mask];
  tet_vtx[3] = hex_vtx[1^mask];
  tetra_pw_integrate(sign, tet_vtx, boxSize, coord, dimen, lbound);
}

extern "C"
{
  void cubeint(float* conc, float* totcoef, int sign, float* result, int nComps,
	       double vtx[1<<dimen][dimen],
	       intercept coord,
	       const int boxSize[dimen],
	       int* compId,  int* pad)
    {
      g_conc = conc;
      g_totcoef = totcoef;
      g_compId = compId;
      g_result = result;
      g_nComps = nComps;
      g_pad = pad;
      Point cube_vtx[1<<dimen];
      for (int k = 0; k < (1<<dimen); ++k)
	for (int j = 0; j < dimen; ++j)
	  cube_vtx[k][j] = vtx[k][j];
      cube_pw_integrate(sign, cube_vtx, coord, boxSize);
    }

  double cubevol(int sign, double vtx[1<<dimen][dimen])
    {
      double volume = 0.;
      int mask = (sign == -1)? 7: 0;
      Point tet_vtx[1+dimen];
      Point hex_vtx[1<<dimen];
      for (int k = 0; k < (1<<dimen); ++k)
	for (int j = 0; j < dimen; ++j)
	  hex_vtx[k][j] = vtx[k][j];
      
      tet_vtx[0] = hex_vtx[0^mask];
      tet_vtx[1] = hex_vtx[1^mask];
      tet_vtx[2] = hex_vtx[2^mask];
      tet_vtx[3] = hex_vtx[4^mask];
      volume += tet_volume(tet_vtx);
      tet_vtx[0] = hex_vtx[3^mask];
      tet_vtx[1] = hex_vtx[2^mask];
      tet_vtx[2] = hex_vtx[1^mask];
      tet_vtx[3] = hex_vtx[7^mask];
      volume += tet_volume(tet_vtx);
      tet_vtx[0] = hex_vtx[5^mask];
      tet_vtx[1] = hex_vtx[4^mask];
      tet_vtx[2] = hex_vtx[7^mask];
      tet_vtx[3] = hex_vtx[1^mask];
      volume += tet_volume(tet_vtx);
      tet_vtx[0] = hex_vtx[6^mask];
      tet_vtx[1] = hex_vtx[7^mask];
      tet_vtx[2] = hex_vtx[4^mask];
      tet_vtx[3] = hex_vtx[2^mask];
      volume += tet_volume(tet_vtx);
      tet_vtx[0] = hex_vtx[7^mask];
      tet_vtx[1] = hex_vtx[4^mask];
      tet_vtx[2] = hex_vtx[2^mask];
      tet_vtx[3] = hex_vtx[1^mask];
      volume += tet_volume(tet_vtx);

      if (sign == -1)
	volume = -volume;
      return volume;
    }
}
