/******************************************************************************
 * Title: padme.c
 *
 * Description: Routines to communicate ghost or padded portions of
 *              rectangular arrays (in a rectangular domain decomposition).
 *              These routines are callable by Fortran.
 *
 *     PadArray(): 
 *       Pads an array already possessing space by pad
 *       cells in each direction.  The order of execution
 *       of message passing is important since we handle 
 *       "corner" points, not just padding of faces.
 *
 *     PadStencil() and RPadStencil(): 
 *       Pads a stencil-array already possessing space by pad
 *       cells in each direction.   The order of execution
 *       of message passing is important since we handle
 *       "corner" points not just padding of faces.
 *
 *     PadVec():
 *       Pads a vector is a spcecific direction.
 *   
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "../Main/msg.h"
#include "../Main/ftn.h"

#define DOWN    4
#define SOUTH  10
#define WEST   12
#define ME     13
#define EAST   14
#define NORTH  16
#define UP     22

#define max(a,b) ((a)>(b) ? (a) : (b))

static int wkSize = 0;
static double* wk = NULL;

void padmeClose() {
  if(wk != NULL) {
    free(wk);
    wkSize = 0;
  }
}

/******************************************************************************
 * ARGUMENTS for PadArray, PadStencil, and RPadStencil
 * 
 * IN/OUT
 *  array: The array, of size declared below, to be padded
 * 
 * IN
 *  stensize: Size of array stencil     
 *  n1,n2,n3: Local size of array
 *  off1,off2,off3 : Offset by 1 if grid-centered, by 0 if cell-centered
 *  pad1,pad2,pad3 : Amount to pad in each direction
 * 
 *  myNbr:   Node numbers of neighboring domains (-1 if no neighbor)
 *  msgRoot: The base number to use for message passing, the
 *           numbers actually used are msgRoot to msgRoot+5
 * 
 * OUT
 *  error: Error code, 0 if no error, -1 otherwise.
 *****************************************************************************/

/* FORTRAN: array(stensize,1-pad1:n1+pad1,1-pad2:n2+pad2,1-pad3:n3+pad3) */
#define array_(m,i1,i2,i3) array[m + (*stensize)*(i1 + nn1*(i2 + nn2*(i3)))]

void PadStencil(double* array,
		int* stensize, int* n1, int* n2, int* n3,
		int* off1, int* off2, int* off3,
		int* pad1, int* pad2, int* pad3,
		int* myNbr, int* msgRoot, int* error) {

  int nn1 = (*n1)+2*(*pad1);
  int nn2 = (*n2)+2*(*pad2);
  int nn3 = (*n3)+2*(*pad3);

  int pp1 = (*pad1)+(*off1);
  int pp2 = (*pad2)+(*off2);
  int pp3 = (*pad3)+(*off3);

  int needed1 = (*stensize)*(*pad1)*nn2*nn3;
  int needed2 = (*stensize)*(*pad2)*nn1*nn3;
  int needed3 = (*stensize)*(*pad3)*nn1*nn2;

  int i1,i2,i3,k,m;
  int wkSizeNeeded;
  int msg = *msgRoot;
  MPI_Status status;

/* Check available memory */

  wkSizeNeeded = max( needed1, max(needed2, needed3) );
  if(wkSizeNeeded > wkSize) {
    if(wk != NULL) free(wk);
    wk = (double*) malloc(wkSizeNeeded*sizeof(double));
    if(wk == NULL) {
      *error = -1;
      fprintf(stderr,
        "PADME FATAL ERROR 1: Not enough memory, proc %d, need %d, have %d\n",
	myNbr[ME],wkSizeNeeded,wkSize);
      exit(-1);
    }
    wkSize = wkSizeNeeded;
  }

/* Communicate East (+ x1 or x direction) */

  if(myNbr[EAST] >= 0) {
    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = nn1 - pp1 - *pad1; i1 < nn1 - pp1; i1++)
    for(m = 0; m < *stensize; m++) {
      wk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed1) fprintf(stderr,"PADME FATAL ERROR 2: wrong count\n");

    MPI_Send((void*)wk,needed1,MPI_DOUBLE,myNbr[EAST],msg,comm);
  }

  if(myNbr[WEST] >= 0) {
    MPI_Recv((void*)wk,needed1,MPI_DOUBLE,myNbr[WEST],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < *pad1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = wk[k];
      k++;
    }
    if(k != needed1) fprintf(stderr,"PADME FATAL ERROR 3: wrong count\n");
  }
  
/* Communicate West (- x1 or x direction) */

  msg++;
  if(myNbr[WEST] >= 0) {
    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = pp1; i1 < pp1 + *pad1; i1++)
    for(m = 0; m < *stensize; m++) {
      wk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed1) fprintf(stderr,"PADME FATAL ERROR 4: wrong count\n");

    MPI_Send((void*)wk,needed1,MPI_DOUBLE,myNbr[WEST],msg,comm);
  }

  if(myNbr[EAST] >= 0) {
    MPI_Recv((void*)wk,needed1,MPI_DOUBLE,myNbr[EAST],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = nn1 - *pad1; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = wk[k];
      k++;
    }
    if(k != needed1) fprintf(stderr,"PADME FATAL ERROR 5: wrong count\n");
  }

/* Communicate North (+ x2 or y direction) */

  msg++;
  if(myNbr[NORTH] >= 0) {
    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = nn2 - pp2 - *pad2; i2 < nn2 - pp2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      wk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed2) fprintf(stderr,"PADME FATAL ERROR 6: wrong count\n");

    MPI_Send((void*)wk,needed2,MPI_DOUBLE,myNbr[NORTH],msg,comm);
  }

  if(myNbr[SOUTH] >= 0) {
    MPI_Recv((void*)wk,needed2,MPI_DOUBLE,myNbr[SOUTH],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < *pad2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = wk[k];
      k++;
    }
    if(k != needed2) fprintf(stderr,"PADME FATAL ERROR 7: wrong count\n");
  }

/* Communicate South (- x2 or y direction) */

  msg++;
  if(myNbr[SOUTH] >= 0) {
    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = pp2; i2 < pp2 + *pad2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      wk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed2) fprintf(stderr,"PADME FATAL ERROR 8: wrong count\n");

    MPI_Send((void*)wk,needed2,MPI_DOUBLE,myNbr[SOUTH],msg,comm);
  }

  if(myNbr[NORTH] >= 0) {
    MPI_Recv((void*)wk,needed2,MPI_DOUBLE,myNbr[NORTH],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = nn2 - *pad2; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = wk[k];
      k++;
    }
     if(k != needed2) fprintf(stderr,"PADME FATAL ERROR 9: wrong count\n");
  }

/* Communicate Up (+ x3 or z direction) */

  msg++;
  if(myNbr[UP] >= 0) {
    k = 0;
    for(i3 = nn3 - pp3 - *pad3; i3 < nn3 - pp3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      wk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed3) fprintf(stderr,"PADME FATAL ERROR 10: wrong count\n");

    MPI_Send((void*)wk,needed3,MPI_DOUBLE,myNbr[UP],msg,comm);
  }

  if(myNbr[DOWN] >= 0) {
    MPI_Recv((void*)wk,needed3,MPI_DOUBLE,myNbr[DOWN],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < *pad3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = wk[k];
      k++;
    }
    if(k != needed3) fprintf(stderr,"PADME FATAL ERROR 11: wrong count\n");
  }
  
/* Communicate Down (- x3 or z direction) */

  msg++;
  if(myNbr[DOWN] >= 0) {
    k = 0;
    for(i3 = pp3; i3 < pp3 + *pad3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      wk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed3) fprintf(stderr,"PADME FATAL ERROR 12: wrong count\n");

    MPI_Send((void*)wk,needed3,MPI_DOUBLE,myNbr[DOWN],msg,comm);
  }

  if(myNbr[UP] >= 0) {
    MPI_Recv((void*)wk,needed3,MPI_DOUBLE,myNbr[UP],msg,comm,&status);

    k = 0;
    for(i3 = nn3 - *pad3; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = wk[k];
      k++;
    }
    if(k != needed3) fprintf(stderr,"PADME FATAL ERROR 13: wrong count\n");
  }
  
  *error = 0;
}

void PadArray(double* array,
	      int* n1, int* n2, int* n3,
	      int* off1, int* off2, int* off3,
	      int* pad1, int* pad2, int* pad3,
	      int* myNbr, int* msgRoot, int* error) {
  int one = 1;
  PadStencil(array,&one,n1,n2,n3,off1,off2,off3,pad1,pad2,pad3,
	     myNbr,msgRoot,error);
}

void RPadStencil(float* array,
		int* stensize, int* n1, int* n2, int* n3,
		int* off1, int* off2, int* off3,
		int* pad1, int* pad2, int* pad3,
		int* myNbr, int* msgRoot, int* error) {

  int nn1 = (*n1)+2*(*pad1);
  int nn2 = (*n2)+2*(*pad2);
  int nn3 = (*n3)+2*(*pad3);

  int pp1 = (*pad1)+(*off1);
  int pp2 = (*pad2)+(*off2);
  int pp3 = (*pad3)+(*off3);

  int needed1 = (*stensize)*(*pad1)*nn2*nn3;
  int needed2 = (*stensize)*(*pad2)*nn1*nn3;
  int needed3 = (*stensize)*(*pad3)*nn1*nn2;

  int i1,i2,i3,k,m;
  int rwkSizeNeeded;
  float* rwk;
  int msg = *msgRoot;
  MPI_Status status;

/* Check available memory */

  rwkSizeNeeded = max( needed1, max(needed2, needed3) );
  if(rwkSizeNeeded > 2*wkSize) {
    if(wk != NULL) free(wk);
    wk = (double*) malloc(((rwkSizeNeeded+1)/2)*sizeof(double));
    if(wk == NULL) {
      *error = -1;
      fprintf(stderr,
        "PADME FATAL ERROR 14: Not enough memory, proc %d, need %d, have %d\n",
	myNbr[ME],(rwkSizeNeeded+1)/2,wkSize);
      exit(-1);
    }
    wkSize = (rwkSizeNeeded+1)/2;
  }
  rwk = (float*)wk;

/* Communicate East (+ x1 or x direction) */

  if(myNbr[EAST] >= 0) {
    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = nn1 - pp1 - *pad1; i1 < nn1 - pp1; i1++)
    for(m = 0; m < *stensize; m++) {
      rwk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed1) fprintf(stderr,"PADME FATAL ERROR 15: wrong count\n");

    MPI_Send((void*)rwk,needed1,MPI_FLOAT,myNbr[EAST],msg,comm);
  }

  if(myNbr[WEST] >= 0) {
    MPI_Recv((void*)rwk,needed1,MPI_FLOAT,myNbr[WEST],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < *pad1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = rwk[k];
      k++;
    }
    if(k != needed1) fprintf(stderr,"PADME FATAL ERROR 16: wrong count\n");
  }
  
/* Communicate West (- x1 or x direction) */

  msg++;
  if(myNbr[WEST] >= 0) {
    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = pp1; i1 < pp1 + *pad1; i1++)
    for(m = 0; m < *stensize; m++) {
      rwk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed1) fprintf(stderr,"PADME FATAL ERROR 17: wrong count\n");

    MPI_Send((void*)rwk,needed1,MPI_FLOAT,myNbr[WEST],msg,comm);
  }

  if(myNbr[EAST] >= 0) {
    MPI_Recv((void*)rwk,needed1,MPI_FLOAT,myNbr[EAST],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = nn1 - *pad1; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = rwk[k];
      k++;
    }
    if(k != needed1) fprintf(stderr,"PADME FATAL ERROR 18: wrong count\n");
  }

/* Communicate North (+ x2 or y direction) */

  msg++;
  if(myNbr[NORTH] >= 0) {
    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = nn2 - pp2 - *pad2; i2 < nn2 - pp2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      rwk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed2) fprintf(stderr,"PADME FATAL ERROR 19: wrong count\n");

    MPI_Send((void*)rwk,needed2,MPI_FLOAT,myNbr[NORTH],msg,comm);
  }

  if(myNbr[SOUTH] >= 0) {
    MPI_Recv((void*)rwk,needed2,MPI_FLOAT,myNbr[SOUTH],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = 0; i2 < *pad2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = rwk[k];
      k++;
    }
    if(k != needed2) fprintf(stderr,"PADME FATAL ERROR 20: wrong count\n");
  }

/* Communicate South (- x2 or y direction) */

  msg++;
  if(myNbr[SOUTH] >= 0) {
    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = pp2; i2 < pp2 + *pad2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      rwk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed2) fprintf(stderr,"PADME FATAL ERROR 21: wrong count\n");

    MPI_Send((void*)rwk,needed2,MPI_FLOAT,myNbr[SOUTH],msg,comm);
  }

  if(myNbr[NORTH] >= 0) {
    MPI_Recv((void*)rwk,needed2,MPI_FLOAT,myNbr[NORTH],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < nn3; i3++)
    for(i2 = nn2 - *pad2; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = rwk[k];
      k++;
    }
    if(k != needed2) fprintf(stderr,"PADME FATAL ERROR 22: wrong count\n");
  }

/* Communicate Up (+ x3 or z direction) */

  msg++;
  if(myNbr[UP] >= 0) {
    k = 0;
    for(i3 = nn3 - pp3 - *pad3; i3 < nn3 - pp3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      rwk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed3) fprintf(stderr,"PADME FATAL ERROR 23: wrong count\n");

    MPI_Send((void*)rwk,needed3,MPI_FLOAT,myNbr[UP],msg,comm);
  }

  if(myNbr[DOWN] >= 0) {
    MPI_Recv((void*)rwk,needed3,MPI_FLOAT,myNbr[DOWN],msg,comm,&status);

    k = 0;
    for(i3 = 0; i3 < *pad3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = rwk[k];
      k++;
    }
    if(k != needed3) fprintf(stderr,"PADME FATAL ERROR 24: wrong count\n");
  }
  
/* Communicate Down (- x3 or z direction) */

  msg++;
  if(myNbr[DOWN] >= 0) {
    k = 0;
    for(i3 = pp3; i3 < pp3 + *pad3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      rwk[k] = array_(m,i1,i2,i3);
      k++;
    }
    if(k != needed3) fprintf(stderr,"PADME FATAL ERROR 25: wrong count\n");

    MPI_Send((void*)rwk,needed3,MPI_FLOAT,myNbr[DOWN],msg,comm);
  }

  if(myNbr[UP] >= 0) {
    MPI_Recv((void*)rwk,needed3,MPI_FLOAT,myNbr[UP],msg,comm,&status);

    k = 0;
    for(i3 = nn3 - *pad3; i3 < nn3; i3++)
    for(i2 = 0; i2 < nn2; i2++)
    for(i1 = 0; i1 < nn1; i1++)
    for(m = 0; m < *stensize; m++) {
      array_(m,i1,i2,i3) = rwk[k];
      k++;
    }
    if(k != needed3) fprintf(stderr,"PADME FATAL ERROR 26: wrong count\n");
  }
  
  *error = 0;
}

/******************************************************************************
 * ARGUMENTS for PadVec
 * 
 * IN/OUT
 *  vec: The vector (array), of Fortran size vec(1-pad:n+pad), to be padded
 * 
 * IN
 *  n: Local size of vector
 *  off: Offset by 1 if grid-centered, by 0 if cell-centered
 *  pad: Amount to pad at each end
 *  dir: Coordinate direction in which vector lies, 1, 2 or 3.
 * 
 *  myNbr:   Node numbers of neighboring domains (-1 if no neighbor)
 *  msgRoot: The base number to use for message passing, the
 *           numbers actually used are msgRoot to msgRoot+1
 * 
 * OUT
 *  error: Error code, 0 if no error, -1 otherwise.
 *****************************************************************************/
void PadVec(double* vec, int* n, int* off, int* pad, int* dir,
	    int* myNbr, int* msgRoot, int* error) {

  int nn = (*n)+2*(*pad);
  int pp = (*pad)+(*off);

  int k,lowend,highend;
  int msg = *msgRoot;
  MPI_Status status;

/* Check available memory */

  if((*pad) > wkSize) {
    if(wk != NULL) free(wk);
    wk = (double*) malloc((*pad)*sizeof(double));
    if(wk == NULL) {
      *error = -1;
      fprintf(stderr,
        "PADME FATAL ERROR 27: Not enough memory, proc %d, need %d, have %d\n",
	myNbr[ME],(*pad),wkSize);
      exit(-1);
    }
    wkSize = (*pad);
  }

/* Define directions */

  switch(*dir) {
  case 1:
    lowend  = WEST;
    highend = EAST;
    break;
  case 2:
    lowend  = SOUTH;
    highend = NORTH;
    break;
  default:
    lowend  = DOWN;
    highend = UP;
  }

/* Communicate highend */
  
  if(myNbr[highend] >= 0) {
    for(k = 0; k < *pad; k++) wk[k] = vec[*n - *off + k];
    MPI_Send((void*)wk,(*pad),MPI_DOUBLE,myNbr[highend],msg,comm);
  }
  if(myNbr[lowend] >= 0) {
    MPI_Recv((void*)wk,(*pad),MPI_DOUBLE,myNbr[lowend],msg,comm,&status);
    for(k = 0; k < *pad; k++) vec[k] = wk[k];
  }

/* Communicate lowend */
  
  msg++;
  if(myNbr[lowend] >= 0) {
    for(k = 0; k < *pad; k++) wk[k] = vec[k+pp];
    MPI_Send((void*)wk,(*pad),MPI_DOUBLE,myNbr[lowend],msg,comm);
  }
  if(myNbr[highend] >= 0) {
    MPI_Recv((void*)wk,(*pad),MPI_DOUBLE,myNbr[highend],msg,comm,&status);
    for(k = 0; k < *pad; k++) vec[nn - *pad + k] = wk[k];
  }

  *error = 0;
}
