c#######################################################################
c Preconditioned conjugate gradient linear solver
c
c Todd Arbogast
c September 1992 and modified February 1997
c
c-----------------------------------------------------------------------
c This is a preconditioned conjugate gradient solver.  It is
c suitable for parallel computation.
c
c The system to solve is (a)(x) = (b)
c   where we have "lagrange multiplier" unknowns on the external
c   boundary.  There are also extra external multipliers on the edges
c   and corners; these solve an identity equation.
c
c=======================================================================
c INPUTS:
c   a           matrix a (padded by 1)
c   x           initial guess at local solution vector (padded by 1)
c   b           local vector b (padded by 1)
c   n1          x1 size of the local system
c   n2          x2 size of the local system
c   n3          x3 size of the local system
c   tolerance   convergence tolerance
c   itmax       maximum number of iterations allowed
c   r           work space, size nn=(n1+2)*(n2+2)*(n3+2)
c                 (residual, padded by 1)
c   q           work space, size nn (search direction)
c   p           work space, size nn (preconditioned r or a times q)
c   pcinv       work space, size nn (preconditioner space)
c   mynbr       neighboring node numbers in the rectangular parallel
c                 connectivity.  That is, mynbr(i,j,k) is the
c                 node number of the node with location in the 3-D
c                 rectangular parallel domain decomposition
c                   i = x-dir, j = y-dir, k = z-dir
c                   -1 = back one index
c                   0  = my parallel processor node coordinate
c                   +1 = forward one index
c   monitor     monitor flag
c   monitorunit monitor fortran unit number
c
c OUTPUTS:
c   x         solution vector
c   ierror    error flag
c               0 = no errors
c               1 = nonconvergence (nonfatal)
c               2 = message passing error (fatal)
c
c-----------------------------------------------------------------------
c CONVERGENCE CRITERON
c   A scale invarient criterion must be used.
c
c   We use the L^2 norm estimate
c     || pcinv*r || <= tolerance * || pcinv*b ||
c
c   If pcinv approximates a well, pcinv*b approximates x, and
c   pcinv*r is the residual converted into the units of x.
c   This choice controls the relative error of || x ||.
c
c   Two alternatives could be used.
c   (1) r*(pcinv*r) <= tolerance**2 * b*(pcinv*b)
c       This choice is as above, except that the conjugate norm
c       defined by the preconditioner is used:
c         || y ||_pc = sqrt(pc * pcinv*y * pcinv*y) is used.
c   (2) We could use L-infinity norms, to control the error in each
c       individual number.
c
c-----------------------------------------------------------------------
c CALLED SUBROUTINES:
c   applya(x,y,a)          gives the product of matrix a and vector x as y
c   setuppc(a,pcinv)       sets up the preconditioner (inverse)
c   applypcinv(x,y,pcinv)  gives the product of the inverse of the
c                            preconditioner and vector x as y
c
c CALLED FUNCTIONS:
c   scdot(x,y)             gives the scalar dot product of two vectors
c
c#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#=#
c declarations

      subroutine pcg(a,x,b,n1,n2,n3,tolerance,itmax,r,q,p,pcinv,
     $     mynbr,monitor,monitorunit,ierror)
      implicit none

      integer n1,n2,n3,itmax,mynbr(-1:1,-1:1,-1:1),
     $        monitor,monitorunit,ierror
      real*8  a(*),x(0:n1+1,0:n2+1,0:n3+1),
     $        b(0:n1+1,0:n2+1,0:n3+1),tolerance,
     $        r(0:n1+1,0:n2+1,0:n3+1),q(0:n1+1,0:n2+1,0:n3+1),
     $        p(0:n1+1,0:n2+1,0:n3+1),pcinv(*)

      integer it,itnewr,i,j,k,ierr,i0,i1,j0,j1,k0,k1
      real*8  alpha,beta,rpdot,rpdoto,qaqdot,tolsq,scdot,errsq
      integer DITNEWR
      parameter(DITNEWR=50)

c=======================================================================
c initialize

      ierror = 0
      it = 0
      itnewr = DITNEWR

c determine active region owned by my processor

      if(mynbr(-1,0,0).eq.-1) then
         i0 = 0
      else
         i0 = 1
      endif
      if(mynbr(1,0,0).eq.-1) then
         i1 = n1+1
      else
         i1 = n1
      endif
      if(mynbr(0,-1,0).eq.-1) then
         j0 = 0
      else
         j0 = 1
      endif
      if(mynbr(0,1,0).eq.-1) then
         j1 = n2+1
      else
         j1 = n2
      endif
      if(mynbr(0,0,-1).eq.-1) then
         k0 = 0
      else
         k0 = 1
      endif
      if(mynbr(0,0,1).eq.-1) then
         k1 = n3+1
      else
         k1 = n3
      endif

c zero out edge x's and b's (possibly in the parallel overlap)

      do i=0,n1+1
         x(i,0,0) = 0
         x(i,0,n3+1) = 0
         x(i,n2+1,0) = 0
         x(i,n2+1,n3+1) = 0
         b(i,0,0) = 0
         b(i,0,n3+1) = 0
         b(i,n2+1,0) = 0
         b(i,n2+1,n3+1) = 0
      enddo
      do j=0,n2+1
         x(0,j,0) = 0
         x(0,j,n3+1) = 0
         x(n1+1,j,0) = 0
         x(n1+1,j,n3+1) = 0
         b(0,j,0) = 0
         b(0,j,n3+1) = 0
         b(n1+1,j,0) = 0
         b(n1+1,j,n3+1) = 0
      enddo
      do k=0,n3+1
         x(0,0,k) = 0
         x(0,n2+1,k) = 0
         x(n1+1,0,k) = 0
         x(n1+1,n2+1,k) = 0
         b(0,0,k) = 0
         b(0,n2+1,k) = 0
         b(n1+1,0,k) = 0
         b(n1+1,n2+1,k) = 0
      enddo

c initial residual

      call applya(x,r,a,n1,n2,n3,mynbr,monitorunit,ierr)
      if(ierr.ne.0) goto 9002
      do 10 k=k0,k1
      do 10 j=j0,j1
      do 10 i=i0,i1
         r(i,j,k) = b(i,j,k) - r(i,j,k)
 10   continue

c set up preconditioner

      call setuppc(a,pcinv,n1,n2,n3,i0,i1,j0,j1,k0,k1,mynbr)

c first search direction q

      call applypcinv(r,q,pcinv,n1,n2,n3,i0,i1,j0,j1,k0,k1)
      rpdot = scdot(r,q,n1,n2,n3,i0,i1,j0,j1,k0,k1)

c set tolerance criteria

      call applypcinv(b,p,pcinv,n1,n2,n3,i0,i1,j0,j1,k0,k1)
      tolsq = tolerance**2*scdot(p,p,n1,n2,n3,i0,i1,j0,j1,k0,k1)

c check initial error

      errsq = scdot(q,q,n1,n2,n3,i0,i1,j0,j1,k0,k1)
      if((monitor.gt.0).and.(mynbr(0,0,0).eq.0)) then
         write(monitorunit,*) '  tol:',dsqrt(tolsq)
         write(monitorunit,*) '  it,err:',it,dsqrt(errsq)
      endif

      if(errsq.le.tolsq) then
         errsq = scdot(r,r,n1,n2,n3,i0,i1,j0,j1,k0,k1)
         if((monitor.gt.0).and.(mynbr(0,0,0).eq.0))
     $        write(monitorunit,*) '  residual:',dsqrt(errsq)
         return
      endif

c=======================================================================
c iteration loop

 9000 continue
      it = it + 1

c update x and r

      call applya(q,p,a,n1,n2,n3,mynbr,monitorunit,ierr)
      if(ierr.ne.0) goto 9002
      qaqdot = scdot(q,p,n1,n2,n3,i0,i1,j0,j1,k0,k1)
      beta = rpdot/qaqdot
      do 20 k=k0,k1
      do 20 j=j0,j1
      do 20 i=i0,i1
         x(i,j,k) = x(i,j,k) + beta*q(i,j,k)
         r(i,j,k) = r(i,j,k) - beta*p(i,j,k)
 20   continue

c correct residual after DITNEWR iterations

      if(it.ge.itnewr) then
         itnewr = itnewr + DITNEWR
         call applya(x,r,a,n1,n2,n3,mynbr,monitorunit,ierr)
         if(ierr.ne.0) goto 9002
         do 40 k=k0,k1
         do 40 j=j0,j1
         do 40 i=i0,i1
            r(i,j,k) = b(i,j,k) - r(i,j,k)
 40      continue
      endif

c check for convergence

      call applypcinv(r,p,pcinv,n1,n2,n3,i0,i1,j0,j1,k0,k1)
      errsq = scdot(p,p,n1,n2,n3,i0,i1,j0,j1,k0,k1)

      if((monitor.gt.0).and.(mynbr(0,0,0).eq.0)) then
         write(monitorunit,*) '  it,err:',it,dsqrt(errsq)
      endif

      if(errsq.le.tolsq) then
         if(monitor.gt.0) then
            call applya(x,r,a,n1,n2,n3,mynbr,monitorunit,ierr)
            if(ierr.ne.0) goto 9002
            do 910 k=k0,k1
            do 910 j=j0,j1
            do 910 i=i0,i1
               r(i,j,k) = b(i,j,k) - r(i,j,k)
 910        continue
            errsq = scdot(r,r,n1,n2,n3,i0,i1,j0,j1,k0,k1)
            if(mynbr(0,0,0).eq.0) write(monitorunit,*)
     $           '  residual:',dsqrt(errsq)
         endif
         return
      endif

c get new search direction q

      rpdoto = rpdot
      rpdot = scdot(r,p,n1,n2,n3,i0,i1,j0,j1,k0,k1)
      alpha = rpdot/rpdoto
      do 30 k=k0,k1
      do 30 j=j0,j1
      do 30 i=i0,i1
         q(i,j,k) = p(i,j,k) + alpha*q(i,j,k)
 30   continue

c end of iteration loop

      if(it.lt.itmax) goto 9000

      ierror = 1
      return
 9002 ierror = 2
      return
      end
c#######################################################################
c Preconditioned conjugate gradient linear solver
c
c Apply a: (a)(x) = (y)
c
c * a is a 19 point symmetric stencil in 3-D
c
c * Since a is symmetric, only the nonzero part of the lower half of the
c   matrix is stored.  So if band > 0,
c   a("eqn",band) = a("reciprocal eqn",-band)
c
c * a is assumed known for the subdomain and all connections out
c   of the subdomain; however, such x's need to be communicated
c
c * The system has "lagrange multiplier" unknowns on the external
c   boundary.  There are also extra external multipliers on the edges
c   and corners; these solve an identity equation.  We assume explicitly
c   that these edge values (x and therefore y) are 0, so that the matrix
c   may include connections to them.
c
c * y is used as a workarray to communicate the padded information
c
c * The reciprocal bands are given as follows:
c   a(i,j,k, 1) = a(i+1, j  , k  ,-1)
c   a(i,j,k, 2) = a(i-1, j+1, k  ,-2)
c   a(i,j,k, 3) = a(i  , j+1, k  ,-3)
c   a(i,j,k, 4) = a(i+1, j+1, k  ,-4)
c   a(i,j,k, 5) = a(i  , j-1, k+1,-5)
c   a(i,j,k, 6) = a(i-1, j  , k+1,-6)
c   a(i,j,k, 7) = a(i  , j  , k+1,-7)
c   a(i,j,k, 8) = a(i+1, j  , k+1,-8)
c   a(i,j,k, 9) = a(i  , j+1, k+1,-9)
c
c * The bands multiply as follows
c   a(i,j,k,-9) * x(i  , j-1, k-1)
c   a(i,j,k,-8) * x(i-1, j  , k-1)
c   a(i,j,k,-7) * x(i  , j  , k-1)
c   a(i,j,k,-6) * x(i+1, j  , k-1)
c   a(i,j,k,-5) * x(i  , j+1, k-1)
c   a(i,j,k,-4) * x(i-1, j-1, k  )
c   a(i,j,k,-3) * x(i  , j-1, k  )
c   a(i,j,k,-2) * x(i+1, j-1, k  )
c   a(i,j,k,-1) * x(i-1, j  , k  )
c   a(i,j,k, 0) * x(i  , j  , k  )
c   a(i,j,k, 1) * x(i+1, j  , k  )
c   a(i,j,k, 2) * x(i-1, j+1, k  )
c   a(i,j,k, 3) * x(i  , j+1, k  )
c   a(i,j,k, 4) * x(i+1, j+1, k  )
c   a(i,j,k, 5) * x(i  , j-1, k+1)
c   a(i,j,k, 6) * x(i-1, j  , k+1)
c   a(i,j,k, 7) * x(i  , j  , k+1)
c   a(i,j,k, 8) * x(i+1, j  , k+1)
c   a(i,j,k, 9) * x(i  , j+1, k+1)

      subroutine applya(x,y,a,n1,n2,n3,mynbr,monitorunit,ierr)
      implicit none
      include 'tmsgtypes.hf'

      integer n1,n2,n3,mynbr(-1:1,-1:1,-1:1),monitorunit,ierr
      real*8  x(0:n1+1,0:n2+1,0:n3+1),y(0:n1+1,0:n2+1,0:n3+1),
     $        a(0:n1+1,0:n2+1,0:n3+1,-9:0)

      integer i,j,k

c Communicate padded data x

      call padarray(x,n1,n2,n3,0,0,0,1,1,1,
     $     mynbr,MSGPADPCG,ierr)

c Compute (a)(x)

      do 20 k = 1, n3
      do 20 j = 1, n2
      do 20 i = 1, n1
         y(i,j,k) =
     $        a(i,j,k,-9) * x(i,j-1,k-1) +
     $        a(i,j,k,-8) * x(i-1,j,k-1) +
     $        a(i,j,k,-7) * x(i,j,k-1) +
     $        a(i,j,k,-6) * x(i+1,j,k-1) +
     $        a(i,j,k,-5) * x(i,j+1,k-1) +
     $        a(i,j,k,-4) * x(i-1,j-1,k) +
     $        a(i,j,k,-3) * x(i,j-1,k) +
     $        a(i,j,k,-2) * x(i+1,j-1,k) +
     $        a(i,j,k,-1) * x(i-1,j,k) +
     $        a(i,j,k, 0) * x(i,j,k) +
     $        a(i+1,j,k,-1) * x(i+1,j,k) +
     $        a(i-1,j+1,k,-2) * x(i-1,j+1,k) +
     $        a(i,j+1,k,-3) * x(i,j+1,k) +
     $        a(i+1,j+1,k,-4) * x(i+1,j+1,k) +
     $        a(i,j-1,k+1,-5) * x(i,j-1,k+1) +
     $        a(i-1,j,k+1,-6) * x(i-1,j,k+1) +
     $        a(i,j,k+1,-7) * x(i,j,k+1) +
     $        a(i+1,j,k+1,-8) * x(i+1,j,k+1) +
     $        a(i,j+1,k+1,-9) * x(i,j+1,k+1)
 20   continue

c external faces

c left face

      if(mynbr(-1,0,0).eq.-1) then
         do k = 1,n3
         do j = 1,n2
            y(0,j,k) =
     $           a(0,j,k,-9) * x(0,j-1,k-1) +
     $           a(0,j,k,-7) * x(0,j,k-1) +
     $           a(0,j,k,-6) * x(1,j,k-1) +
     $           a(0,j,k,-5) * x(0,j+1,k-1) +
     $           a(0,j,k,-3) * x(0,j-1,k) +
     $           a(0,j,k,-2) * x(1,j-1,k) +
     $           a(0,j,k, 0) * x(0,j,k) +
     $           a(1,j,k,-1) * x(1,j,k) +
     $           a(0,j+1,k,-3) * x(0,j+1,k) +
     $           a(1,j+1,k,-4) * x(1,j+1,k) +
     $           a(0,j-1,k+1,-5) * x(0,j-1,k+1) +
     $           a(0,j,k+1,-7) * x(0,j,k+1) +
     $           a(1,j,k+1,-8) * x(1,j,k+1) +
     $           a(0,j+1,k+1,-9) * x(0,j+1,k+1)
         enddo
         enddo
      endif

c right face

      if(mynbr(1,0,0).eq.-1) then
         do k = 1,n3
         do j = 1,n2
            y(n1+1,j,k) =
     $           a(n1+1,j,k,-9) * x(n1+1,j-1,k-1) +
     $           a(n1+1,j,k,-8) * x(n1,j,k-1) +
     $           a(n1+1,j,k,-7) * x(n1+1,j,k-1) +
     $           a(n1+1,j,k,-5) * x(n1+1,j+1,k-1) +
     $           a(n1+1,j,k,-4) * x(n1,j-1,k) +
     $           a(n1+1,j,k,-3) * x(n1+1,j-1,k) +
     $           a(n1+1,j,k,-1) * x(n1,j,k) +
     $           a(n1+1,j,k, 0) * x(n1+1,j,k) +
     $           a(n1,j+1,k,-2) * x(n1,j+1,k) +
     $           a(n1+1,j+1,k,-3) * x(n1+1,j+1,k) +
     $           a(n1+1,j-1,k+1,-5) * x(n1+1,j-1,k+1) +
     $           a(n1,j,k+1,-6) * x(n1,j,k+1) +
     $           a(n1+1,j,k+1,-7) * x(n1+1,j,k+1) +
     $           a(n1+1,j+1,k+1,-9) * x(n1+1,j+1,k+1)
         enddo
         enddo
      endif

c front face

      if(mynbr(0,-1,0).eq.-1) then
         do k = 1,n3
         do i = 1,n1
            y(i,0,k) =
     $           a(i,0,k,-8) * x(i-1,0,k-1) +
     $           a(i,0,k,-7) * x(i,0,k-1) +
     $           a(i,0,k,-6) * x(i+1,0,k-1) +
     $           a(i,0,k,-5) * x(i,1,k-1) +
     $           a(i,0,k,-1) * x(i-1,0,k) +
     $           a(i,0,k, 0) * x(i,0,k) +
     $           a(i+1,0,k,-1) * x(i+1,0,k) +
     $           a(i-1,1,k,-2) * x(i-1,1,k) +
     $           a(i,1,k,-3) * x(i,1,k) +
     $           a(i+1,1,k,-4) * x(i+1,1,k) +
     $           a(i-1,0,k+1,-6) * x(i-1,0,k+1) +
     $           a(i,0,k+1,-7) * x(i,0,k+1) +
     $           a(i+1,0,k+1,-8) * x(i+1,0,k+1) +
     $           a(i,1,k+1,-9) * x(i,1,k+1)
         enddo
         enddo
      endif

c back face

      if(mynbr(0,1,0).eq.-1) then
         do k = 1,n3
         do i = 1,n1
            y(i,n2+1,k) =
     $           a(i,n2+1,k,-9) * x(i,n2,k-1) +
     $           a(i,n2+1,k,-8) * x(i-1,n2+1,k-1) +
     $           a(i,n2+1,k,-7) * x(i,n2+1,k-1) +
     $           a(i,n2+1,k,-6) * x(i+1,n2+1,k-1) +
     $           a(i,n2+1,k,-4) * x(i-1,n2,k) +
     $           a(i,n2+1,k,-3) * x(i,n2,k) +
     $           a(i,n2+1,k,-2) * x(i+1,n2,k) +
     $           a(i,n2+1,k,-1) * x(i-1,n2+1,k) +
     $           a(i,n2+1,k, 0) * x(i,n2+1,k) +
     $           a(i+1,n2+1,k,-1) * x(i+1,n2+1,k) +
     $           a(i,n2,k+1,-5) * x(i,n2,k+1) +
     $           a(i-1,n2+1,k+1,-6) * x(i-1,n2+1,k+1) +
     $           a(i,n2+1,k+1,-7) * x(i,n2+1,k+1) +
     $           a(i+1,n2+1,k+1,-8) * x(i+1,n2+1,k+1) 
         enddo
         enddo
      endif

c bottom face

      if(mynbr(0,0,-1).eq.-1) then
         do j = 1,n2
         do i = 1,n1
            y(i,j,0) =
     $           a(i,j,0,-4) * x(i-1,j-1,0) +
     $           a(i,j,0,-3) * x(i,j-1,0) +
     $           a(i,j,0,-2) * x(i+1,j-1,0) +
     $           a(i,j,0,-1) * x(i-1,j,0) +
     $           a(i,j,0, 0) * x(i,j,0) +
     $           a(i+1,j,0,-1) * x(i+1,j,0) +
     $           a(i-1,j+1,0,-2) * x(i-1,j+1,0) +
     $           a(i,j+1,0,-3) * x(i,j+1,0) +
     $           a(i+1,j+1,0,-4) * x(i+1,j+1,0) +
     $           a(i,j-1,1,-5) * x(i,j-1,1) +
     $           a(i-1,j,1,-6) * x(i-1,j,1) +
     $           a(i,j,1,-7) * x(i,j,1) +
     $           a(i+1,j,1,-8) * x(i+1,j,1) +
     $           a(i,j+1,1,-9) * x(i,j+1,1)
         enddo
         enddo
      endif

c top face

      if(mynbr(0,0,1).eq.-1) then
         do j = 1,n2
         do i = 1,n1
            y(i,j,n3+1) =
     $           a(i,j,n3+1,-9) * x(i,j-1,n3) +
     $           a(i,j,n3+1,-8) * x(i-1,j,n3) +
     $           a(i,j,n3+1,-7) * x(i,j,n3) +
     $           a(i,j,n3+1,-6) * x(i+1,j,n3) +
     $           a(i,j,n3+1,-5) * x(i,j+1,n3) +
     $           a(i,j,n3+1,-4) * x(i-1,j-1,n3+1) +
     $           a(i,j,n3+1,-3) * x(i,j-1,n3+1) +
     $           a(i,j,n3+1,-2) * x(i+1,j-1,n3+1) +
     $           a(i,j,n3+1,-1) * x(i-1,j,n3+1) +
     $           a(i,j,n3+1, 0) * x(i,j,n3+1) +
     $           a(i+1,j,n3+1,-1) * x(i+1,j,n3+1) +
     $           a(i-1,j+1,n3+1,-2) * x(i-1,j+1,n3+1) +
     $           a(i,j+1,n3+1,-3) * x(i,j+1,n3+1) +
     $           a(i+1,j+1,n3+1,-4) * x(i+1,j+1,n3+1) 
         enddo
         enddo
      endif

c external edge and corner lagrange multipliers

      do i=0,n1+1
         if((mynbr(0,-1,0).eq.-1).and.(mynbr(0,0,-1).eq.-1))
     $        y(i,0,0) = x(i,0,0)
         if((mynbr(0,-1,0).eq.-1).and.(mynbr(0,0,1).eq.-1))
     $        y(i,0,n3+1) = x(i,0,n3+1)
         if((mynbr(0,1,0).eq.-1).and.(mynbr(0,0,-1).eq.-1))
     $        y(i,n2+1,0) = x(i,n2+1,0)
         if((mynbr(0,1,0).eq.-1).and.(mynbr(0,0,1).eq.-1))
     $        y(i,n2+1,n3+1) = x(i,n2+1,n3+1)
      enddo

      do j=0,n2+1
         if((mynbr(-1,0,0).eq.-1).and.(mynbr(0,0,-1).eq.-1))
     $        y(0,j,0) = x(0,j,0)
         if((mynbr(-1,0,0).eq.-1).and.(mynbr(0,0,1).eq.-1))
     $        y(0,j,n3+1) = x(0,j,n3+1)
         if((mynbr(1,0,0).eq.-1).and.(mynbr(0,0,-1).eq.-1))
     $        y(n1+1,j,0) = x(n1+1,j,0)
         if((mynbr(1,0,0).eq.-1).and.(mynbr(0,0,1).eq.-1))
     $        y(n1+1,j,n3+1) = x(n1+1,j,n3+1)
      enddo

      do k=0,n3+1
         if((mynbr(-1,0,0).eq.-1).and.(mynbr(0,-1,0).eq.-1))
     $        y(0,0,k) = x(0,0,k)
         if((mynbr(-1,0,0).eq.-1).and.(mynbr(0,1,0).eq.-1))
     $        y(0,n2+1,k) = x(0,n2+1,k)
         if((mynbr(1,0,0).eq.-1).and.(mynbr(0,-1,0).eq.-1))
     $        y(n1+1,0,k) = x(n1+1,0,k)
         if((mynbr(1,0,0).eq.-1).and.(mynbr(0,1,0).eq.-1))
     $        y(n1+1,n2+1,k) = x(n1+1,n2+1,k)
      enddo

      return
      end
c#######################################################################
c Preconditioned conjugate gradient linear solver
c
c Perform a global scalar dot product: (x).(y)

      real*8 function scdot(x,y,n1,n2,n3,i0,i1,j0,j1,k0,k1)
      implicit none
      include 'mpif.h'
      include 'tmsgtypes.hf'

      integer n1,n2,n3,i0,i1,j0,j1,k0,k1,i,j,k
      real*8  x(0:n1+1,0:n2+1,0:n3+1),y(0:n1+1,0:n2+1,0:n3+1)

      real*8  temp
      integer error

      temp = 0.d0
      do 10 k=k0,k1
      do 10 j=j0,j1
      do 10 i=i0,i1
         temp = temp + x(i,j,k)*y(i,j,k)
 10   continue

cmpi      call gdsumx(MSGDOTPCG,temp,1)
      call MSG_f_Gdsum1(temp,MPI_COMM_WORLD)
      scdot = temp

      return
      end
c#######################################################################
c Preconditioned conjugate gradient linear solver
c
c Set up diagonal preconditioning

      subroutine setuppc(a,pcinv,n1,n2,n3,i0,i1,j0,j1,k0,k1,mynbr)
      implicit none

      integer n1,n2,n3,i0,i1,j0,j1,k0,k1,mynbr(-1:1,-1:1,-1:1),i,j,k
      real*8  a(0:n1+1,0:n2+1,0:n3+1,10), pcinv(0:n1+1,0:n2+1,0:n3+1)
      real*8 EPS
      parameter(EPS = 1.d-8)

      do 10 k=k0,k1
      do 10 j=j0,j1
      do 10 i=i0,i1
         if(a(i,j,k,10).le.EPS) then
            if(a(i,j,k,10).ge.-EPS) then
               pcinv(i,j,k) = 1./EPS
            else
               pcinv(i,j,k) = 1.
            endif
         else
            pcinv(i,j,k) = 1./a(i,j,k,10)
         endif
 10   continue

      do i=0,n1+1
         if((mynbr(0,-1,0).eq.-1).and.(mynbr(0,0,-1).eq.-1))
     $        pcinv(i,0,0) = 1.
         if((mynbr(0,-1,0).eq.-1).and.(mynbr(0,0,1).eq.-1))
     $        pcinv(i,0,n3+1) = 1.
         if((mynbr(0,1,0).eq.-1).and.(mynbr(0,0,-1).eq.-1))
     $        pcinv(i,n2+1,0) = 1.
         if((mynbr(0,1,0).eq.-1).and.(mynbr(0,0,1).eq.-1))
     $        pcinv(i,n2+1,n3+1) = 1.
      enddo

      do j=0,n2+1
         if((mynbr(-1,0,0).eq.-1).and.(mynbr(0,0,-1).eq.-1))
     $        pcinv(0,j,0) = 1.
         if((mynbr(-1,0,0).eq.-1).and.(mynbr(0,0,1).eq.-1))
     $        pcinv(0,j,n3+1) = 1.
         if((mynbr(1,0,0).eq.-1).and.(mynbr(0,0,-1).eq.-1))
     $        pcinv(n1+1,j,0) = 1.
         if((mynbr(1,0,0).eq.-1).and.(mynbr(0,0,1).eq.-1))
     $        pcinv(n1+1,j,n3+1) = 1.
      enddo

      do k=0,n3+1
         if((mynbr(-1,0,0).eq.-1).and.(mynbr(0,-1,0).eq.-1))
     $        pcinv(0,0,k) = 1.
         if((mynbr(-1,0,0).eq.-1).and.(mynbr(0,1,0).eq.-1))
     $        pcinv(0,n2+1,k) = 1.
         if((mynbr(1,0,0).eq.-1).and.(mynbr(0,-1,0).eq.-1))
     $        pcinv(n1+1,0,k) = 1.
         if((mynbr(1,0,0).eq.-1).and.(mynbr(0,1,0).eq.-1))
     $        pcinv(n1+1,n2+1,k) = 1.
      enddo

      return
      end
c#######################################################################
c Preconditioned conjugate gradient linear solver
c
c Apply diagonal preconditioning: (pcinv)(x) = (y)

      subroutine applypcinv(x,y,pcinv,n1,n2,n3,i0,i1,j0,j1,k0,k1)
      implicit none

      integer n1,n2,n3,i0,i1,j0,j1,k0,k1,i,j,k
      real*8  x(0:n1+1,0:n2+1,0:n3+1), y(0:n1+1,0:n2+1,0:n3+1),
     $        pcinv(0:n1+1,0:n2+1,0:n3+1)

      do 10 k=k0,k1
      do 10 j=j0,j1
      do 10 i=i0,i1
         y(i,j,k) = pcinv(i,j,k)*x(i,j,k)
 10   continue

      return
      end
