/*****************************************************************************
 * DIFFERENTIAL EQUATIONS OF REACTIONS
 *
 * For mobile species with nonequilibrium reactions
 *                                      d c_i
 *   ( (porosity) + (alpha_i)(sigma_i) )----- = (porosity) R_i(c_1,...,c_n)
 *                                        dt
 * For immobile species with nonequilibrium reactions
 *   d c_i
 *   ----- = R_i(c_1,...,c_n)
 *     dt
 * For species with equilibrium reactions (must be analytically solved!)
 *    c_i = R_i(c_1,...,c_{i-1},c_{i+1},...,c_n)
 * Where
 *   c_i = total concentration of species i in the fluid phase,
 *         the phase within which reactions take place.
 *   R_i = the (batch) reaction function
 *   totCoef_j = (porosity) + (alpha_i)(sigma_i), defined by trace type j(i)
 *
 * First solve nonequilibrium reactions, and then equilibrate.
 *
 *****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "../Main/driverGlobals.h"
#include "transportGlobals.h"
#include "reaction.h"

#ifndef MIN
#define MIN(x,y) ((x) < (y)) ? (x) : (y)
#endif
#ifndef MAX
#define MAX(x,y) ((x) > (y)) ? (x) : (y)
#endif

void reactionHeader(int step, double time, double dt, FILE* debugRxnFile) {
  fprintf(debugRxnFile,"REACTIONS: step %d, time %g, dt %g\n",step,time,dt);
}

static int rungeKuttaFehlberg(double* x, int n, double t, double tEnd,
    rxnFcn* f, int nDParams, double* dParam, int nIParams, int* iParam,
    double* multiplier, double dt, double dtMax, double tol) {
  const double c11 = 1./4.;

  const double c21 = 3./32.;
  const double c22 = 9./32.;

  const double c31 = 1932./2197.;
  const double c32 = -7200./2197.;
  const double c33 = 7296./2197.;

  const double c41 = 439./216.;
  const double c42 = -8.;
  const double c43 = 3680./513.;
  const double c44 = -845./4104.;

  const double c51 = -8./27.;
  const double c52 = 2.;
  const double c53 = -3544/2565.;
  const double c54 = 1859./4104.;
  const double c55 = -11./40.;

  const double c61 = 1./360.;
  const double c62 = 0.;
  const double c63 = -128./4275.;
  const double c64 = -2197./75240.;
  const double c65 = 1./50.;
  const double c66 = 2./55.;

  const double c1 = 25./216.;
  const double c2 = 0.;
  const double c3 = 1408./2565.;
  const double c4 = 2197./4104.;
  const double c5 = -1./5.;

  const double ct1 = 1./4.;
  const double ct2 = 3./8.;
  const double ct3 = 12./13.;
  const double ct4 = 1.;
  const double ct5 = 1./2.;

  static double dtOld = -1;
  int error;

  double *k1,*k2,*k3,*k4,*k5,*k6;
  double* x1 = (double*) malloc(7*n*sizeof(double));
  if(x1 == NULL) return ERR_RXN_MEMORY;
  k1 = &x1[n];
  k2 = &k1[n];
  k3 = &k2[n];
  k4 = &k3[n];
  k5 = &k4[n];
  k6 = &k5[n];

  if(dt > dtOld && dtOld > 0) dt = dtOld;

  do {
    int i;
    double errorEst, delta;

    dt = MIN(dt,dtMax);
    if(dt >= tEnd-t) {
      dtOld = dt;
      dt = tEnd-t;
    }

    /* Initialize slopes to 0 (to omit null reactions) */

    for(i=0; i<6*n; i++) k1[i] = 0;

    /* Determine RKF slopes */

    #define MULT(k,dt) for(i=0; i<n; i++) { k[i] *= multiplier[i]*dt; }

    error = f(x,k1,n,t,dt,nDParams,dParam,nIParams,iParam);
    if(error) return error;
    MULT(k1,dt);

    for(i=0; i<n; i++) x1[i] = x[i] + c11*k1[i];
    error = f(x1,k2,n,t+ct1*dt,dt,nDParams,dParam,nIParams,iParam);
    if(error) return error;
    MULT(k2,dt);

    for(i=0; i<n; i++) x1[i] = x[i] + c21*k1[i] + c22*k2[i];
    error = f(x1,k3,n,t+ct2*dt,dt,nDParams,dParam,nIParams,iParam);
    if(error) return error;
    MULT(k3,dt);

    for(i=0; i<n; i++) x1[i] = x[i] + c31*k1[i] + c32*k2[i] + c33*k3[i];
    error = f(x1,k4,n,t+ct3*dt,dt,nDParams,dParam,nIParams,iParam);
    if(error) return error;
    MULT(k4,dt);

    for(i=0; i<n; i++)
      x1[i] = x[i] + c41*k1[i] + c42*k2[i] + c43*k3[i] + c44*k4[i];
    error = f(x1,k5,n,t+ct4*dt,dt,nDParams,dParam,nIParams,iParam);
    if(error) return error;
    MULT(k5,dt);

    for(i=0; i<n; i++)
      x1[i] = x[i] + c51*k1[i] + c52*k2[i] + c53*k3[i] + c54*k4[i] + c55*k5[i];
    error = f(x1,k6,n,t+ct5*dt,dt,nDParams,dParam,nIParams,iParam);
    if(error) return error;
    MULT(k6,dt);

    #undef MULT

    /* Determine error estimate */

    errorEst = 0;
    for(i=0; i<n; i++) {
      double r =
	c61*k1[i] + c62*k2[i] + c63*k3[i] + c64*k4[i] + c65*k5[i] + c66*k6[i];
      errorEst += r*r;
    }
    errorEst = sqrt(errorEst)/dt;

    /* Check acceptance criteria */

    if(errorEst <= tol) {
      /* Accept step */
      t += dt;
      for(i=0; i<n; i++)
	x[i] += c1*k1[i] + c2*k2[i] + c3*k3[i] + c4*k4[i] + c5*k5[i];
    }

    /* Time step control */

    delta = MIN(4,MAX(0.1, 0.84*sqrt(sqrt(tol/errorEst) )));
    dt *= delta;

  } while(t < tEnd);

  free(x1);
  return 0;
}

int reaction(double deltaT_t, FILE* debugRxnFile) {
  int i,j,k,m,comp,step,error;
  double dt,t,rxn_dt,tEnd;

  double* c = (double*) malloc(rxn_nComps*sizeof(double));
  double* sorpFactor = (double*) malloc(rxn_nComps*sizeof(double));
  if(c == NULL || sorpFactor == NULL) return ERR_RXN_MEMORY;

  dt = deltaT_t/(double)rxn_nMicroSteps;

  /* Loop over the grid cells */

  for(k = 0; k<nz; k++)
  for(j = 0; j<ny; j++)
  for(i = 0; i<nx; i++) {

    /* Set up local array of concentrations */

    for(m=0; m < rxn_nComps; m++) {
      int ttype;
      comp = rxn_comp[m];
      ttype = traceType[comp] - 1;
      c[m] = cbar_(i,j,k,comp);
      if(ttype != -1) {
	sorpFactor[m] = porosity_(i,j,k) / totCoef_(i,j,k,ttype);
      } else {
	sorpFactor[m] = 1.;
      }
    }

    if(debug >= 1) {
      fprintf(debugRxnFile,"Cell = (%d,%d,%d), time = %g\n",i,j,k,time_t);
      fprintf(debugRxnFile,"  initial:");
      comp=0;
      do {
	fprintf(debugRxnFile,"  %g",c[comp]);
	if(!(++comp%5)) fprintf(debugRxnFile,"\n          ");
      } while(comp<rxn_nComps);
    }

    /* React with an explicit 4th order Runga-Kutta-Fehlberg scheme */

    t = time_t;
    rxn_dt = dt/rxn_nMicroSteps;

    for(step = 0; step < rxn_nMicroSteps; step++) {

      /* Nonequilibrium reactions using user defined function "rxn_react" */

      tEnd = t + rxn_dt;
      error = rungeKuttaFehlberg(c,rxn_nComps,t,tEnd,&rxn_react,
                rxn_nDParams,rxn_dParam,rxn_nIParams,rxn_iParam,
                sorpFactor,rxn_dt,rxn_dt,rxn_tolerance);
      if(error) return error;
      t = tEnd;

      /* Equilibrium reactions using user defined function "rxn_equil" */

      error = rxn_equil(c,rxn_nComps,t,rxn_nDParams,rxn_dParam,
			rxn_nIParams,rxn_iParam,rxn_tolerance);
      if(error) return error;

      if(debug >= 1) {
	fprintf(debugRxnFile,"  reacted:");
	comp=0;
	do {
	  fprintf(debugRxnFile,"  %g",c[comp]);
	  if(!(++comp%5)) fprintf(debugRxnFile,"\n          ");
	} while(comp<rxn_nComps);
      }
    }

    /* Replace cbar concentration */

    for(m=0; m < rxn_nComps; m++) {
      comp = rxn_comp[m];
      cbar_(i,j,k,comp) = c[m];
    }
  }

  free(sorpFactor);
  free(c);
  return 0;
}
