c=======================================================================
c driver for rnd
c
c Structure of Iparm:
c   Iparm(1) = number of radionuclide decay chains
c   Iparm(2) = number of components in the first chain
c   Iparm(3:2+Iparm(2)) = the components in the first chain
c   Iparm(3+Iparm(2)) = number of components in the second chain
c   etc.
c
c Structure of Rparm:
c   Rparm(1:Iparm(2)) = atomic numbers of components in the first chain
c   Rparm(1+Iparm(2):2*Iparm(2))
c     = decay constants for the components in the first chain
c   On first call:
c     Rparm(1+2*Iparm(2):1+(2+Iparm(2))*Iparm(2))
c       = matrix of decay constants and branching ratios times decay
c         constants for the first chain
c   On output and on subsequent calls:
c     Rparm(1+2*Iparm(2):1+(2+Iparm(2))*Iparm(2))
c       = matrix of eigenvectors of the original system
c   etc.
c
c N is real*8 work space:  Need  ncomp  locations
c-----------------------------------------------------------------------
      subroutine decay(cbar,deltat,Iparm,Rparm,nxc,nyc,nzc,N)
      implicit none

      integer Iparm(*),nxc,nyc,nzc
      real*8  cbar(nxc,nyc,nzc,*),deltat,Rparm(*),N(*)

      integer i,j,k,chain,comp,nChains,nComps,iI,iR
      integer firstcall
      data firstcall /1/

      nChains = Iparm(1)

      if(firstcall.eq.1) then
         iI = 2
         iR = 0
         do chain=1,nChains
            nComps = Iparm(iI)
            call findEigenVect(nComps,Rparm(iR+2*nComps+1))
            iI = iI + nComps + 1
            iR = iR + nComps*(nComps + 2)
         enddo
         firstcall = 0
      endif

      do 10 k=1,nzc
      do 10 j=1,nyc
      do 10 i=1,nxc
         iI = 2
         iR = 0
         do 20 chain=1,nChains
            nComps = Iparm(iI)
            do 30 comp=1,nComps
               N(comp) = cbar(i,j,k,Iparm(iI+comp))*Rparm(iR+comp)
 30         continue

            call forwardSolve(nComps,Rparm(iR+2*nComps+1),N)
            call rndScale(nComps,Rparm(iR+nComps+1),deltat,N)
            call multMatVec(nComps,Rparm(iR+2*nComps+1),N)

           do 40 comp=1,nComps
               cbar(i,j,k,Iparm(iI+comp)) = N(comp)/Rparm(iR+comp)
 40         continue
            iI = iI + nComps + 1
            iR = iR + nComps*(nComps + 2)
 20      continue
 10   continue

      return
      end
c=======================================================================
c This subroutine finds the eigenvectors of a lower triangular matrix
c-----------------------------------------------------------------------
      subroutine findEigenVect(n,a)
      implicit none
      INTRINSIC sqrt

      integer n,i,j,k
      real*8  a(n,n),sum,norm

c compute k-th normalized eigenvector

      do k = 1,n
         norm = 1.0
         do i = k+1,n
            sum = a(i,k)
            do j = k+1,i-1
               sum = sum + a(i,j)*a(j,k)
            enddo
            if(sum.ne.0.) then
               a(i,k) = -sum/(a(i,i)-a(k,k))
            else
               a(i,k) = 0.
            endif
            norm = norm + a(i,k)**2
         enddo

         norm = 1./sqrt(norm)
         a(k,k) = norm
         do i = k+1,n
            a(i,k) = a(i,k)*norm
         enddo
      enddo

      return
      end
c=======================================================================
      subroutine forwardSolve(n,a,b)
      implicit none

      integer n,i,j
      real*8  a(n,n),b(n)

      b(1) = b(1)/a(1,1)
      do i = 2,n
         do j = 1,i-1
            b(i) = b(i) - a(i,j)*b(j)
         enddo
         b(i) = b(i)/a(i,i)
      enddo

      return
      end
c=======================================================================
      subroutine rndScale(n,D,t,b)
      implicit none
      INTRINSIC exp

      integer n,i
      real*8  D(n),t,b(n)

      do i = 1,n
         b(i) = b(i)*exp(D(i)*t)
      enddo

      return
      end
c=======================================================================
c lower triangular matrix-vector multiplication
c-----------------------------------------------------------------------
      subroutine multMatVec(n,a,b)
      implicit none

      integer n,i,j
      real*8  a(n,n),b(n)

      do i = n,1,-1
         b(i) = a(i,i)*b(i)
         do j = 1,i-1
            b(i) = b(i) + a(i,j)*b(j)
         enddo
      enddo

      return
      end

