/*=====================================================================
 * Integrate a trilinear (for conc) or constant (for totcoef)
 *   function in a tetrahedron
 *
 * The quadrature rule (exact on quadratics) used is
 *
 *   int_T f(x) dx = vol(T)/20 * ( sum_{4 vertices} f(vertex)
 *                          + 16 * f( .25*sum_{4 vertices} vertex) )
 *
 * INPUTS:
 *   vtx          are the 4 vertices of the tetrahedron T
 *   sign         is the sign for adding the integral to or subtracting
 *                the integral from result
 *   label        defines the logical mesh location containing T
 *                starting from 0, not 1
 *   conc         are the average concentrations on the mesh.  We call
 *                triConc to get the 8 trilinear values
 *   totCoef      are the average totCoef values on the mesh
 *   concvtx      define the 2 near and far grid points of the mesh
 *   concsize     are the dimensions for totCoef
 *   compId       are the component id's (from 0, not 1) of the
 *                components of conc that are to be integrated,
 *                except the last value is the traceType, used to
 *                integrate totCoef
 *   n            is the number of integrations, compId's, and results
 *
 *----------------------------------------------------------------------*/

#include "../Main/ftn.h"

void tetint(vtx, sign, label, pad, conc, totCoef,
	     concvtx, concsize, compId, result, n)
     int sign, label[3], pad[3], concsize[3], n;
     int compId[];
     double vtx[4][3], concvtx[2][3];
     float  result[];
     float  conc[];
     float  totCoef[];
{
  int i,index;
  double d1[3][5],d2[3][5],sum,vol;
  double tconc[8];
  int lbl[3];

  double by20 = .05;
  
  double tetvol();

  /*-----------------------------------------------------------------------
   *  inline function
   */
#define interpConc(i)							\
  ((((tconc[0]-tconc[1])*d1[0][i] + tconc[1]) * d1[1][i] +		\
    ((tconc[2]-tconc[3])*d1[0][i] + tconc[3]) * d2[1][i]) * d1[2][i] +	\
   (((tconc[4]-tconc[5])*d1[0][i] + tconc[5]) * d1[1][i] +		\
    ((tconc[6]-tconc[7])*d1[0][i] + tconc[7]) * d2[1][i]) * d2[2][i])
    ;
  /*----------------------------------------------------------------------*/

  vol = sign * tetvol(vtx);

  for (i = 0; i < 3; ++i)
    {
      double byD = 1. / (concvtx[1][i] - concvtx[0][i]);
      double centroid = .25 * (vtx[0][i] + vtx[1][i] + vtx[2][i] + vtx[3][i]);
      d1[i][0] = byD * (concvtx[1][i] - vtx[0][i]);
      d1[i][1] = byD * (concvtx[1][i] - vtx[1][i]);
      d1[i][2] = byD * (concvtx[1][i] - vtx[2][i]);
      d1[i][3] = byD * (concvtx[1][i] - vtx[3][i]);
      d1[i][4] = byD * (concvtx[1][i] - centroid);

      d2[i][0] = 1. - d1[i][0];
      d2[i][1] = 1. - d1[i][1];
      d2[i][2] = 1. - d1[i][2];
      d2[i][3] = 1. - d1[i][3];
      d2[i][4] = 1. - d1[i][4];
    }
  
  for (i = 0; i < n-1; ++i)
    {
      int cmp = compId[i]+1;
      lbl[0] = label[0]+1-pad[0];
      lbl[1] = label[1]+1-pad[1];
      lbl[2] = label[2]+1-pad[2];

      Trilinear(conc, tconc, &lbl[0], &lbl[1], &lbl[2], &cmp);

      sum = interpConc(0) + interpConc(1) + interpConc(2)
	+ interpConc(3) + 16.*interpConc(4);
      result[i] += vol * by20 * sum;
    }

  lbl[0] = ( label[0] < 0 ? 0 : label[0] );
  lbl[0] = ( lbl[0] > concsize[0] ? concsize[0] : lbl[0] );
  lbl[1] = ( label[1] < 0 ? 0 : label[1] );
  lbl[1] = ( lbl[1] > concsize[1] ? concsize[1] : lbl[1] );
  lbl[2] = ( label[2] < 0 ? 0 : label[2] );
  lbl[2] = ( lbl[2] > concsize[2] ? concsize[2] : lbl[2] );
  index = lbl[0] + concsize[0]*
    (lbl[1] + concsize[1]*(lbl[2] + concsize[2]*(compId[n-1]-1)));
  result[n-1] += vol * totCoef[index];
}

#define det(a1,a2,a3,b1,b2,b3,c1,c2,c3) \
  (a1*(b2*c3-c2*b3)-b1*(a2*c3-c2*a3)+c1*(a2*b3-b2*a3))

double tetvol(vtx)
     double vtx[4][3];
{
  double nvtx[3][3];

  nvtx[0][0] = vtx[0][0] - vtx[3][0];
  nvtx[0][1] = vtx[0][1] - vtx[3][1];
  nvtx[0][2] = vtx[0][2] - vtx[3][2];
  	    	        	    
  nvtx[1][0] = vtx[1][0] - vtx[3][0];
  nvtx[1][1] = vtx[1][1] - vtx[3][1];
  nvtx[1][2] = vtx[1][2] - vtx[3][2];
  	    	        	    
  nvtx[2][0] = vtx[2][0] - vtx[3][0];
  nvtx[2][1] = vtx[2][1] - vtx[3][1];
  nvtx[2][2] = vtx[2][2] - vtx[3][2];
  
  return det(nvtx[0][0],nvtx[1][0],nvtx[2][0],
	     nvtx[0][1],nvtx[1][1],nvtx[2][1],
	     nvtx[0][2],nvtx[1][2],nvtx[2][2] ) / -6.;
}
