/******************************************************************************
Transport

Parallel 3D Reactive Transport using the
  Characteristics-Mixed Method
  Higher order Godunov Method

As examples, this routine can simulate transport for the following:

     (1)  3-D Miscible Displacement as may be encountered
          in oil production;

     (2)  Multi-phase solute transport with linear sorption,
          radionuclide decay, and simple biodegradation, as may
          be encountered in ground-water contamination;

     (3)  General purpose geochemistry.

Serial code begun December 1991
Parallel code begun June 1992
Simple bioremediation added January 1993
Multi-phase capability added August 1993
General geometry capability added December 1993
Radionuclide decay added December 1993
Godunov added January 1996
Geochemistry added July 1997
Driver reformulated August 1997
Chemistry Work Load measuring added December 1997

Primary authors:
  Todd Arbogast
  Doug Moore
  Ashokkumar Chilakapati
  Clint Dawson (HOG)
  Fredrik Saaf (Geochemistry)

Complete list of authors:
  Todd Arbogast
  Steve Bryant
  Ashokkumar Chilakapati
  Lawrence Cowsar
  Clint Dawson
  Joe Eaton
  Frederic d'Hennezel
  Robert McLay
  Douglas Moore
  Fredrik Saaf
  Carol San Soucie (Woodward)
  Shuyu Sun
  Mary Wheeler
  Ivan Yotov

  Before September 1, 1995:
  Department of Computational and Applied Mathematics
  Rice University
  Houston, TX  77251--1892

  After September 1, 1995:
  Center for Subsurface Modeling
  Texas Institute for Computational and Applied Mathematics
  The University of Texas at Austin
  Austin, Texas  78712

NOTE: Only subdomain information need be passed in,
      i.e., initial padded information will be communicated.

BRIEF DEFINITIONS OF SELECTED VARIABLES (OUT OF DATE?):
  step_t            time step number (0=initial state, 1=first call)
  time_t            time level [at step_t-1 on call, at step_t on exit]
  dt                time step size [desired on input, taken on output]
  cflFactor_t       for automatic time step control with CMM: the proportion of
                    the padded region we are willing to trace into
                    (a negative number means use dt as is)
  computeTransport  indicates method used for the advection step.
                    (1=characteristic tracing, 2=higher-order
                    Godunov method, 3=first-order Godunov method)
  conc              The average primaryPhase concentration (padded)
                       Postprocess to get the trilinear
                       at the eight corners
                       Corner: 1=(lo,lo,lo), 2=(hi,lo,lo),
                               3=(lo,hi,lo), 4=(hi,hi,lo),
                               5=(lo,lo,hi), 6=(hi,lo,hi),
                               7=(lo,hi,hi), 8=(hi,hi,hi).
                      Internally, conc is sometimes the total
                      concentration (in all phases)
  totCoef           phase partitioning coefficient for each mobile specie:
                    mole-volume relation between primaryPhase and
                    total concentration for each mobile trace type
  nSpecies          total number of species
  NPHASES           total number of fluid phases (compilation macro)
  phaseDist         Henry's or Raoult's constants for the linear
                    phase partitioning
  xpComp,ypComp,    (computational) concentration grid points (padded)
    zpComp
  nx,ny,nz          local number of grid cells
  padx,pady,        padding distance (in cells) on each face
    padz,
  molDiff           molecular diffusion
  longDisp          longitudinal dispersion
  transDisp         transverse dispersion
  effIdentity       effective identity tensor for dispersion
  zeroD             flag for zero dispersion
  computeRND        flag for radionuclide decay
  iParmRND          int parameters for the radionuclide decay
  rParmRND          double parameters for the radionuclide decay
  computeRXN        flag for reaction kinetics
  nMicroStepsRXN    number of Runga-Kutta steps for reactions
  nrParmRXN         size of rParmRXN array
  rParmRXN          double parameters for the reaction kinetics
  niParmRXN         size of iParmRXN array
  iParmRXN          int parameters for the reaction kinetics
  steadyFlow        flag for steady flow (time independent velocity)
  sorpType          sorption type by component index array
                    (type 0 means immobile)
  nSorpTypes        number of mobile sorption types
  bcRegion?         boundary condition regions
                    faces (1=x,2=y,3=z), by cells;
  bcType_t          boundary condition types (see the include files)
  nBCRegions        number of BC regions
  ntCutMax_t        maximun number of times the time step may be cut
  relTol_t          solver relative tolerance factor
  maxIter_t         maximun number of iterations the solver may take
  volTol_t          maximum traceback volume relative error
  volRefine_t       level of traceback volume refinement
  outFlags_t[?]     output flags
  monitorUnit       fortran unit number for monitoring output
  debugUnit         fortran unit number for debugging output
  traceUnit         fortran unit number for output of traceback points
  error             error flag
                      1 = nonconvergence of solver (nonfatal)
                      2 = message passing error (fatal)
                      3 = not enough work space (fatal)
  myNbr             my neighbors' node numbers
                      myNbr_(0,0,0) = my node number
                      myNbr_(i,j,k) = node number of neighbor offset
                         in direction by (i,j,k), |i|,|j|,|k| <= 1.
                      value is -1 if no neighbor
  dwk               double work space array
  iwk               integer work space array
  dimDwk            dimension of dwk array
  dimIwk            dimension of iwk array

NOTES:
  1. Padding is by a fixed amount on all faces.
  2. Trace-back and post-processing quantities are padded.
     Padded by padx,pady,padz on each x,y,z face, respectively:
        byhx, byhy, byhz,
        byDistx, byDisty, byDistz, totcoef
     Padded by padx+1,pady+1,padz+1 on each x,y,z face:
        hx, hy, hz, xpComp, ypComp, zpComp, conc
  3. Matrix application quantities are padded to width 1 on each face:
        cmatrix, csol
  4. For the flags: 0=false, 1=true.
  5. If steadyFlow is set, dt cannot change between calls.
     This is not checked.
  6. A consistent set of units are assumed.
  7. It is not assumed that padded info in conc is communicated on
     first call; however, padded info in conc is communicated
     before exit and assumed to be correct on subsequent calls.

SOME INTERNAL VARIABLES
  hx,hy,hz          cell sizes
  byDistx,byDisty,  inverses of distances between cell centers
    byDistz
  byhx,byhy,byhz    inverses of cell sizes
  cbar              trace-back integrals and/or advected conc
  cdmatrix          diagonal of linear system matrix
  cmatrix           linear system matrix
  crhs              right-hand side of linear system
  csol              solution of linear system
  d11,d12,d13,      dispersion coefficients
    d22,d23,d33
  dispType          distinct dispersion type index (0=no dispersion)
  nDispTypes        number of nonzero distinct dispersion types
  nTraceTypes       number of nonzero distinct trace types
  totCoefInt        integral of molar-volume relation between
                    primaryPhase and total concentration at the
                    advanced time level on the concentration grid
                    for each mobile traceType (smoothed)
  totCoef           average integral of molar-volume relation between
                    primaryPhase and total concentration at the
                    advanced time level on the concentration grid
                    for each mobile traceType (not smoothed)
  traceType         distinct trace type index (0=immobile)

EXTERNAL ROUTINES NEEDED
  double function evalPorosity(x,y,z)
    Evaluate the porosity at the given point
  double function evalSorp(x,y,z,sorpType)
    Evaluate the given sorption factor at the given point (sorpType=0
    is the same as porosity)
  subroutine evalVel(x,y,z,t,phaseVel): phaseVel(3,nphase)
    Evaluate the phase velocities at the given point and time
  subroutine evalPorSat(x,y,z,t,sat): sat(nphase)
    Evaluate porosity times the phase saturations at the given
    point and time
  double function evalRxnVolInt(xpComp(i),ypComp(j),zpComp(k))
    Evaluate porosity times the water saturations at the given
    point and time
  subroutine evalSorpInt(xpComp(i),ypComp(j),zpComp(k),sorpint)
    Integrate the sorption factors over the given cube
  subroutine evalTotCoefInt(xpComp(i),ypComp(j),zpComp(k),totcoefint)
    Integrate totCoef over the given cube

DIFFERENTIAL EQUATIONS

Transport equation (with reactions and RND)

  conc_i = molar concentration of species i in primaryPhase_i
  D((saturation),vel) = diffusion/dispersion tensor

  totCoef_i = sum_{j=1}^NPHASES [ (porosity)*(saturation)_j*phaseDist(j,i)
                      + sorp_i*phaseDist(0,i) ]
  velocity_i = sum_{j=1}^NPHASES vel_j*phaseDist(j,i)
  D_i = sum_{j=1}^NPHASES D((saturation)_j,vel_j)

  totCoef_i * (d conc_i/dt) + div { velocity_i * conc_i - D_i grad conc_i }
   = (porosity)*(saturation)_{primaryPhase_i}*R_i^C(conc_1,...,conc_{nSpecies})
     + totCoef_i * R_i^N(conc_1,...,conc_{nSpecies})
     + cwell(i)*(well+) - conc_i*(well-)

Notes: (1) we input q = well*d(vol)
       (2) we assumed small concentrations, so the density is
          (nearly) constant wrt concentration changes

Immobile components satisfy

  d conc_i/dt
    = R_i^C(conc_1,...,conc_{nSpecies}) + R_i^N(conc_1,...,conc_{nSpecies})
      + cwell(i)*(well+) - conc_i*(well-)

In the Characteristics-Mixed Method (CMM), on each element E
  we use the approximately integrated form

  [ (totCoef_i, 1)_E * conc_i - (totCoef_i * conc_i, 1)_{E-traceback} ] / (dt)
    = (div D_i grad conc_i, 1)_E
      + ((porosity)*(saturation)_{primaryPhase_i, 1)_E
        * R_i^C(conc_1,...,conc_{nSpecies})
      + (totCoef_i, 1)_E * R_i^N(conc_1,...,conc_{nSpecies})

  where wells and BC's are in the traceback integral.

BASIC ALGORITHM STEPS

  A time splitting algorithm is used to solve these equations
  in several basic steps.

Leaks:

  We add a specified amount of moles to specified cells, half before
  advection, and half just before we equilibrate reactions.

  conc_i is incremented by leakRate * dt / totCoef_i

  Remark: leakRate is in moles (mass) per time per volume

Advection:

  CMM

    Shift to total concentration:

      T_i^n = totCoef_i * conc_i^n (call it conc_i in the code)

    Remark: Currently, we define totCoef_i by assuming there is
            only 1 flowing phase

    Solve by characteristics over each element:

      d T_i
      ----- + div [ traceVel_i * T_i ] = (sources and BC's),
        dt

    where

      traceVel_i = velocity_i / totCoef_i

    We assume traceVel_i is continuous, so this quantity is smoothed.  Then
     
      tbarInt_i - int_{E,TB} T_i^n = dt * int_{E,TB} (sources and BC's)^n

   Shift back to concentrations:

      cbar1_i = tbarInt_i / totCoefInt_i (we let tbar = cbar in the code)

   Remark: The CMM post-processing step is incorporated into the traceback
     integral evaluation to save storage.  Thus conc_i is only the average
     concentration in the cell.

  HOG/FOG

    totCoef_i( cbar1_i - c_i^n ) + div ( velocity_i^n * conc_i^n )
      = dt * (sources and BC's)

  NONE OR IMMOBILE

    cbar1_i = c_i^n

Radionuclide decay:

  cbar2_i - cbar1_i = dt * R_i^N(cbar1_1,...,cbar1_{nSpecies})

Reactions:

  cbar3_i - cbar2_i
    = (( porosity)*(saturation)_{primaryPhase_i} * dt / totCoef_i )
      * R_i^C(cbar2_1,...,cbar2_{nSpecies})

Diffusion/dispersion:

  MOBILE
    totCoef_i * ( csol_i - cbar3_i ) = dt * div ( D grad csol_i)

  IMMOBILE
    csol_i = cbar3_i

  The solution is conc_i = csol_i.

Reaction equilibration.

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "../Main/msgs.h"
#include "../Main/debug.h"
#include "../Main/driverGlobals.h"
#define TRN_ALLOCATE_MEMORY
#include "transportGlobals.h"
#include "transport.h"
#include "reaction.h"

/* ROUTINES */
static int chemistryRxn(double dt, int equilibrateOnly, double* time_chem);
void transportClose();
static void debugOutConc(int printPad);
static void debugOutTB();
static void debugOutCRHS();
static int updateConc(int updateWells, int updateBCS, int updatePad);
int defineTotCoef();

void derivedSizes();
int initTrans();

/* MACROS */
#define ABORT(err) do{ error = err; goto exit; }while(0)
#define max(i,j) ( (i>j) ? (i) : (j) )

/* STATIC VARIABLES */
static int monitorRoutines;
static double* localConc;
static FILE* debugRxnFile;

/******************************************************************************
 * Initialize the Transport driver
 *****************************************************************************/
int transportInit() {
  int error = 0;
  int icell;

  int dimGrid0 = nx*ny*nz;
  int dimGrid1 = (nx+1)*(ny+1)*(nz+1);
  int dimGrid2 = (nx+2)*(ny+2)*(nz+2);

  monitorRoutines =  ( (outFlags_t[OF_ROUTINES] >= 1) && !myProc );

  if(monitorRoutines) printf("Initializing Transport\n");

  /* MISCELLANEOUS */

  zeroD = !computeTransport
    || (molDiff == 0 && longDisp == 0 && transDisp == 0);

  monitorUnit = 6;
  traceUnit = 16;
  debugUnit = 0;
  if( outFlags_t[OF_SOLUTION] || outFlags_t[OF_INITIALIZATION]
     || outFlags_t[OF_TRACEBACK_INTEGRALS] || outFlags_t[OF_REACTIONS]
     || outFlags_t[OF_LINEARSYSTEM] ) debugUnit = 14;

  if(debugUnit) {
    if(nProcs > 9999) {
      fprintf(stderr,
	      "Too many processors to open transport debug files\n");
      debugUnit = 0;
    } else {
      OpenTransDebug(&myProc,&debugUnit);
    }
  }

  /* define some sizes */
  derivedSizes();

  /* ALLOCATE MEMORY */

  if(monitorRoutines) printf("Allocating memory\n");
  
  if(nTraceTypes) {
    totCoef = (float*)
      malloc(((nx+2*padx)*(ny+2*pady)*(nz+2*padz)*nTraceTypes)*sizeof(float));
    totCoefInt = (double*) malloc(dimGrid0*nTraceTypes*sizeof(double));
    if(totCoef == NULL || totCoefInt == NULL) ABORT(ERR_MEMORY);
  }
  traceType = (int*) malloc(nSpecies*sizeof(int));
  dispType  = (int*) malloc(nSpecies*sizeof(int));
  if(traceType == NULL || dispType == NULL) ABORT(ERR_MEMORY);

  if(nWells) {
    wellInjConc    = (double*) malloc(nSpecies*nWells*sizeof(double));
    wellProd       = (double*) malloc(nSpecies*nWells*sizeof(double));
    wellTotCoefInt = (double*) malloc(nSpecies*nWells*sizeof(double));
    if(wellInjConc == NULL || wellProd == NULL
       || wellTotCoefInt == NULL) ABORT(ERR_MEMORY);
  }

  {
    int size = (nx+2*(padx+1)) + (ny+2*(pady+1)) + (nz+2*(padz+1))
      + (nx+2*padx) + (ny+2*pady) + (nz+2*padz)
	+ (nx+2*padx+1) + (ny+2*pady+1) + (nz+2*padz+1);
    hx = (double*) malloc(size*sizeof(double));
  }
  if(hx == NULL) ABORT(ERR_MEMORY);
  hy      = hx + (nx+2*(padx+1));
  hz      = hy + (ny+2*(pady+1));
  byhx    = hz + (nz+2*(padz+1));
  byhy    = byhx + (nx+2*padx);
  byhz    = byhy + (ny+2*pady);
  byDistx = byhz + (nz+2*padz);
  byDisty = byDistx + (nx+2*padx+1);
  byDistz = byDisty + (ny+2*pady+1);

  if(nDispTypes) {
    int cmatrix_s  = 10*dimGrid2;
    int cdmatrix_s = dimGrid0;
    int face1_s    = ny*nz*5;
    int face2_s    = nx*nz*5;
    int face3_s    = nx*ny*5;
    if(steadyFlow) {
      cmatrix_s  = cmatrix_s*nDispTypes;
      cdmatrix_s = cdmatrix_s*nDispTypes;
      face1_s    = face1_s*nDispTypes;
      face2_s    = face2_s*nDispTypes;
      face3_s    = face3_s*nDispTypes;
    }
    cmatrix  = (double*) malloc(cmatrix_s*sizeof(double));
    cdmatrix = (double*) malloc(cdmatrix_s*sizeof(double));
    face10   = (double*) malloc(face1_s*sizeof(double));
    face11   = (double*) malloc(face1_s*sizeof(double));
    face20   = (double*) malloc(face2_s*sizeof(double));
    face21   = (double*) malloc(face2_s*sizeof(double));
    face30   = (double*) malloc(face3_s*sizeof(double));
    face31   = (double*) malloc(face3_s*sizeof(double));
    if(cmatrix == NULL || cdmatrix == NULL
       || face10 == NULL || face11 == NULL
       || face20 == NULL || face21 == NULL
       || face30 == NULL || face31 == NULL) ABORT(ERR_MEMORY);
  }

  /* Allocate transitory memory */
  if(nDispTypes) {
    d11  = (double*) malloc(6*dimGrid1*sizeof(double));
    d12  = d11 + dimGrid1;
    d13  = d12 + dimGrid1;
    d22  = d13 + dimGrid1;
    d23  = d22 + dimGrid1;
    d33  = d23 + dimGrid1;
    if(d11 == NULL) ABORT(ERR_MEMORY);
  }

  csol = (double*) malloc(dimGrid2*sizeof(double));
  crhs = (double*) malloc(dimGrid2*sizeof(double));
  cbar = (double*) malloc(dimGrid0*nSpecies*sizeof(double));
  if(csol == NULL || crhs == NULL || cbar == NULL) ABORT(ERR_MEMORY);

  if(computeTransport == ADV_HOG || computeTransport == ADV_GODUNOV) {
    int size = 5*dimGrid2 + (nx+3)*(ny+2)*(nz+2)
      + (nx+2)*(ny+3)*(nz+2) + (nx+2)*(ny+2)*(nz+3);
    slx_hog = (double*) malloc(size*sizeof(double));
    if(slx_hog == NULL) ABORT(ERR_MEMORY);
    sly_hog  = slx_hog + dimGrid2;
    slz_hog  = sly_hog + dimGrid2;
    velx_hog = slz_hog + dimGrid2;
    vely_hog = velx_hog + (nx+3)*(ny+2)*(nz+2);
    velz_hog = vely_hog + (nx+2)*(ny+3)*(nz+2);
    phio_hog = velz_hog + (nx+2)*(ny+2)*(nz+3);
    phi_hog  = phio_hog + dimGrid2;

    flagWell = (int*) malloc(dimGrid2*sizeof(int));
    if(flagWell == NULL) ABORT(ERR_MEMORY);
  }

  if(computeChemistry) {
    localConc = (double*) malloc(nSpecies*sizeof(double));
    if(localConc == NULL) ABORT(ERR_MEMORY);

/* SLB, JE adding stuff for workload computation */
/*    set up array to store workload as if work were a pressure  */
    locIter = (float*) malloc(nx*ny*nz*sizeof(float));   
    if (locIter == NULL ) ABORT(ERR_MEMORY);
    for( icell = 0; icell < nx*ny*nz; icell++)
        locIter[icell] = 0.0;
    recv_list = (int*) malloc(2*nProcs*sizeof(int));
    if (recv_list == NULL ) ABORT(ERR_MEMORY);
    send_list = (int*) malloc(2*nProcs*sizeof(int));
    if (send_list == NULL ) ABORT(ERR_MEMORY);
    cell_list = (int*) malloc((1+nx*ny*nz)*sizeof(int));
    if (cell_list == NULL ) ABORT(ERR_MEMORY);
    for( icell = 0; icell < nx*ny*nz; icell++)
        cell_list[icell] = icell; 

    moles = (double*)malloc(nSpecies*sizeof(double));
    if(moles == NULL) ABORT(ERR_MEMORY);
    InitGuess = (double*)malloc(nSpecies*sizeof(double));
    if(InitGuess == NULL) ABORT(ERR_MEMORY);
    rates = (double*)malloc(4*nProducts*sizeof(double));
    if(rates == NULL) ABORT(ERR_MEMORY);
    refrate = (double*)malloc(nProducts*sizeof(double));
    if(refrate == NULL) ABORT(ERR_MEMORY);
    avgrate = (double*)malloc(nProducts*sizeof(double));
    if(avgrate == NULL) ABORT(ERR_MEMORY);
  }

  dimDwk = max(dimDwk,nSpecies);
  dimDwk = max(dimDwk,5*N_FLOWING_PHASES);
  dimDwk = max(dimDwk,(padz+1)*(nx+2*(padx+1))*(ny+2*(pady+1))+2);
  dimDwk = max(dimDwk,(pady+1)*(nx+2*(padx+1))*(nz+2*(padz+1))+2);
  dimDwk = max(dimDwk,(padx+1)*(ny+2*(pady+1))*(nz+2*(padz+1))+2);
  dimDwk = max(dimDwk,4*dimGrid2);
  dwk = (double*) realloc((void*)dwk,dimDwk*sizeof(double));
  if(dwk == NULL) ABORT(ERR_MEMORY);

  if( computeRXN && (outFlags_t[OF_REACTIONS] >= 1) ) {
    char name[14];
    sprintf(name,"rxn_debug%4.4d",myProc);
    debugRxnFile = fopen(name,"w");
  }

  /* INITTRANS */

  if(monitorRoutines) printf("Calling initTrans\n");
  error = initTrans(); if(error) goto exit;

  if(error) ABORT(-1);

  /* EQUILIBRATE IC's (so simulation starts from equilibrium conditions) */

  if(computeChemistry && nRQ != 0) {
    int specie;
    double time_chem_dummy = 0;
    if(monitorRoutines) printf("Equilibrating initial conditions\n");
    error = updateConc(1,0,0); if(error) ABORT(-1);
    SetCBar(cbar,conc,&nSpecies,&nx,&ny,&nz,&padx,&pady,&padz);
    error = chemistryRxn(1,1,&time_chem_dummy); if(error) goto exit;
    for(specie=0; specie<nSpecies; specie++)
      AdvanceCBar(&cbar_array_(specie),&conc_array_(specie),
		  &nx,&ny,&nz,&padx,&pady,&padz);
    error = updateConc(1,1,0); if(error) ABORT(-1);
   }

  /* Initial padding of conc array */

  error = updateConc(0,1,1); if(error) ABORT(-1);

 exit:
  switch(error) {
  case 0:
    break;
  case ERR_MEMORY:
    fprintf(stderr,"Memory allocation error on processor %d\n",myProc);
    break;
  case ERR_IN_CODE:
    fprintf(stderr,"Coding error in the transport code!\n");
    break;
  case ERR_CHEM_EQUILIBRIUM:
  case ERR_CHEM_MAX_ITER:
  case ERR_CHEM_MAX_BACKTRACKS:
  case ERR_CHEM_ILL_CONDITIONED:
  case ERR_CHEM_INFEASIBLE:
  case ERR_CHEM_INTEGRATION:
  case ERR_CHEM:
    if(!myProc)
      fprintf(stderr,
	      "Geochemistry error: non-convergence of IC's (error code %d)\n",
	      error);
    error = ERR_CHEM_EQUILIBRIUM;
    break;
  default:
    fprintf(stderr,"Fatal error on processor %d\n",myProc);
  }
  return error;
}

/******************************************************************************
 * Transport driver
 *
 * Remark on workspace array dwk
 *   The sizes needed in the routines follows.
 *     (1) pcg: 4*(nx+2)*(ny+2)*(nz+2)
 *     (2) radionuclide decay: nSpecies (if active)
 *     (3) evalTotCoefInt: nTraceTypes (<= nSpecies)
 *     (4) disp: 5*N_FLOWING_PHASES
 *****************************************************************************/
int transport(double* dt , double* time_chem) {
  int error = 0;

  int specie,well,newDT;
  int dimGrid2 = (nx+2)*(ny+2)*(nz+2);
  double byDt;
  static double dtOld = -1;
  double ElapsedTime = 0;

  int maxTCut = ntCutMax_t, tcut = 0;

  /* BEGIN *******************************************************************/

  /* ElapsedTime = dclockX(); */
  ElapsedTime = MPI_Wtime();
  /* if(monitorRoutines) printf("Computing Reactive Transport\n"); */

  newDT = (*dt != dtOld);

  /* Set well array values */

  for(well=0; well<nWells; well++)
    for(specie=0; specie <nSpecies; specie++)
      wellInjConc_(specie,well) = cWellValue(time_t,specie,well);

  /* Communicate padded information and set well and exterior values */

  error = updateConc(1,computeTransport,computeTransport); if(error) ABORT(-1);
  if(outFlags_t[OF_INITIALIZATION] >= 1) debugOutConc(1);

  /* SELECT TIME STEP SIZE ***************************************************/

  if(computeTransport && cflFactor_t > 0) {
    if(monitorRoutines) printf("Timestep control\n");
    switch(computeTransport) {
    case ADV_CMM:
      CflTimeStep(&cflFactor_t,dt,&padDist,
		  &time_t,&nTraceTypes,&xpComp_(0),&ypComp_(0),&zpComp_(0),
		  &nx,&ny,&nz,&padx,&pady,&padz);
      break;
    case ADV_HOG:
    case ADV_GODUNOV:
      {      
	int cflTimeOnly = 1;
	int msg = MSG_HOG1;
	Hog(cbar,conc,hx,hy,hz,dt,&time_t,&time_f,&cflTimeOnly,&nWells,
	    wellType,hWellIndex,zWellIndexLo,zWellIndexHi,wellInjConc,
	    &padx,&pady,&padz,&nx,&ny,&nz,&nSpecies,
	    phaseDist,myNbr,xpComp,ypComp,zpComp,
	    traceType,sorpType,&computeTransport,
	    slx_hog,sly_hog,slz_hog,velx_hog,vely_hog,velz_hog,
	    phio_hog,phi_hog,velX,velY,velZ,&msg,
	    flagWell,&monitorRoutines);
      }
      break;
    case ADV_DIFF_ONLY:
      break;
    default:
      ABORT(ERR_UNSUPPORTED_ADVECTION);
    }
  }

  /* LEAKS (FIRST HALF) ******************************************************/

  /* Note: if the time step is cut below, this must be corrected! */
  if(nLeaks) {
    double half_dt = *dt/2;
    dtOld = *dt;
    maxTCut = 0;
    UpdateLeaks(conc,totCoef,&time_t,&half_dt,
		leakIndexLo,leakIndexHi,traceType,&nLeaks,&nSpecies,
		jacInv,&nx,&ny,&nz,&padx,&pady,&padz,myNbr);
  }

  /* ADVECTION ***************************************************************/

  /* Advection Step (computeTransport)
       (ADV_CMM) The characteristics method, involving trace-backs along
         characteristics of the flow field. Recommended when large time-steps
	 are feasible (in the absence of fast reactions, etc.)
       (ADV_HOG) The higher-order Godunov method.
       (ADV_GODUNOV) First order Godunov method.
       (ADV_DIFF_ONLY) Diffusion (and dispersion) only.
 
    Repeat Advection-Reaction until an acceptable time-step is found:
    (1) Advection step is attempted using current time-step;
        if necessary it is cut within the traceback routine.
    (2) Reaction step is attempted using the advection time-step.
        If convergence is not attained, time-step is again cut,
	and we repeat the Advection-Reaction process.
  */

 startAdvection:
  if(monitorRoutines) printf("Desired dt = %g\n",*dt);

  switch(computeTransport) {
  case ADV_DIFF_ONLY:
  case 0:
    SetCBar(cbar,conc,&nSpecies,&nx,&ny,&nz,&padx,&pady,&padz);
    break;
  case ADV_CMM:
    if(monitorRoutines) printf("Advection step (characteristics)\n");

    /* Define total concentrations of mobile species (call them conc!) */
    DefineTotConc(conc,totCoef,&nx,&ny,&nz,&nSpecies,traceType,
		  &padx,&pady,&padz,myNbr);
    error = updateConc(0,0,1); if(error) ABORT(-1);

    {
      int tooManyDTCuts
	= traceBack(cbar,totCoefInt,conc,totCoef,volTol_t,
		    outFlags_t[OF_DELTAT],outFlags_t[OF_TRACEBACK_POINTS],
		    traceUnit,maxTCut,step_t-1,time_t,dt,newDT);
      if(tooManyDTCuts && !nLeaks) ABORT(ERR_TOO_MANY_DT_CUTS);
    }
    if(outFlags_t[OF_TRACEBACK_INTEGRALS] >= 1) debugOutTB();

    /* Return to primary concentrations */
    DefinePrimaryCBar(cbar,totCoef,&byhx_(0),&byhy_(0),&byhz_(0),
		      &nx,&ny,&nz,&padx,&pady,&padz,&nSpecies,traceType);
    break;
  case ADV_HOG:
  case ADV_GODUNOV:
    if(monitorRoutines) printf("Advection step (Godunov)\n");
    {
      int cflTimeOnly = 0;
      int msg = MSG_HOG2;
      Hog(cbar,conc,hx,hy,hz,dt,&time_t,&time_f,&cflTimeOnly,&nWells,
	  wellType,hWellIndex,zWellIndexLo,zWellIndexHi,wellInjConc,
	  &padx,&pady,&padz,&nx,&ny,&nz,&nSpecies,
	  phaseDist,myNbr,xpComp,ypComp,zpComp,
	  traceType,sorpType,&computeTransport,
	  slx_hog,sly_hog,slz_hog,velx_hog,vely_hog,velz_hog,
	  phio_hog,phi_hog,velX,velY,velZ,&msg,
	  flagWell,&monitorRoutines);
    }
    break;
  default:
    ABORT(ERR_UNSUPPORTED_ADVECTION);
  }

  if(nLeaks && *dt != dtOld) {
    double prev_t = time_t + dtOld;
    double dtCorrection = (*dt-dtOld)/2;
    dtOld = *dt;
    if(computeTransport==ADV_CMM) {
      tcut++;
      if(tcut > ntCutMax_t) ABORT(ERR_TOO_MANY_DT_CUTS);
    }
    if(computeTransport==ADV_CMM) {
      UnDefineTotConc(conc,totCoef,&nx,&ny,&nz,&nSpecies,traceType,
		      &padx,&pady,&padz,myNbr);
    }
    UpdateLeaks(conc,totCoef,&prev_t,&dtCorrection,
		leakIndexLo,leakIndexHi,traceType,&nLeaks,&nSpecies,
		jacInv,&nx,&ny,&nz,&padx,&pady,&padz,myNbr);
    newDT = 1;
    goto startAdvection;
  }

  byDt = 1/(*dt);

  /* Define totCoef (leaks incorrect if N_FLOWING_PHASES > 1 && dt changes) */
  if(N_FLOWING_PHASES > 1) {
    if(computeTransport != ADV_CMM)
    DefineTotCoefInt(totCoefInt,&time_t,&xpComp_(0),&ypComp_(0),&xpComp_(0),
		     &byhx_(0),&byhy_(0),&byhz_(0),
		     &nx,&ny,&nz,&padx,&pady,&padz,&nTraceTypes,dwk);
    /* DefineTotCoefSimple(totCoef,totCoefInt,&byhx_(0),&byhy_(0),&byhz_(0),
			&nx,&ny,&nz,&padx,&pady,&padz,
			&nTraceTypes,myNbr,&error); */
    error = defineTotCoef();
    if(error) goto exit;
  }
    

  /* RADIONUCLIDE DECAY ******************************************************/

  if(computeRND) {
    if(monitorRoutines) printf("Radionuclide decay\n");
    Decay(cbar,dt,iParmRND,rParmRND,&nx,&ny,&nz,dwk);
  }

  /* REACTIONS ***************************************************************/

  /* Specialized chemical reactions */
  if(computeRXN) {
    if(outFlags_t[OF_REACTIONS] >= 1)
      reactionHeader(step_t, time_t, *dt, debugRxnFile);
    if(monitorRoutines) printf("Specialized chemical reactions\n");
    reaction(*dt,debugRxnFile);
  }

  if(outFlags_t[OF_ROUTINES] >= 1) {
    /* ElapsedTime -= dclockX(); */
    ElapsedTime = MPI_Wtime() - ElapsedTime;
    printf("Proc %d: Time in advection = %g\n", myProc , ElapsedTime);
  }
  /* ElapsedTime = dclockX(); */

  /* Reactions through general chemistry package */
  if(computeChemistry) {
    double dt_chem = 0;
    if(outFlags_t[OF_REACTIONS] >= 1)
      ReactionHeader(&step_t, &time_t, dt, &debugUnit);
    if(monitorRoutines) printf("General chemical reactions\n");
    /* dt_chem = dclockX(); */
    dt_chem = MPI_Wtime();
    error = chemistryRxn(*dt,0,time_chem);
    if(outFlags_t[OF_ROUTINES] >= 1) {
      /* dt_chem -= dclockX(); */
      dt_chem -= MPI_Wtime();
      printf("Proc %d: Total time in chemistry = %g\n",myProc , dt_chem);
    }
    if(error) {
      *dt /= 2;
      newDT = 1;
      if(nLeaks) {
	double prev_t = time_t + dtOld;
	double dtCorrection = -*dt/4;
	dtOld = *dt;
	if(computeTransport==ADV_CMM) {
	  UnDefineTotConc(conc,totCoef,&nx,&ny,&nz,&nSpecies,traceType,
			  &padx,&pady,&padz,myNbr);
	}
	UpdateLeaks(conc,totCoef,&prev_t,&dtCorrection,
		    leakIndexLo,leakIndexHi,traceType,&nLeaks,&nSpecies,
		    jacInv,&nx,&ny,&nz,&padx,&pady,&padz,myNbr);
	tcut= 0;
      }
      goto startAdvection;
    }
  }

  /* UPDATE TIME *************************************************************/

  time_t = time_t + (*dt);
  dtOld = *dt;

  /* LEAKS (SECOND HALF) *****************************************************/

  if(nLeaks) {
    double half_dt = *dt/2;
    double prev_t = time_t - half_dt;
    UpdateLeaks(conc,totCoef,&prev_t,&half_dt,
		leakIndexLo,leakIndexHi,traceType,&nLeaks,&nSpecies,
		jacInv,&nx,&ny,&nz,&padx,&pady,&padz,myNbr);
  }

  /* DISPERSION **************************************************************/

  /* Loop over each dispersion type and component */

  if(!computeTransport) {
    for(specie=0; specie<nSpecies; specie++)
      AdvanceCBar(&cbar_array_(specie),&conc_array_(specie),
		  &nx,&ny,&nz,&padx,&pady,&padz);
  } else {
    int index = 0;
    static int needLinSyst = 1;
    int computeLinSyst = needLinSyst;
    int type;

    for(type = 0; type<=nDispTypes; type++) {
      if(monitorRoutines) printf("Dispersion type: %d\n",type);

      if(steadyFlow) index = type-1;

      for(specie=0; specie<nSpecies; specie++) {
	if(dispType[specie] != type)
	  continue;

	if(type == 0) {
	  AdvanceCBar(&cbar_array_(specie),&conc_array_(specie),
		      &nx,&ny,&nz,&padx,&pady,&padz);
	  /* gsyncX(); */
	  MPI_Barrier(comm);
	  continue;
	}

	/* Evaluate dispersion tensor and set up dispersive part of */
	/*   concentration matrix (if necessary) */

	if(computeLinSyst) {
	  int specie1 = specie + 1;

	  if(monitorRoutines) printf("  Calling disp/csetupmat\n");

	  if(uniformDispersion) {
	    Disp(d11,d12,d13,d22,d23,d33,
		 &xpComp_(0),&ypComp_(0),&zpComp_(0),
		 &time_t,&molDiff,&longDisp,&transDisp,
		 phaseDist,&nSpecies,&specie1,effIdentity,
		 dwk,&dwk[3*N_FLOWING_PHASES],&dwk[4*N_FLOWING_PHASES],
		 &nx,&ny,&nz,&outFlags_t[OF_LINEARSYSTEM],&debugUnit);
	  } else {
	    DispVarD(d11,d12,d13,d22,d23,d33,
		     &xpComp_(0),&ypComp_(0),&zpComp_(0),
		     &time_t,molDiffAry,longDispAry,transDispAry,
		     phaseDist,&nSpecies,&specie1,effIdentity,
		     dwk,&dwk[3*N_FLOWING_PHASES],&dwk[4*N_FLOWING_PHASES],
		     &nx,&ny,&nz,&outFlags_t[OF_LINEARSYSTEM],&debugUnit);
	  }
	  SetupBCFlux(&face10_array_(index),&face20_array_(index),
		      &face30_array_(index),
		      &xpComp_(0),&ypComp_(0),&zpComp_(0),&nx,&ny,&nz,
		      bcRegion1,bcRegion2,bcRegion3,bcType_t,
		      &specie1,&time_t);
	  CSetupMat(&cmatrix_array_(index),&nx,&ny,&nz,
		    d11,d12,d13,d22,d23,d33,
		    &hx_(-1),&hy_(-1),&hz_(-1),
		    &byDistx_(0),&byDisty_(0),&byDistz_(0),
		    &cdmatrix_array_(index),
		    &face10_array_(index),&face11_array_(index),
		    &face20_array_(index),&face21_array_(index),
		    &face30_array_(index),&face31_array_(index),
		    bcRegion1,bcRegion2,bcRegion3,bcType_t,
		    &outFlags_t[OF_LINEARSYSTEM],&debugUnit,myNbr);
	  computeLinSyst = 0;
	}

	/* Set up diagonal of matrix, initial guess, and right-hand-side */

	CSetupRHS(&cmatrix_array_(index),&nx,&ny,&nz,
		  &hx_(-1),&hy_(-1),&hz_(-1),
		  &face10_array_(index),&face11_array_(index),
		  &face20_array_(index),&face21_array_(index),
		  &face30_array_(index),&face31_array_(index),
		  &cdmatrix_array_(index),crhs,csol,&conc_array_(specie),
		  &padx,&pady,&padz,bcRegion1,bcRegion2,bcRegion3,
		  bcType_t,&cbar_array_(specie),
		  &totCoef_array_(traceType[specie]-1),&byDt);
	if(outFlags_t[OF_LINEARSYSTEM] >= 1) debugOutCRHS();

	/* Solve the linear system to compute the average concentration */

	if(monitorRoutines) printf("  Calling pcg, specie = %d\n",specie+1);
	Pcg(&cmatrix_array_(index),csol,crhs,&nx,&ny,&nz,
	    &relTol_t,&maxIter_t,dwk,&dwk[dimGrid2],
	    &dwk[2*dimGrid2],&dwk[3*dimGrid2],myNbr,
	    &outFlags_t[OF_PCG],&monitorUnit,&error);
	if(error) {
	  if(error == 1) ABORT(ERR_PCG_CONVERGENCE);
	  ABORT(-1);
	}

	Smooth(csol,&nx,&ny,&nz);
	SetConc(&conc_array_(specie),csol,&nx,&ny,&nz,&padx,&pady,&padz);
      }
      computeLinSyst = needLinSyst;
    }
    if(steadyFlow) needLinSyst = 0;
  }

  /* END *********************************************************************/

  /* Update so wells and external boundary region is correct */
  error = updateConc(1,1,0); if(error) ABORT(-1);

  /* Chemical Equilibrium */
  if(computeChemistry && !zeroD && nRQ != 0) {
    if(monitorRoutines) printf("Equilibrating final conditions\n");
    SetCBar(cbar,conc,&nSpecies,&nx,&ny,&nz,&padx,&pady,&padz);
    error = chemistryRxn(1,1,time_chem);
    for(specie=0; specie<nSpecies; specie++)
      AdvanceCBar(&cbar_array_(specie),&conc_array_(specie),
		  &nx,&ny,&nz,&padx,&pady,&padz);
    if(error) goto exit;
  }

  /* Update so external boundary region and padding is correct */
  if(periodicBC[0] || periodicBC[1] || periodicBC[2]) {
     error = updateConc(0,1,1); if(error) ABORT(-1);
  }

  /* ERROR CONTROL AND EXIT **************************************************/

 exit:
  if(outFlags_t[OF_SOLUTION] >= 1) debugOutConc(0);

  switch(error) {
  case 0:
    break;
  case ERR_MEMORY:
    fprintf(stderr,"Memory allocation error on processor %d\n",myProc);
    break;
  case ERR_UNSUPPORTED_ADVECTION:
    if(!myProc) fprintf(stderr,"Unsupported advection option\n");
    break;
  case ERR_TOO_MANY_DT_CUTS:
    if(!myProc) fprintf(stderr,"Cutting time-step too many times\n");
    break;
  case ERR_PCG_CONVERGENCE:
    if(!myProc) fprintf(stderr,
			"Nonconvergence of the dispersion linear solver\n");
    break;
  case ERR_CHEM_EQUILIBRIUM:
  case ERR_CHEM_MAX_ITER:
  case ERR_CHEM_MAX_BACKTRACKS:
  case ERR_CHEM_ILL_CONDITIONED:
  case ERR_CHEM_INFEASIBLE:
  case ERR_CHEM_INTEGRATION:
  case ERR_CHEM:
    if(!myProc)
      fprintf(stderr,
	      "Geochemistry equilibrium not reached (error code %d)\n",
	      error);
    error = ERR_CHEM_EQUILIBRIUM;
    break;
  default:
    fprintf(stderr,"Fatal error on processor %d\n",myProc);
  }
  return error;
}

/******************************************************************************
 * Reactions through the general chemistry package
 * Input:  dt                Time step size
 *         equilibrateOnly   Flag to equilibrate species only (such as IC's)
 * Return: error flag
 *****************************************************************************/
static int chemistryRxn(double dt, int equilibrateOnly, double* time_chem) {

  int cumNtTaken,cumIter,peakIter;
  int exceptionLoc[3];
  double avgIter,avgNtTaken;
  double imbalance;
  double timeStart;
  int error = 0;

  /* react */
  /* timeStart = dclockX(); */
  timeStart = MPI_Wtime();
  error = reactionStep(dt, chmLoadBalFlag , equilibrateOnly,
		       &cumNtTaken,&cumIter,&peakIter,
		       exceptionLoc);
  /* *time_chem += dclockX() - timeStart; */
  *time_chem += MPI_Wtime() - timeStart;

  /* process error flag */
  /* gimaxX(MSG_RXN,&error,1); */
  MSG_Gimax1(&error,comm);
  if(error) {
    switch(error) {
    case 1: error = ERR_CHEM_MAX_ITER; break;
    case 2: error = ERR_CHEM_MAX_BACKTRACKS; break;
    case 3: error = ERR_CHEM_ILL_CONDITIONED; break;
    case 4: error = ERR_CHEM_INFEASIBLE; break;
    case 5: error = ERR_CHEM_INTEGRATION; break;
    default: error = ERR_CHEM;
    }
    if(monitorRoutines) {
      if(equilibrateOnly) {
	printf("Equilibration unsuccessful");
      } else {
	printf("Speciation unsuccessful");
      }
      switch(error) {
      case ERR_CHEM_MAX_ITER:
	printf(": Exceeded max number of non-linear iterations\n");
	break;
      case ERR_CHEM_MAX_BACKTRACKS:
	printf(": Exceeded max number of back-tracks in linesearch\n");
	break;
      case ERR_CHEM_ILL_CONDITIONED:
	printf(": Ill-conditioned linear system\n");
	break;
      case ERR_CHEM_INFEASIBLE:
	printf(": Infeasible initial point\n");
	break;
      case ERR_CHEM_INTEGRATION:
	printf(": Time-integration error\n");
	break;
      case ERR_CHEM:
	printf(".\n");
	break;
      }
      goto exit;
    }
  }

  /* statistics */
  {
    double nBlocks;
    int buf[3];

    imbalance = cumIter + cumNtTaken;

    buf[0] = cumNtTaken;
    buf[1] = cumIter;
    buf[2] = nx*ny*nz;
  
    if(outFlags_t[OF_REACTIONS] ==2) {
       printf("%d Local cumIter: %d  cumNtTaken: %d\n",
               myProc,cumIter, cumNtTaken);
    }
    /* gimaxX(MSG_RXN,&peakIter,1); */
    /* gisumX(MSG_RXN,buf,3); */
    MSG_Gimax1(&peakIter,comm);
    { int buf_wk[3]; MSG_Gisum2(buf,buf_wk,3,comm); }

    cumNtTaken = buf[0];
    cumIter = buf[1];
    nBlocks = buf[2];

    avgIter = cumIter/nBlocks;
    avgNtTaken = cumNtTaken/nBlocks;
    if (cumIter > 0) {
       imbalance = imbalance/(cumIter+cumNtTaken)*nProcs;
    } else {
       imbalance = 1.0;
    }

  }

  if (outFlags_t[OF_ROUTINES] >= 2 && nRQ > 0)  {
      printf("%d: chemistry imbalance factor = %g \n",
                 myProc,imbalance);
  }

  if( monitorRoutines && (nRK > 0 || nRQ > 0) ) {
    if(equilibrateOnly) {
      printf("Equilibration successful:\n");
    } else {
      printf("Speciation successful:\n");
    }
    if(nRK > 0 && !equilibrateOnly)
      printf("  Average number of reaction time-steps = %g\n",avgNtTaken);
    if(nRQ > 0) {
      printf("  Average nonlinear iterations = %g, Max = %d\n",avgIter,
	     peakIter);
      if(nProcs == 1 && peakIter > 0)
	printf("  Max nonlinear iterations in cell (%d,%d,%d)\n",
	       exceptionLoc[0],exceptionLoc[1],exceptionLoc[2]);
    }
  }

exit:
  return error;
}

/******************************************************************************
 * Close down the Transport driver
 *****************************************************************************/
void transportClose() {
  if(computeTransport == ADV_CMM) {
    cleanupEvalTraceVel();
    traceCleanup();
    traceback_cleanup();
  }

  if(debugUnit) CloseTransDebug(&debugUnit);
}

/******************************************************************************
 * Debugging output of the concentration values
 *****************************************************************************/
static void debugOutConc(int printPad) {
  int padx1 = padx+1, pady1 = pady+1, padz1 = padz+1;
  int one = 1, four = 4;

  MyRDebugOut(&four,&step_t,conc,&one,&nx,&ny,&nz,&nSpecies,
	      &padx1,&pady1,&padz1,&printPad,&debugUnit);
}

/******************************************************************************
 * Debugging output of cbar and totCoefInt
 *****************************************************************************/
static void debugOutTB() {
  int zero = 0, one = 1, two = 2;

  MyDebugOut(&one,&step_t,cbar,&one,&nx,&ny,&nz,&nSpecies,
	     &zero,&zero,&zero,&zero,&debugUnit);
  MyDebugOut(&two,&step_t,totCoefInt,&one,&nx,&ny,&nz,&nTraceTypes,
	     &zero,&zero,&zero,&zero,&debugUnit);
}

/******************************************************************************
 * Debugging output of crhs
 *****************************************************************************/
static void debugOutCRHS() {
  int one = 1, three = 3;

  MyDebugOut(&three,&step_t,crhs,&one,&nx,&ny,&nz,&one,
	     &one,&one,&one,&one,&debugUnit);
}

/******************************************************************************
 * Parallel communication (padding) of the concentration array, and filling in
 * of BC and well cells.
 *****************************************************************************/
static int updateConc(int updateWells, int updateBCS, int updatePad) {
  int specie,error=0;
  int padx1 = padx+1, pady1 = pady+1, padz1 = padz+1;
  int one = 1;
  int zero = 0;

  if(updateWells) {
    FillWellConc(conc,&time_t,
		 &nWells,wellType,hWellIndex,zWellIndexLo,zWellIndexHi,
		 &nSpecies,&nx,&ny,&nz,&padx,&pady,&padz,myNbr);
  }

  if(updateBCS) {
    if(computeTransport) {
      PadinFlow(conc,&time_t,
		bcType_t,bcRegion1,bcRegion2,bcRegion3,
		&nSpecies,&nx,&ny,&nz,&padx,&pady,&padz,myNbr);
      if(error) return error;
    } else {
      for(specie=0; specie<nSpecies; specie++) {
	RPadExtFace(&conc_array_(specie),
		    &one,&nx,&ny,&nz,&padx1,&pady1,&padz1,
		    myNbr,&zero,&zero);
      }
    }
  }

  if(updatePad) {
    for(specie=0; specie<nSpecies; specie++) {
      int msg = MSG_PADCONC1;

      RPadStencil(&conc_array_(specie),
		  &one,&nx,&ny,&nz,&zero,&zero,&zero,&padx1,&pady1,&padz1,
		  myNbr,&msg,&error);
      if(error) return error;
      /* gsyncX(); */
      MPI_Barrier(comm);
    }
  }

  return 0;
}
