/******************************************************************************
1) POSTPROCESSING:
   Find the best local linear approximation to the concentration.
2) SLOPE LIMITING:
   Keep the concentration between the maximum and minimum average values on
   the neighboring cells.   Look at the eight corners.  If any is too
   big or too small, change the bad value to be in range and then
   adjust the diagonally opposity value to keep the total mass constant.
   This is very fast, and reasonable; however, it does not prevent local
  extrema from being created.
******************************************************************************/

#include <stdio.h>
/* #include <assert.h> */

#include "../Main/ftn.h"
#include "../Main/mymisc.h"
#include "../Main/float_types.h"
#include "../Main/cglobals.h"

static double * hc[3];
static double * byDist[3];
static int * MYNBR;

/* precomputed temporary variables to speed up indexing into conc[:] */
static int IDX0; 
static int IDX1;
static int IDX2;
static int IDX3;
static int IDX4;

#define MIN(A,B) ( (A) < (B) ? (A) : (B) )
#define MAX(A,B) ( (A) > (B) ? (A) : (B) )

/* assume fortran indexing */

#define mynbr(I,J,K) (MYNBR[(I)+1+3*((J)+1)+9*((K)+1)])
#define concLookUp(I,J,K,COMP) (((I)+concPad[0]) + (pl_concSize[0]+2)* \
				(((J)+concPad[1]) + (pl_concSize[1]+2)* \
				 (((K)+concPad[2]) + (pl_concSize[2]+2)* \
				  ((COMP)-1))))



#define newconcLookUp(I,J,K,COMP) \
 (  (IDX0)+(COMP)*(IDX4) + (K)*(IDX3) + (J)*(IDX2) + (I)  )

#define conc(I,J,K,COMP) CONC[newconcLookUp(I,J,K,COMP)]
#define hxc(I) ( hc[0][(I)+concPad[0]] )
#define hyc(J) ( hc[1][(J)+concPad[1]] )
#define hzc(K) ( hc[2][(K)+concPad[2]] )
#define byDistx(I) ( byDist[0][(I)+concPad[0]] )
#define byDisty(J) ( byDist[1][(J)+concPad[1]] )
#define byDistz(K) ( byDist[2][(K)+concPad[2]] )

void setupTrilinear(double* hxcP, double* hycP, double* hzcP,
		    double* byDistxP, double* byDistyP, double* byDistzP,
		    int* mynbrP) {
  hc[0] = hxcP;
  hc[1] = hycP;
  hc[2] = hzcP;
  byDist[0] = byDistxP;
  byDist[1] = byDistyP;
  byDist[2] = byDistzP;
  MYNBR = mynbrP;

/* precompute index variables (hope this is correct and can
   be done only once) to simplify index computation into conc[:] */

  IDX0 = concLookUp(0,0,0,0);
  IDX1 = concLookUp(1,0,0,0) - IDX0;
  IDX2 = concLookUp(0,1,0,0) - IDX0;
  IDX3 = concLookUp(0,0,1,0) - IDX0;
  IDX4 = concLookUp(0,0,0,1) - IDX0;

  /*  assert( IDX1 == 1 ); */

  return;
}

void Trilinear(CONC,tconc,i,j,k,comp)
     conc_t* CONC;
     double* tconc;
     integer i,j,k,comp;
{

/* temporary variables, hopefully, compiler can generate better code */

  int	ii = (*i);
  int 	jj = (*j);
  int 	kk = (*k);
  int	ccomp = (*comp);

  int    im1 = ii-1;
  int    ip1 = ii+1;
  int    jm1 = jj-1;
  int    jp1 = jj+1;
  int    km1 = kk-1;
  int    kp1 = kk+1;
  int    m0,m1,m2;

  /* minimize mixed mode real->double and double->real conversion,
     use double instead of conc_t for temp variables */

  double cavg = conc(ii,jj,kk,ccomp);
  double cmin,cmax;
  double maxChange,change;
  double gradX,gradY,gradZ;

  if (mynbr(-1,0,0) == -1)
    {
      if (ii < 1)
	goto returncavg;
      else if (ii == 1)
	im1 = 1;
    }
  if (mynbr( 1,0,0) == -1)
    {
      if (ii > l_concSize[0])
	goto returncavg;
      else if (ii == l_concSize[0])
	ip1 = l_concSize[0];
    }
  if (mynbr(0,-1,0) == -1)
    {
      if (jj < 1)
	goto returncavg;
      else if (jj == 1)
	jm1 = 1;
    }
  if (mynbr(0, 1,0) == -1)
    {
      if (jj > l_concSize[1])
	goto returncavg;
      else if (jj == l_concSize[1])
	jp1 = l_concSize[1];
    }
  if (mynbr(0,0,-1) == -1)
    {
      if (kk < 1)
	goto returncavg;
      else if (kk == 1)
	km1 = 1;
    }
  if (mynbr(0,0, 1) == -1)
    {
      if (kk > l_concSize[2])
	goto returncavg;
      else if (kk == l_concSize[2])
	kp1 = l_concSize[2];
    }

  cmin = cmax = cavg;
  for (m2=km1; m2 <= kp1; ++m2) {
    /* use temporary variable for common expression
	to encourage better code */
    conc_t ctmp;

    for (m1=jm1; m1 <= jp1; ++m1) {
      for (m0=im1; m0 <= ip1; ++m0)
	{
	  ctmp = conc(m0,m1,m2,ccomp);
	  cmin = MIN(cmin,ctmp);
	  cmax = MAX(cmax,ctmp);
	}
    };
   };
  maxChange = MIN(cavg-cmin,cmax-cavg);

/* ORIGINAL
  gradX = .5*( ( conc(ip1,*j,*k,*comp) - cavg )*byDistx(*i)
	      + ( cavg - conc(im1,*j,*k,*comp) )*byDistx(im1) );
  gradY = .5*( ( conc(*i,jp1,*k,*comp) - cavg )*byDisty(*j)
	      + ( cavg - conc(*i,jm1,*k,*comp) )*byDisty(jm1) );
  gradZ = .5*( ( conc(*i,*j,kp1,*comp) - cavg )*byDistz(*k)
	      + ( cavg - conc(*i,*j,km1,*comp) )*byDistz(km1) );
  gradX = .5*gradX*hxc(*i)*(3-(ip1-im1));
  gradY = .5*gradY*hyc(*j)*(3-(jp1-jm1));
  gradZ = .5*gradZ*hzc(*k)*(3-(kp1-km1));
*/


  /* merge computation expression, absorb 0.5*0.5 into 0.25 */

  gradX =0.25*hxc(ii)*(3-(ip1-im1))*
	( ( conc(ip1,jj,kk,ccomp) - cavg )*byDistx(ii)
              + ( cavg - conc(im1,jj,kk,ccomp) )*byDistx(im1) );
  gradY =0.25*hyc(jj)*(3-(jp1-jm1))*
	( ( conc(ii,jp1,kk,ccomp) - cavg )*byDisty(jj)
              + ( cavg - conc(ii,jm1,kk,ccomp) )*byDisty(jm1) );
  gradZ = 0.25*hzc(kk)*(3-(kp1-km1))*
	( ( conc(ii,jj,kp1,ccomp) - cavg )*byDistz(kk)
              + ( cavg - conc(ii,jj,km1,ccomp) )*byDistz(km1) );

  change = gradX+gradY+gradZ;
  change = MAX( -maxChange, MIN( maxChange, change ) );
  tconc[0] = cavg - change;
  tconc[7] = cavg + change;

  change = -gradX+gradY+gradZ;
  change = MAX( -maxChange, MIN( maxChange, change ) );
  tconc[1] = cavg - change;
  tconc[6] = cavg + change;

  change = gradX-gradY+gradZ;
  change = MAX( -maxChange, MIN( maxChange, change ) );
  tconc[2] = cavg - change;
  tconc[5] = cavg + change;

  change = -gradX-gradY+gradZ;
  change = MAX( -maxChange, MIN( maxChange, change ) );
  tconc[3] = cavg - change;
  tconc[4] = cavg + change;

  return;

 returncavg:
  {
    int m;
    for ( m = 0; m < 8; ++m)
      tconc[m] = cavg;
  }
  return;
}


