/*
  This file provides a module for the represention of velocity and
  porosity * saturation data.  We can set the RICHARDS approximation (to 1),
  in which case velocities are given for NPHASES-1 ( = NPHASES-RICHARDS )
  phases, and only NPHASES-1 porsats need be returned.  NPHASES must be
  at least 2, but one phase can be simulated with NPHASES=2 and RICHARDS=1.

  The principal components are:
  allocfcnmem_ - allocates space for dynamic data structures
  freefcnmem_  - releases space
  sendporosdata_
  sendporsatdata_
  sendsorpdata_
  sendveldata_ - provides velocity data for a given time.  Subsequent
                 calculations use last two velocities and times to
		 interpolate
  evalvel_ - produces an interpolated velocity value for a given point
            and time; previous location point is remembered and
            locality of reference is exploited
  evalporosity_
  evalsorp_
  evalsat_
  setcell_ - produces a list of n linear functions that assume
            positive values over a region containing a given point
            and which defines a region over which the velocity is
            differentiable
  evalcellface_ - produces a list of n linear functions that assume
                 positive values everywhere in the domain and only in
		 the domain
*/

#define NPHASES 2
#define RICHARDS 1

#include <values.h>
#include <math.h>
#include "ftn.h"
#include "bool.h"
#include "mybits.h"
#include "mymem.h"
#include "mymisc.h"
#include "float_types.h"
#include <stdio.h>
#include <stdlib.h>
#include "msg.h"
#include "debug.h"

/*
The data encapsulated in this module represents the velocity grid, its
size and spacings, its data, and the loction of the last query point.
*/

static int newestData = 0;
static int foundPresCell[3];
static fcn_t* velData[NPHASES-RICHARDS][2][3];
static fcn_t timeData[2];
static int np[3];
static int presSize[3];
static int pad[3];
static double* presPlane[3];
static fcn_t bound[6];
static int nbr[3][3][3];
static int msgRoot;
static bool evenBlock;
static float* dMap;
static int identityMap;
static int nr[3];

static fcn_t** sorpData;
static int nSorps;
static fcn_t* porsatData;

void AllocFcnmem(   xp,    yp,    zp,
     nxp,   nyp,   nzp,
   padxp, padyp, padzp,
   nsorps,
   mynbr, mymsgroot, parity,
   identityMp, dMp, nxr, nyr, nzr)
real8      xp,    yp,    zp;
integer   nxp,   nyp,   nzp;
integer   padxp, padyp, padzp;
integer nsorps;
integer mynbr, mymsgroot, parity;
integer identityMp;
real4 dMp;
integer nxr, nyr, nzr;
{
  int i, p;
  nSorps = *nsorps;
  msgRoot = *mymsgroot;
  for (i = 0; i < 27; ++i)
    nbr[i/9][i/3%3][i%3] = mynbr[i];
  evenBlock = *parity & 1;
  identityMap = *identityMp;
  dMap = dMp;
  nr[0] = *nxr;
  nr[1] = *nyr;
  nr[2] = *nzr;

  np[0] = *nxp;
  pad[0] = *padxp;
  presSize[0] = *nxp + 2 * *padxp;
  presPlane[0] = xp;
  np[1] = *nyp;
  pad[1] = *padyp;
  presSize[1] = *nyp + 2 * *padyp;
  presPlane[1] = yp;
  np[2] = *nzp;
  pad[2] = *padzp;
  presSize[2] = *nzp + 2 * *padzp;
  presPlane[2] = zp;

  for (i = 0; i < 3; ++i)
    {
      foundPresCell[i] = pad[i];

      ++presSize[i];
      /* Allocate space for velocity data */
      {
	int size = presSize[0]*presSize[1]*presSize[2];
	for (p = 0; p < NPHASES-RICHARDS; ++p)
	  {
	    velData[p][0][i] = (fcn_t*) MYMALLOC(2*size*sizeof(fcn_t));
	    velData[p][1][i] = velData[p][0][i] + size;
	  }
      }
      --presSize[i];
    }

  /* allocate storage for sorption and porosity (sorp 0) data */
  {
    int s;
    int size = presSize[0] * presSize[1] * presSize[2];
    sorpData = (fcn_t**) MYMALLOC((1+nSorps) * sizeof(fcn_t*));
    sorpData[0] = (fcn_t*) MYMALLOC((1+nSorps) * size * sizeof(fcn_t));
    for (s = 1; s <= nSorps; ++s)
      sorpData[s] = sorpData[0] + s*size;
    
    porsatData = (fcn_t*) MYMALLOC(2*(NPHASES-1)*size*sizeof(fcn_t));
  }
}

void FreeFcnmem()
{
  int j, p;
  for (j = 0; j < 3; ++j)
    for (p = 0; p < NPHASES-RICHARDS; ++p)
      MYFREE((char*) velData[p][0][j]);
  MYFREE((char*) sorpData[0]);
  MYFREE((char*) sorpData);
  MYFREE((char*) porsatData);
}

static int packPresGrid(loose, tight, tightLo, tightHi, looseSize)
     fcn_t* loose;
     double* tight;
     int tightLo[3], tightHi[3], looseSize[3];
{
  int i0, i1, i2;
  int k = 0;
  for (i2 = tightLo[2]; i2 < tightHi[2]; ++i2)
    for (i1 = tightLo[1]; i1 < tightHi[1]; ++i1)
      for (i0 = tightLo[0]; i0 < tightHi[0]; ++i0)
	tight[k++] = loose[i0 +
			   looseSize[0] * (i1 +
					   looseSize[1] * i2)];
  return k;
}

static int unpackPresGrid(loose, tight, tightLo, tightHi, looseSize)
     fcn_t* loose;
     double* tight;
     int tightLo[3], tightHi[3], looseSize[3];
{
  int i0, i1, i2;
  int k = 0;
  for (i2 = tightLo[2]; i2 < tightHi[2]; ++i2)
    for (i1 = tightLo[1]; i1 < tightHi[1]; ++i1)
      for (i0 = tightLo[0]; i0 < tightHi[0]; ++i0)
	{
	  loose[i0 +
		looseSize[0] * (i1 +
				looseSize[1] * i2)] = tight[k++];
	}
  return k;
}

static void padPresGrid(loose, dim, coord, tightLo, tightHi, looseSize)
     fcn_t* loose;
     int dim, coord;
     int tightLo[3], tightHi[3], looseSize[3];
{
  int ii[3];
  int i0 = dim;
  int i1 = (i0 < 2)? i0+1: 0;
  int i2 = (i1 < 2)? i1+1: 0;

  ii[i0] = coord;
  for (ii[i2] = tightLo[i2]; ii[i2] < tightHi[i2]; ++ii[i2])
    for (ii[i1] = tightLo[i1]; ii[i1] < tightHi[i1]; ++ii[i1])
      {
	fcn_t val = loose[ii[0] +
			   looseSize[0] * (ii[1] +
					   looseSize[1] * ii[2])];
	
	for (ii[i0] = tightLo[i0]; ii[i0] < tightHi[i0]; ++ii[i0])
	  loose[ii[0] +
		looseSize[0] * (ii[1] +
				looseSize[1] * ii[2])] = val;
	ii[i0] = coord;
      }
}

static void localUnpackPresGrid(loose, tight)
     fcn_t* loose;
     double* tight;
{
  int j, hi[3];
  for (j = 0; j < 3; ++j)
    hi[j] = presSize[j] - pad[j];
  unpackPresGrid(loose, tight, pad, hi, presSize);
}

static void fillPresGrid(dst, src)
     fcn_t* dst;
     double* src;
{
  int start[3], end[3], ndir[3];
  int i;
  
  localUnpackPresGrid(dst, src);
  
  /* exchange data with neighbors */
  for (i = 0; i < 3; ++i)
    {
      ndir[i] = 1;
      start[i] = pad[i];
      end[i] = presSize[i] - pad[i];
    }
  
  for (i = 0; i < 3; ++i)
    /* communicate boundary data in dimension i */
    {
      int j;
      bool way = evenBlock;
      double* msg, *msg2;
      int msgSize = 1;
      start[i] = 0;
      end[i] = pad[i];
      for (j = 0; j < 3; ++j)
	msgSize *= end[j] - start[j];
      msg = (double*) MYMALLOC(msgSize*sizeof(*msg));
      msg2 = (double*) MYMALLOC(msgSize*sizeof(*msg));
      
      do
	/* communicate boundary data in direction way */
	{
	  int buddy = ((way?
			(ndir[i] = 0, start[i] = pad[i]):
			(ndir[i] = 2, start[i] = presSize[i] - 2*pad[i])),
		       nbr[ndir[2]][ndir[1]][ndir[0]]);
	  end[i] = start[i] + pad[i];
	  
	  if (buddy != -1)
	    {
	      int msgType1 = msgRoot + 2*i + way;
	      int msgType2 = msgRoot + 2*i + !way;
	      packPresGrid(dst, msg, start, end, presSize);

	      /* asendRealX(msgType1,msg, msgSize,  buddy); */
	      {
		MPI_Status status;
		MPI_Sendrecv((void*)msg, msgSize, MPI_DOUBLE, buddy, msgType1,
			     (void*)msg2, msgSize, MPI_DOUBLE, buddy, msgType2,
			     comm, &status);
	      }
	      start[i] += (way? -pad[i]: pad[i]);
	      end[i] = start[i] + pad[i];
	      /* recvRealX(msgType2, msg2, msgSize); */

	      unpackPresGrid(dst, msg2, start, end, presSize);
	    }
	  else
	    {
	      start[i] += (way? -pad[i]: pad[i]);
	      end[i] = start[i] + pad[i];
	      padPresGrid(dst, i, start[i]+(way? pad[i]: -1), start, end, presSize);
	    }
	  way = !way;
	} 
      while (way != evenBlock);
      MYFREE ((char*) msg);
      MYFREE ((char*) msg2);
      ndir[i] = 1;
      start[i] = 0;
      end[i] = presSize[i];
    }
}

#if NPHASES==2 && RICHARDS==0
void SendVelPorSatData(velx0,vely0,velz0, velx1,vely1,velz1, porsat, t)
#else
void SendVelPorSatData(velx0,vely0,velz0, porsat, t)
#endif
     real8 velx0, vely0, velz0;
#if NPHASES==2 && RICHARDS==0
     real8 velx1, vely1, velz1;
#endif
     real8 porsat;
     real8 t;
{
  int start[3], end[3], ndir[3];
  int i, j;
  int p;
  double* velDir[NPHASES-RICHARDS][3];

  velDir[0][0] = velx0;
  velDir[0][1] = vely0;
  velDir[0][2] = velz0;
#if NPHASES==2 && RICHARDS==0
  velDir[1][0] = velx1;
  velDir[1][1] = vely1;
  velDir[1][2] = velz1;
#endif

  /* record the updated time */
  newestData = !newestData;
  timeData[newestData] = *t;

  /* record the local velocity data */
  for (p = 0; p < NPHASES-RICHARDS; ++p)
    for (j = 0; j < 3; ++j)
      {
	++presSize[j];
	localUnpackPresGrid(velData[p][newestData][j], velDir[p][j]);
	--presSize[j];
      }
 
  /* record the local porsat data */
  for (p = 0; p < NPHASES-1; ++p)
    {
      fcn_t* dst = &porsatData[presSize[0]*presSize[1]*presSize[2]*
			       (p + (NPHASES-1) * newestData)];
      double* src = porsat + p * ((presSize[0]-2*pad[0])*
				  (presSize[1]-2*pad[1])*
				  (presSize[2]-2*pad[2]));
      localUnpackPresGrid(dst, src);
    }

  /* exchange data with neighbors */
  for (i = 0; i < 3; ++i)
    {
      ndir[i] = 1;
      start[i] = pad[i];
      end[i] = presSize[i] - pad[i];
    }

  for (i = 0; i < 3; ++i)
    /* communicate boundary vel data in dimension i */
    {
      int j;
      bool way = evenBlock;
      double* msg, *msg2;
      int msgSize, vMsgSize = 0, sMsgSize = 1;
      start[i] = 0;
      end[i] = pad[i];
      for (j = 0; j < 3; ++j)
	{
	  int ii, dimProd = 1;
	  if (i != j) ++end[j];
	  for (ii = 0; ii < 3; ++ii)
	    dimProd *= end[ii] - start[ii];
	  vMsgSize += dimProd;
	  if (i != j) --end[j];
	  sMsgSize *= end[j] - start[j];
	}
      vMsgSize *= NPHASES-RICHARDS;
      sMsgSize *= NPHASES-1;
      msgSize = (vMsgSize + sMsgSize);
      msg = (double*) MYMALLOC(msgSize*sizeof(*msg));
      msg2 = (double*) MYMALLOC(msgSize*sizeof(*msg));

      do
	/* communicate boundary vel data in direction way */
	{
	  int buddy = ((way?
			(ndir[i] = 0, start[i] = pad[i]):
			(ndir[i] = 2, start[i] = presSize[i] - 2*pad[i])),
		       nbr[ndir[2]][ndir[1]][ndir[0]]);
	  end[i] = start[i] + pad[i];
	  
	  if (buddy != -1)
	    {
	      int msgType1 = msgRoot + 2*i + way;
	      int msgType2 = msgRoot + 2*i + !way;
	      int k = 0;
	      for (j = 0; j < 3; ++j)
		/* pack dimension j velocities */
		{
		  if (i == j)
		    {
		      if (way) ++start[j];
		      else --end[j];
		    }
		  ++end[j];
		  ++presSize[j];
		  for (p = 0; p < NPHASES-RICHARDS; ++p)
		    k += packPresGrid(velData[p][newestData][j],
				      &msg[k], start, end, presSize);
		  --presSize[j];
		  --end[j];
		  if (i == j)
		    {
		      if (way) --start[j];
		      else ++end[j];
		    }
		}

	      for (p = 0; p < NPHASES-1; ++p)
		{
		  fcn_t* dst = &porsatData[presSize[0]*presSize[1]*presSize[2]*
					   (p + (NPHASES-1) * newestData)];
		  k += packPresGrid(dst, &msg[k], start, end, presSize);
		}
	      /* asendRealX(msgType1, msg, msgSize, buddy); */
	      {
		MPI_Status status;
		MPI_Sendrecv((void*)msg, msgSize, MPI_DOUBLE, buddy, msgType1,
			     (void*)msg2, msgSize, MPI_DOUBLE, buddy, msgType2,
			     comm, &status);
	      }
	      start[i] += (way? -pad[i]: pad[i]);
	      end[i] = start[i] + pad[i];
	      /* recvRealX(msgType2, msg2, msgSize); */
	      k = 0;
	      for (j = 0; j < 3; ++j)
		/* unpack dimension j velocities */
		{
		  if (i == j)
		    {
		      if (way) --end[j];
		      else ++start[j];
		    }
		  ++end[j];
		  ++presSize[j];
		  for (p = 0; p < NPHASES-RICHARDS; ++p)
		    k += unpackPresGrid(velData[p][newestData][j],
					&msg2[k], start, end, presSize);
		  --presSize[j];
		  --end[j];
		  if (i == j)
		    {
		      if (way) ++end[j];
		      else --start[j];
		    }
		}
	      for (p = 0; p < NPHASES-1; ++p)
		{
		  fcn_t* dst = &porsatData[presSize[0]*presSize[1]*presSize[2]*
					   (p + (NPHASES-1) * newestData)];
		  k += unpackPresGrid(dst, &msg2[k], start, end, presSize);
		}
	    }
	  else
	    {
	      start[i] = way? 0: presSize[i] - pad[i];
	      end[i] = start[i] + pad[i];
	      for (j = 0; j < 3; ++j)
		{
		  ++end[j];
		  ++presSize[j];
		  for (p = 0; p < NPHASES-RICHARDS; ++p)
		    padPresGrid
		      (velData[p][newestData][j], i,
		       (way? pad[i]: presSize[i]-pad[i]-1), start, end, presSize);
		  --presSize[j];
		  --end[j];
		}
	      for (p = 0; p < NPHASES-1; ++p)
		{
		  fcn_t* dst = &porsatData[presSize[0]*presSize[1]*presSize[2]*
					   (p + (NPHASES-1) * newestData)];
		  padPresGrid(dst, i, start[i]+(way? pad[i]: -1), start, end, presSize);
		}
	    }
	  way = !way;
	} while (way != evenBlock);
      MYFREE ((char*) msg);
      MYFREE ((char*) msg2);
      ndir[i] = 1;
      start[i] = 0;
      end[i] = presSize[i];
    }
}

void SendSorpData(sorp)
     real8 sorp;
{
  int s;
  for (s = 0; s < nSorps; ++s)
    fillPresGrid(sorpData[s+1],
		 sorp+s*
		 (presSize[0]-2*pad[0])*
		 (presSize[1]-2*pad[1])*
		 (presSize[2]-2*pad[2]));
}

void SendPorosData(porosity)
     real8 porosity;
{
  fillPresGrid(sorpData[0], porosity);
}

static bool findPresCell(sought)
     fcn_t sought[3];
{
  int j;
  for (j = 0; j < 3; ++j)
    {
      double* plane = presPlane[j];
      double goal = sought[j];
      int lBound = 0;
      int uBound = presSize[j];
      int k0 = foundPresCell[j]/2;
      int k1 = k0 + 1;
      int step = 1;
      while (goal < plane[k0])
	{
	  k0 -= step;
	  if (k0 < lBound)
	    {
	      k0 = lBound;
	      break;
	    }
	  step *= 2;
	}
      step = 1;
      while (plane[k1] <= goal)
	{
	  k1 += step;
	  if (uBound < k1)
	    {
	      k1 = uBound;
	      break;
	    }
	  step *= 2;
	}
      while (k0 + 1 != k1)
	{
	  int mid = (k0+k1)/2;
	  if (goal < plane[mid])
	    k1 = mid;
	  else
	    k0 = mid;
	}
      foundPresCell[j] = 2*k0;
      if (goal >= 0.5*(plane[k0]+plane[k1]))
	++foundPresCell[j];
      if (foundPresCell[j] == lBound || foundPresCell[j] == 2*uBound-1)
	return FALSE;
    }
  return TRUE;
}

bool EvalCornerVels(xPos, yPos, zPos,
   t,
   vel)
real8 xPos, yPos, zPos;
real8 t;
double (*vel)[8][3];
{
  int j, p;
  fcn_t sought[3];
  double relTime;
  sought[0] = *xPos;
  sought[1] = *yPos;
  sought[2] = *zPos;
  if (!findPresCell(sought))
    return FALSE;
  
  relTime = ((*t          - timeData[0]) /
	     (timeData[1] - timeData[0]));

  for (j = 3; j-- > 0;)
    for (p = 0; p < NPHASES-RICHARDS; ++p)
      {
	int i, k;
	int bb[3];
	fcn_t v[16];
	++presSize[j];
	++foundPresCell[j];
	/* Find data velocities at 16 corners of containing hyperbox */
	for (i = 0; i < 3; ++i)
	  bb[i] = (foundPresCell[i]-1)/2;
	for (k = 0; k < 16; ++k)
	  v[k] = velData[p][bit(k,3)][j]
	    [bb[0]+bit(k,0) +
	     presSize[0] * (bb[1]+bit(k,1) +
			    presSize[1] * (bb[2]+bit(k,2)))];

	/* Adjust velocites for bounding sub-hyperbox */
	for (i = 0; i < 3; ++i)
	  {
	    double* plane = presPlane[i];
	    fcn_t w = (i == j)? 0.5 :
	      ((plane[bb[i]+1] - plane[bb[i]]) /
	       (plane[bb[i]+2] - plane[bb[i]]));
	    int kbit = !!(foundPresCell[i]%2)? setbit(0,i): 0;
	    int mask = clrbit(0xf, i);
	    int k = 0;
	    do
	      v[k | kbit] = v[k] + w*(v[setbit(k,i)] - v[k]);
	    while (!!maskedIncrement(mask,k));
	  }
	--foundPresCell[j];
	--presSize[j];

	/* Fold corner velocities in time dimension */
	for (k = 0; k < 8; ++k)
	  vel[p][k][j] = v[k] + relTime * (v[k+8]-v[k]);
      }
  return TRUE;
}

bool EvalVel(xPos, yPos, zPos,
   t,
   vel)
real8 xPos, yPos, zPos;
real8 t;
double (*vel)[3];
{
  int j, p;
  fcn_t baryPos[4];
  fcn_t sought[3];
  sought[0] = *xPos;
  sought[1] = *yPos;
  sought[2] = *zPos;
  if (!findPresCell(sought))
    return FALSE;

  baryPos[3] = ((*t          - timeData[0]) /
		(timeData[1] - timeData[0]));

  for (j = 3; j-- > 0;)
    {
      int bigCell = foundPresCell[j] / 2;
      double* plane = presPlane[j];
      baryPos[j] = ((sought[j]        - plane[bigCell]) /
		    (plane[bigCell+1] - plane[bigCell]));
      baryPos[j] *= 2.;
      if (baryPos[j] >= 1.)
	baryPos[j] -= 1.;
    }

  for (j = 3; j-- > 0;)
    for (p = 0; p < NPHASES-RICHARDS; ++p)
      {
	int i, k;
	int bb[3];
	fcn_t v[16];
	++presSize[j];
	++foundPresCell[j];
	/* Find data velocities at 16 corners of containing hyperbox */
	for (i = 0; i < 3; ++i)
	  bb[i] = (foundPresCell[i]-1)/2;
	for (k = 0; k < 16; ++k)
	  v[k] = velData[p][bit(k,3)][j]
	    [bb[0]+bit(k,0) +
	     presSize[0] * (bb[1]+bit(k,1) +
			    presSize[1] * (bb[2]+bit(k,2)))];

	/* Adjust velocites for bounding sub-hyperbox */
	for (i = 0; i < 3; ++i)
	  {
	    double* plane = presPlane[i];
	    fcn_t w = (i == j)? 0.5 :
	    ((plane[bb[i]+1] - plane[bb[i]]) /
	     (plane[bb[i]+2] - plane[bb[i]]));
	    int kbit = !!(foundPresCell[i]%2)? setbit(0,i): 0;
	    int mask = clrbit(0xf, i);
	    int k = 0;
	    do
	      v[k | kbit] = v[k] + w*(v[setbit(k,i)] - v[k]);
	    while (!!maskedIncrement(mask,k));
	  }
	--foundPresCell[j];
	--presSize[j];

	/* Fold corner velocities into velocity at point */
 	for (i = 4; i-- > 0 ;)
	  {
	    int gap = 1 << i;
	    for (k = 0; k < gap; ++k)
	      v[k] += baryPos[i] * (v[k+gap]-v[k]);
	  }
	vel[p][j] = v[0];
      }
  return TRUE;
}

bool SetCell(xPos, yPos, zPos, t,
   nFaces)
real8 xPos, yPos, zPos;
real8 t;
integer nFaces;
{
  int j;
  bool found;
  fcn_t sought[3];
  sought[0] = *xPos;
  sought[1] = *yPos;
  sought[2] = *zPos;
  found = findPresCell(sought);
  for (j = 0; j < 3; ++j)
    {
      double alpha;
      bound[2*j] = 0.5 * (presPlane[j][(foundPresCell[j]+0)/2] +
			  presPlane[j][(foundPresCell[j]+1)/2]);
      bound[2*j+1] = 0.5 * (presPlane[j][(foundPresCell[j]+1)/2] +
			    presPlane[j][(foundPresCell[j]+2)/2]);
      alpha = (sought[j] - bound[2*j]) / (bound[2*j+1] - bound[2*j]);
      if (alpha < 1e-4)
	bound[2*j] = 0.5 * (presPlane[j][(foundPresCell[j]-1)/2] +
			    presPlane[j][(foundPresCell[j]+0)/2]);
      else if (1.-alpha < 1e-4)
	bound[2*j+1] = 0.5 * (presPlane[j][(foundPresCell[j]+2)/2] +
			      presPlane[j][(foundPresCell[j]+3)/2]);
    }
  *nFaces = found? 6: 0;
  return found;
}

double EvalCellSize(xPos, yPos, zPos)
real8 xPos, yPos, zPos;
{
  int j;
  fcn_t sought[3];
  double min = MAXDOUBLE;
  sought[0] = *xPos;
  sought[1] = *yPos;
  sought[2] = *zPos;
  findPresCell(sought);
  for (j = 0; j < 3; ++j)
    {
      double bound1;
      double bound0 = 0.5 * (presPlane[j][(foundPresCell[j]+0)/2] +
			     presPlane[j][(foundPresCell[j]+1)/2]);
      if (bound0 == sought[j])
	bound0 = 0.5 * (presPlane[j][(foundPresCell[j]-1)/2] +
			presPlane[j][(foundPresCell[j]+0)/2]);
      bound1 = 0.5 * (presPlane[j][(foundPresCell[j]+1)/2] +
		      presPlane[j][(foundPresCell[j]+2)/2]);
      if (min > bound1-bound0)
	min = bound1-bound0;
    }
  return min;
}

double EvalCellFace(xPos, yPos, zPos, i)
real8 xPos, yPos, zPos;
integer i;
{
  switch (*i)
    {
    case 0:
      return *xPos - bound[0];
    case 1:
      return bound[1] - *xPos;
    case 2:
      return *yPos - bound[2];
    case 3:
      return bound[3] - *yPos;
    case 4:
      return *zPos - bound[4];
    case 5:
      return bound[5] - *zPos;
    }
  return 0.;
}


bool presGridIntegrate(xPos, yPos, zPos,
		       nDatasets, dataset, result)
     double *xPos, *yPos, *zPos;
     int nDatasets;
     fcn_t* dataset;
     double* result;
{
  typedef fcn_t datacube[8];
  datacube* dat = (datacube*) MYMALLOC(nDatasets*sizeof(datacube));
  int i, j, k[3], m;
  int index0[3], index1[3];
  fcn_t baryPos0[3], baryPos1[3];
  fcn_t corner0[3], corner1[3];
  fcn_t width[3];
  
  corner0[0] = xPos[0];
  corner0[1] = yPos[0];
  corner0[2] = zPos[0];
  if (!findPresCell(corner0)) return FALSE;
  index0[0] = (foundPresCell[0]-1) / 2;
  index0[1] = (foundPresCell[1]-1) / 2;
  index0[2] = (foundPresCell[2]-1) / 2;

  corner1[0] = xPos[1];
  corner1[1] = yPos[1];
  corner1[2] = zPos[1];
  if (!findPresCell(corner1)) return FALSE;
  index1[0] = (foundPresCell[0]-1) / 2;
  index1[1] = (foundPresCell[1]-1) / 2;
  index1[2] = (foundPresCell[2]-1) / 2;

  for (i = 0; i < nDatasets; ++i)
    result[i] = 0.;

  for (k[2] = index0[2]; k[2] <= index1[2]; ++k[2])
    {
      width[2] = 0.5*(presPlane[2][k[2]+2] - presPlane[2][k[2]]);
      if (k[2] == index0[2])
	{
	  double* plane = presPlane[2];
	  baryPos0[2] =
	    (corner0[2] - 0.5*(plane[k[2]+1] + plane[k[2]])) / width[2];
	}
      else baryPos0[2] = 0.;
      if (k[2] == index1[2])
	{
	  double* plane = presPlane[2];
	  baryPos1[2] =
	    (corner1[2] - 0.5*(plane[k[2]+1] + plane[k[2]])) / width[2];
	}
      else baryPos1[2] = 1.;
      
      for (k[1] = index0[1]; k[1] <= index1[1]; ++k[1])
	{
	  width[1] = 0.5*(presPlane[1][k[1]+2] - presPlane[1][k[1]]);
	  if (k[1] == index0[1])
	    {
	      double* plane = presPlane[1];
	      baryPos0[1] =
		(corner0[1] - 0.5*(plane[index0[1]+1] + plane[index0[1]])) / width[1];
	    }
	  else baryPos0[1] = 0.;
	  if (k[1] == index1[1])
	    {
	      double* plane = presPlane[1];
	      baryPos1[1] =
		(corner1[1] - 0.5*(plane[index1[1]+1] + plane[index1[1]])) / width[1];
	    }
	  else baryPos1[1] = 1.;
      
	  for (k[0] = index0[0]; k[0] <= index1[0]; ++k[0])
	    {
	      width[0] = 0.5*(presPlane[0][k[0]+2] - presPlane[0][k[0]]);
	      if (k[0] == index0[0])
		{
		  double* plane = presPlane[0];
		  baryPos0[0] =
		    (corner0[0] - 0.5*(plane[index0[0]+1] + plane[index0[0]])) / width[0];
		}
	      else baryPos0[0] = 0.;
	      if (k[0] == index1[0])
		{
		  double* plane = presPlane[0];
		  baryPos1[0] =
		    (corner1[0] - 0.5*(plane[index1[0]+1] + plane[index1[0]])) / width[0];
		}
	      else baryPos1[0] = 1.;
	      {
		fcn_t vol = ((baryPos1[2]-baryPos0[2]) * width[2] *
			     (baryPos1[1]-baryPos0[1]) * width[1] *
			     (baryPos1[0]-baryPos0[0]) * width[0]);

		/* Find data at 8 corners of containing box */
		for (i = 0; i < nDatasets; ++i)
		  for (m = 0; m < 8; ++m)
		    {
		      int index = (k[0]+bit(m,0) + presSize[0] *
				   (k[1]+bit(m,1) + presSize[1] *
				    (k[2]+bit(m,2) + presSize[2] *
				     (i))));
		      dat[i][m] = dataset[index];
		    }
	      
		/* Adjust data for bounding sub-box */
		for (j = 3; j-- > 0;)
		  {
		    if (baryPos0[j] > 0.)
		      {
			fcn_t w = baryPos0[j];
			int mbit = 0;
			int mask = clrbit(0x7, j);
			int m = 0;
			do
			  {
			    int dest = m | mbit;
			    int m2 = setbit(m,j);
			    for (i = 0; i < nDatasets; ++i)
			      dat[i][dest] = dat[i][m] + w*(dat[i][m2] - dat[i][m]);
			  }
			while (!!maskedIncrement(mask,m));
		      }
		    if (baryPos1[j] < 1.)
		      {
			fcn_t w = baryPos1[j];
			int mbit = setbit(0,j);
			int mask = clrbit(0x7, j);
			int m = 0;
			do
			  {
			    int dest = m | mbit;
			    int m2 = setbit(m,j);
			    for (i = 0; i < nDatasets; ++i)
			      dat[i][dest] = dat[i][m] + w*(dat[i][m2] - dat[i][m]);
			  }
			while (!!maskedIncrement(mask,m));
		      }
		  }
		for (i = 0; i < nDatasets; ++i)
		  {
		    int m;
		    fcn_t sum = 0.;
		    for (m = 0; m < 8; ++m)
		      sum += dat[i][m];
		    sum *= 0.125 * vol;
		    result[i] += sum;
		  }
	      }
	    }
	}
    }
  MYFREE((char*)dat);
  return TRUE;
}

void BarytoNatural(bary, xPos, yPos, zPos)
     double bary[3];
     real8 xPos;
     real8 yPos;
     real8 zPos;
{
  double p[3];
  int j;
  for (j = 3; j-- > 0;)
    {
      int bigCell = foundPresCell[j] / 2;
      double* plane = presPlane[j];
      double lo = plane[bigCell];
      double hi = plane[bigCell+1];
      if (foundPresCell[j] & 1)
	lo = 0.5*(lo+hi);
      else
	hi = 0.5*(lo+hi);
      p[j] = lo + bary[j]*(hi-lo);
    }
  *xPos = p[0];
  *yPos = p[1];
  *zPos = p[2];
}

static bool presGridInterp(xPos, yPos, zPos, nDatasets, dataset, result)
     fcn_t xPos, yPos, zPos;
     int nDatasets;
     fcn_t* dataset;
     double* result;
{
  int i, j, k;
  typedef fcn_t datacube[8];
  datacube* dat = (datacube*) MYMALLOC(nDatasets*sizeof(datacube));
  int bb[3];
  fcn_t baryPos[3];
  fcn_t sought[3];
  sought[0] = xPos;
  sought[1] = yPos;
  sought[2] = zPos;
  if (!findPresCell(sought)) return FALSE;
  for (j = 3; j-- > 0;)
    {
      int bigCell = foundPresCell[j] / 2;
      double* plane = presPlane[j];
      bb[j] = (foundPresCell[j]-1)/2;
      baryPos[j] = ((sought[j]        - plane[bigCell]) /
		    (plane[bigCell+1] - plane[bigCell]));
      if (baryPos[j] >= 0.5)
	baryPos[j] = 2*baryPos[j] - 1.;
    }
  
  /* Find data at 8 corners of containing box */
  for (i = 0; i < nDatasets; ++i)
    for (k = 0; k < 8; ++k)
      {
	int index = (bb[0]+bit(k,0) + presSize[0] *
		     (bb[1]+bit(k,1) + presSize[1] *
		      (bb[2]+bit(k,2) + presSize[2] *
		       (i))));
	dat[i][k] = dataset[index];
      }
  
  /* Adjust data for bounding sub-box */
  for (j = 3; j-- > 0;)
    {
      double* plane = presPlane[j];
      fcn_t w = ((plane[bb[j]+1] - plane[bb[j]]) /
		  (plane[bb[j]+2] - plane[bb[j]]));
      int kbit = !!(foundPresCell[j]%2)? setbit(0,j): 0;
      int mask = clrbit(0x7, j);
      int k = 0;
      do
	{
	  int dest = k | kbit;
	  int k2 = setbit(k,j);
	  for (i = 0; i < nDatasets; ++i)
	    dat[i][dest] = dat[i][k] + w*(dat[i][k2] - dat[i][k]);
	}
      while (!!maskedIncrement(mask,k));
    }

  /* Fold corner sorptions into sorption at point */
  for (j = 3; j-- > 0 ;)
    {
      int gap = 1 << j;
      for (k = 0; k < gap; ++k)
	for (i = 0; i < nDatasets; ++i)
	  dat[i][k] += baryPos[j] * (dat[i][k+gap]-dat[i][k]);
    }
  for (i = 0; i < nDatasets; ++i)
    result[i] = dat[i][0];
  MYFREE((char*)dat);
  return TRUE;
}

bool EvalSorp(xPos, yPos, zPos, sorp)
real8 xPos, yPos, zPos, sorp;
{
  return presGridInterp(*xPos, *yPos, *zPos, nSorps+1, sorpData[0], sorp);
}

double EvalOneSorp(xPos, yPos, zPos, sorpid)
real8 xPos, yPos, zPos;
integer sorpid;
{
  double result;
  presGridInterp(*xPos, *yPos, *zPos, 1, sorpData[*sorpid], &result);
  return result;
}

double EvalPorosity(xPos, yPos, zPos)
     real8 xPos, yPos, zPos;
{
  double result;
  presGridInterp(*xPos, *yPos, *zPos, 1, sorpData[0], &result);
  return result;
}

bool EvalPorSat(xPos, yPos, zPos, t, porsat)
real8 xPos, yPos, zPos, t;
real8 porsat;
{
  int p;
  double results[2*(NPHASES-1)];
  bool ok = presGridInterp
    (*xPos, *yPos, *zPos, 2*(NPHASES-1), porsatData, results);
  fcn_t alpha = ((*t          - timeData[0]) /
		  (timeData[1] - timeData[0]));
#if RICHARDS==0
  fcn_t sum = EvalPorosity(xPos, yPos, zPos);
#endif
  for (p = 0; p < NPHASES-1; ++p)
    {
      porsat[p] = results[p] + alpha*(results[p+NPHASES-1] - results[p]);
#if RICHARDS==0
      sum -= porsat[p];
#endif
    }
#if RICHARDS==0
  porsat[NPHASES-1] = sum;
#endif
  return ok;
}

bool EvalSorpInt(xpos, ypos, zpos, result)
     double *xpos, *ypos, *zpos;
     double *result;
{
  return presGridIntegrate(xpos, ypos, zpos,
			   nSorps+1, sorpData[0],
			   result);
}

double EvalOneSorpInt(xpos, ypos, zpos, sorpid)
     double *xpos, *ypos, *zpos;
     int* sorpid;
{
  double result;
  presGridIntegrate(xpos, ypos, zpos, 1, sorpData[*sorpid], &result);
  return result;
}

bool EvalPorSatInt(xpos, ypos, zpos, t, porsat)
     double *xpos, *ypos, *zpos;
     double *t;
     double *porsat;
{
  double results[2*(NPHASES-1)];
  bool ok = presGridIntegrate(xpos, ypos, zpos,
			      2*(NPHASES-1), porsatData,
			      results);
  fcn_t alpha = ((*t          - timeData[0]) /
		  (timeData[1] - timeData[0]));
#if RICHARDS==0
  int zero = 0;
  fcn_t sum = EvalOneSorpInt(xpos, ypos, zpos, &zero);
#endif
  int p;
  for (p = 0; p < NPHASES-1; ++p)
    {
      porsat[p] = results[p] + alpha*(results[p+NPHASES-1] - results[p]);
#if RICHARDS==0
      sum -= porsat[p];
#endif
    }
#if RICHARDS==0
  porsat[NPHASES-1] = sum;
#endif
  return ok;
}

void EvalEffSpeed(vel,ii,jj,kk,speed)
     double (*vel)[3];
     integer ii,jj,kk; /* these are fortran indices on the conc
			  grid for vertices starting at 1 */
     real8 speed;
{
  int p;

  if(identityMap)
    {
      for (p = 0; p < NPHASES-RICHARDS; ++p)
	{
	  int m;

	  speed[p] = 0.0;
	  for(m = 0; m < 3; ++m)
	    speed[p] += vel[p][m]*vel[p][m];
	  speed[p] = sqrt(speed[p]);
	}
    }
  else
    {
      int i = (*ii - 1)*nr[0];
      int j = (*jj - 1)*nr[1];
      int k = (*kk - 1)*nr[2];

      for (p = 0; p < NPHASES-RICHARDS; ++p)
	{
	  int m;

	  speed[p] = 0.0;
	  for(m = 0; m < 3; ++m)
	    {
	      int n;
	      double effV = 0.;

	      for (n = 0; n < 3; ++n)
		effV += dMap[m + 3*(n + 3*(i + (np[0]+1)*(j + (np[1]+1)*k)))]*vel[p][n];
	      speed[p] += effV*effV;
	    }
	  speed[p] = sqrt(speed[p]);
	}
    }
}
